//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef f_AUDIOSOURCE_H
#define f_AUDIOSOURCE_H

#include <windows.h>
#include <vfw.h>

#include "DubSource.h"

class IAVIReadHandler;
class IAVIReadStream;

class AudioSource : public DubSource {
public:
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	AudioSource(void) {
		m_SourceType = SOURCE_STD_FILE | SOURCE_AUDIO;
		m_SourceName = L"Not available [baseclass AudioSource]";
		m_SourceDesc = L"baseclass AudioSource";
	};
	// This function should be used to extend/change the way the stream is
	// handled
	// By default the stream is handled as for AVI, i.e. assuming it is
	// Constant Bitrate and using the byterate as main rate of the stream
	// In this case VDub will request a certain amount of blocks
	// that you need to give (no more)
	// Using this function can put the handling in a 'VBR' state : then
	// the SampleRate will be used as main rate
	// You can also tell if you prefer (if possible) to group audio frames
	// or not. The value of groupFrames tells what to do :
	//   0  : treat only one audio frame per read request
	//   1  : treat as many audio frames as necessary for the read request
	//  n>1 : treat a minimum of n frames per read request
	// Note that the groupFrames parameter can sometimes be inefective : if VBR
	// handling is off and audio frames don't all have the same length in bytes
	virtual void extend(bool bVBRHandling, sint32 groupFrames) {};
// END ****************************************************************
	WAVEFORMATEX *getWaveFormat() {
		return (WAVEFORMATEX *)getFormat();
	}
};

class AudioSourceWAV : public AudioSource {
private:
	HMMIO				hmmioFile;
	MMCKINFO			chunkRIFF;
	MMCKINFO			chunkDATA;
	LONG				lCurrentSample;
	LONG				bytesPerSample;

	~AudioSourceWAV();

public:
	AudioSourceWAV(const wchar_t *fn, LONG inputBufferSize);

	bool init();
	virtual int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);
};

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
class AudioSourceAVI : public AudioSource, public IStreamIndex, public IMP3Source {
private:
	IAVIReadHandler *pAVIFile;
	IAVIReadStream *pAVIStream;
	bool bQuiet;
	int m_StreamIndex;
	bool m_bFixVBRAudio;

	BOOL _isKey(LONG lSample);

	~AudioSourceAVI();

public:
	AudioSourceAVI(IAVIReadHandler *pAVIFile, bool bAutomated, int streamIndex, bool bFixVBRAudio);

	void setRate(const VDFraction& f) { streamInfo.dwRate = f.getHi(); streamInfo.dwScale = f.getLo(); }

	void Reinit();
	bool isStreaming();

	void streamBegin(bool fRealTime);
	void streamEnd();

	bool init();
	int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lSamplesRead, LONG *lBytesRead);

	bool QueryInterface(InterfaceID ID, void **destination);
	IAVIReadHandler *GetReadHandler(void);
	IAVIReadStream *GetReadStream(void);
	DWORD GetStreamType(void);
	long GetStreamIndex(void);
	bool UseVBRalaNandub(void);
	bool IsVBR(void);
	bool FixVBR(bool bFix);
};
// END ****************************************************************

#endif
