//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef f_DUBSOURCE_H
#define f_DUBSOURCE_H

#include <windows.h>
#include <vfw.h>

#include <vd2/system/fraction.h>
#include <vd2/system/refcount.h>

class InputFile;

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
#include <vd2/system/vdtypes.h>
#include <vd2/system/VDString.h>
#include "ModClassInterfaces.h"
// Some types
#define SOURCE_CONTAINER	0xF000
#define SOURCE_MAIN_TYPE	0x0F00
#define SOURCE_SUBTYPE		0x00FF
#define SOURCE_TYPE			(SOURCE_MAIN_TYPE|SOURCE_SUBTYPE)
#define CONTAINER(x)		((x)&SOURCE_CONTAINER)
#define MAIN_TYPE(x)		((x)&SOURCE_MAIN_TYPE)
#define SUBTYPE(x)			((x)&SOURCE_SUBTYPE)
#define TYPE(x)				((x)&SOURCE_TYPE)
#define CONTAINER_TO_INDEX(x)	(CONTAINER(x)>>12)
#define MAIN_TYPE_TO_INDEX(x)	(MAIN_TYPE(x)>>8)
#define SUBTYPE_TO_INDEX(x)		SUBTYPE(x)
#define TYPE_TO_INDEX(x)		TYPE(x)
#define INDEX_TO_CONTAINER(x)	(CONTAINER(x<<12))
#define INDEX_TO_MAIN_TYPE(x)	(MAIN_TYPE(x<<8))
#define INDEX_TO_SUBTYPE(x)		SUBTYPE(x)
#define INDEX_TO_TYPE(x)		TYPE(x)
// Some containers
#define SOURCE_STD_FILE		0x0000
#define SOURCE_AVI			0x1000
#define SOURCE_WAV			0x2000
#define SOURCE_MPEG			0x3000
#define SOURCE_OGM			0x4000
#define SOURCE_MKV			0x5000
// Unknown ^^
#define SOURCE_UNKNOWN		0x0000
// Video
#define SOURCE_VIDEO		0x0100
// Audio
#define SOURCE_AUDIO		0x0200
#define SOURCE_AUDIO_WAV		0x0201
#define SOURCE_AUDIO_MP3		0x0202
#define SOURCE_AUDIO_AC3		0x0203
#define SOURCE_AUDIO_VORBIS		0x0204
#define SOURCE_AUDIO_DTS		0x0205
// Text
#define SOURCE_TEXT			0x0300
#define SOURCE_TEXT_SRT			0x0301
#define SOURCE_TEXT_SSA			0x0302


class IVDStreamSource : public IVDRefCount, public ExposeInterface {
protected:
	VDStringW	m_SourceName;
	VDStringW	m_SourceDesc;
	uint32		m_SourceType;
	void clearSourceInfo(void)
		{ m_SourceName.clear(); m_SourceDesc.clear(); };
// END ****************************************************************
public:
	virtual bool init() = 0;

	virtual VDPosition	getLength() = 0;
	virtual VDPosition	getStart() = 0;
	virtual VDPosition	getEnd() = 0;
	virtual const VDFraction getRate() = 0;

	virtual const AVISTREAMINFO& getStreamInfo() = 0;		// DELETE ME I'M STUPID

	virtual int read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) = 0;

	virtual void *getFormat() const = 0;
	virtual int getFormatLen() const = 0;

	virtual bool isStreaming() = 0;

	virtual void streamBegin( bool fRealTime) = 0;
	virtual void streamEnd() = 0;

	enum ErrorMode {
		kErrorModeReportAll = 0,
		kErrorModeConceal,
		kErrorModeDecodeAnyway,
		kErrorModeCount
	};

	virtual void setDecodeErrorMode(ErrorMode mode) = 0;
	virtual bool isDecodeErrorModeSupported(ErrorMode mode) = 0;

	virtual LONG msToSamples(LONG lMs) const = 0;
	virtual LONG samplesToMs(LONG lSamples) const = 0;

//	virtual LONG samplesToSamples(const DubSource *source, LONG lSamples) const = 0;
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	// Returns the position in file of a sample
	virtual sint64 sampleOffset(sint64 lSample)
		{ return 0; };
	// Returns how many samples we would miss if asking to read starting a sample lStart
	// (useless with AVI sources, but not with OGM)
	virtual sint32 samplesLag(sint64 lStart)
		{ return 0; };
	// Returns the original offset (if any; in ms) of the stream
	virtual sint32 sourceOffsetMS(void)
		{ return 0; };
	VDStringW& getSourceName(void)
		{ return m_SourceName; };
	VDStringW& getSourceDesc(void)
		{ return m_SourceDesc; };
	uint32 getSourceFullType(void)
		{ return m_SourceType; };
	uint32 getSourceContainer(void)
		{ return CONTAINER(m_SourceType); };
	uint32 getSourceType(void)
		{ return TYPE(m_SourceType); };
	uint32 getSourceMainType(void)
		{ return MAIN_TYPE(m_SourceType); };
	uint32 getSourceSubType(void)
		{ return SUBTYPE(m_SourceType); };
	virtual VDStringW writeReport(void);
// END ****************************************************************
};

class DubSource : public vdrefcounted<IVDStreamSource> {
private:
	void *	format;
	int		format_len;

protected:
	void *allocFormat(int format_len);

	LONG lSampleFirst, lSampleLast;
	AVISTREAMINFO	streamInfo;

	DubSource();
	virtual ~DubSource();

public:
	virtual bool init();

	virtual VDPosition getLength() {
		return lSampleLast - lSampleFirst;
	}

	virtual VDPosition getStart() {
		return lSampleFirst;
	}

	virtual VDPosition getEnd() {
		return lSampleLast;
	}

	virtual const VDFraction getRate() {
		return VDFraction(streamInfo.dwRate, streamInfo.dwScale);
	}

	virtual const AVISTREAMINFO& getStreamInfo() {
		return streamInfo;
	}

	virtual int read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead);
	virtual int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead) = 0;

	void *getFormat() const { return format; }
	int getFormatLen() const { return format_len; }

	virtual bool isStreaming();


	virtual void streamBegin( bool fRealTime);
	virtual void streamEnd();

	virtual void setDecodeErrorMode(ErrorMode mode) {}
	virtual bool isDecodeErrorModeSupported(ErrorMode mode) { return false; }

	LONG msToSamples(LONG lMs) const {
		return (LONG)(((__int64)lMs * streamInfo.dwRate + (__int64)500 * streamInfo.dwScale) / ((__int64)1000 * streamInfo.dwScale));
	}
	LONG samplesToMs(LONG lSamples) const {
		return (LONG)(
				(((__int64)lSamples * streamInfo.dwScale) * 1000 + streamInfo.dwRate/2) / streamInfo.dwRate
			);
	}

	// This is more accurate than AVIStreamSampleToSample(), which does a conversion to
	// milliseconds and back.

	static LONG samplesToSamples(const AVISTREAMINFO *dest, const AVISTREAMINFO *source, LONG lSamples) {
		__int64 divisor = (__int64)source->dwRate * dest->dwScale;

		return (LONG)((((__int64)lSamples * source->dwScale) * dest->dwRate + divisor/2)
				/ divisor);
	}


	LONG samplesToSamples(const DubSource *source, LONG lSamples) const {
		return samplesToSamples(&streamInfo, &source->streamInfo, lSamples);
	}
};

#endif