//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODCLASSINTERFACES_H__
#define __MODCLASSINTERFACES_H__

//#include "AviReadHandler.h"
#include "ModList.h"
#include "ModTextSource.h"
#include "ModTextUtil.h"
#include <ogg/ogg.h>
#include <vector>

// some things borrowed from COM code :p
// (a bit adapted to our needs)

typedef unsigned long InterfaceID;

class ExposeInterface {
public:
	virtual bool QueryInterface(InterfaceID ID, void** destination) { return false; };
};


enum {
	IID_Unknown,
	IID_StreamIndex,
	//IID_AVIHandler,
	IID_CommentList,
	IID_SubtitleHandler,
	IID_MP3Source,
	IID_FramedSource,
	IID_OggStream,
	IID_MatroskaStream,
	IID_OverlappingOutput
};

class IStreamIndex {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_StreamIndex; };
	virtual DWORD		GetStreamType(void) = 0;
	virtual long		GetStreamIndex(void) = 0;
};

//#pragma message(__TODO__ "Is this still used ?!?!")
/*class IAVIHandler {
public:
	static InterfaceID		GetInterfaceID(void) { return IID_AVIHandler; };
	virtual IAVIReadHandler	*GetReadHandler(void) = 0;
	virtual IAVIReadStream	*GetReadStream(void) = 0;
};*/

class ModStreamComment;
class ModChapterList;

class ICommentList {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_CommentList; };
	virtual void		GetCommentList(ModList<ModStreamComment>& list) = 0;
	virtual void		GetChapterList(ModChapterList& list) = 0;
};

class ISubtitleHandler {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_SubtitleHandler; };
	virtual bool		IsASCII(void) = 0;
	virtual bool		HasBOM(void) = 0;
	virtual eTextType	GetOutputType(void) = 0;
	virtual ModSubtitleProvider	*GetSubtitleProvider(void) = 0;
};

class IMP3Source {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_MP3Source; };
	virtual bool		UseVBRalaNandub(void) = 0;
	virtual bool		IsVBR(void) = 0;
	virtual bool		FixVBR(bool bFix) = 0;
};

class IFramedSource {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_FramedSource; };
	virtual bool		UseFrameMode(void) = 0;
};

class IOggStream {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_OggStream; };
	virtual uint32		GetOggStreamType(void) = 0;
	virtual uint16		GetNbHeaderPackets(void) = 0;
	virtual ogg_packet	*GetHeaderPacket(uint16 index) = 0;
};

class IMatroskaStream {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_MatroskaStream; };
	virtual std::string	GetCodecID(void) = 0;
	virtual std::vector<uint8>	GetCodecPrivate(void) = 0;
	virtual void *GetKaxTrackEntry(void) = 0;
};

class IOverlappingOutput {
public:
	static InterfaceID	GetInterfaceID(void) { return IID_OverlappingOutput; };
	virtual BOOL		WriteData(LONG dwIndexFlags, LPVOID lpBuffer, LONG cbBuffer, LONG lSampleStart, LONG lSampleCount) = 0;
};

#endif // __MODCLASSINTERFACES_H__