//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODCOMPACTLIST_H__
#define __MODCOMPACTLIST_H__

#include "ModList.h"

// Same as ModList except we try to save memory by 'packaging' the data
// This version use less memory and need less time to process, but is far
// less flexible than ModList. Use only with big (thousands of elements)
// lists with small data.

// ModCompactListNode definition

template<class _N> class ModCompactList;

template<class _N>
class ModCompactListNode : public ModListNode < ModCompactListNode<_N> > {
	friend ModCompactList<_N>;
protected:
	unsigned char m_Count;
	_N            m_Array[256];
public:
	ModCompactListNode<_N>(void);
	_N *GetElementAt(unsigned char pos);
	unsigned char GetCount(void);
};


template<class _N>
ModCompactListNode<_N>::ModCompactListNode<_N>(void) {
	m_Count = 0;
}

template<class _N>
_N *ModCompactListNode<_N>::GetElementAt(unsigned char pos) {
	return (m_Array+pos);
}

template<class _N>
unsigned char ModCompactListNode<_N>::GetCount(void) {
	return m_Count;
}


// ModCompactList definition


template<class _N>
class ModCompactList {
protected:
	ModList < ModCompactListNode<_N> > m_List;
	size_t m_Count;

public:
	ModCompactList<_N>(void);
	_N *AddLast(void);
	size_t GetCount(void);
	void Clear(void);
	ModCompactListNode<_N> *GetFirstCompact(void);
	ModCompactListNode<_N> *GetLastCompact(void);
	_N *GetFirst(void);
	_N *GetLast(void);
	bool RemoveLast(void);

	class iterator {
	protected:
		ModCompactList<_N> *m_List;
		ModCompactListNode<_N> *m_Node;
		unsigned char m_Index;

	public:
		iterator(void) {
			m_List = NULL;
			m_Node = NULL;
			m_Index = 0;
		};
		iterator(ModCompactList<_N>& list) {
			m_List = &list;
			m_Node = list.GetFirstCompact();
			m_Index = 0;
		};
		iterator(const iterator& src) throw() {
			m_List = src.m_List;
			m_Node = src.m_Node;
			m_Index = src.m_Index;
		};

		void begin(ModCompactList<_N>& list) {
			m_List = &list;
			m_Node = list.GetFirstCompact();
			m_Index = 0;
		};
		void end(ModCompactList<_N>& list) {
			m_List = &list;
			if(m_Node = list.GetLastCompact())
				m_Index = m_Node->GetCount();
			else
				m_Index = 0;
		};
		_N *previous(void) {
			if(m_Index > 0)
				return (m_Node->GetElementAt(m_Index-1));
			else {
				ModCompactListNode<_N> *previous = m_Node->GetPrevious();
				if(previous)
					return previous->GetElementAt(previous->GetCount());
				else
					return NULL;
			}
		};
		_N *next(void) {
			if(m_Index < m_Node->GetCount())
				return m_Node->GetElementAt(m_Index+1);
			else {
				ModCompactListNode<_N> *next = m_Node->GetNext();
				if(next)
					return (next->GetElementAt(0));
				else
					return NULL;
			}
		};
		iterator& operator=(const iterator& src) throw() {
			m_List = src.m_List;
			m_Node = src.m_Node;
			m_Index = src.m_Index;
			return *this;
		};
		bool operator!() const throw()
			{ return (m_Node==NULL); };
		//_N *operator->() const throw()
		//	{ return (m_Node->m_Array+m_Index); };
		operator bool() const throw()
			{ return (m_Node!=NULL); };
		_N *operator *() const throw()
			{ return m_Node->GetElementAt(m_Index); };
		iterator& operator++() throw() {
			if(m_Index < m_Node->GetCount())
				m_Index++;
			else {
				m_Index = 0;
				m_Node = m_Node->GetNext();
			}
			return *this;
		};
		void operator++(int) throw()
			{ ++*this; };
		iterator& operator--() throw() {
			if(m_Index > 0)
				m_Index--;
			else {
				m_Node = m_Node->GetPrevious();
				if(m_Node)
					m_Index = m_Node->GetCount();
			}
			return *this;
		};
		void operator--(int) throw()
			{ --*this; };
	};
};


template<class _N>
ModCompactList<_N>::ModCompactList<_N>(void) {
	m_Count = 0;
}

template<class _N>
size_t ModCompactList<_N>::GetCount(void) {
	return m_Count;
}

template<class _N>
_N *ModCompactList<_N>::AddLast(void) {
	ModCompactListNode<_N> *last = m_List.GetLast();
	if(last && (last->m_Count<255)) {
		last->m_Count++;
		m_Count++;
		return (last->m_Array+last->m_Count);
	} else {
		ModCompactListNode<_N> *node = m_List.AddLast();
		m_Count++;
		return node->m_Array;
	}
}

template<class _N>
ModCompactListNode<_N> *ModCompactList<_N>::GetFirstCompact(void) {
	return m_List.GetFirst();
}

template<class _N>
ModCompactListNode<_N> *ModCompactList<_N>::GetLastCompact(void) {
	return m_List.GetLast();
}

template<class _N>
_N *ModCompactList<_N>::GetFirst(void) {
	return m_List.GetFirst();
}

template<class _N>
_N *ModCompactList<_N>::GetLast(void) {
	ModCompactListNode<_N> *last = m_List.GetLast();
	if(!last)
		return NULL;
	return (last->m_Array+last->m_Count);
}

template<class _N>
bool ModCompactList<_N>::RemoveLast(void) {
	ModCompactListNode<_N> *last = m_List.GetLast();
	if(!last)
		return false;
	if(last->m_Count) {
		last->m_Count--;
		return true;
	} else {
		ModCompactListNode<_N> *previous = last->GetPrevious();
		m_List.DeleteNode(last);
		return !!previous;
	}
}

template<class _N>
void ModCompactList<_N>::Clear(void) {
	m_List.DeleteContents(true);
	m_List.Clear();
	m_Count = 0;
}

#endif // __MODCOMPACTLIST_H__