//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODDUBSTREAM_H__
#define __MODDUBSTREAM_H__

#include <vd2/system/vdtypes.h>
#include <vd2/system/atomic.h>
#include <vd2/system/VDString.h>
#include <vd2/system/VDRingBuffer.h>
#include <AVIPipe.h>
#include "DubInfo.h"
#include "DubUtils.h"
#include "AudioFilterSystem.h"
#include "ModList.h"
#include "ModStream.h"
#include "ModTextUtil.h"
#include "ModClassInterfaces.h"


class AudioStream;

class ModAudioPipe {
protected:
	void	*m_Buffer;
	long	m_Buffer_Bytes;
	sint64	m_Buffer_Start;
	sint64	m_Buffer_Samples;
	int		m_Buffer_Handle;

public:
	ModAudioPipe(void);
	virtual ~ModAudioPipe(void);

	VDSignal&	getReadSignal(void);
	VDSignal&	getWriteSignal(void);

	void	Init(int iSpace);
	int		getSpace(void);
	int		getLevel(void);
	bool	full(void);
	void*	BeginWrite(sint32 iRequested, sint32& iActual, sint32& iHandle);
	void	EndWrite(sint32 iActualBytes, sint32 iActualSamples, sint32 iActualRealSamples, sint32 iHandle, sint32 iStartSample=0);
	void	StoreBuffer(void *buffer, long iBytes, sint64 iStart, sint64 iSamples, int iHandle);
	void	*RetrieveBuffer(long& iBytes, sint64& iStart, sint64& iSamples, int& iHandle);
	void	Read(sint32 iRealSamples);
	void	Shutdown(void);

	sint32				m_iSpace;
	VDAtomicInt		m_iLevel;
	bool				m_bCBRPipe;
	VDAudioPipeline		m_CBRPipe;
	AVIPipe				m_VBRPipe;
	sint64				m_SamplesRead;
};


class ModDubStream : public ModListNode<ModDubStream> {
public:
	AudioSource		*m_Source;
	bool			m_bAccepted;
	bool			m_bInputFile;
	uint32			m_Serial;
	ModList<ModStreamComment>	m_Comments;
	ModChapterList	m_SourceChapters;	// Original (unmodified) chapters
	ModChapterList	m_Chapters;			// Modified chapters (according to subsets and clipping information)
	VDFraction		m_SamplesPerBlock;	// When coming from AVI/WAV
	bool			m_bIsNandubMP3VBR;	// Well I think you understand this one :P
	bool			m_bSeemsVBR;
	bool			m_bHandleVBR;		/* Whether or not we would like the source to treat VBR data */
	uint16			m_GroupFrames;		/* Group frames option */
	AudioStream		*m_AudioStream;
	AudioStream		*m_AudioTimingStream;
	AudioStream		*m_AudioStatusStream;
	AudioStreamL3Corrector	*m_AudioCorrector;
	std::vector<AudioStream *>	m_AudioStreams;
	vdautoptr<VDAudioFilterGraph>	m_AudioFilterGraph;
	BOOL			m_bEndReached;
	DubAudioOptions	m_Opts;		// Options
	DubAudioStreamInfo	m_aInfo;
	VDFormatStruct<WAVEFORMATEX> m_AudioCompressionFormat;
	//DWORD			m_ACompressionFormatSize;
	sint32			m_InitialSamplesLag;
	sint32			m_OffsetSamples;
	//sint64			m_StartUS;
	//sint64			m_SpillStreamPoint;
	//sint64			m_SpillStreamOk;
	//sint64			m_StreamPoint;
	//sint64			m_Sample;
	std::vector<char>	m_Frame;
	long				m_FrameSize;
	IOverlappingOutput	*m_OverlappingOutput;
	bool				m_bShortcutSubtitles;
	ModSubtitleProvider	m_SubtitleProvider;
	bool				m_bText;
	// 1.5.4 merging
	ModAudioPipe		m_AudioPipe;
	//VDSignal			m_sigAudioPipeRead;
	//VDSignal			m_sigAudioPipeWrite;
	//volatile bool		m_bAudioPipeFinalized;
	//volatile bool		m_bAudioPipeFinalizeAcked;

	ModDubStream(void);
	virtual ~ModDubStream(void);
	void ChainAudioStream(AudioStream *audioStream);
	void RemoveAudioStreams(void);
	ModDubStream& operator =(ModStream& stream);

	void EndReached(void);
};

#endif // __MODDUBSTREAM_H__