//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODTEXTSOUCE_H__
#define __MODTEXTSOUCE_H__

#include <vd2/system/File64.h>
#include <vd2/system/VDString.h>
#include <vector>

// ===================================================================
// How to identify Unicode encoded files, or the Byte-Order Mask.
//
// The byte 0x00 means the end of the file for a text file
// And you won't find a lot of plain text file having 0xEF, 0xFE or 0xFF
//
// Unicode file with a BOM (Byte-Order Mask) start with the Unicode
// character U+FEFF (UTF-16) / U+0000FEFF (UTF-32).
// Unicode is preferrably big-endian, but Intel machines are little endian
// and so here is what you will find in the file header (*cough* i.e. the
// first bytes for a text file ;)) for each case :
//	0xEF 0xBB 0xBF 0x??	-> UTF-8
//	0xFF 0xFE 0x?? 0x?? -> UTF-16/UCS-2, Little Endian
//	0xFE 0xFF 0x?? 0x?? -> UTF-16/UCS-2, Big Endian
//	0xFF 0xFE 0x00 0x00	-> UTF-32/UCS-4, Little endian
//	0x00 0x00 0xFE 0xFF	-> UTF-32/UCS-4, Big Endian
//
// If there is no BOM there are still ways to guess the format :
//	- if there are 0x00 bytes then it is most likely to be a UTF-16/32 file
//	- if there are 0x00 0x00 bytes then it is an UTF-32 file
//	- search for carriage returns may help :)
//	- the position of the 0x00 bytes tell us the byte ordering
// ===================================================================

enum eTextType {
	kTextType_Unknown = 0,
	kTextType_ASCII,
	kTextType_UTF8,
	kTextType_UTF16LE,
	kTextType_UTF16BE,
	kTextType_UTF32LE,
	kTextType_UTF32BE
};

bool GetBOMInformation(eTextType textType, void * & BOM, uint8& BOMLength);

class ModTextSource {
protected:
	VDStringW		m_FileName;
	VDFile			m_VDFile;
	eTextType		m_TextType;
	bool			m_bHasBOM;
	union {
		uint8			m_U8[5];
		uint16			m_U16[2];
		uint32			m_U32;
		wchar_t			m_WCHAR;
	};
	VDStringW		m_TextBuffer;
	const wchar_t	*m_TextBufferPos;
	std::vector<uint8>	m_Buffer;
	uint8				*m_pBufferReadPos;
	uint32				m_iBufferLevel;

	sint32	readData(void *pBuffer, sint32 iBytes);
	virtual bool getLine(const wchar_t * & buffer, VDStringW& line);

public:
	ModTextSource(void);
	virtual ~ModTextSource(void);
	virtual bool Open(VDStringW fileName);
	virtual void Close(void);
	virtual bool getLine(VDStringW& line);
	virtual bool eof(void);
	virtual eTextType GetTextType(void);
	virtual bool HasBOM(void);
};

#endif // __MODTEXTSOUCE_H__