//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODTEXTUTIL_H__
#define __MODTEXTUTIL_H__

#include <vd2/system/Fraction.h>
#include "FrameSubset.h"
#include "ModTextSource.h"
#include "ModList.h"

class ModTextUtil {
public:
	ModTextUtil(eTextType inputType=kTextType_ASCII, eTextType outputType=kTextType_ASCII, bool bWriteBOM=false, bool bFormatAsSubtiDS=false);
	virtual ~ModTextUtil(void);

	static void Shorten(VDStringW& text);
	static void Trim(VDStringW& text);
	static bool IsBlank(VDStringW& text);

	void SetInputType(eTextType type);
	void SetOutputType(eTextType type, bool bPrefixBOM, bool bAsSubtiDS);

	VDStringW GetText(void *pBuffer, long iBufferLen);
	int SetText(VDStringW text, void *pBuffer, long iBufferLen, long *iBytesWritten);

protected:
	eTextType		m_TextType;
	eTextType		m_OutputInType;
	bool			m_bPrefixBOMInOutput;
	bool			m_bFormatAsSubtiDS;
};

//////////////////////////////////////////////////////////////////////

class SubtitleFrame : public ModListNode<SubtitleFrame> {
public:
	VDStringW m_Subtitle;
	sint32 m_StartMS;
	sint32 m_DurationMS;

	SubtitleFrame(void) {
		m_StartMS = 0;
		m_DurationMS = 0;
	};
};

class ModSubtitleProvider {
protected:
	std::list<SubtitleFrame>			m_SourceSubtitles;
	std::list<SubtitleFrame>			m_AdjustedSubtitles;
	std::list<SubtitleFrame>::iterator	m_SubtitleIterator;
	bool								m_bIteratorReady;
	bool								m_bCollisions;

public:
	ModSubtitleProvider(void);
	virtual ~ModSubtitleProvider(void);

	ModSubtitleProvider& operator =(ModSubtitleProvider& subProvider);

	void Clear();

	std::list<SubtitleFrame>&	GetSourceSubtitles(void);
	std::list<SubtitleFrame>&	GetAdjustedSubtitles(void);
	bool					HasCollisions(void);

	void ApplySubsets(long lStartMS, long lEndMS, FrameSubset *subsets, VDFraction frameRate);

	bool AddSubtitle(VDStringW subtitle, sint32 startMS, sint32 durationMS);
	bool GetSubtitle_Linear(VDStringW& subtitle, sint32 startMS, sint32& durationMS);
	void ResetIterator(void);
	bool GetNextSubtitle(VDStringW& subtitle, sint32& startMS, sint32& durationMS);
	bool IsEnd(void);
};

#endif // __MODTEXTUTIL_H__