//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MODTREECONTROL_H__
#define __MODTREECONTROL_H__

#include "commctrl.h"
#include <vector>
#include <list>
#include <map>
#include <string>
#include <tchar.h>


typedef std::basic_string<TCHAR>	tcstring;

class ModTreeControl;
class ModTreeControl_RadioButton;

//////////////////////////////////////////////////////////////////////

class ModTreeControl_Control {
friend class ModTreeControl;
friend class ModTreeControl_RadioButton;
public:
	ModTreeControl_Control(ModTreeControl& treeControl);
	virtual ~ModTreeControl_Control(void);
	HTREEITEM		GetItem(void);
	virtual bool	Expand(bool bRecursive);
	virtual bool	Collapse(bool bRecursive);
	bool			BranchEnabled(void);
	virtual bool	SubBranchEnabled(void);
	bool			Enabled(void);
	void			Enable(void);
	void			Disable(void);
	virtual void	IconHit(void);
	virtual void	LabelHit(void);
	virtual void	SpaceKeyHit(void);
	bool			SetLabel(LPCTSTR pszToolTip);
	LPCTSTR			GetLabel(void);
	void			SetToolTip(LPCTSTR pszToolTip);
	LPCTSTR			GetToolTip(void);
	virtual tcstring	GetInfoTip(void);
	virtual COLORREF	GetTextColor(void);

protected:
	ModTreeControl&						m_TreeControl;
	ModTreeControl_Control				*m_Parent;
	std::list<ModTreeControl_Control *>	m_Children;
	std::list<ModTreeControl_RadioButton *>	m_RadioButtons;
	HTREEITEM							m_Item;
	bool								m_bEnabled;
	tcstring							m_Label;
	tcstring							m_ToolTip;

	void			SetItem(HTREEITEM item);
	void			SetParent(ModTreeControl_Control *parent);
	void			AddChild(ModTreeControl_Control& child);
	void			AddRadioButton(ModTreeControl_RadioButton& radioButton);
	void			CheckRadioButton(ModTreeControl_RadioButton& radioButton);
	virtual bool	RefreshString(void);
	void			RefreshState(void);
	virtual void	RefreshStateImage(void);
	virtual void	ItemHit(void);
};

//////////////////////////////////////////////////////////////////////

class ModTreeControl_CheckBox : public ModTreeControl_Control {
friend class ModTreeControl;
public:
	ModTreeControl_CheckBox(ModTreeControl& treeControl);
	bool	SubBranchEnabled(void);
	void	IconHit(void);
	void	SpaceKeyHit(void);
	bool	Checked(void);
	void	Check(bool bCheck);

protected:
	bool	m_bChecked;
	
	void	RefreshStateImage(void);
};

//////////////////////////////////////////////////////////////////////

class ModTreeControl_RadioButton : public ModTreeControl_Control {
friend class ModTreeControl;
public:
	ModTreeControl_RadioButton(ModTreeControl& treeControl);
	bool	SubBranchEnabled(void);
	void	IconHit(void);
	void	SpaceKeyHit(void);
	bool	Checked(void);
	void	Check(bool bCheck);

protected:
	bool	m_bChecked;
	
	void	RefreshStateImage(void);
};

//////////////////////////////////////////////////////////////////////

class ModTreeControl_EditBox : public ModTreeControl_Control {
friend class ModTreeControl;
public:
	ModTreeControl_EditBox(ModTreeControl& treeControl);
	bool	SetValue(LPCTSTR pszValue);
	virtual LPCTSTR	GetValue(void);
	void	SetWidth(int nWidth);
	int		GetWidth(void);

protected:
	tcstring	m_Value;
	int			m_Width;
	HWND		m_hWnd;

	bool	RefreshString(void);
	void	RefreshStateImage(void);
	void	ItemHit(void);
	virtual HWND	StartEdit(HWND hParent, RECT& rPos);
	virtual void	EndEdit(void);
};

//////////////////////////////////////////////////////////////////////

class ModTreeControl_ComboBox : public ModTreeControl_EditBox {
friend class ModTreeControl;
public:
	ModTreeControl_ComboBox(ModTreeControl& treeControl);
	LPCTSTR	GetValue(void);
	void	AddValue(LPCTSTR pszValue, bool bSelected);
	void	SelectValue(int nIndex);
	int		GetSelected(void);
	LPCTSTR	GetSelectedValue(void);
	void	SetShownValues(int nNumber);

protected:
	int		m_SelectedValue;
	std::vector<tcstring>	m_ValueList;
	int		m_ShownValues;

	void	RefreshStateImage(void);
	HWND	StartEdit(HWND hParent, RECT& rPos);
	void	EndEdit(void);
};

//////////////////////////////////////////////////////////////////////

class ModTreeControl_FileSelect : public ModTreeControl_EditBox {
friend class ModTreeControl;
public:
	ModTreeControl_FileSelect(ModTreeControl& treeControl);
	void SetFilters(LPCTSTR szFilters);

protected:
	LPCTSTR	m_Filters;

	void	RefreshStateImage(void);
	HWND	StartEdit(HWND hParent, RECT& rPos);
	void	EndEdit(void);
};

//////////////////////////////////////////////////////////////////////

ATOM RegisterModTreeControl(void);

class ModTreeControl : public ModTreeControl_Control {
friend ModTreeControl_Control;
friend ModTreeControl_CheckBox;
friend ModTreeControl_RadioButton;
friend ModTreeControl_EditBox;
friend ModTreeControl_ComboBox;
friend ModTreeControl_FileSelect;
public:
	// ctor/dtor
	ModTreeControl(HWND hWnd);
	virtual ~ModTreeControl(void);

	// static functions
	static LRESULT CALLBACK	StaticWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam);
	static ModTreeControl	*Create(HWND hwndParent, int x, int y, int cx, int cy, UINT id);
	static bool SetItemLabel(HWND hwndTreeView, HTREEITEM hItem, LPCTSTR pszLabel);
	static bool SetItemData(HWND hwndTreeView, HTREEITEM hItem, LPARAM itemData);
	static LPARAM GetItemData(HWND hwndTreeView, HTREEITEM hItem);
	static bool	SetStateImages(HWND hwndTreeView, HTREEITEM hItem, int iImage, int iSelectedImage);

	ModTreeControl_Control	*AddGroup(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszToolTip);
	ModTreeControl_CheckBox	*AddCheckBox(ModTreeControl_Control *parent, LPCTSTR pszLabel, bool bChecked, LPCTSTR pszToolTip);
	ModTreeControl_RadioButton	*AddRadioButton(ModTreeControl_Control *parent, LPCTSTR pszLabel, bool bChecked, LPCTSTR pszToolTip);
	ModTreeControl_EditBox	*AddEditBox(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszValue, LPCTSTR pszToolTip);
	ModTreeControl_ComboBox	*AddComboBox(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszToolTip);
	ModTreeControl_FileSelect	*AddFileSelect(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszValue, LPCTSTR pszToolTip);

	bool		Expand(bool bRecursive);
	bool		Collapse(bool bRecursive);

	void		Flush(void);

protected:
	HWND	m_hWnd;
	HWND	m_hInputWnd;
	ModTreeControl_EditBox	*m_InputEdit;
	SIZE	m_ScrollPos;
	HWND	m_hTreeList;
	SIZE	m_ClientSize;
	SIZE	m_TreeListSize;
	SIZE	m_TreeListActualSize;
	HTREEITEM	m_LargestItem;
	int		m_VScrollBarWidth;
	int		m_HScrollBarHeight;
	std::map<std::string, int>			m_StateImageList;
	COLORREF	m_BKColor;
	COLORREF	m_SelectedColor;
	HTREEITEM	m_hSelectedItem;

	bool	SetStateImages(ModTreeControl_Control *control, std::string sImage, std::string sSelectedImage);
	bool	SetItemLabel(ModTreeControl_Control *control, tcstring label);

	LRESULT	WndProc(UINT msg, WPARAM wParam, LPARAM lParam);
	bool	Init(void);
	bool	Expand(HTREEITEM item);
	bool	Collapse(HTREEITEM item);
	bool	ExpandCollapse(HTREEITEM item);
	void	EditValue(ModTreeControl_EditBox& editBox);
	void	ClearInputControl(void);
	void	UpdateScrollBars(void);
	void	SetTreeListSize(int nWidth, int nHeight, bool bOnlyShowOff=false);
	void	RefreshNeededSize(void);
};

ModTreeControl *GetModTreeControl(HWND hwnd);

#endif // __MODTREECONTROL_H__