//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <windows.h>
#include <commdlg.h>
#include <vd2/system/Error.h>
#include <vd2/system/filesys.h>
#include "oshelper.h"
#include "InputFileMatroska.h"
#include "VideoSource.h"
#include "StreamSourceMatroska.h"
#include "MatroskaReadHandler.h"
#include "prefs.h"


extern HWND	g_hWnd;
extern const wchar_t fileFiltersAppendMKV[]=
		L"Matroska segment (*.mkv)\0"	L"*.mkv\0"
		L"All files (*.*)\0"			L"*.*\0"
		;

static bool fileExists(const char *fn) {
	DWORD dwAttrib = GetFileAttributes(fn);

	return dwAttrib != -1 && !(dwAttrib & FILE_ATTRIBUTE_DIRECTORY);
}

class VDInputDriverMKV : public vdrefcounted<IVDInputDriver> {
public:
	const wchar_t *GetSignatureName() { return L"Matroska input driver (internal)"; }

	int GetDefaultPriority() {
		return 0;
	}

	uint32 GetFlags() { return kF_Video | kF_Audio | kF_Text; }

	const wchar_t *GetFilenamePattern() {
		return L"Matroska file (*.mkv)\0*.mkv\0";
	}

	bool DetectByFilename(const wchar_t *pszFilename) {
		return false;
	}

	int DetectBySignature(const void *pHeader, sint32 nHeaderSize, const void *pFooter, sint32 nFooterSize, sint64 nFileSize) {
		if (nHeaderSize >= 4) {
			if (*(long *)pHeader==0xA3DF451A)
				return 1;
		}

		return -1;
	}

	InputFile *CreateInputFile(uint32 flags) {
		InputFileMatroska *pf = new_nothrow InputFileMatroska;

		if (!pf)
			throw MyMemoryError();

		if (flags & kOF_Quiet)
			pf->setAutomated(true);

		if (flags & kOF_AutoSegmentScan)
			pf->EnableSegmentAutoscan();

		return pf;
	}
};

extern IVDInputDriver *VDCreateInputDriverMKV() { return new VDInputDriverMKV; }

//////////////////////////////////////////////////////////////////////

char InputFileMatroska::szME[] = "Matroska File import filter";

InputFileMatroska::InputFileMatroska(void)
: InputFileAVI() {
	pAVIFile = NULL;
}

InputFileMatroska::~InputFileMatroska(void) {
}

void InputFileMatroska::Init(const wchar_t *szFile) {
	AddFilename(szFile);

	if(!(pAVIFile = new MatroskaReadHandler(szFile)))
		throw MyMemoryError();

	if(!(videoSrc = new VideoSourceMatroska((MatroskaReadHandler *)pAVIFile)))
		throw MyMemoryError();

	if(!videoSrc->init())
		throw MyError("%s: problem opening video stream", szME);

	ModList<MatroskaReadNode>::iterator it(((MatroskaReadHandler *)pAVIFile)->GetStreamList());
	int index = 0;
	while(it) {
		MatroskaStream *aStream = (*it)->m_Segments.GetFirst()->m_MatroskaStream;
		if(MAIN_TYPE(aStream->m_StreamType) != SOURCE_VIDEO) {
			vdrefptr<AudioSource> newSource;
			if(MAIN_TYPE(aStream->m_StreamType) == SOURCE_TEXT)
				newSource = new TextSourceMatroska((MatroskaReadHandler *)pAVIFile, index);
			else
				newSource = new StreamSourceMatroska((MatroskaReadHandler *)pAVIFile, index);
			if(!newSource)
				throw MyMemoryError();
			if(!newSource->init()) {
				newSource = NULL;
				break;
			}
			m_Sources.push_back(newSource);
		}
		it++;
		index++;
	}

	if(fAutoscanSegments) {
	}
}

bool InputFileMatroska::Append(const wchar_t *szFile) {
	if(!szFile)
		return true;

	if(pAVIFile->AppendFile(szFile)) {
		if(videoSrc)
			((VideoSourceMatroska *)&*videoSrc)->Reinit();
		std::vector< vdrefptr<AudioSource> >::const_iterator it;
		for(it=m_Sources.begin() ; it!=m_Sources.end() ; it++)
				((StreamSourceMatroska *)&*(*it))->Reinit();

		AddFilename(szFile);

		return true;
	}

	return false;
}

VDStringW InputFileMatroska::writeReport(void) {
	VDStringW report = InputFile::writeReport();
	report += L"\r\n\r\n";
	report += ((MatroskaReadHandler *)pAVIFile)->writeReport();
	return report;
}