//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __MATROSKAOUTPUTFILESYSTEM_H__
#define __MATROSKAOUTPUTFILESYSTEM_H__

#include "ModList.h"
#include "DubOutput.h"

class MatroskaOutputFileSystem : public IVDDubberOutputSystem {
protected:
	VDStringW	mSegmentBaseName;
	VDStringW	mSegmentExt;
	int			mSegmentDigits;
	int			mCurrentSegment;
	int			mBufferSize;
	bool		mbAllowCaching;
	bool		m_bAcceptsVideo;
	bool					m_bLastSegment;
	AVIStreamHeader_fixed	mVideoStreamInfo;
	std::vector<char>		mVideoFormat;
	AVIStreamHeader_fixed	mAudioStreamInfo;
	std::vector<char>		mAudioFormat;
	ModVideoStream				*m_VideoStream;
	ModList<ModAVIDubStream>	m_Streams;
	ModList<ModStreamComment>	m_Comments;

public:
	MatroskaOutputFileSystem();
	~MatroskaOutputFileSystem();
	AVIOutput *CreateSegment();
	void SetCaching(bool bAllowOSCaching);
	void SetBuffer(int bufferSize);
	void SetFilename(const wchar_t *pszFilename);
	void SetFilenamePattern(const wchar_t *pszSegmentPrefix, const wchar_t *pszExt, int nMinimumDigits, int nCurrentSegment, bool bLastSegment);
	void CloseSegment(AVIOutput *pSegment, bool bLast);
	void SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat);
	void SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved);
	bool AcceptsVideo();
	void SetAcceptsVideo(bool bAccept);
	bool AcceptsAudio();
	void SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat);
	void AddStream(ModDubStream& stream, const AVIStreamHeader_fixed&, const void *pFormat, int cbFormat, bool bInterleaved);
	bool AcceptsText();
	bool IsRealTime();
};

#endif // __MATROSKAOUTPUTFILESYSTEM_H__