//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __STREAMSOURCEMATROSKA_H__
#define __STREAMSOURCEMATROSKA_H__


#include <vd2/system/thread.h>
#include <vd2/system/VDString.h>
#include <ogg/ogg.h>
#include "VideoSource.h"
#include "AudioSource.h"
#include "ModClassInterfaces.h"
#include "ModList.h"
#include "ModTextSource.h"
#include "MatroskaReadHandler.h"

//class MatroskaReadHandler;
//class MatroskaReadStream;

class VideoSourceMatroska : public VideoSourceAVI/*, public IOggStream*/, public IMatroskaStream {
protected:
	bool	m_bHasBFrames;
	std::vector<uint8>	m_ReadBuffer;
	void	*m_pReadBuffer;
	long	m_ReadBufferLen;
	long	m_ReadBufferNum;
	long	m_ReadBufferBytes;
	VDCriticalSection	m_CSQueue;
	std::vector<ModFrameIndex>	m_ListFrames;
	ModFrameIndex		*m_NextFrame;
	uint32	m_FrameRemaining;

	void AllocReadBuffer(long size);
	void *_StreamGetFrame(void *inputBuffer, long data_len, BOOL is_key, BOOL is_preroll, long frame_num);

public:
	VideoSourceMatroska(MatroskaReadHandler *readHandler);

	bool init();
	void Reinit();
	int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lBytesRead, LONG *lSamplesRead);
	char getFrameTypeChar(long lFrameNum);
	bool _isB(LONG lSample);
	bool hasBFrames(void);
	void *streamGetFrame(void *inputBuffer, long data_len, BOOL is_key, BOOL is_preroll, long frame_num);
	void *getFrame(LONG frameNum);
	//void streamSetDesiredFrame(long frame_num);
	//long streamGetNextRequiredFrame(BOOL *is_preroll);

	void	GetCommentList(ModList<ModStreamComment>& list);
	void	GetChapterList(ModChapterList& list);

	bool	QueryInterface(InterfaceID ID, void **destination);
	//uint32	GetOggStreamType(void);
	//uint16	GetNbHeaderPackets(void);
	//ogg_packet	*GetHeaderPacket(uint16 index);
	VDStringW	writeReport(void);
	std::string	GetCodecID(void);
	std::vector<uint8>	GetCodecPrivate(void);
	void *GetKaxTrackEntry(void);
};

class StreamSourceMatroska : public AudioSource, public IStreamIndex, public ICommentList, public IOggStream, public IMatroskaStream {
protected:
	MatroskaReadHandler	*m_ReadHandler;
	MatroskaReadStream	*m_ReadStream;
	long			m_StreamIndex;
	bool			m_bVBRHandling;

public:
	StreamSourceMatroska(MatroskaReadHandler *readHandler, long index);
	virtual ~StreamSourceMatroska(void);

	void extend(bool bVBRHandling, sint32 groupFrames);
	bool init();
	void Reinit();
	int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lpBytesRead, LONG *lpSamplesRead);
	bool isStreaming();
	void streamBegin(bool fRealTime);
	void streamEnd();
	sint64 sampleOffset(sint64 lSample);
	sint32 samplesLag(sint64 lStart);
	sint32 sourceOffsetMS(void);
	VDStringW	writeReport(void);

	bool	QueryInterface(InterfaceID ID, void **destination);
	DWORD	GetStreamType(void);
	long	GetStreamIndex(void);
	void	GetCommentList(ModList<ModStreamComment>& list);
	void	GetChapterList(ModChapterList& list);
	uint32	GetOggStreamType(void);
	uint16	GetNbHeaderPackets(void);
	ogg_packet	*GetHeaderPacket(uint16 index);
	std::string	GetCodecID(void);
	std::vector<uint8>	GetCodecPrivate(void);
	void *GetKaxTrackEntry(void);
};


class TextSourceMatroska : public StreamSourceMatroska, public ISubtitleHandler {
public:
	TextSourceMatroska(MatroskaReadHandler *readHandler, long index);

	bool	QueryInterface(InterfaceID ID, void **destination);
	bool	IsASCII(void);
	bool	HasBOM(void);
	eTextType GetOutputType(void);
	ModSubtitleProvider *GetSubtitleProvider(void);
};

#endif // __STREAMSOURCEMATROSKA_H__