//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include <tchar.h>
#include "misc.h"
#include "ModAudioSourceVorbis.h"
#include "OGMReadHandler.h"


bool ModAudioSourceVorbis::QueryInterface(InterfaceID ID, void **destination) {
	if(ID == ICommentList::GetInterfaceID()) {
		*destination = static_cast<ICommentList *>(this);
		return true;
	} else if(ID == IOggStream::GetInterfaceID()) {
		*destination = static_cast<IOggStream *>(this);
		return true;
	}
	return false;
}

ModAudioSourceVorbis::ModAudioSourceVorbis(VDStringW fileName)
: AudioSource() {
	m_ReadHandler = NULL;
	m_ReadStream = NULL;

	m_SourceType = SOURCE_OGM;
	clearSourceInfo();
	m_SourceName = fileName;
	m_SourceDesc = L"Ogg Vorbis audio stream";
}

ModAudioSourceVorbis::~ModAudioSourceVorbis(void) {
	if(m_ReadStream)
		delete m_ReadStream;
	if(m_ReadHandler)
		delete m_ReadHandler;
}


bool ModAudioSourceVorbis::init() {
	LONG format_len;

	if(m_SourceName.empty())
		return false;

	m_ReadHandler = new OGMReadHandler(m_SourceName.c_str());
	if(!m_ReadHandler)
		throw MyMemoryError();

	m_ReadStream = (OGMReadStream *)m_ReadHandler->GetStream(streamtypeAUDIO, 0);

	if(!m_ReadStream)
		return false;
	if(m_ReadStream->GetOggStreamType() != SOURCE_AUDIO_VORBIS)
		return false;

	if (m_ReadStream->Info(&streamInfo, sizeof streamInfo))
		return false;

	m_ReadStream->FormatSize(0, &format_len);

	if (!allocFormat(format_len)) return FALSE;

	if (m_ReadStream->ReadFormat(0, getFormat(), &format_len))
		return false;

	lSampleFirst = m_ReadStream->Start();
	lSampleLast = m_ReadStream->End();

	m_SourceType |= m_ReadStream->GetOggStreamType();

	return true;
}

bool ModAudioSourceVorbis::isStreaming() {
	return m_ReadStream->isStreaming();
}

void ModAudioSourceVorbis::streamBegin(bool fRealTime) {
	m_ReadStream->BeginStreaming(lSampleFirst, lSampleLast, fRealTime ? 1000 : 2000);
}

void ModAudioSourceVorbis::streamEnd() {
	m_ReadStream->EndStreaming();
}

sint64 ModAudioSourceVorbis::sampleOffset(sint64 lSample) {
	return m_ReadStream->sampleOffset(lSample);
}

sint32 ModAudioSourceVorbis::samplesLag(sint64 lStart) {
	return m_ReadStream->samplesLag(lStart);
}

sint32 ModAudioSourceVorbis::sourceOffsetMS(void) {
	return m_ReadStream->sourceOffsetMS();
}

int ModAudioSourceVorbis::_read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lpBytesRead, LONG *lpSamplesRead) {
	return m_ReadStream->Read(lStart, lCount, lpBuffer, cbBuffer, lpBytesRead, lpSamplesRead);
}

uint32 ModAudioSourceVorbis::GetOggStreamType(void) {
	return m_ReadStream->GetOggStreamType();
}

uint16 ModAudioSourceVorbis::GetNbHeaderPackets(void) {
	return m_ReadStream->GetNbHeaderPackets();
}

ogg_packet *ModAudioSourceVorbis::GetHeaderPacket(uint16 index) {
	return m_ReadStream->GetHeaderPacket(index);
}

void ModAudioSourceVorbis::GetCommentList(ModList<ModStreamComment>& list) {
	m_ReadStream->GetCommentList(list);
}

void ModAudioSourceVorbis::GetChapterList(ModChapterList& list) {
	m_ReadStream->GetChapterList(list);
}

VDStringW ModAudioSourceVorbis::writeReport(void) {
	VDStringW report = AudioSource::writeReport();
	report += L"\r\n";
	report += m_ReadHandler->writeReport();
	report += L"\r\n";
	report += m_ReadStream->writeReport();
	return report;
}