//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __OGMOUTPUTFILE_H__
#define __OGMOUTPUTFILE_H__

#include "AVIOutput.h"
#include "ModList.h"
#include "ModTextUtil.h"
#include <list>

class OGMPacketizer;
class OGMOutputFile;

class OGMVideoOutputStream : public AVIOutputStream {
protected:
	OGMOutputFile	*m_Parent;
	OGMPacketizer	*m_Packetizer;

public:
	OGMVideoOutputStream(OGMOutputFile *parent);
	virtual ~OGMVideoOutputStream(void);
	void setPacketizer(OGMPacketizer *packetizer);
	void write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples);
	void finalize(void);
	void setEndOfStream(void);
	void setStreamLag(sint32 samplesLag);
	OGMPacketizer *GetPacketizer(void);
};


class OGMOutputStream : public AVIOutputStream {
protected:
	OGMOutputFile	*m_Parent;
	OGMPacketizer	*m_Packetizer;

public:
	bool			m_bPureData;
	bool			m_bFlushData;
	bool			m_bWriteEOSOnEnd;

	OGMOutputStream(OGMOutputFile *parent);
	virtual ~OGMOutputStream(void);
	void setPacketizer(OGMPacketizer *packetizer);
	void write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples);
	void finalize(void);
	void setEndOfStream(void);
	void setStreamLag(sint32 samplesLag);
};


class OGMSubtitleOutputStream : public OGMOutputStream {
protected:
	ModTextUtil	m_TextUtil;

public:
	OGMSubtitleOutputStream(OGMOutputFile *parent, eTextType inputType);
	virtual ~OGMSubtitleOutputStream(void);
	void write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples);
};


class OGMOutputStreamNode : public ModListNode<OGMOutputStreamNode> {
public:
	sint32			m_Serial;
	ModDubStream	*m_Input;
	OGMOutputStream	*m_Output;
	ModList<ModStreamComment>	m_Comments;
	ModChapterList	m_Chapters;

	OGMOutputStreamNode(void) {
		m_Serial = -1;
		m_Input = NULL;
		m_Output = NULL;
	};

	virtual ~OGMOutputStreamNode(void) {
		if(m_Output)
			delete m_Output;
		m_Comments.DeleteContents(true);
		m_Comments.Clear();
		m_Chapters.clear();
	};
};

class FastWriteStream;
class OGMMuxerPage;

class OGMPacketizerNode : public ModListNode<OGMPacketizerNode> {
public:
	OGMPacketizer	*m_Packetizer;
	bool			m_bWriteEOSOnEnd;
	sint32			m_Serial;
	bool			m_bHasFewPackets;
	uint64			m_FirstPagePos;
	uint32			m_FirstPageLength;
	uint64			m_LastPagePos;
	uint32			m_LastPageLength;

	OGMPacketizerNode(void) {
		m_Packetizer = NULL;
		m_bWriteEOSOnEnd = true;
		m_Serial = -1;
		m_bHasFewPackets = false;
		m_FirstPagePos = 0;
		m_FirstPageLength = 0;
		m_LastPagePos = 0;
		m_LastPageLength = 0;
	};
};

class OGMOutputFile : public AVIOutput {
protected:
	static char szME[];
	ModVideoStream		*m_VideoStream;
	sint32				m_VideoSerial;
	ModList<OGMOutputStreamNode>	m_Outputs;
	ModList<OGMOutputStreamNode>::iterator	m_OutputIterator;
	ModList<OGMPacketizerNode>	m_PacketizerList;
	ModList<OGMPacketizerNode>::iterator	m_PacketizerIterator;
	VDStringW			m_FileName;
	FastWriteStream		*m_fastIO;
	VDFile				m_File;
	sint32				mBufferSize;
	__int64				m_i64FilePosition;
	bool				m_bCaching;
	bool				m_bEOSWritten;
	bool				m_bCreateVideo;

	void				_seekDirect(__int64 i64NewPos);
	bool				PendingHeaderPage(void);
	OGMMuxerPage		*GetHeaderPage(void);
	void				WriteEOS(void);
	bool				PendingMuxerPage(void);
	OGMMuxerPage		*GetMuxerPage(void);
	OGMMuxerPage		*GetFirstMuxerPage(void);
	void				WriteMuxerPage(OGMMuxerPage *muxerPage);
	OGMPacketizerNode	*GetPacketizerNode(sint32 serial);
	void				FlushPacketizers(void);
	void				ProcessHeaderPages(void);
	void				RewriteHeaderPages(void);
	void				RewriteLastPages(void);

public:
	OGMOutputFile(bool bCreateVideo=true);
	virtual ~OGMOutputFile(void);

	void disable_os_caching();
	void ProcessPendingMuxerPages(void);

	void	setBufferSize(sint32 iBufferSize);
	IVDMediaOutputStream *createVideoStream();
	IVDMediaOutputStream *createAudioStream();
	bool initOutputStreams(ModVideoStream& videoStream, ModList<ModDubStream>& streams);
	IVDMediaOutputStream *getOutputStream(uint32 serial);
	bool init(const wchar_t *szFile);
	void finalize(void);

	void writeIndexedChunk(FOURCC ckid, LONG dwIndexFlags, LPVOID lpBuffer, LONG cbBuffer);
};

#endif // __OGMOUTPUTFILE_H__