//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include "OGMOutputFileSystem.h"
#include "OGMOutputFile.h"


OGMOutputFileSystem::OGMOutputFileSystem()
	: mbAllowCaching(false)
	, mCurrentSegment(0)
{
	m_VideoStream = NULL;
	m_bAcceptsVideo = true;
	m_bLastSegment = false;
}

OGMOutputFileSystem::~OGMOutputFileSystem() {
	m_Streams.DeleteContents(true);
	m_Streams.Clear();
}

AVIOutput *OGMOutputFileSystem::CreateSegment() {
	vdautoptr<OGMOutputFile> pOutput(new OGMOutputFile(m_bAcceptsVideo));

	if(!mbAllowCaching)
		pOutput->disable_os_caching();

	VDStringW s(mSegmentBaseName);

	if (mSegmentDigits) {
		s += VDFastTextPrintfW(L".%0*d", mSegmentDigits, mCurrentSegment++);
		VDFastTextFree();
		s += mSegmentExt;

		//pOutput->setSegmentHintBlock(true, NULL, 1);
	}

	ModList<ModDubStream> tmpList;
	tmpList.DeleteContents(false);
	ModList<ModAVIDubStream>::iterator it(m_Streams);
	while(it) {
		tmpList.Append((*it)->m_Stream);
		it++;
	}
	pOutput->initOutputStreams(*m_VideoStream, tmpList);

	if (!mVideoFormat.empty()) {
		IVDMediaOutputStream *pVideoOut = pOutput->getVideoOutput();
		pVideoOut->setFormat(&mVideoFormat[0], mVideoFormat.size());
		pVideoOut->setStreamInfo(mVideoStreamInfo);
	}

	it.begin(m_Streams);
	ModAVIDubStream *curStream = NULL;
	while(it) {
		curStream = *it;
		IVDMediaOutputStream *streamOut = pOutput->getOutputStream(curStream->m_Stream->m_Serial);
		if(!streamOut)
			throw MyError("Unable to retrieve output stream %lu", curStream->m_Stream->m_Serial);
		if (!curStream->m_Format.empty()) {
			streamOut->setFormat(&curStream->m_Format[0], curStream->m_Format.size());
			streamOut->setStreamInfo(curStream->m_StreamInfo);
		}
		it++;
	}

	pOutput->setBufferSize(mBufferSize);

	pOutput->init(s.c_str());

	return pOutput.release();
}

void OGMOutputFileSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	/*AVIOutputFile *pFile = (AVIOutputFile *)pSegment;
	if (mSegmentDigits)
		pFile->setSegmentHintBlock(bLast & m_bLastSegment, NULL, 1);*/
	pSegment->finalize();
	delete pSegment;
}

void OGMOutputFileSystem::SetCaching(bool bAllowOSCaching) {
	mbAllowCaching = bAllowOSCaching;
}

void OGMOutputFileSystem::SetBuffer(int bufferSize) {
	mBufferSize = bufferSize;
}

void OGMOutputFileSystem::SetFilename(const wchar_t *pszFilename) {
	mSegmentBaseName	= pszFilename;
	mSegmentDigits		= 0;
}

void OGMOutputFileSystem::SetFilenamePattern(const wchar_t *pszFilename, const wchar_t *pszExt, int nMinimumDigits, int nCurrentSegment, bool bLastSegment) {
	mSegmentBaseName	= pszFilename;
	mSegmentExt			= pszExt;
	mSegmentDigits		= nMinimumDigits;
	mCurrentSegment		= nCurrentSegment;
	m_bLastSegment		= bLastSegment;
}

void OGMOutputFileSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void OGMOutputFileSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool OGMOutputFileSystem::AcceptsVideo() {
	return m_bAcceptsVideo;
}

void OGMOutputFileSystem::SetAcceptsVideo(bool bAccept) {
	m_bAcceptsVideo = bAccept;
}

bool OGMOutputFileSystem::AcceptsAudio() {
	return true;
}

void OGMOutputFileSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	m_VideoStream = &stream;
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void OGMOutputFileSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	ModAVIDubStream *newStream = m_Streams.AddLast();
	newStream->m_Stream = &stream;
	newStream->m_StreamInfo = asi;
	newStream->m_Format.resize(cbFormat);
	memcpy(&newStream->m_Format[0], pFormat, cbFormat);
}

bool OGMOutputFileSystem::AcceptsText() {
	return true;
}

bool OGMOutputFileSystem::IsRealTime() {
	return false;
}