//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __OGMPACKETIZER_H__
#define __OGMPACKETIZER_H__

#include <vd2/system/Fraction.h>
#include <ogg/ogg.h>
#include "OGMSpecs.h"
#include "ModList.h"
#include "ModStream.h"

// Maximum overhead for a Packet : 9 bytes
// for data Packet : 1 byte (header) + 8 bytes (samples count)
#define OGM_PACKET_MAXOVERHEAD	9

class OGMMuxerPage : public ModListNode<OGMMuxerPage> {
public:
	ogg_page	*m_Page;
	sint64		m_TimeUS;
	uint8		m_Priority;
	sint32		m_Serial;

	OGMMuxerPage(void) {
		m_Page = NULL;
		m_TimeUS = 0;
		m_Priority = 0;
		m_Serial = -1;
	};

	~OGMMuxerPage(void) {
		if(m_Page) {
			VDASSERT(m_Page->header && m_Page->body);
			free(m_Page->header);
			free(m_Page->body);
			delete m_Page;
		}
	};
};

class OGMPacketizer {
protected:
	sint32			m_Serial;
	void			*m_Header;
	uint16			m_HeaderLength;
	stream_header	*m_StreamHeader;
	bool			m_bStandardHeader;
	ModList<OGMMuxerPage>	m_PageList;
	ogg_stream_state	m_StreamState;
	VDFraction		m_SampleRate;
	bool			m_bEOS;
	bool			m_bInvalid;
	unsigned char	*m_Buffer;
	uint32			m_BufferSize;
	sint32			m_MaxFrameLength;
	ogg_int64_t		m_StartGranulepos;
	ogg_int64_t		m_LastGranulepos;
	sint64			m_Timestamp;
	bool			m_bGranuleposMeansEnd;
	ogg_packet		m_CachePacket;
	bool			m_bCachedPacket;
	uint64			m_NbPackets;
	sint16			m_HeaderPages;
	bool			m_bPriorityHeaders;
	bool			m_bPrioritySet;

	void			ResizeBuffer(void);
	void			_ProcessPacket(ogg_packet *packet, bool bFlush, bool bHeader);
	void			SetMuxerPagePriority(void);
	void			AddPage(ogg_page *page);
	ogg_page		*CopyOggPage(ogg_page *page);
	void			ComputeTimestamp(ogg_int64_t granulepos);

public:
	OGMPacketizer(sint32 serial, uint32 bufferSize, ogg_int64_t startGranulepos=1, bool bGranuleposMeansEnd=true, bool bPriorityHeaders=false);
	virtual ~OGMPacketizer(void);

	bool IsEOS(void)
		{ return m_bEOS; };
	bool IsInvalid(void)
		{ return m_bInvalid; };

	bool	CopyHeader(void *header, uint16 headerLen);
	bool	WriteHeader(void);
	bool	WriteComments(ModList<ModStreamComment>& comments, ModChapterList& chapters, bool bSourceIsOgg, bool bFlush);
	void	WriteEOS(void);
	uint32	GetBufferSize(void);

	void	ProcessFrame(void *frame, uint32 frameLength, ogg_int64_t frameSize, bool bPureData=true, bool bKeyFrame=false, bool bFlush=false);
	void	ProcessPacket(ogg_packet *packet, bool bFlush, bool bHeader);
	void	ProcessCachedPacket(void);
	void	ProcessPages(bool bFlush);
	void	ProcessHeaderPages(bool bFlush);

	void	AddLagGranulepos(sint64 granuleposDiff);
	void	AddLagTimeUS(sint64 timeUSDiff);

	OGMMuxerPage	*PendingMuxerPage(void);
	OGMMuxerPage	*GetMuxerPage(void);
	OGMMuxerPage	*PendingHeaderPage(void);
	OGMMuxerPage	*GetHeaderPage(void);
	sint64			GetFirstTimestamp(void);
	sint64			GetEndTimestamp(void);
	bool			RewriteHeaderPage(void *lpBuffer, uint32 pageLength);
	bool			RewriteLastPage(void *lpBuffer, uint32 pageLength, sint64 timestampEndUS);
};

#endif // __OGMPACKETIZER_H__