//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __OGMREADHANDLER_H__
#define __OGMREADHANDLER_H__

#include <vd2/system/VDString.h>
#include <vd2/system/File64.h>
#include <vd2/system/Fraction.h>
#include <ogg/ogg.h>
#include <vorbis/codec.h>
#include <vector>
#include "ModClassInterfaces.h"
#include "AVIReadHandler.h"
#include "OGMSpecs.h"
#include "ModList.h"
#include "ModCompactList.h"
#include <list>

#define OGM_FLAGS			VD64(0xC000000000000000)
#define OGM_GETFLAGS(x)		((x)&OGM_FLAGS)
#define OGM_FILEPOS(x)		((x)&VD64(0x3FFFFFFFFFFFFFFF))
#define OGM_FLAG_CONT		VD64(0x8000000000000000)
#define OGM_GETFLAG_CONT(x)	((x)&OGM_FLAG_CONT)
#define OGM_FLAG_RESETED	VD64(0x4000000000000000)
#define OGM_GETFLAG_RESETED(x)	((x)&OGM_FLAG_RESETED)

#pragma pack(push)
#pragma pack(4)
class OggPage {
public:
	// Beware packing & alignment
	// Fields have been ordered for a lower memory consumption
	// Bit 63 of m_FilePos will tell if this is a continued page or not
	// the 63 other bits should be enough for a file size :)
	uint64	m_FilePos;
	ogg_int64_t	m_Granulepos;
	// Bit 31 of m_NbBytes will tell if this is a key or not
	// the 31 other bits should be enough for a packet size :)
	uint32	*m_NbBytes;
	uint32	*m_NbSamples;
	uint16	m_Size;
	uint8	m_NbPackets;

	OggPage(void) {
		m_FilePos = 0;
		m_Size = 0;
		m_Granulepos = -1;
		m_NbPackets = 0;
		m_NbSamples = NULL;
		m_NbBytes = NULL;
	};

	~OggPage(void) {
		if(m_NbSamples)
			delete[] m_NbSamples;
		if(m_NbBytes)
			delete[] m_NbBytes;
	};

	void SetNbPackets(uint8 nbPackets) {
		if(m_NbPackets = nbPackets) {
			m_NbSamples = new uint32[nbPackets];
			m_NbBytes = new uint32[nbPackets];
			memset(m_NbSamples, 0, nbPackets*4);
			memset(m_NbBytes, 0, nbPackets*4);
		}
	};

	void ResetNbPackets(void) {
		if(m_NbPackets) {
			delete[] m_NbSamples;
			delete[] m_NbBytes;
		}
		m_NbPackets = 0;
		m_NbSamples = NULL;
		m_NbBytes = NULL;
	};

	void UpdateNbPackets(uint8 nbPackets) {
		// We shouldn't come here to often
		if(!nbPackets) {
			ResetNbPackets();
			return;
		}
		uint32 *new_NbSamples = new uint32[nbPackets];
		uint32 *new_NbBytes = new uint32[nbPackets];
		memset(new_NbSamples, 0, nbPackets*4);
		memset(new_NbBytes, 0, nbPackets*4);
		if(m_NbPackets <= nbPackets) {
			memcpy(new_NbSamples, m_NbSamples, m_NbPackets * 4);
			memcpy(new_NbBytes, m_NbBytes, m_NbPackets * 4);
			delete[] m_NbSamples;
			delete[] m_NbBytes;
		} else if(m_NbPackets) {
			memcpy(new_NbSamples, m_NbSamples, nbPackets * 4);
			memcpy(new_NbBytes, m_NbBytes, nbPackets * 4);
			delete[] m_NbSamples;
			delete[] m_NbBytes;
		}
		m_NbPackets = nbPackets;
		m_NbSamples = new_NbSamples;
		m_NbBytes = new_NbBytes;
	};
};
#pragma pack(pop)

class OggWeirdPage {
public:
	uint64	m_FilePos;
	uint64	m_TimePosMS;
	uint8	m_NbPacketsExpected;
	uint16	m_NbPacketsReceived;
	uint16	m_NbPacketsLostInAction;

	OggWeirdPage(void) {
		m_FilePos = 0;
		m_TimePosMS = 0;
		m_NbPacketsExpected = 0;
		m_NbPacketsReceived = 0;
		m_NbPacketsLostInAction = 0;
	};
};

class OggWeirdPacket {
public:
	uint64	m_FilePos;
	uint64	m_TimePosMS;
	uint8	m_PacketIndex;
	uint64	m_NbSamplesAnnounced;
	uint32	m_NbSamplesCorrected;

	OggWeirdPacket(void) {
		m_FilePos = 0;
		m_TimePosMS = 0;
		m_PacketIndex = 0;
		m_NbSamplesAnnounced = 0;
		m_NbSamplesCorrected = 0;
	};
};

class OggCorruptedData {
public:
	uint64	m_FilePos;
	uint32	m_Length;
	uint64	m_TimePosMS;

	OggCorruptedData(void) {
		m_FilePos = 0;
		m_Length = 0;
		m_TimePosMS = 0;
	};
};

class OggStream {
public:
	sint32				m_RefCount;
	sint32				m_Serial;
	uint32				m_StreamType;
	VDFraction			m_SampleRate;
	ogg_stream_state	m_StreamState;
	void				*m_Header;
	stream_header		*m_StreamHeader;
	vorbis_info			m_VorbisInfo;
	vorbis_comment		m_VorbisComment;
	std::vector<ogg_packet *> m_HeaderPacketList;
	bool				m_bHasComments;
	sint32				m_PreviousW;
	ModCompactList<OggPage>	m_PageList;
	uint16				m_PacketIndex;
	ModCompactList<OggWeirdPage>	m_WeirdPageList;
	ModCompactList<OggWeirdPacket>	m_WeirdPacketList;
	sint32				m_MaxPacketSize;
	sint64				m_StartOffsetMS;
	sint64				m_FirstGranulepos;
	sint64				m_FirstGranuleposPage;
	sint64				m_Bytes;
	sint64				m_NbPackets;
	sint64				m_NbSamples;
	bool				m_bKeyIndexBuilt;
	uint8				*m_KeyFrames;
	uint8				*m_KeyFramesEnd;
	sint64				m_TmpPageGranulepos_0;
	sint64				m_TmpPageGranulepos_1;
	uint16				m_CheckPage;
	sint16				m_PacketsExpected;
	uint16				m_SplittedPages;

	OggStream(void);
	~OggStream(void);
	void AddRef(void);
	void Release(void);
	void SetHeaderLen(uint16 size);
	void AddHeaderPacket(ogg_packet *packet);
	void AddWeirdPage(uint64 filePos, uint64 timePosMS, uint8 nbPacketsExpected, uint16 nbPacketsReceived);
	void SetLastWeirdPageLostPackets(uint16 nbPacketsLostInAction);
	void AddWeirdPacket(uint64 filePos, uint64 timePosMS, uint8 packetIndex, uint64 nbSamplesAnnounced, uint32 nbSamplesCorrected);
	VDStringW writeReport(void);
	void GetCommentList(ModList<ModStreamComment>& list);
	void GetChapterList(ModChapterList& list, uint32 offsetMS);
};


class OGMReadHandler;
class OGMReadStream;

class OggStreamSegment : public ModListNode<OggStreamSegment> {
public:
	OGMReadHandler		*m_ReadHandler;
	OggStream			*m_OggStream;
	uint8				m_FirstPacketIndex;
	sint64				m_Start;
	sint64				m_End;
	ogg_stream_state	m_StreamState;
	sint64				m_LastFilePos;
	ModCompactList<OggPage>::iterator	m_FirstDataPage;

	OggStreamSegment(void);
	~OggStreamSegment(void);
	void UpdateState(bool bVBRHandling=true);
};

class OGMReadNode : public ModListNode<OGMReadNode> {
public:
	sint64	m_Start;
	sint64	m_End;
	sint64	m_LastFilePos;
	sint64	m_NbSamples;
	sint64	m_Bytes;
	sint64	m_FirstKeyFrame;
	sint16	m_BlockAlign;
	sint32	m_MaxPacketSize;
	bool	m_bKeyFrameOnly;
	ModList<OggStreamSegment>	m_Segments;
	std::list<OGMReadStream *>	m_RegisteredReadStreams;

	OGMReadNode(void);
	~OGMReadNode(void);
	void AppendSegment(OggStreamSegment *segment);
	void UpdateState(bool bVBRHandling=true);
	void AddRef(void);
	void Release(void);
	void RegisterReadStream(OGMReadStream *readStream);
	void UnregisterReadStream(OGMReadStream *readStream);
	void ReinitReadStreams(void);
};

class OGMReadHandler : public IAVIReadHandler {
friend class OGMReadStream;
protected:
	sint32		m_RefCount;
	VDFile		m_File;
	VDStringW	m_FileName;
	ModList<OGMReadNode>	m_StreamList;
	ModList<OGMReadNode>::iterator	m_StreamIterator;
	OggStream	*m_CurrentStream;
	ModCompactList<OggCorruptedData>	m_CorruptedDataList;
	ogg_sync_state	m_SyncState;
	sint64		m_BytesRead;
	sint64		m_BytesSkipped;
	ogg_page	m_Page;
	ogg_packet	m_Packet;

	void		_construct(const wchar_t *pszFile);
	void		_updateStreamStates(void);
	void		_parseFile(void);
	int			_getNextPage(ogg_page *page);
	void		_addStream(ogg_page *page);
	OggStream	*_getStream(sint32 serial);
	void		_addPage(ogg_page *page);
	void		_addPacket(ogg_packet *packet);
	void		_destruct(void);
	void		_addCorruptedData(uint64 filePos, uint32 length);
	bool		_getCurrentTimeMS(bool bDiscardLastPage, uint64& nMillisecond);
	void		_setCorruptedFlags(void);
	long		ReadData(void *buffer, long nBytes);
	void		SeekFile(sint64 filePos);

public:
	OGMReadHandler(const wchar_t *pszFile);
	virtual ~OGMReadHandler(void);

	void AddRef();
	void Release();
	IAVIReadStream *GetStream(DWORD fccType, LONG lParam);
	IAVIReadStream *GetStream(LONG lParam);
	ModList<OGMReadNode>& GetStreamList(void);
	long GetStreamIndexByType(OGMReadNode *stream);
	void EnableFastIO(bool);
	bool isOptimizedForRealtime();
	bool isStreaming();
	bool isIndexFabricated();
	bool AppendFile(const wchar_t *pszFile);
	bool getSegmentHint(const char **ppszPath);
	void GetCommentList(ModList<ModStreamComment>& list);
	void GetChapterList(ModChapterList& list);
	VDStringW writeReport(void);
	VDStringW GetFileName(void);
};


class OGMReadStream : public IAVIReadStream {
protected:
	OGMReadHandler	*m_Parent;
	OggStream		*m_Stream;
	OGMReadNode		*m_ReadNode;
	OggStreamSegment	*m_CurrentStreamSegment;
	void			*m_Format;
	sint32			m_FormatSize;
	sint16			m_BlockAlign;
	ogg_sync_state	m_SyncState;
	ModCompactList<OggPage>::iterator	m_PageIterator;
	ModCompactList<OggPage>::iterator	m_TmpIterator;
	bool			m_bVBRHandling;
	sint64			m_PageByte;
	OggPage			*m_Page;
	sint64			m_Length;
	sint64			m_LastFilePos;
	uint8			*m_KeyFrames;
	uint8			*m_KeyFramesEnd;
	bool			m_bIsVideo;
	bool			m_bCanTrimPacket;
	ogg_page		m_OggPage;
	ogg_packet		m_OggPacket;

	sint32			_getNextPage(ogg_page *page, uint16 size);
	long			FindStreamSegment(long lStart);
	inline void		SynchSegmentInformation(void);
	void			SeekFirstStreamSegment(void);
	void			SeekLastStreamSegment(void);
	bool			SeekPreviousSegment(void);
	bool			SeekNextSegment(void);
	HRESULT			ReadCBR(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples);

public:
	OGMReadStream(OGMReadNode *readNode);
	virtual ~OGMReadStream();

	void SetVBRHandling(bool bVBRHandling);
	void Reinit(void);

	HRESULT BeginStreaming(long lStart, long lEnd, long lRate);
	HRESULT EndStreaming();
	HRESULT Info(AVISTREAMINFO *pasi, long lSize);
	bool IsKeyFrame(long lFrame);
	VDFraction getDAR(void);
	HRESULT Read(long lStart, long lSamples, void *lpBuffer, long cbBuffer, long *plBytes, long *plSamples);
	long Start();
	long End();
	long PrevKeyFrame(long lFrame);
	long NextKeyFrame(long lFrame);
	long NearestKeyFrame(long lFrame);
	HRESULT FormatSize(long lFrame, long *plSize);
	HRESULT ReadFormat(long lFrame, void *pFormat, long *plSize);
	bool isStreaming();
	bool isKeyframeOnly();

	long GetStreamIndexByType(void);

	bool getVBRInfo(double& bitrate_mean, double& bitrate_stddev, double& maxdev);
	void fixVBRAudio(bool bFix, double averageBitrate);
	sint64 sampleOffset(sint64 lFrame);
	sint32 samplesLag(sint64 lFrame);
	sint32 sourceOffsetMS(void);

	void GetCommentList(ModList<ModStreamComment>& list);
	void GetChapterList(ModChapterList& list);
	uint32 GetOggStreamType(void);
	uint16 GetNbHeaderPackets(void);
	ogg_packet *GetHeaderPacket(uint16 index);
	VDStringW writeReport(void);
};

#endif // __OGMREADHANDLER_H__