//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#ifndef __STREAMSOURCEOGM_H__
#define __STREAMSOURCEOGM_H__

#include <vd2/system/VDString.h>
#include <ogg/ogg.h>
#include "VideoSource.h"
#include "AudioSource.h"
#include "ModClassInterfaces.h"
#include "ModList.h"
#include "ModTextSource.h"

class OGMReadHandler;
class OGMReadStream;

class VideoSourceOGM : public VideoSourceAVI, public IOggStream {
public:
	VideoSourceOGM(OGMReadHandler *readHandler);

	void	GetCommentList(ModList<ModStreamComment>& list);
	void	GetChapterList(ModChapterList& list);

	bool	QueryInterface(InterfaceID ID, void **destination);
	uint32	GetOggStreamType(void);
	uint16	GetNbHeaderPackets(void);
	ogg_packet	*GetHeaderPacket(uint16 index);
	VDStringW	writeReport(void);
};

class StreamSourceOGM : public AudioSource, public IStreamIndex, public ICommentList, public IOggStream {
protected:
	OGMReadHandler	*m_ReadHandler;
	OGMReadStream	*m_ReadStream;
	long			m_StreamIndex;
	bool			m_bVBRHandling;

public:
	StreamSourceOGM(OGMReadHandler *readHandler, long index);
	virtual ~StreamSourceOGM(void);

	void extend(bool bVBRHandling, sint32 groupFrames);
	bool init();
	void Reinit();
	int _read(LONG lStart, LONG lCount, LPVOID lpBuffer, LONG cbBuffer, LONG *lpBytesRead, LONG *lpSamplesRead);
	bool isStreaming();
	void streamBegin(bool fRealTime);
	void streamEnd();
	sint64 sampleOffset(sint64 lSample);
	sint32 samplesLag(sint64 lStart);
	sint32 sourceOffsetMS(void);
	VDStringW	writeReport(void);

	bool	QueryInterface(InterfaceID ID, void **destination);
	DWORD	GetStreamType(void);
	long	GetStreamIndex(void);
	void	GetCommentList(ModList<ModStreamComment>& list);
	void	GetChapterList(ModChapterList& list);
	uint32	GetOggStreamType(void);
	uint16	GetNbHeaderPackets(void);
	ogg_packet	*GetHeaderPacket(uint16 index);
};


class TextSourceOGM : public StreamSourceOGM, public ISubtitleHandler {
public:
	TextSourceOGM(OGMReadHandler *readHandler, long index);

	bool	QueryInterface(InterfaceID ID, void **destination);
	bool	IsASCII(void);
	bool	HasBOM(void);
	eTextType GetOutputType(void);
	ModSubtitleProvider *GetSubtitleProvider(void);
};

#endif // __STREAMSOURCEOGM_H__