//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include "IDGenerator.h"
#include <time.h>
#include <limits.h>

#ifndef U32_MAX
#define U32_MAX	((uint32)0xFFFFFFFF)
#endif


IDGenerator::IDGenerator(bool bRandom) {
	m_bRandom = bRandom;
	m_bScaleRandom = ((uint64)RAND_MAX < (uint64)U32_MAX);
	srand(time(NULL));
}

IDGenerator::~IDGenerator(void) {
	m_IDList.clear();
}

bool IDGenerator::IsRegistered(uint32 id) {
	std::list<uint32>::const_iterator it;
	for(it=m_IDList.begin() ; it!=m_IDList.end() ; it++) {
		if((*it) == id)
			return true;
	}
	return false;
}

bool IDGenerator::RegisterID(uint32 id) {
	if(IsRegistered(id))
		return false;
	m_IDList.push_back(id);
	return true;
}

uint32 IDGenerator::GenerateID(bool bRegister) {
	uint32 newID = 0;
	uint32 count = U32_MAX;

	if(m_bRandom) {
		do {
			if(m_bScaleRandom)
				newID = ((uint64)rand() * U32_MAX) / RAND_MAX;
			else
				newID = rand();
			count--;
		} while(IsRegistered(newID) && count);
	} else {
		newID = 0;
		while(IsRegistered(newID) && count) {
			newID++;
			count--;
		}
	}

	if(!count)
		throw MyError("All the available IDs seems to be registered");
	if(IsRegistered(newID))
		throw MyError("The chosen ID (%lu) has already been registered :/", newID);
	if(bRegister && !RegisterID(newID))
		throw MyError("The chosen ID (%lu) hasn't been registered but couldn't be either :(", newID);

	return newID;
}
