//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "ISO-639.h"


#define DEFINE_LANGUAGE_CODE(englishName, ISO639_2_Code, ISO639_1_Code)	\
	{englishName, ISO639_2_Code, ISO639_1_Code},

ISO639_Language s_LanguageList[] = {
#include "ISO-639-List.h"
	{NULL,	NULL,	NULL}
};


bool ISO639_IsValidIndex_1(sint32 index) {
	if(!ISO639_IsValidIndex_2(index))
		return false;
	return !!s_LanguageList[index].m_ISO639_1_Code;
}

bool ISO639_IsValidIndex_2(sint32 index) {
	if(index < 0)
		return false;
	if(index >= sizeof(s_LanguageList)/sizeof(s_LanguageList[0])-1)
		return false;
	return true;
}

sint32 ISO639_GetRootIndex(sint32 index) {
	if(!ISO639_IsValidIndex_2(index))
		return -1;
	do {
		if(s_LanguageList[index].m_EnglishName)
			return index;
	} while(--index);
	return -1;
}

wchar_t *ISO639_GetEnglishName(sint32 index) {
	if(!ISO639_IsValidIndex_2(index))
		return NULL;
	do {
		if(s_LanguageList[index].m_EnglishName)
			return s_LanguageList[index].m_EnglishName;
	} while(--index);
	return NULL;
}

wchar_t *ISO639_GetEnglishName_1(const wchar_t *ISO639_1_Code) {
	sint32 index = ISO639_GetIndex_1(ISO639_1_Code);
	if(index >= 0) {
		do {
			if(s_LanguageList[index].m_EnglishName)
				return s_LanguageList[index].m_EnglishName;
		} while(--index);
	}
	return NULL;
}

wchar_t *ISO639_GetEnglishName_2(const wchar_t *ISO639_2_Code) {
	sint32 index = ISO639_GetIndex_2(ISO639_2_Code);
	if(index >= 0) {
		do {
			if(s_LanguageList[index].m_EnglishName)
				return s_LanguageList[index].m_EnglishName;
		} while(--index);
	}
	return NULL;
}

sint32 ISO639_GetIndex(const wchar_t *ISO639_EnglishName) {
	if(!ISO639_EnglishName)
		return NULL;
	sint32 i;
	for(i=0;
		s_LanguageList[i].m_EnglishName || s_LanguageList[i].m_ISO639_1_Code || s_LanguageList[i].m_ISO639_2_Code;
		i++)
	{
		if(!s_LanguageList[i].m_EnglishName)
			continue;
		if(!wcsicmp(s_LanguageList[i].m_EnglishName, ISO639_EnglishName))
			return i;
	}
	return -1;
}

sint32 ISO639_GetIndex_1(const wchar_t *ISO639_1_Code) {
	if(!ISO639_1_Code)
		return NULL;
	sint32 i;
	for(i=0;
		s_LanguageList[i].m_EnglishName || s_LanguageList[i].m_ISO639_1_Code || s_LanguageList[i].m_ISO639_2_Code;
		i++)
	{
		if(!s_LanguageList[i].m_ISO639_1_Code)
			continue;
		if(!wcsicmp(s_LanguageList[i].m_ISO639_1_Code, ISO639_1_Code))
			return i;
	}
	return -1;
}

sint32 ISO639_GetIndex_2(const wchar_t *ISO639_2_Code) {
	if(!ISO639_2_Code)
		return NULL;
	sint32 i;
	for(i=0;
		s_LanguageList[i].m_EnglishName || s_LanguageList[i].m_ISO639_1_Code || s_LanguageList[i].m_ISO639_2_Code;
		i++)
	{
		if(!s_LanguageList[i].m_ISO639_2_Code)
			continue;
		if(!wcsicmp(s_LanguageList[i].m_ISO639_2_Code, ISO639_2_Code))
			return i;
	}
	return -1;
}

sint32 ISO639_GetNearestIndex(const wchar_t *ISO639_EnglishName) {
	if(!ISO639_EnglishName)
		return -1;

	sint32 index = ISO639_GetIndex(ISO639_EnglishName);
	if(index >= 0)
		return index;

	index = -1;
	sint32 length = wcslen(ISO639_EnglishName);
	sint32 i;
	sint32 count;
	wchar_t *pos;
	for(i=0;
		s_LanguageList[i].m_EnglishName || s_LanguageList[i].m_ISO639_1_Code || s_LanguageList[i].m_ISO639_2_Code;
		i++)
	{
		pos = s_LanguageList[i].m_EnglishName;
		if(!pos)
			continue;
		count = wcslen(pos) - length;
		while(count>=0) {
			if(!wcsnicmp(pos, ISO639_EnglishName, length)) {
				if(pos == s_LanguageList[i].m_EnglishName)
					return i;
				if(index < 0)
					index = i;
			}
			count--;
			pos++;
		}
	}

	return index;
}
