//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "DubSource.h"
#include "DubInfo.h"
#include "ModDubStream.h"
#include "prefs.h"

extern DubOptions g_dubOpts;


ModDubStream::ModDubStream(void)
: m_SamplesPerBlock(1) {
	m_bAccepted = true;
	m_Source = NULL;
	m_bInputFile = false;
	m_Serial = 0;
	m_bIsNandubMP3VBR = false;
	m_bSeemsVBR = false;
	m_bHandleVBR = true;
	m_GroupFrames = g_VDMPrefs.m_GroupFrames;
	m_AudioStream = NULL;
	m_AudioTimingStream = NULL;
	m_AudioStatusStream = NULL;
	m_AudioCorrector = NULL;
	m_bEndReached = FALSE;
	memset(&m_Opts, 0, sizeof(m_Opts));
	m_Opts.volume = 0;
	m_Opts.preload = g_VDMPrefs.m_DefaultPreload;
	m_Opts.interval = 1;	// interleave every unit (frame by default)
	m_Opts.new_rate = 0;
	m_Opts.offset = 0;
	m_Opts.is_ms = false;	// interleave every ? frames (and not every ? ms)
	m_Opts.enabled = true;
	m_Opts.fStartAudio = g_dubOpts.audio.fStartAudio;
	m_Opts.fEndAudio = g_dubOpts.audio.fEndAudio;
	m_Opts.integral_rate = false;
	m_Opts.fHighQuality = false;
	m_Opts.newPrecision = DubAudioOptions::P_NOCHANGE;
	m_Opts.newChannels = DubAudioOptions::C_NOCHANGE;
	m_Opts.mode = DubAudioOptions::M_NONE;
	m_Opts.bUseAudioFilterGraph = false;
	m_AudioFilterGraph = NULL;
	//m_ACompressionFormat = NULL;
	//m_ACompressionFormatSize = 0;
	memset(&m_aInfo, 0, sizeof(m_aInfo));
	m_InitialSamplesLag = 0;
	m_OffsetSamples = 0;
	//m_StartUS = 0;
	//m_SpillStreamPoint = 0;
	//m_SpillStreamOk = 0;
	//m_StreamPoint = 0;
	//m_Sample = 0;
	m_FrameSize = 0;
	// 1.5.4 merging
	//m_bAudioPipeFinalized = false;
	//m_bAudioPipeFinalizeAcked = false;
	m_OverlappingOutput = NULL;
	m_bShortcutSubtitles = false;
	m_bText = false;
}

ModDubStream::~ModDubStream(void) {
	m_Comments.DeleteContents(true);
	m_Comments.Clear();
	m_Chapters.clear();
	m_SourceChapters.clear();
	//if(m_ACompressionFormat)
	//	freemem(m_ACompressionFormat);
	//if(m_Source)
	//	delete m_Source;
	//if(m_AudioStream)
	//	delete m_AudioStream;
	if(m_AudioFilterGraph)
		//delete m_AudioFilterGraph;
		m_AudioFilterGraph = NULL;
	RemoveAudioStreams();
	m_Frame.clear();
}

void ModDubStream::ChainAudioStream(AudioStream *audioStream) {
	m_AudioStream = audioStream;
	m_AudioStreams.push_back(audioStream);
}

void ModDubStream::RemoveAudioStreams(void) {
	std::vector<AudioStream *>::const_iterator it(m_AudioStreams.begin()), itEnd(m_AudioStreams.end());
	for(; it!=itEnd; ++it)
		delete *it;
	m_AudioStreams.clear();
	//if(m_AudioCorrector)
	//	delete m_AudioCorrector;
	m_AudioCorrector = NULL;
	m_AudioStream = NULL;
}

ModDubStream& ModDubStream::operator =(ModStream &stream) {
	m_Comments = stream.m_Comments;
	m_SourceChapters = stream.m_Chapters;
	m_Chapters = stream.m_Chapters;
	m_Source = stream.m_Source;
	m_bInputFile = stream.m_bInputFile;
	//m_Serial = 0;
	//m_SamplesPerBlock = 1;
	//m_bIsNandubMP3VBR = false;
	//m_AudioStream = NULL;
	//m_bEndReached = false;
	m_Opts = stream.m_Opts;
	if(m_Opts.bUseAudioFilterGraph)
		m_AudioFilterGraph = new VDAudioFilterGraph(stream.m_AudioFilterGraph);
	else
		m_AudioFilterGraph = NULL;
	if(stream.m_ACompressionFormat)
		m_AudioCompressionFormat.assign(stream.m_ACompressionFormat, stream.m_ACompressionFormatSize);
	else
		m_AudioCompressionFormat.clear();
	//m_ACompressionFormat = stream.m_ACompressionFormat;
	//m_ACompressionFormatSize = stream.m_ACompressionFormatSize;
	//memset(&m_aInfo, 0, sizeof(m_aInfo));
	//m_StartOffsetMS = 0;
	return *this;
}

void ModDubStream::EndReached(void) {
	m_bEndReached = true;
	m_AudioPipe.m_CBRPipe.CloseInput();
	m_AudioPipe.m_VBRPipe.finalize();
}

//////////////////////////////////////////////////////////////////////

ModAudioPipe::ModAudioPipe(void)
: m_VBRPipe(100, 4096)
{
	m_bCBRPipe = false;
	m_iSpace = 0;
	m_iLevel = 0;
	m_SamplesRead = 0;
	m_Buffer = 0;
}

ModAudioPipe::~ModAudioPipe(void) {
}

VDSignal& ModAudioPipe::getReadSignal(void) {
	if(m_bCBRPipe)
		return m_CBRPipe.getReadSignal();

	return m_VBRPipe.getReadSignal();
}

VDSignal& ModAudioPipe::getWriteSignal(void) {
	if(m_bCBRPipe)
		return m_CBRPipe.getWriteSignal();

	return m_VBRPipe.getWriteSignal();
}

void ModAudioPipe::Init(int iSpace) {
	m_iSpace = iSpace;
	if(m_bCBRPipe)
		m_CBRPipe.Init(iSpace);
}

int ModAudioPipe::getSpace(void) {
	if(m_bCBRPipe)
		return m_CBRPipe.getSpace();

	return (m_iSpace - m_iLevel);
}

int ModAudioPipe::getLevel(void) {
	if(m_bCBRPipe)
		return m_CBRPipe.getLevel();

	return m_iLevel;
}

bool ModAudioPipe::full(void) {
	if(m_bCBRPipe)
		return m_CBRPipe.full();

	if(m_iLevel >= m_iSpace)
		return true;
	return m_VBRPipe.full();
}

void *ModAudioPipe::BeginWrite(sint32 iRequested, sint32& iActual, sint32& iHandle) {
	if(m_bCBRPipe)
		return m_CBRPipe.BeginWrite(iRequested, iActual);

	iActual = iRequested;
	return m_VBRPipe.getWriteBuffer(iRequested, &iHandle);
}

void ModAudioPipe::EndWrite(sint32 iActualBytes, sint32 iActualSamples, sint32 iActualRealSamples, sint32 iHandle, sint32 iStartSample) {
	if(m_bCBRPipe) {
		m_CBRPipe.EndWrite(iActualBytes);
	} else {
		m_VBRPipe.postBuffer(iActualBytes, iActualSamples, iActualRealSamples, iStartSample, 0, 0, iHandle, false);
		m_iLevel += iActualRealSamples;
	}
}

void ModAudioPipe::Read(sint32 iRealSamples) {
	m_iLevel -= iRealSamples;

	//VDASSERT(m_iLevel >= 0);
}

void ModAudioPipe::Shutdown(void) {
	if(m_bCBRPipe)
		m_CBRPipe.Shutdown();
	//else
	//	m_VBRPipe.finalizeAck();
}

void ModAudioPipe::StoreBuffer(void *buffer, long iBytes, sint64 iStart, sint64 iSamples, int iHandle) {
	VDASSERT(!m_Buffer);

	m_Buffer = buffer;
	m_Buffer_Bytes = iBytes;
	m_Buffer_Start = iStart;
	m_Buffer_Samples = iSamples;
	m_Buffer_Handle = iHandle;
}

void *ModAudioPipe::RetrieveBuffer(long& iBytes, sint64& iStart, sint64& iSamples, int& iHandle) {
	if(m_Buffer) {
		iBytes = m_Buffer_Bytes;
		iStart = m_Buffer_Start;
		iSamples = m_Buffer_Samples;
		iHandle = m_Buffer_Handle;
	}
	void *result = m_Buffer;
	m_Buffer = 0;
	return result;
}
