//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "ModFileSelectControl.h"
#include "resource.h"
#include <windows.h>
#include <commdlg.h>
#include <utility>
#include <windowsx.h>

extern HINSTANCE	g_hInst;
extern const char	g_szModFileSelectControlName[] = "vdModFileSelectControl";

//////////////////////////////////////////////////////////////////////

ATOM RegisterModFileSelectControl() {
	WNDCLASS wc;

	wc.style		= 0;
	wc.lpfnWndProc	= ModFileSelectControl::StaticWndProc;
	wc.cbClsExtra	= 0;
	wc.cbWndExtra	= sizeof(ModFileSelectControl *);
	wc.hInstance	= g_hInst;
	wc.hIcon		= NULL;
	wc.hCursor		= NULL;
	wc.hbrBackground= (HBRUSH)/*(COLOR_BACKGROUND+1)*/0;
	wc.lpszMenuName	= NULL;
	wc.lpszClassName= g_szModFileSelectControlName;

	return RegisterClass(&wc);
}

ModFileSelectControl *GetModFileSelectControl(HWND hwnd) {
	return reinterpret_cast<ModFileSelectControl *>(GetWindowLong(hwnd, 0));
}

ModFileSelectControl *ModFileSelectControl::Create(HWND hwndParent, int x, int y, int cx, int cy, UINT id) {
	HWND hwnd = CreateWindowEx(WS_EX_TRANSPARENT, g_szModFileSelectControlName, "Mew", WS_VISIBLE | WS_CHILD, x, y, cx, cy, hwndParent, (HMENU)id, g_hInst, NULL);

	if(hwnd)
		return GetModFileSelectControl(hwnd);

	return NULL;
}

ModFileSelectControl::ModFileSelectControl(HWND hwnd)
: m_hWnd(hwnd)
, m_hEditBox(NULL)
, m_hButton(NULL)
{
	m_Filters = _T("All files (*.*)\0")	_T("*.*\0");
}

ModFileSelectControl::~ModFileSelectControl(void) {
}

HWND ModFileSelectControl::GetWindow(void) {
	return m_hWnd;
}

HFONT ModFileSelectControl::GetFont(void) {
	if(m_hEditBox)
		return (HFONT)SendMessage(m_hEditBox, WM_GETFONT, 0, 0);
	return (HFONT)NULL;
}

HFONT ModFileSelectControl::SetFont(HFONT hFont, BOOL bRedraw) {
	if(m_hButton)
		SendMessage(m_hButton, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(bRedraw, 0));
	if(m_hEditBox)
		return (HFONT)SendMessage(m_hEditBox, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(bRedraw, 0));
	return (HFONT)NULL;
}

LPCTSTR ModFileSelectControl::GetFileName(void) {
	if(m_hEditBox) {
		TCHAR szFile[MAX_PATH];
		if(GetWindowText(m_hEditBox, szFile, sizeof(szFile) / sizeof(szFile[0])) >= 0)
			m_FileName = szFile;
	}
	return m_FileName.c_str();
}

void ModFileSelectControl::SetFileName(LPCTSTR szFileName) {
	m_FileName = szFileName;
	if(m_hEditBox)
		SetWindowText(m_hEditBox, szFileName);
}

void ModFileSelectControl::SetFilters(LPCTSTR szFilters) {
	m_Filters = szFilters;
}

LRESULT CALLBACK ModFileSelectControl::StaticWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	ModFileSelectControl *pThis = (ModFileSelectControl *)GetWindowLong(hwnd, 0);

	switch(msg) {
	case WM_NCCREATE:
		if (!(pThis = new ModFileSelectControl(hwnd)))
			return FALSE;
		SetWindowLong(hwnd, 0, (DWORD)pThis);
		break;

	case WM_NCDESTROY:
		delete pThis;
		return DefWindowProc(hwnd, msg, wParam, lParam);
	}

	return pThis->WndProc(msg, wParam, lParam);
}

LRESULT ModFileSelectControl::WndProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
	case WM_CREATE:
		if(!Init())
			return -1;
		break;
	//case WM_SIZE:
	//	break;
	case WM_DESTROY:
		{
			TCHAR szFile[MAX_PATH];
			if(GetWindowText(m_hEditBox, szFile, sizeof(szFile) / sizeof(szFile[0])) >= 0)
				m_FileName = szFile;
		}
		break;
	case WM_COMMAND:
		if( (HIWORD(wParam)==BN_CLICKED) && ((HWND)lParam==m_hButton) ) {
			OPENFILENAME ofn;
			ofn.lStructSize			= OPENFILENAME_SIZE_VERSION_400;
			ofn.hwndOwner			= (HWND)m_hWnd;
			ofn.lpstrCustomFilter	= NULL;
			ofn.nFilterIndex		= 0;
			ofn.lpstrFileTitle		= NULL;
			ofn.lpstrInitialDir		= NULL;
			ofn.Flags				= /*OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | */OFN_ENABLESIZING | OFN_EXPLORER/* | OFN_OVERWRITEPROMPT*/;

			TCHAR szFile[MAX_PATH];

			ofn.lpstrFilter		= m_Filters;
			ofn.lpstrFile		= szFile;
			ofn.nMaxFile		= sizeof(szFile) / sizeof(szFile[0]);
			ofn.lpstrTitle		= _T("Select File");
			ofn.lpstrDefExt		= NULL;

			GetWindowText(m_hEditBox, szFile, ofn.nMaxFile);

			if(GetOpenFileName(&ofn)) {
				m_FileName = szFile;
				SetWindowText(m_hEditBox, szFile);
			}
			break;
		}
	default:
		return DefWindowProc(m_hWnd, msg, wParam, lParam);
	}
	return FALSE;
}

bool ModFileSelectControl::Init(void) {
	RECT cRect;
	if(!GetClientRect(m_hWnd, &cRect)) {
		cRect.left = 0;
		cRect.top = 0;
		cRect.right = CW_USEDEFAULT;
		cRect.bottom = CW_USEDEFAULT;
	}

	DWORD editBoxStyle = WS_VISIBLE | WS_CHILD | WS_TABSTOP
		| ES_AUTOHSCROLL;
	DWORD buttonStyle = WS_VISIBLE | WS_CHILD | WS_TABSTOP
		| BS_PUSHBUTTON /*| BS_BITMAP */| BS_TEXT | BS_CENTER | BS_VCENTER;

	int iButtonBorderX = 2*GetSystemMetrics(SM_CXEDGE);
	int iButtonBorderY = 2*GetSystemMetrics(SM_CYEDGE);
	int iButtonSizeX = 16 + iButtonBorderX;
	int iButtonSizeY = 16 + iButtonBorderY;
	int iButtonMarginY = (cRect.bottom-cRect.top-iButtonSizeY) / 2;

	m_hEditBox = CreateWindowEx(WS_EX_CLIENTEDGE, _T("EDIT"), _T("")
		, editBoxStyle
		, 0, 0, cRect.right-cRect.left - /*5 -*/ iButtonSizeX, cRect.bottom-cRect.top
		, m_hWnd, (HMENU)1, g_hInst, NULL);

	m_hButton = CreateWindowEx(0, _T("BUTTON"), _T("...")
		, buttonStyle
		, cRect.right - iButtonSizeX, iButtonMarginY, iButtonSizeX -1, iButtonSizeY -1
		, m_hWnd, (HMENU)2, g_hInst, NULL);
/*
	if(m_hButton) {
		HBITMAP hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_FILESELECT));
		SendMessage(m_hButton, BM_SETIMAGE, (WPARAM)IMAGE_BITMAP, (LPARAM)hBitmap);
	}
*/
	return (m_hEditBox && m_hButton);
}