//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include "DubSource.h"
#include "DubInfo.h"
#include "ModStream.h"
#include "ModTextSource.h"
#include "prefs.h"

extern DubOptions g_dubOpts;


ModStream::ModStream(void) {
	Reset(false, false, true);
}

ModStream::~ModStream(void) {
	Reset(true, true, true);
}

void ModStream::Reset(bool bFreeSource, bool bFreeData, bool bResetValues) {
	if(bFreeData) {
		if(m_ACompressionFormat)
			freemem(m_ACompressionFormat);
		m_ACompressionFormat = NULL;
	}
	if(bFreeSource) {
		if(!m_bInputFile && m_Source)
			m_Source = NULL;
		m_Source = NULL;
	}
	if(bResetValues) {
		m_Comments.DeleteContents(true);
		m_Comments.Clear();
		m_Chapters.clear();

		m_bEnabled = true;
		m_bToDemux = false;
		m_bToPreview = false;
		m_Source = NULL;
		m_bInputFile = false;
		memset(&m_Opts, 0, sizeof(m_Opts));
		m_Opts.volume = 0;
		m_Opts.preload = g_VDMPrefs.m_DefaultPreload;
		m_Opts.interval = 1;	// interleave every unit (frame by default)
		m_Opts.new_rate = 0;
		m_Opts.offset = 0;
		m_Opts.is_ms = false;	// interleave every ? frames (and not every ? ms)
		m_Opts.enabled = true;
		m_Opts.fStartAudio = g_dubOpts.audio.fStartAudio;
		m_Opts.fEndAudio = g_dubOpts.audio.fEndAudio;
		m_Opts.integral_rate = false;
		m_Opts.fHighQuality = false;
		m_Opts.newPrecision = DubAudioOptions::P_NOCHANGE;
		m_Opts.newChannels = DubAudioOptions::C_NOCHANGE;
		m_Opts.mode = DubAudioOptions::M_NONE;
		m_Opts.bUseAudioFilterGraph = false;
		m_ACompressionFormat = NULL;
		m_ACompressionFormatSize = 0;
		m_StreamType = 0;
		m_StreamIndex = 0;
	}
}

ModVideoStream::ModVideoStream(void)
: m_DAR(0)
{
	m_Source = NULL;
	m_bDARScaleUp = true;
}

ModVideoStream::~ModVideoStream(void) {
	m_Comments.DeleteContents(true);
	m_Comments.Clear();
	m_Chapters.clear();
}

using namespace nsModStream;

ModStreamComment::ModStreamComment(void) {
	m_Type = cGeneric;
}

sint16 nsModStream::GetCommentIndex(eCommentType type) {
	if(type >= cLanguage)
		type = cLanguage;
	sint16 i=0;
	while( (s_CommentTypes[i].m_Type!=cVoid) || (s_CommentTypes[i].m_Name) || (s_CommentTypes[i].m_Tip) ) {
		if(s_CommentTypes[i].m_Type == type)
			return i;
		i++;
	}
	return -1;
}


wchar_t *nsModStream::GetCommentName(eCommentType type) {
	sint16 index = GetCommentIndex(type);
	if(index != -1)
		return s_CommentTypes[index].m_Name;
	return NULL;
}

wchar_t *nsModStream::GetCommentTip(eCommentType type) {
	sint16 index = GetCommentIndex(type);
	if(index != -1)
		return s_CommentTypes[index].m_Tip;
	return NULL;
}

bool nsModStream::CanUseComment(eCommentType type) {
	sint16 index = GetCommentIndex(type);
	if(index != -1)
		return s_CommentTypes[index].m_bCanUse;
	return false;
}

bool nsModStream::CanEditCommentTag(eCommentType type) {
	sint16 index = GetCommentIndex(type);
	if(index != -1)
		return s_CommentTypes[index].m_bCanEditTag;
	return true;
}

bool nsModStream::CanEditCommentValue(eCommentType type) {
	if(type > cLanguage)
		return false;
	sint16 index = GetCommentIndex(type);
	if(index != -1)
		return s_CommentTypes[index].m_bCanEditValue;
	return true;
}

ModStreamComment::ModStreamComment(ModStreamComment& otherComment) {
	*this = otherComment;
}

ModStreamComment& ModStreamComment::operator =(ModStreamComment& otherComment) {
	m_Type = otherComment.m_Type;
	m_Tag = otherComment.m_Tag;
	m_Value = otherComment.m_Value;
	return *this;
}

bool ModStreamComment::operator ==(ModStreamComment& otherComment) {
	if(m_Type != otherComment.m_Type)
		return false;
	if(m_Tag != otherComment.m_Tag)
		return false;
	if(m_Value != otherComment.m_Value)
		return false;
	return true;
}

bool ModStreamComment::operator !=(ModStreamComment& otherComment) {
	if(m_Type != otherComment.m_Type)
		return true;
	if(m_Tag != otherComment.m_Tag)
		return true;
	if(m_Value != otherComment.m_Value)
		return true;
	return false;
}

//////////////////////////////////////////////////////////////////////

void RemoveDuplicata(ModList<ModStreamComment>& comments) {
	ModList<ModStreamComment>::iterator commentIt;
	ModList<ModStreamComment>::iterator_safer otherIt;
	commentIt.begin(comments);
	while(commentIt) {
		ModStreamComment *pComment = (*commentIt);
		otherIt = commentIt;
		while(++otherIt) {
			ModStreamComment *pOtherComment = (*otherIt);
			if((*pOtherComment) == (*pComment))
				comments.DeleteNode(pOtherComment);
		}
		commentIt++;
	}
}

//////////////////////////////////////////////////////////////////////

ModStreamChapter::ModStreamChapter(void)
: m_TimeMS(0)
{
}

ModStreamChapter::ModStreamChapter(const ModStreamChapter& otherChapter) {
	*this = otherChapter;
}

ModStreamChapter& ModStreamChapter::operator =(const ModStreamChapter& otherChapter) {
	m_TimeMS = otherChapter.m_TimeMS;
	m_Name = otherChapter.m_Name;
	return *this;
}

bool ModStreamChapter::SeemsSimilar(const ModStreamChapter& otherChapter) const {
	if(m_Name != otherChapter.m_Name)
		return false;
	uint32 iTimeDiffMS;
	if(m_TimeMS <= otherChapter.m_TimeMS)
		iTimeDiffMS = otherChapter.m_TimeMS - m_TimeMS;
	else
		iTimeDiffMS = m_TimeMS - otherChapter.m_TimeMS;
	if(iTimeDiffMS == 0)
		return true;
	// If there is less than 2ms difference in the time chapter
	// (and that the chapter name is the same) then assume chapters
	// as similar
	if(iTimeDiffMS <= 2)
		return true;
	return false;
}

bool ModStreamChapter::operator ==(const ModStreamChapter& otherChapter) const {
	if(m_TimeMS != otherChapter.m_TimeMS)
		return false;
	if(m_Name != otherChapter.m_Name)
		return false;
	return true;
}

bool ModStreamChapter::operator !=(const ModStreamChapter& otherChapter) const {
	if(m_TimeMS != otherChapter.m_TimeMS)
		return true;
	if(m_Name != otherChapter.m_Name)
		return true;
	return false;
}

uint32 ModStreamChapter::GetTimeMS(void) const {
	return uint32(m_TimeMS);
}

VDStringW ModStreamChapter::GetTime(void) const {
	sint64 iTimeMS = m_TimeMS;
	int nMilliseconds = iTimeMS % 1000;
	int nSeconds = (iTimeMS / 1000) % 60;
	int nMinutes = (iTimeMS / 1000 / 60) % 60;
	int nHours = iTimeMS / 1000 / 60 / 60;
	VDStringW cTime = VDStringW::setf(L"%02d:%02d:%02d.%03d", nHours, nMinutes, nSeconds, nMilliseconds);
	return cTime;
}

const wchar_t *ModStreamChapter::GetName(void) const {
	return m_Name.c_str();
}

void ModStreamChapter::SetTimeMS(uint32 iTimeMS) {
	m_TimeMS = iTimeMS;
}

bool ModStreamChapter::SetTime(const char *time) {
	/*int nHours, nMinutes, nSeconds, nMilliseconds;
	if(sscanf(time, "%d:%d:%d.%d", &nHours, &nMinutes, &nSeconds, &nMilliseconds) != 4)
		return false;
	uint32 lTime = nMilliseconds
		+ (uint32)nSeconds * 1000
		+ (uint32)nMinutes * 1000 * 60
		+ (uint32)nHours * 1000 * 60 * 60;
	if(lTime < 0)
		return false;
	m_iTime = VDFraction((unsigned long)lTime);
	return true;*/
	return SetTime(VDTextAToW(time).c_str());
}

bool ModStreamChapter::SetTime(const wchar_t *time) {
	int nHours, nMinutes, nSeconds, nMilliseconds;
	if(swscanf(time, L"%d:%d:%d.%d", &nHours, &nMinutes, &nSeconds, &nMilliseconds) != 4)
		return false;
	uint32 iTimeMS = nMilliseconds
		+ (uint32)nSeconds * 1000
		+ (uint32)nMinutes * 1000 * 60
		+ (uint32)nHours * 1000 * 60 * 60;
	if(iTimeMS < 0)
		return false;
	m_TimeMS = iTimeMS;
	return true;
}

void ModStreamChapter::SetName(const wchar_t *wszName) {
	m_Name = wszName;
}

bool ImportModStreamChapters(VDStringW fileName, ModChapterList& chapters) {
	ModTextSource textSource;
	if(!textSource.Open(fileName))
		return false;

	VDStringW line;
	ModChapterList tmpList;
	sint16 nbNames = 0;
	sint16 nbTimes = 0;
	if(!textSource.getLine(line))
		return false;

	if(!wcsncmp(line.c_str(), L"CHAPTER01", wcslen(L"CHAPTER01"))) {	// chapterXtractor / DVDDecrypter
		while(!textSource.eof()) {
			VDStringW newName;
			uint32 newTimeMS = 0;
			const wchar_t *chapterTime = line.c_str();
			while( (*chapterTime) && (*chapterTime != L'='))
				chapterTime++;
			if(!*chapterTime)
				break;
			chapterTime++;
			int nHour=0, nMinute=0, nSecond=0, nMillisecond=0;
			if(wcslen(chapterTime) < 9) {	// Old format HH:MM:SS
				if(swscanf(chapterTime, L"%d:%d:%d", &nHour, &nMinute, &nSecond) != 3)
					break;
			} else {	// New format HH:MM:SS.mmm
				if(swscanf(chapterTime, L"%d:%d:%d.%d", &nHour, &nMinute, &nSecond, &nMillisecond) != 4)
					break;
			}
			nbTimes++;
			newTimeMS = nMillisecond
				+ (1000 * nSecond)
				+ (1000 * 60 * nMinute)
				+ (1000 * 60 * 60 * nHour);
			if(!textSource.getLine(line))
				break;
			const wchar_t *chapterName = line.c_str();
			while( (*chapterName) && (*chapterName != L'='))
				chapterName++;
			if(*chapterName) {
				chapterName++;
				newName = chapterName;
			}
			nbNames++;
			tmpList.AddChapter(newTimeMS, newName.c_str());
			textSource.getLine(line);
		}
	} else if(!wcsncmp(line.c_str(), L"*************", wcslen(L"*************"))) {	// SmartRipper
		// Chap#=Frame# Name
		// Chap#=Time
		while(!textSource.eof() && wcsncmp(line.c_str(), L"Chap#=Frame# Name", wcslen(L"Chap#=Frame# Name")))
			textSource.getLine(line);

		std::vector<ModChapterList::tChapterNode> tmpChapters;

		textSource.getLine(line);
		while(!textSource.eof() && line.length()) {
			const wchar_t *chapterName = line.c_str();
			while( (*chapterName) && (*chapterName != L' '))
				chapterName++;
			ModChapterList::tChapterNode newChapter;
			if(*chapterName) {
				chapterName++;
				newChapter.SetName(chapterName);
			}
			tmpChapters.push_back(newChapter);
			nbNames++;
			textSource.getLine(line);
		}

		while(!textSource.eof() && wcsncmp(line.c_str(), L"Chap#=Time", wcslen(L"Chap#=Time")))
			textSource.getLine(line);

		textSource.getLine(line);
		while(!textSource.eof() && line.length()) {
			const wchar_t *chapterTime = line.c_str();
			while( (*chapterTime) && (*chapterTime != L'='))
				chapterTime++;
			if(!*chapterTime)
				break;
			chapterTime++;
			ModStreamChapter *chapter = NULL;
			if(nbTimes < tmpChapters.size())
				chapter = &(tmpChapters[nbTimes]);
			if(!chapter)
				break;
			int nHour, nMinute, nSecond, nMillisecond;
			if(swscanf(chapterTime, L"%d:%d:%d.%d", &nHour, &nMinute, &nSecond, &nMillisecond) != 4)
				break;
			nMillisecond *= 10;
			uint32 time = nMillisecond
				+ (1000 * nSecond)
				+ (1000 * 60 * nMinute)
				+ (1000 * 60 * 60 * nHour);
			chapter->SetTimeMS(time);
			nbTimes++;
			textSource.getLine(line);
		}

		std::vector<ModChapterList::tChapterNode>::iterator chapterIt;
		for(chapterIt=tmpChapters.begin() ; chapterIt!=tmpChapters.end() ; chapterIt++)
			tmpList.AddChapter(*chapterIt);
	}

	if(!nbNames || (nbNames!=nbTimes) )
		return false;

	chapters = tmpList;
	chapters.Sort();
	textSource.Close();

	return true;
}

//////////////////////////////////////////////////////////////////////

ModChapterList::ModChapterList(void)
: m_bAdjustChapters(true)
{
}

ModChapterList::ModChapterList(const ModChapterList& otherList) {
	*this = otherList;
}

ModChapterList& ModChapterList::operator =(const ModChapterList& otherList) {
	*this = otherList.m_Chapters;
	m_bAdjustChapters = otherList.m_bAdjustChapters;

	return *this;
}

ModChapterList& ModChapterList::operator =(const tChapterList& otherList) {
	m_Chapters = otherList;

	return *this;
}

ModChapterList::tChapterNode& ModChapterList::AddChapter(void) {
	tChapterNode newChapter;

	m_Chapters.push_back(newChapter);

	return m_Chapters.back();
}

ModChapterList::tChapterNode& ModChapterList::AddChapter(uint32 iTimeMS, const wchar_t *wszName) {
	tChapterNode& newChapter = AddChapter();

	newChapter.SetTimeMS(iTimeMS);
	newChapter.SetName(wszName);

	return newChapter;
}

ModChapterList::tChapterNode& ModChapterList::AddChapter(const char *szTime, const wchar_t *wszName) {
	tChapterNode& newChapter = AddChapter();

	newChapter.SetTime(szTime);
	newChapter.SetName(wszName);

	return newChapter;
}

ModChapterList::tChapterNode& ModChapterList::AddChapter(tChapterNode& chapter) {
	tChapterNode& newChapter = AddChapter();

	newChapter = chapter;

	return newChapter;
}

std::list<ModChapterList::tChapterNode *> ModChapterList::GetPtrs(void) {
	std::list<tChapterNode *> listPtrs;
	iterator chapterIt;
	for(chapterIt=begin() ; chapterIt!=end() ; chapterIt++) {
		tChapterNode *pNode = &(*chapterIt);
		listPtrs.push_back(pNode);
	}
	return listPtrs;
}

bool ModChapterList::GetAdjustChapters(void) {
	return m_bAdjustChapters;
}

void ModChapterList::SetAdjustChapters(bool bAdjust) {
	m_bAdjustChapters = bAdjust;
}

void ModChapterList::Sort(void) {
	m_Chapters.sort(std::greater<ModStreamChapter>());
}
