//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <windows.h>
#include <commdlg.h>
#include <commctrl.h>
#include <tchar.h>
#include "projectui.h"
#include "ModStreamsSetup.h"
#include <vd2/Dita/services.h>
#include <vd2/system/registry.h>
#include <vd2/system/filesys.h>
#include <vd2/system/error.h>
#include <vd2/system/VDString.h>
#include "Dub.h"
#include "InputFile.h"
#include "resource.h"
#include "gui.h"
#include "misc.h"
#include "optdlg.h"
#include "AudioSource.h"
#include "ModAudioSourceAC3.h"
#include "ModAudioSourceDTS.h"
#include "ModAudioSourceMP3.h"
#include "ModTextSourceSRT.h"
#include "ogg&co/ModAudioSourceVorbis.h"
#include "prefs.h"
#include "command.h"
#include "job.h"
#include "ISO-639.h"


sint16 InsertListBoxItem(HWND hwndList, LPSTR lpstr, LPARAM lparam, sint16 index=-1);
template<class _N>
	void RefreshList(HWND hwndList, ModList<_N>& list, sint16 select=-1);
template<class _N>
	void RefreshListData(HWND hwndList, std::list<_N *>& list, sint16 select=-1);
void RefreshStreamList(HWND hwndList, sint16 select=-1);
void RefreshAudioMenu(void);
LRESULT CALLBACK StreamListHookMouseProc(int nCode, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK ListHookMouseProc(int nCode, WPARAM wParam, LPARAM lParam);
sint32 AddModStream(AudioSource *source, bool bInputFile, sint64 offset, bool bRetrieveComments, bool bRetrieveOffset);

void DrawTransparentBitmap(HDC dc, HBITMAP hBmp, int xStart, int yStart, COLORREF cTransparentColor);

extern DubSource::ErrorMode	g_audioErrorMode;
extern WAVEFORMATEX *AudioChooseCompressor(HWND hwndParent, WAVEFORMATEX *, WAVEFORMATEX *);
extern void ChooseCompressor(HWND hwndParent, COMPVARS *lpCompVars, BITMAPINFOHEADER *bihInput);
extern void CPUTest();

namespace {
	bool g_bUseUnicodeWinAPI = (LONG)GetVersion() >= 0;
}

enum {
	kFileDialog_ModInputFile	= 'imod',
	kFileDialog_ChapterFileIn	= 'chap',
	kFileDialog_SaveAs			= 'saas'
};

extern vdautoptr<VDProjectUI> g_projectui;
extern HWND			g_hWnd;
extern HINSTANCE	g_hInst;
extern vdrefptr<InputFile>	inputAVI;
extern DubOptions	g_dubOpts;
ModList<ModStream>	g_Streams;
ModVideoStream		g_VideoStream;
HMENU				g_hAudioMenu;
int					g_iSelectedStream = 0;
extern vdrefptr<VideoSource>	inputVideoAVI;
extern vdrefptr<AudioSource>	inputAudio;

extern const wchar_t g_fileFiltersModInputs[]=
		L"Supported formats (*.wav,*.mp?,*.ac3,*.ogg,*.dts,*.srt)\0"	L"*.wav;*.mp?;*.ac3;*.ogg;*.dts;*.srt\0"
		L"WAV file (*.wav)\0"	L"*.wav\0"
		L"Mpeg Audio file (*.mp?)\0"	L"*.mp?\0"
		L"AC3 file (*.ac3)\0"	L"*.ac3\0"
		L"Ogg Vorbis (*.ogg)\0"	L"*.ogg\0"
		L"DTS file (*.dts)\0"	L"*.dts\0"
		L"SRT file (*.srt)\0"	L"*.srt\0"
;

extern const wchar_t g_fileFiltersModChapterFile[]=
		L"Chapter file (*.txt)\0"	L"*.txt\0"
		L"All files (*.*)\0"		L"*.*\0"
;

HHOOK	g_MouseHook		= NULL;
HWND	g_hListBox		= NULL;
HWND	g_hDlg			= NULL;
HBITMAP	g_hSpeakerOn	= NULL;
HBITMAP	g_hSpeakerOff	= NULL;
HBITMAP	g_hMovie		= NULL;

//////////////////////////////////////////////////////////////////////

class ModDialogStreamListW32 : public VDDialogBaseW32 {
public:
	inline ModDialogStreamListW32(void) : VDDialogBaseW32(IDD_STREAMS_SETUP) {}

	void Activate(VDGUIHandle hParent) {
		ActivateDialog(hParent);
	};

protected:
	HWND	m_hListBox;
	void	RefreshStates(bool bDubbing=false);
	BOOL	DlgProc(UINT msg, UINT wParam, LONG lParam);
	void	UpdateListBoxHorizontalExtent(void);
	sint32	PrintStreamInfo(ModStream *stream, HDC hDC, RECT rcItem, bool bGetHorizontalExtent);
};

void ModDisplayStreamListDialog(VDGUIHandle hwndParent) {
	ModDialogStreamListW32 dlg;
	dlg.Activate(hwndParent);
}

//////////////////////////////////////////////////////////////////////

sint16 InsertListBoxItem(HWND hwndList, LPSTR lpstr, LPARAM lparam, sint16 index) {
	int nItem = SendMessage(hwndList, LB_INSERTSTRING, (WPARAM)index, (LPARAM)lpstr);
	if( (nItem==LB_ERR) || (nItem==LB_ERRSPACE) )
		return -1;
	if(SendMessage(hwndList, LB_SETITEMDATA, (WPARAM)nItem, lparam) == LB_ERR)
		return -1;
	return nItem;
}

static void VDCheckMenuItemW32(HMENU hMenu, UINT opt, bool en) {
	CheckMenuItem(hMenu, opt, en ? (MF_BYCOMMAND|MF_CHECKED) : (MF_BYCOMMAND|MF_UNCHECKED));
}

static void VDEnableMenuItemW32(HMENU hMenu, UINT opt, bool en) {
	EnableMenuItem(hMenu, opt, en ? (MF_BYCOMMAND|MF_ENABLED) : (MF_BYCOMMAND|MF_GRAYED));
}

static void ModSetSelectedStream(sint16 select) {
	g_iSelectedStream = select;
	// Uncomments this line once all the changes have been done
	ModStream *selectedStream = g_Streams[g_iSelectedStream];
	inputAudio = NULL;
	if(selectedStream)
		inputAudio = selectedStream->m_Source;
}

static void ModRefreshSelectedStream(void) {
	ModSetSelectedStream(g_iSelectedStream);
}

static void InstallMouseHook(HOOKPROC hookFunction, HWND currentDlg, int nIDListBox) {
	if(g_MouseHook)
		UnhookWindowsHookEx(g_MouseHook);
	g_MouseHook = NULL;
	g_hDlg = currentDlg;
	g_hListBox = (currentDlg ? GetDlgItem(currentDlg, nIDListBox) : NULL);
	if(hookFunction)
		g_MouseHook = SetWindowsHookEx(WH_MOUSE, hookFunction, NULL, GetCurrentThreadId());
}

static void GetHorizontalExtent(VDStringW& text, HDC hDC, sint32& currentExtent, sint32 extraExtent) {
	SIZE newExtent;
	if(g_bUseUnicodeWinAPI) {
		if(!GetTextExtentPoint32W(hDC, text.c_str(), text.length(), &newExtent))
			newExtent.cx = 0;
		newExtent.cx += extraExtent;
	} else {
		VDStringA textA = VDTextWToA(text);
		if(!GetTextExtentPoint32A(hDC, textA.c_str(), textA.length(), &newExtent))
			newExtent.cx = 0;
		newExtent.cx += extraExtent;
	}
	if(newExtent.cx > currentExtent)
		currentExtent = newExtent.cx;
}

//////////////////////////////////////////////////////////////////////

template<class _N>
void RefreshList(HWND hwndList, ModList<_N>& list, sint16 select) {
	TCHAR tchBuffer[1024];

	if(select < 0)
		select = SendMessage(hwndList, LB_GETCURSEL, 0, 0);

	SendMessage(hwndList, LB_RESETCONTENT, 0, 0);
	ModList<_N>::iterator it(list);
	sint32 index = 0;
	while(it) {
		index++;
		_sntprintf(tchBuffer, sizeof(tchBuffer)-1, _T("%ld"), index);
		InsertListBoxItem(hwndList, tchBuffer, (LPARAM)(*it));
		it++;
	}
	SetFocus(hwndList);
	if(select < 0)
		return;
	if(select >= index)
		select = index-1;
	SendMessage(hwndList, LB_SETCURSEL, select, 0);
}

template<class _N>
void RefreshListData(HWND hwndList, std::list<_N *>& list, sint16 iSelect) {
	TCHAR tchBuffer[1024];

	if(iSelect < 0)
		iSelect = SendMessage(hwndList, LB_GETCURSEL, 0, 0);

	SendMessage(hwndList, LB_RESETCONTENT, 0, 0);
	std::list<_N *>::iterator it;
	sint32 iIndex = 0;
	for(it=list.begin() ; it!=list.end() ; it++) {
		iIndex++;
		_sntprintf(tchBuffer, sizeof(tchBuffer)-1, _T("%ld"), iIndex);
		InsertListBoxItem(hwndList, tchBuffer, (LPARAM)(*it));
	}
	SetFocus(hwndList);
	if(iSelect < 0)
		return;
	if(iSelect >= iIndex)
		iSelect = iIndex-1;
	SendMessage(hwndList, LB_SETCURSEL, iSelect, 0);
}

void RefreshStreamList(HWND hwndList, sint16 select) {
	RefreshList<ModStream>(hwndList, g_Streams, select);
	ModSetSelectedStream(select);
	RefreshAudioMenu();
}

#define TOP_INDEX			2
#define LEFT_INDEX			20
#define LEFT_FIRSTLINE		48
#define LEFT_OTHERLINE		55
#define ICON_SPEAKER_LEFT	24
#define ICON_SPEAKER_BOTTOM	0

static LRESULT CALLBACK StreamListHookMouseProc(int nCode, WPARAM wParam, LPARAM lParam) {
	if(nCode == HC_ACTION) {
		switch(wParam) {
		case WM_RBUTTONDOWN:
			{
				RECT listRect;
				RECT itemRect;
				HMENU audioMenu;
				MOUSEHOOKSTRUCT *pStruct = (MOUSEHOOKSTRUCT *)lParam;
				if(pStruct->hwnd != g_hListBox)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				GetWindowRect(g_hListBox, &listRect);
				long relX = pStruct->pt.x - listRect.left;
				long relY = pStruct->pt.y - listRect.top;
				int index = SendMessage(g_hListBox, LB_ITEMFROMPOINT, 0, MAKEWPARAM(relX, relY));
				if(SendMessage(g_hListBox, LB_GETITEMRECT, index, (WPARAM)&itemRect) == LB_ERR)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				if( (relX < itemRect.left) || (relX > itemRect.right)
				|| (relY < itemRect.top) || (relY > itemRect.bottom) )
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				ModSetSelectedStream(index);
				SendMessage(g_hListBox, LB_SETCURSEL, index, 0);
				ModStream *stream = g_Streams[g_iSelectedStream];
				if(!stream || !stream->m_Source /*|| (stream->m_Source->getSourceMainType()!=SOURCE_AUDIO)*/
				|| !(audioMenu = GetSubMenu(g_hAudioMenu, 0)))
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				RefreshAudioMenu();
				TrackPopupMenu(audioMenu
					, TPM_LEFTALIGN | TPM_TOPALIGN
						| TPM_LEFTBUTTON
					, pStruct->pt.x, pStruct->pt.y, 0, g_hDlg, NULL);
				return TRUE;
			}

		case WM_LBUTTONDOWN:
			{
				RECT listRect;
				RECT itemRect;
				MOUSEHOOKSTRUCT *pStruct = (MOUSEHOOKSTRUCT *)lParam;
				if(pStruct->hwnd != g_hListBox)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				GetWindowRect(g_hListBox, &listRect);
				long relX = pStruct->pt.x - listRect.left;
				long relY = pStruct->pt.y - listRect.top;
				int index = SendMessage(g_hListBox, LB_ITEMFROMPOINT, 0, MAKEWPARAM(relX, relY));
				if(SendMessage(g_hListBox, LB_GETITEMRECT, index, (WPARAM)&itemRect) == LB_ERR)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				if( (relX < itemRect.left) || (relX > itemRect.right)
				|| (relY < itemRect.top) || (relY > itemRect.bottom) )
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				ModSetSelectedStream(index);
				SendMessage(g_hListBox, LB_SETCURSEL, index, 0);
				ModStream *stream = g_Streams[g_iSelectedStream];
				if(!stream || (stream->m_Source->getSourceMainType()!=SOURCE_AUDIO)
				|| (stream->m_bToPreview && !g_hSpeakerOn) || (!stream->m_bToPreview && !g_hSpeakerOff))
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				BITMAP bm;
				GetObject(stream->m_bToPreview ? g_hSpeakerOn : g_hSpeakerOff, sizeof(bm), &bm);
				if( (relX < itemRect.left + ICON_SPEAKER_LEFT) || (relX > itemRect.left + ICON_SPEAKER_LEFT + bm.bmWidth)
				|| (relY < itemRect.bottom - ICON_SPEAKER_BOTTOM - bm.bmHeight) || (relY > itemRect.bottom - ICON_SPEAKER_BOTTOM) )
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				// Someone clicked on the Speaker icon
				if(stream->m_bToPreview)
					stream->m_bToPreview = false;
				else {
					ModList<ModStream>::iterator it(g_Streams);
					while(it) {
						(*it)->m_bToPreview = false;
						it++;
					}
					stream->m_bToPreview = true;
				}
				RedrawWindow(g_hDlg, NULL, NULL, RDW_INTERNALPAINT | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
				return TRUE;
			}

		case WM_LBUTTONDBLCLK:
			{
				RECT listRect;
				RECT itemRect;
				MOUSEHOOKSTRUCT *pStruct = (MOUSEHOOKSTRUCT *)lParam;
				if(pStruct->hwnd != g_hListBox)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				GetWindowRect(g_hListBox, &listRect);
				long relX = pStruct->pt.x - listRect.left;
				long relY = pStruct->pt.y - listRect.top;
				int index = SendMessage(g_hListBox, LB_ITEMFROMPOINT, 0, MAKEWPARAM(relX, relY));
				if(SendMessage(g_hListBox, LB_GETITEMRECT, index, (WPARAM)&itemRect) == LB_ERR)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				if( (relX < itemRect.left) || (relX > itemRect.right)
				|| (relY < itemRect.top) || (relY > itemRect.bottom) )
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				ModSetSelectedStream(index);
				SendMessage(g_hListBox, LB_SETCURSEL, index, 0);
				ModStream *stream = g_Streams[g_iSelectedStream];
				if(!stream)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				stream->m_bEnabled = !stream->m_bEnabled;
				RedrawWindow(g_hDlg, NULL, NULL, RDW_INTERNALPAINT | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
				return TRUE;
			}
		}
	}
	return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
}

static void RefreshAudioMenu(void) {
	HMENU hMenu = g_hAudioMenu;
	ModStream *selectedStream = g_Streams[g_iSelectedStream];
	if(!selectedStream)
		return;

	CheckMenuRadioItem(hMenu, ID_AUDIO_MODE_DIRECT, ID_AUDIO_MODE_FULL, ID_AUDIO_MODE_DIRECT+selectedStream->m_Opts.mode, MF_BYCOMMAND);

	VDCheckMenuItemW32(hMenu, ID_AUDIO_ADVANCEDFILTERING,			selectedStream->m_Opts.bUseAudioFilterGraph);

	const bool bIsAudio = (selectedStream->m_Source->getSourceMainType() == SOURCE_AUDIO);
	const bool bAudioProcessingEnabled			= bIsAudio && (selectedStream->m_Opts.mode == DubAudioOptions::M_FULL);
	const bool bUseFixedFunctionAudioPipeline	= bIsAudio && bAudioProcessingEnabled && !selectedStream->m_Opts.bUseAudioFilterGraph;
	const bool bUseProgrammableAudioPipeline	= bIsAudio && bAudioProcessingEnabled && selectedStream->m_Opts.bUseAudioFilterGraph;

	VDEnableMenuItemW32(hMenu,ID_AUDIO_COMPRESSION			, bIsAudio);
	VDEnableMenuItemW32(hMenu,ID_AUDIO_MODE_DIRECT			, bIsAudio);
	VDEnableMenuItemW32(hMenu,ID_AUDIO_MODE_FULL			, bIsAudio);
	VDEnableMenuItemW32(hMenu,ID_AUDIO_ADVANCEDFILTERING	, bAudioProcessingEnabled);
	VDEnableMenuItemW32(hMenu,ID_AUDIO_COMPRESSION			, bAudioProcessingEnabled);
	VDEnableMenuItemW32(hMenu,ID_AUDIO_CONVERSION			, bUseFixedFunctionAudioPipeline);
	VDEnableMenuItemW32(hMenu,ID_AUDIO_VOLUME				, bUseFixedFunctionAudioPipeline);
	VDEnableMenuItemW32(hMenu,ID_AUDIO_FILTERS				, bUseProgrammableAudioPipeline);
}

void ModDialogStreamListW32::RefreshStates(bool bDubbing) {
	ModStream *stream = g_Streams[g_iSelectedStream];
	if(!bDubbing)
		bDubbing = !!g_dubber;

	EnableWindow(GetDlgItem(mhdlg, IDC_ADD),				!bDubbing);
	EnableWindow(GetDlgItem(mhdlg, IDC_MOVEUP),				!!stream/* && !bDubbing*/);
	EnableWindow(GetDlgItem(mhdlg, IDC_MOVEDOWN),			!!stream/* && !bDubbing*/);
	EnableWindow(GetDlgItem(mhdlg, IDC_STREAM_COMMENTS),	!!stream && !bDubbing);
	EnableWindow(GetDlgItem(mhdlg, IDC_STREAM_CHAPTERS),	!!stream && !bDubbing);
	EnableWindow(GetDlgItem(mhdlg, IDC_SAVEWAV),			!!stream && !bDubbing);
	EnableWindow(GetDlgItem(mhdlg, IDC_DEMUX),				!!stream && !bDubbing);
	TCHAR buttonText[256];
	memset(buttonText, 0, sizeof(buttonText));
	if(stream && stream->m_bInputFile) {
		if(stream->m_bEnabled)
			LoadString(g_hInst, IDS_INFO_DISABLE, buttonText, sizeof(buttonText)/sizeof(buttonText[0])-1);
		else
			LoadString(g_hInst, IDS_INFO_ENABLE, buttonText, sizeof(buttonText)/sizeof(buttonText[0])-1);
	} else
		LoadString(g_hInst, IDS_INFO_DELETE, buttonText, sizeof(buttonText)/sizeof(buttonText[0])-1);
	SetDlgItemText(mhdlg, IDC_DELETE, buttonText);
	EnableWindow(GetDlgItem(mhdlg, IDC_DELETE),				!!stream && /*!stream->m_bInputFile &&*/ !bDubbing);
}

static struct ReposItem StreamListCtlPosData[]={
	{ IDC_STREAM_LIST		, REPOS_SIZERIGHT | REPOS_SIZEDOWN },
	{ IDOK					, REPOS_MOVERIGHT },
	{ IDC_ADD				, REPOS_MOVERIGHT },
	{ IDC_MOVEUP			, REPOS_MOVERIGHT },
	{ IDC_MOVEDOWN			, REPOS_MOVERIGHT },
	{ IDC_STREAM_COMMENTS	, REPOS_MOVERIGHT },
	{ IDC_STREAM_CHAPTERS	, REPOS_MOVERIGHT },
	{ IDC_SAVEWAV			, REPOS_MOVERIGHT | REPOS_MOVEDOWN },
	{ IDC_DEMUX				, REPOS_MOVERIGHT | REPOS_MOVEDOWN },
	{ IDC_DELETE			, REPOS_MOVERIGHT | REPOS_MOVEDOWN },
	{ 0 }
};

POINT StreamListCtlPos[10];

BOOL ModDialogStreamListW32::DlgProc(UINT msg, UINT wParam, LONG lParam) {
	TCHAR tchBuffer[1024];
	TEXTMETRIC tm;
	int y;
	LPMEASUREITEMSTRUCT lpmis;
	LPDRAWITEMSTRUCT lpdis;
	ModStream *selectedStream = NULL;
	static RECT rInitial;
	static BOOL returnValue;

	switch(msg) {
	case WM_INITDIALOG:
		{
			returnValue = FALSE;
			g_hSpeakerOn = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_SPEAKER1));
			g_hSpeakerOff = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_SPEAKER0));
			g_hMovie = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_MOVIE));
			GetWindowRect(mhdlg, &rInitial);
			g_hListBox = m_hListBox = GetDlgItem(mhdlg, IDC_STREAM_LIST);
			ModSetSelectedStream(g_iSelectedStream);
			InstallMouseHook(StreamListHookMouseProc, mhdlg, IDC_STREAM_LIST);
			SendMessage(mhdlg, WM_SETICON, (WPARAM)ICON_SMALL, (LPARAM)LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_VIRTUALDUB)));
			guiReposInit(mhdlg, StreamListCtlPosData, StreamListCtlPos);
			RefreshStreamList(m_hListBox, g_iSelectedStream);
			UpdateListBoxHorizontalExtent();
			RefreshStates();
		}
		return TRUE;

	case WM_SIZE:
		guiReposResize(mhdlg, StreamListCtlPosData, StreamListCtlPos);
		return TRUE;

	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO lpmmi = (LPMINMAXINFO)lParam;

			lpmmi->ptMinTrackSize.x = rInitial.right - rInitial.left;
			lpmmi->ptMinTrackSize.y = rInitial.bottom - rInitial.top;
		}
		return TRUE;

	case WM_MEASUREITEM:
		{
			if(wParam != IDC_STREAM_LIST)
				return FALSE;
			lpmis = (LPMEASUREITEMSTRUCT)lParam;
			HDC hDC = GetDC(m_hListBox);
			HFONT hFont = (HFONT)SendMessage(m_hListBox, WM_GETFONT, 0, 0);
			if(!hFont)
				hFont = (HFONT)GetStockObject(SYSTEM_FONT);
			HGDIOBJ hOld = SelectObject(hDC, hFont);
			GetTextMetrics(hDC, &tm);
			SelectObject(hDC, hOld);
			ReleaseDC(m_hListBox, hDC);
			lpmis->itemHeight = 3 * tm.tmHeight + 8;
		}
		return TRUE;

	case WM_DRAWITEM:
		if(wParam != IDC_STREAM_LIST)
			return FALSE;
		lpdis = (LPDRAWITEMSTRUCT) lParam;
		if(lpdis->itemID == -1)
			break;
		switch (lpdis->itemAction) {
		case ODA_SELECT:
		case ODA_DRAWENTIRE:
		case ODA_FOCUS:
			{
				HGDIOBJ hOld;
				ModStream *stream = g_Streams[lpdis->itemID];
				COLORREF color;
				HBRUSH brush;
				if(lpdis->itemState & ODS_SELECTED) {
					if(lpdis->itemID != g_iSelectedStream) {
						ModSetSelectedStream(lpdis->itemID);
						RefreshStates(false);
					}
					color = RGB(0xE0, 0xE0, 0xE0);
				} else
					color = RGB(0xFF, 0xFF, 0xFF);
				SetBkColor(lpdis->hDC, color);
				SetBkMode(lpdis->hDC, TRANSPARENT);
				brush = CreateSolidBrush(color);
				FillRect(lpdis->hDC, &lpdis->rcItem, brush);
				DeleteObject(brush);
				if(stream && stream->m_bInputFile && g_hMovie) {
					DrawTransparentBitmap(lpdis->hDC, g_hMovie
						, lpdis->rcItem.left, lpdis->rcItem.top+2
						, RGB( 0xFF, 0xFF, 0xFF));
				}
				if(!stream || !stream->m_bEnabled) {
					brush = CreateHatchBrush(HS_BDIAGONAL, RGB(0x80, 0x80, 0x80));
					hOld = SelectObject(lpdis->hDC, brush);
					//FillRect(lpdis->hDC, &lpdis->rcItem, brush);
					PatBlt(lpdis->hDC, lpdis->rcItem.left, lpdis->rcItem.top, lpdis->rcItem.right-lpdis->rcItem.left, lpdis->rcItem.bottom-lpdis->rcItem.top, 0xA000C9);
					DeleteObject(SelectObject(lpdis->hDC, hOld));
				}
				if(stream && (stream->m_Source->getSourceMainType()==SOURCE_AUDIO) ) {
					if(stream && stream->m_bToPreview && g_hSpeakerOn) {
						DrawTransparentBitmap(lpdis->hDC, g_hSpeakerOn
							, lpdis->rcItem.left + ICON_SPEAKER_LEFT, lpdis->rcItem.bottom - 24 - ICON_SPEAKER_BOTTOM
							, RGB( 0xFF, 0xFF, 0xFF));
					} else if(stream && !stream->m_bToPreview && g_hSpeakerOff) {
						DrawTransparentBitmap(lpdis->hDC, g_hSpeakerOff
							, lpdis->rcItem.left + ICON_SPEAKER_LEFT, lpdis->rcItem.bottom - 24 - ICON_SPEAKER_BOTTOM
							, RGB( 0xFF, 0xFF, 0xFF));
					}
				}
				_sntprintf(tchBuffer, sizeof(tchBuffer)-1, _T("%u"), lpdis->itemID+1);
				y = lpdis->rcItem.top;
				TextOut(lpdis->hDC, LEFT_INDEX, y + TOP_INDEX, tchBuffer, _tcslen(tchBuffer));
				PrintStreamInfo(stream, lpdis->hDC, lpdis->rcItem, false);
				hOld = SelectObject(lpdis->hDC, CreatePen(PS_SOLID, 2, RGB(0,0,0)));
				MoveToEx(lpdis->hDC, lpdis->rcItem.left, lpdis->rcItem.top, NULL);
				LineTo(lpdis->hDC, lpdis->rcItem.right, lpdis->rcItem.top);
				MoveToEx(lpdis->hDC, lpdis->rcItem.left, lpdis->rcItem.bottom, NULL);
				LineTo(lpdis->hDC, lpdis->rcItem.right, lpdis->rcItem.bottom);
				DeleteObject(SelectObject(lpdis->hDC, CreatePen(PS_DOT, 1, RGB(0xA0,0xA0,0xA0))));
				MoveToEx(lpdis->hDC, LEFT_FIRSTLINE, lpdis->rcItem.top, NULL);
				LineTo(lpdis->hDC, LEFT_FIRSTLINE, lpdis->rcItem.bottom);
				DeleteObject(SelectObject(lpdis->hDC, hOld));
			}
			break;
		}
	return TRUE;

	case WM_VKEYTOITEM:
		{
			if((HWND)lParam == m_hListBox) {
				if(LOWORD(wParam) == VK_DELETE) {
					SendMessage(mhdlg, WM_COMMAND, IDC_DELETE, (LPARAM)GetDlgItem(mhdlg, IDC_DELETE));
					return -2;	// Tell the listbox we did the necessary handling
				}
			}
			return -1; // Tell the listbox to handle the keystroke as usual
		}
		break;

	case WM_COMMAND:
		selectedStream = g_Streams[g_iSelectedStream];
		switch(LOWORD(wParam)) {
		case IDC_ADD:
			{
				DWORD nFilterIndex = 1;
				static const VDFileDialogOption sOptions[]={
					{ VDFileDialogOption::kSelectedFilter, 0, NULL, 0, 0 },
					{0}
				};

				int optVals[1]={
					nFilterIndex
				};

				VDStringW fileName = VDGetLoadFileName(kFileDialog_ModInputFile, (VDGUIHandle)mhdlg, L"Open external source", g_fileFiltersModInputs, g_prefs.main.fAttachExtension ? L"wav" : NULL, sOptions, optVals);

				nFilterIndex = optVals[0];
				if(!fileName.empty()) {
					int type = SOURCE_UNKNOWN;
					bool valid = true;
					const wchar_t *wszName = fileName.c_str();
					if( (nFilterIndex < 2) || (nFilterIndex > 7) ) {
						const wchar_t *wszExt = wszName + fileName.length();
						while( (wszExt >= wszName) && (*wszExt != L'\\') && (*wszExt!=L'.') )
							wszExt--;
						if(*wszExt != L'.')
							return TRUE;
						wszExt++;
						if(!wcsicmp(wszExt, L"wav"))
							nFilterIndex = 2;
						else if(!wcsnicmp(wszExt, L"mp", 2) && (wcslen(wszExt)==3))
							nFilterIndex = 3;
						else if(!wcsicmp(wszExt, L"ac3"))
							nFilterIndex = 4;
						else if(!wcsicmp(wszExt, L"ogg"))
							nFilterIndex = 5;
						else if(!wcsicmp(wszExt, L"dts"))
							nFilterIndex = 6;
						else if(!wcsicmp(wszExt, L"srt"))
							nFilterIndex = 7;
						else
							return TRUE;
					}
					//VDStringA errorMsgFormat("\"%s\" doesn't seem to ba a valid %s file.");
					{
						VDAutoLogDisplay logDisp;

						switch(nFilterIndex) {
							// TODO : make only one error message and replace %s by WAV/Vorbis/...
						case 2: // WAV file
							type = SOURCE_WAV | SOURCE_AUDIO_WAV;
							if(!(valid=OpenModInputFile(fileName, type, 0, true, false, true)))
								guiMessageBoxF(mhdlg, IDS_WARN_INVALID_X_FILE, IDS_WARN_CAPTION, MB_ICONWARNING | MB_OK, VDTextWToA(fileName).c_str(), "WAV");
							break;
						case 3: // MP3
							type = SOURCE_STD_FILE | SOURCE_AUDIO_MP3;
							if(!(valid=OpenModInputFile(fileName, type, 0, true, false, true)))
								guiMessageBoxF(mhdlg, IDS_WARN_INVALID_X_FILE, IDS_WARN_CAPTION, MB_ICONWARNING | MB_OK, VDTextWToA(fileName).c_str(), "Mpeg Audio");
							break;
						case 4: // AC3
							type = SOURCE_STD_FILE | SOURCE_AUDIO_AC3;
							if(!(valid=OpenModInputFile(fileName, type, 0, true, false, true)))
								guiMessageBoxF(mhdlg, IDS_WARN_INVALID_X_FILE, IDS_WARN_CAPTION, MB_ICONWARNING | MB_OK, VDTextWToA(fileName).c_str(), "AC3");
							break;
						case 5: // Audio stream in Ogg file (likely to be a Vorbis stream)
							type = SOURCE_OGM | SOURCE_AUDIO_VORBIS;
							if(!(valid=OpenModInputFile(fileName, type, 0, true, false, true)))
								guiMessageBoxF(mhdlg, IDS_WARN_INVALID_X_FILE, IDS_WARN_CAPTION, MB_ICONWARNING | MB_OK, VDTextWToA(fileName).c_str(), "OGG");
							break;
						case 6: // DTS
							type = SOURCE_STD_FILE | SOURCE_AUDIO_DTS;
							if(!(valid=OpenModInputFile(fileName, type, 0, true, false, true)))
								guiMessageBoxF(mhdlg, IDS_WARN_INVALID_X_FILE, IDS_WARN_CAPTION, MB_ICONWARNING | MB_OK, VDTextWToA(fileName).c_str(), "DTS");
							break;
						case 7: // SRT
							type = SOURCE_STD_FILE | SOURCE_TEXT_SRT;
							if(!(valid=OpenModInputFile(fileName, type, 0, true, false, true)))
								guiMessageBoxF(mhdlg, IDS_WARN_INVALID_X_FILE, IDS_WARN_CAPTION, MB_ICONWARNING | MB_OK, VDTextWToA(fileName).c_str(), "SRT");
							break;
						}

						logDisp.Post((VDGUIHandle)g_hWnd);
					}

					if(valid) {
						RefreshStreamList(m_hListBox, g_iSelectedStream >= 0 ? g_iSelectedStream : 0);
						UpdateListBoxHorizontalExtent();
						RefreshStates();
					}
				}
			}
			return TRUE;

		case IDC_DELETE:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=g_Streams.GetCount()) )
					return TRUE;
				ModStream *stream = g_Streams[index];
				if(stream) {
					if(!stream->m_bInputFile) {
						g_Streams.DeleteNode(stream);
						if(SendMessage(m_hListBox, LB_DELETESTRING, index, 0) == LB_ERR)
							RefreshStreamList(m_hListBox, index);
						else {
							if(index == g_Streams.GetCount())
								index--;
							SendMessage(m_hListBox, LB_SETCURSEL, index, 0);
							ModSetSelectedStream(index);
						}
						UpdateListBoxHorizontalExtent();
					} else {
						stream->m_bEnabled = !stream->m_bEnabled;
						RefreshStreamList(m_hListBox, g_iSelectedStream >= 0 ? g_iSelectedStream : 0);
					}
					RefreshStates();
				}
			}
			return TRUE;

		case IDC_MOVEUP:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if(index<=0)
					return TRUE;
				ModStream *previous = g_Streams[index-1];
				ModStream *current = g_Streams[index];
				if(previous && current) {
					g_Streams.Swap(previous, current);
					if( (SendMessage(m_hListBox, LB_SETITEMDATA, index-1, (LPARAM)current)==LB_ERR)
						|| (SendMessage(m_hListBox, LB_SETITEMDATA, index, (LPARAM)previous)==LB_ERR) )
						RefreshStreamList(m_hListBox, index-1);
					else
						SendMessage(m_hListBox, LB_SETCURSEL, index-1, 0);
				}
			}
			return TRUE;

		case IDC_MOVEDOWN:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=g_Streams.GetCount()-1) )
					return TRUE;
				ModStream *current = g_Streams[index];
				ModStream *next = g_Streams[index+1];
				if(current && next) {
					g_Streams.Swap(current, next);
					if( (SendMessage(m_hListBox, LB_SETITEMDATA, index, (LPARAM)next)==LB_ERR)
						|| (SendMessage(m_hListBox, LB_SETITEMDATA, index+1, (LPARAM)current)==LB_ERR) )
						RefreshStreamList(m_hListBox, index+1);
					else
						SendMessage(m_hListBox, LB_SETCURSEL, index+1, 0);
				}
			}
			return TRUE;

		case IDC_STREAM_COMMENTS:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=g_Streams.GetCount()) )
					return TRUE;
				ModStream *stream = g_Streams[index];
				if(stream) {
					ModDisplayStreamCommentsDialog((VDGUIHandle)mhdlg, stream->m_Comments);
					// Reinstall our mouse hook :)
					InstallMouseHook(StreamListHookMouseProc, mhdlg, IDC_STREAM_LIST);
				}
			}
			return TRUE;

		case IDC_STREAM_CHAPTERS:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=g_Streams.GetCount()) )
					return TRUE;
				ModStream *stream = g_Streams[index];
				if(stream) {
					ModDisplayStreamChaptersDialog((VDGUIHandle)mhdlg, stream->m_Chapters);
					// Reinstall our mouse hook :)
					InstallMouseHook(StreamListHookMouseProc, mhdlg, IDC_STREAM_LIST);
				}
			}
			return TRUE;

		case IDC_SAVEWAV:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=g_Streams.GetCount()) )
					return TRUE;
				ModStream *stream = g_Streams[index];
				ModList<ModStream>::iterator it(g_Streams);
				while(it) {
					(*it)->m_bToDemux = false;
					it++;
				}
				if(stream) {
					stream->m_bToDemux = true;
					RefreshStates(true);
					g_projectui->SaveWAVAsk();
					RefreshStates();
				}
			}
			break;

		case IDC_DEMUX:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=g_Streams.GetCount()) )
					return TRUE;
				ModStream *stream = g_Streams[index];
				ModList<ModStream>::iterator it(g_Streams);
				while(it) {
					(*it)->m_bToDemux = false;
					it++;
				}
				if(stream) {
					stream->m_bToDemux = true;
					RefreshStates(true);
					g_projectui->DemuxStreamAsk();
					RefreshStates();
				}
			}
			break;

		case IDOK:
			returnValue = TRUE;
		case IDCANCEL:
			InstallMouseHook(NULL, NULL, NULL);
			if(g_hSpeakerOn)
				DeleteObject(g_hSpeakerOn);
			g_hSpeakerOn = NULL;
			if(g_hSpeakerOff)
				DeleteObject(g_hSpeakerOff);
			g_hSpeakerOff = NULL;
			if(g_hMovie)
				DeleteObject(g_hMovie);
			g_hMovie = NULL;
			End(returnValue);
			return TRUE;

		////
		// Here start messages for the audio streams
		////

		case ID_AUDIO_ADVANCEDFILTERING:
			selectedStream->m_Opts.bUseAudioFilterGraph = !selectedStream->m_Opts.bUseAudioFilterGraph;
			break;

		case ID_AUDIO_FILTERS:
			extern void VDDisplayAudioFilterDialog(VDGUIHandle, VDAudioFilterGraph&);
			//SetAudioSource();
			CPUTest();
			VDDisplayAudioFilterDialog((VDGUIHandle)mhdlg, selectedStream->m_AudioFilterGraph);
			break;

		case ID_AUDIO_CONVERSION:
			ActivateDubAudioDialog(g_hInst, MAKEINTRESOURCE(IDD_AUDIO_CONVERSION), mhdlg, AudioConversionDlgProc, selectedStream->m_Opts);
			break;
		case ID_AUDIO_INTERLEAVE:
			ActivateDubAudioDialog(g_hInst, MAKEINTRESOURCE(IDD_INTERLEAVE), mhdlg, AudioInterleaveDlgProc, selectedStream->m_Opts);
			break;
		case ID_AUDIO_COMPRESSION:
			//SetAudioSource();

			if(!selectedStream->m_Source)
				selectedStream->m_ACompressionFormat = AudioChooseCompressor(mhdlg, selectedStream->m_ACompressionFormat, NULL);
			else {
		
				PCMWAVEFORMAT wfex;

				AudioSource *inputAudio = (AudioSource *)selectedStream->m_Source;

				memcpy(&wfex, inputAudio->getWaveFormat(), sizeof(PCMWAVEFORMAT));
				// Say 16-bit if the source was compressed.

				if (wfex.wf.wFormatTag != WAVE_FORMAT_PCM)
					wfex.wBitsPerSample = 16;

				wfex.wf.wFormatTag = WAVE_FORMAT_PCM;

				switch(selectedStream->m_Opts.newPrecision) {
				case DubAudioOptions::P_8BIT:	wfex.wBitsPerSample = 8; break;
				case DubAudioOptions::P_16BIT:	wfex.wBitsPerSample = 16; break;
				}

				switch(selectedStream->m_Opts.newChannels) {
				case DubAudioOptions::C_MONO:	wfex.wf.nChannels = 1; break;
				case DubAudioOptions::C_STEREO:	wfex.wf.nChannels = 2; break;
				}

				if (selectedStream->m_Opts.new_rate) {
					long samp_frac;

					if (selectedStream->m_Opts.integral_rate)
						if (selectedStream->m_Opts.new_rate > wfex.wf.nSamplesPerSec)
							samp_frac = 0x10000 / ((selectedStream->m_Opts.new_rate + wfex.wf.nSamplesPerSec/2) / wfex.wf.nSamplesPerSec); 
						else
							samp_frac = 0x10000 * ((wfex.wf.nSamplesPerSec + selectedStream->m_Opts.new_rate/2) / selectedStream->m_Opts.new_rate);
					else
						samp_frac = MulDiv(wfex.wf.nSamplesPerSec, 0x10000L, selectedStream->m_Opts.new_rate);

					wfex.wf.nSamplesPerSec = MulDiv(wfex.wf.nSamplesPerSec, 0x10000L, samp_frac);
				}

				wfex.wf.nBlockAlign = (wfex.wBitsPerSample+7)/8 * wfex.wf.nChannels;
				wfex.wf.nAvgBytesPerSec = wfex.wf.nSamplesPerSec * wfex.wf.nBlockAlign;

				selectedStream->m_ACompressionFormat = AudioChooseCompressor(mhdlg, selectedStream->m_ACompressionFormat, (WAVEFORMATEX *)&wfex);

			}

			if (selectedStream->m_ACompressionFormat) {
				selectedStream->m_ACompressionFormatSize = sizeof(WAVEFORMATEX) + selectedStream->m_ACompressionFormat->cbSize;
			}
			break;

		case ID_AUDIO_VOLUME:
			ActivateDubAudioDialog(g_hInst, MAKEINTRESOURCE(IDD_AUDIO_VOLUME), mhdlg, AudioVolumeDlgProc, selectedStream->m_Opts);
			break;

		case ID_AUDIO_MODE_DIRECT:
			selectedStream->m_Opts.mode = DubAudioOptions::M_NONE;
			break;
		case ID_AUDIO_MODE_FULL:
			selectedStream->m_Opts.mode = DubAudioOptions::M_FULL;
			break;
		}
		break;
	}
	return FALSE;
}

sint32 ModDialogStreamListW32::PrintStreamInfo(ModStream *stream, HDC hDC, RECT rcItem, bool bGetHorizontalExtent) {
	TEXTMETRIC tm;
	HGDIOBJ hOld;
	int y;
	sint32 horExtent = LEFT_FIRSTLINE;

	if(bGetHorizontalExtent) {
		HFONT hFont = (HFONT)SendMessage(m_hListBox, WM_GETFONT, 0, 0);
		if(!hFont)
			hFont = (HFONT)GetStockObject(SYSTEM_FONT);
		hOld = SelectObject(hDC, hFont);
	}
	GetTextMetrics(hDC, &tm);
	if(stream) {
		VDStringW sourceName(L"Source : ");
		if(stream->m_Source)
			sourceName += stream->m_Source->getSourceName();
		else
			sourceName += L"No source!";
		VDStringW sourceDesc(L"Desc. : ");
		if(stream->m_Source)
			sourceDesc += stream->m_Source->getSourceDesc();
		VDStringW sourceDuration;
		sint64 duration = 0;
		if(stream->m_Source)
			duration = stream->m_Source->samplesToMs(stream->m_Source->getEnd());
		if(duration > 0) {
			sint32 milliseconds = duration % 1000;
			sint32 seconds = (duration / 1000) % 60;
			sint32 minutes = (duration / 1000 / 60) % 60;
			sint32 hours = duration / 1000 / 60 / 60;
			sourceDuration = VDStringW::setf(L"Length : %02d:%02d:%02d.%03d"
				, hours, minutes, seconds, milliseconds);
		} else
			sourceDuration = L"Length : (Unknown)";
		y = rcItem.top;
		if(bGetHorizontalExtent) {
			GetHorizontalExtent(sourceName, hDC, horExtent, LEFT_FIRSTLINE+5 + 5);
			GetHorizontalExtent(sourceDesc, hDC, horExtent, LEFT_OTHERLINE+5 + 5);
			GetHorizontalExtent(sourceDuration, hDC, horExtent, LEFT_OTHERLINE+5 + 5);
		} else {
			if(g_bUseUnicodeWinAPI) {
				TextOutW(hDC, LEFT_FIRSTLINE+5, y + 2, sourceName.c_str(), sourceName.length());
				TextOutW(hDC, LEFT_OTHERLINE+5, y + 4 + tm.tmHeight, sourceDesc.c_str(), sourceDesc.length());
				TextOutW(hDC, LEFT_OTHERLINE+5, y + 6 + 2*tm.tmHeight, sourceDuration.c_str(), sourceDuration.length());
			} else {
				VDStringA sourceNameA = VDTextWToA(sourceName);
				VDStringA sourceDescA = VDTextWToA(sourceDesc);
				VDStringA sourceDurationA = VDTextWToA(sourceDuration);
				TextOutA(hDC, LEFT_FIRSTLINE+5, y + 2, sourceNameA.c_str(), sourceNameA.length());
				TextOutA(hDC, LEFT_OTHERLINE+5, y + 4 + tm.tmHeight, sourceDescA.c_str(), sourceDescA.length());
				TextOutA(hDC, LEFT_OTHERLINE+5, y + 6 + 2*tm.tmHeight, sourceDurationA.c_str(), sourceDurationA.length());
			}
		}
	} else {
		y = (rcItem.top + rcItem.bottom - tm.tmHeight) / 2;
		if(bGetHorizontalExtent)
			GetHorizontalExtent(VDStringW(L"Could not retrieve info"), hDC, horExtent, LEFT_FIRSTLINE+5 + 5);
		else
			TextOut(hDC, LEFT_FIRSTLINE+5, y, _T("Could not retrieve info"), _tcslen(_T("Could not retrieve info")));
	}
	if(bGetHorizontalExtent) {
		SelectObject(hDC, hOld);
		return horExtent;
	}
	return -1;
}

void ModDialogStreamListW32::UpdateListBoxHorizontalExtent() {
	HDC hDC = GetDC(m_hListBox);
	sint32 horExtent = 0;
	sint32 tmpExtent = 0;
	RECT rcItem = { 0 };
	ModList<ModStream>::iterator it(g_Streams);
	while(it) {
		ModStream *stream = *it;
		tmpExtent = PrintStreamInfo(stream, hDC, rcItem, true);
		if(tmpExtent > horExtent)
			horExtent = tmpExtent;
		it++;
	}
	SendMessage(m_hListBox, LB_SETHORIZONTALEXTENT, horExtent, 0);
	ReleaseDC(m_hListBox, hDC);
}

//////////////////////////////////////////////////////////////////////

bool OpenModInputFile(VDStringW fileName, uint32 type, sint64 offset, bool bRetrieveComments, bool bShutUp, bool bRetrieveOffset) {
	vdrefptr<AudioSource> source;
	bool valid = false;
	switch(TYPE(type)) {
	case SOURCE_AUDIO_WAV:	// WAV file
		{
			source = new AudioSourceWAV(fileName.c_str(), g_dubOpts.perf.waveBufferSize);
			if(!source)
				throw MyMemoryError();
			if(!source->init()) {
				valid = FALSE;
			} else {
				valid = TRUE;
			}
		}
		break;
	case SOURCE_AUDIO_VORBIS: // Audio stream in Ogg file (likely to be a Vorbis stream)
		{
			source = new ModAudioSourceVorbis(fileName);
			if(!source)
				throw MyMemoryError();
			if(!source->init()) {
				valid = FALSE;
			} else {
				valid = TRUE;
			}
		}
		break;
	case SOURCE_AUDIO_MP3: // MP3
		{
			MPAOptions opts;
			opts.m_bNandubTrick = !!g_VDMPrefs.m_bNandubCompatibility;
			opts.m_bKeepCorruptedData = !!g_VDMPrefs.m_bKeepCorruptedData;
			opts.m_bKeepTags = false;
			source = new ModAudioSourceMP3(fileName.c_str(), opts, bShutUp);
			if(!source)
				throw MyMemoryError();
			if(!source->init()) {
				valid = FALSE;
			} else {
				valid = TRUE;
			}
		}
		break;
	case SOURCE_AUDIO_AC3: // AC3
		{
			AC3Options opts;
			opts.m_bFrameAccuracy = !!g_VDMPrefs.m_bAudioFrameMode;
			opts.m_bKeepCorruptedData = !!g_VDMPrefs.m_bKeepCorruptedData;
			source = new ModAudioSourceAC3(fileName.c_str(), opts);
			if(!source)
				throw MyMemoryError();
			if(!source->init()) {
				valid = FALSE;
			} else {
				valid = TRUE;
			}
		}
		break;
	case SOURCE_AUDIO_DTS: // DTS
		{
			DTSOptions opts;
			opts.m_bFrameAccuracy = !!g_VDMPrefs.m_bAudioFrameMode;
			opts.m_bKeepCorruptedData = !!g_VDMPrefs.m_bKeepCorruptedData;
			source = new ModAudioSourceDTS(fileName.c_str(), opts);
			if(!source)
				throw MyMemoryError();
			if(!source->init()) {
				valid = FALSE;
			} else {
				valid = TRUE;
			}
		}
		break;
	case SOURCE_TEXT_SRT: // SRT
		source = new ModTextSourceSRT(fileName.c_str());
		if(!source)
			throw MyMemoryError();
		if(!source->init()) {
			valid = FALSE;
		} else {
			valid = TRUE;
		}
		break;
	}
	if(valid)
		AddModStream(source, false, offset, bRetrieveComments, bRetrieveOffset);
	else if(source)
		source = NULL;
	return valid;
}

sint32 AddModStream(AudioSource *source, bool bInputFile, sint64 offset, bool bRetrieveComments, bool bRetrieveOffset) {
	ModList<ModStream>::iterator it(g_Streams);
	bool bOneToPreview = false;
	while(it) {
		if((*it)->m_bToPreview) {
			bOneToPreview = true;
			break;
		}
		it++;
	}
	ModStream *stream = g_Streams.AddLast();
	VDASSERT(source);

	stream->m_Source = source;
	stream->m_bInputFile = bInputFile;
	source->setDecodeErrorMode(g_audioErrorMode);
	if(bRetrieveOffset)
		offset = source->sourceOffsetMS();
	stream->m_Opts.offset = offset;
	if(bRetrieveComments) {
		ICommentList *cList;
		if(stream->m_Source && stream->m_Source->QueryInterface(ICommentList::GetInterfaceID(), (void **)&cList)) {
			cList->GetCommentList(stream->m_Comments);
			cList->GetChapterList(stream->m_Chapters);
		}
	}
	IStreamIndex *sIndex;
	if(stream->m_Source && stream->m_Source->QueryInterface(IStreamIndex::GetInterfaceID(), (void **)&sIndex)) {
		stream->m_StreamType = sIndex->GetStreamType();
		stream->m_StreamIndex = sIndex->GetStreamIndex();
	}
	stream->m_bToPreview = !bOneToPreview && source && (source->getSourceMainType()==SOURCE_AUDIO);
	return g_Streams.GetCount()-1;
}

void AddModInputStreams(bool bRetrieveComments, bool bRetrieveOffset) {
	ICommentList *cList;
	if(inputVideoAVI && inputVideoAVI->QueryInterface(ICommentList::GetInterfaceID(), (void **)&cList)) {
		cList->GetCommentList(g_VideoStream.m_Comments);
		cList->GetChapterList(g_VideoStream.m_Chapters);
	}
	int index = 0;
	// AddModInputStream returns the stream index (starting @ 0)
	// -1 means the stream was found but wasn't added (because unwanted, e.g. a video stream)
	// -2 means the stream couldn't be found (i.e. there is no more streams to come)
	while(AddModInputStream(index++, 0, bRetrieveComments, false, bRetrieveOffset) >= -1);
}

sint32 AddModInputStream(DWORD fccType, LONG index, sint64 offset, bool bRetrieveComments, bool bShutUp, bool bRetrieveOffset) {
	if(inputAVI) {
		inputAVI->setAutomated(bShutUp);
		AudioSource *source = (AudioSource *)inputAVI->GetSource(fccType, index);
		if(!source)
			return -2;
		/*if(!source->init()) {
			delete source;
			return -2;
		}*/
		return AddModStream(source, true, offset, bRetrieveComments, bRetrieveOffset);
	}
	return -2;
}

sint32 AddModInputStream(LONG index, sint64 offset, bool bRetrieveComments, bool bShutUp, bool bRetrieveOffset) {
	if(inputAVI) {
		inputAVI->setAutomated(bShutUp);
		AudioSource *source = (AudioSource *)inputAVI->GetSource(index);
		if(!source)
			return -2;
		/*if(!source->init()) {
			delete source;
			return -2;
		}*/
		if(source->getSourceMainType() == SOURCE_VIDEO) {
			//delete source;
			return -1;
		}
		return AddModStream(source, true, offset, bRetrieveComments, bRetrieveOffset);
	}
	return -2;
}

void ClearVideoStream(void) {
	g_VideoStream.m_Source = NULL;
	g_VideoStream.m_Comments.DeleteContents(true);
	g_VideoStream.m_Comments.Clear();
	g_VideoStream.m_Chapters.clear();
	g_VideoStream.m_DAR = VDFraction(0);
	g_VideoStream.m_bDARScaleUp = true;
}

void RemoveModStreams(void) {
	g_Streams.DeleteContents(true);
	g_Streams.Clear();
}

void CloseModInputFiles(void) {
	ModList<ModStream>::iterator it(g_Streams);
	while(it) {
		if(!(*it)->m_bInputFile)
			g_Streams.DeleteNode(*it);
		it++;
	}
}

void PrepareRefreshModInputStreams(void) {
	ModList<ModStream>::iterator_safer it(g_Streams);
	while(it) {
		ModStream *stream = *it;
		if(stream->m_bInputFile) {
			bool bKeepStream = false;
			DubSource *newSource = NULL;
			long index = -1;
			DWORD streamType;
			sint32 streamIndex;

			IStreamIndex *sIndex;
			if(inputAVI && stream->m_Source && stream->m_Source->QueryInterface(IStreamIndex::GetInterfaceID(), (void **)&sIndex)) {
				streamType = sIndex->GetStreamType();
				streamIndex = sIndex->GetStreamIndex();
				bKeepStream = true;
			}
			stream->Reset(true, false, false);
			if(bKeepStream) {
				stream->m_Source = NULL;
				stream->m_bInputFile = true;
				stream->m_StreamType = streamType;
				stream->m_StreamIndex = streamIndex;
			} else
				g_Streams.DeleteNode(stream);
		}
		it++;
	}
	ModRefreshSelectedStream();
}

void RefreshModInputStreams(void) {
	ModList<ModStream>::iterator_safer it(g_Streams);
	while(it) {
		ModStream *stream = *it;
		if(stream->m_bInputFile) {
			bool bKeepStream = false;
			AudioSource *newSource = NULL;
			long index = -1;

			if(!stream->m_Source)
				newSource = (AudioSource *)inputAVI->GetSource(stream->m_StreamType, stream->m_StreamIndex);
			if(newSource) {
				/*if(!newSource->init())
					delete newSource;
				else*/
					bKeepStream = true;
			}
			stream->Reset(true, false, false);
			if(bKeepStream) {
				stream->m_Source = newSource;
				stream->m_bInputFile = true;
			} else
				g_Streams.DeleteNode(stream);
		}
		it++;
	}
	ModRefreshSelectedStream();
}

void RefreshStreamsComments(void) {
	ICommentList *cList;
	if(inputVideoAVI && inputVideoAVI->QueryInterface(ICommentList::GetInterfaceID(), (void **)&cList)) {
		g_VideoStream.m_Comments.DeleteContents(true);
		g_VideoStream.m_Comments.Clear();
		g_VideoStream.m_Chapters.clear();
		cList->GetCommentList(g_VideoStream.m_Comments);
		cList->GetChapterList(g_VideoStream.m_Chapters);
	}
	ModList<ModStream>::iterator it(g_Streams);
	while(it) {
		ModStream *curStream = *it;
		if(curStream->m_Source && curStream->m_Source->QueryInterface(ICommentList::GetInterfaceID(), (void **)&cList)) {
			curStream->m_Comments.DeleteContents(true);
			curStream->m_Comments.Clear();
			curStream->m_Chapters.clear();
			cList->GetCommentList(curStream->m_Comments);
			cList->GetChapterList(curStream->m_Chapters);
		}
		it++;
	}
}

//////////////////////////////////////////////////////////////////////

//
// This pretty useful function was found here : http://home.tvd.be/ws36637/transbmp.html
// Near to no changes were required to make it work in Win32 =)
//
// Disclaimer : no DC nor BITMAP was hurted or ill-treated in the process.
//
void DrawTransparentBitmap(HDC dc, HBITMAP hBmp, int xStart, int yStart, COLORREF cTransparentColor) {
	BITMAP		bm;
	COLORREF	cColor;
	HBITMAP		bmAndBack, bmAndObject, bmAndMem, bmSave;
	HBITMAP		bmBackOld, bmObjectOld, bmMemOld, bmSaveOld;
	HDC			hdcMem, hdcBack, hdcObject, hdcTemp, hdcSave;
	POINT		ptSize;

	hdcTemp = CreateCompatibleDC(dc);
	SelectObject(hdcTemp, hBmp);		// Select the bitmap

	GetObject(hBmp, sizeof(bm), &bm);
	ptSize.x	= bm.bmWidth;			// Get width of bitmap
	ptSize.y	= bm.bmHeight;			// Get height of bitmap
	DPtoLP(hdcTemp, &ptSize, 1);		// Convert from device
										// to logical points

	// Create some DCs to hold temporary data.
	hdcBack		= CreateCompatibleDC(dc);
	hdcObject	= CreateCompatibleDC(dc);
	hdcMem		= CreateCompatibleDC(dc);
	hdcSave		= CreateCompatibleDC(dc);

	// Create a bitmap for each DC. DCs are required for a number of
	// GDI functions.

	// Monochrome DC
	bmAndBack	= CreateBitmap(ptSize.x, ptSize.y, 1, 1, NULL);

	// Monochrome DC
	bmAndObject	= CreateBitmap(ptSize.x, ptSize.y, 1, 1, NULL);

	bmAndMem	= CreateCompatibleBitmap(dc, ptSize.x, ptSize.y);
	bmSave		= CreateCompatibleBitmap(dc, ptSize.x, ptSize.y);

	// Each DC must select a bitmap object to store pixel data.
	bmBackOld	= (HBITMAP)SelectObject(hdcBack, bmAndBack);
	bmObjectOld	= (HBITMAP)SelectObject(hdcObject, bmAndObject);
	bmMemOld	= (HBITMAP)SelectObject(hdcMem, bmAndMem);
	bmSaveOld	= (HBITMAP)SelectObject(hdcSave, bmSave);

	// Set proper mapping mode.
	SetMapMode(hdcTemp, GetMapMode(dc));

	// Save the bitmap sent here, because it will be overwritten.
	BitBlt(hdcSave, 0, 0, ptSize.x, ptSize.y, hdcTemp, 0, 0, SRCCOPY);

	// Set the background color of the source DC to the color.
	// contained in the parts of the bitmap that should be transparent
	cColor	= SetBkColor(hdcTemp, cTransparentColor);

	// Create the object mask for the bitmap by performing a BitBlt
	// from the source bitmap to a monochrome bitmap.
	BitBlt(hdcObject, 0, 0, ptSize.x, ptSize.y, hdcTemp, 0, 0,
		SRCCOPY);

	// Set the background color of the source DC back to the original
	// color.
	SetBkColor(hdcTemp, cColor);

	// Create the inverse of the object mask.
	BitBlt(hdcBack, 0, 0, ptSize.x, ptSize.y, hdcObject, 0, 0,
		NOTSRCCOPY);

	// Copy the background of the main DC to the destination.
	BitBlt(hdcMem, 0, 0, ptSize.x, ptSize.y, dc, xStart, yStart,
		SRCCOPY);

	// Mask out the places where the bitmap will be placed.
	BitBlt(hdcMem, 0, 0, ptSize.x, ptSize.y, hdcObject, 0, 0, SRCAND);

	// Mask out the transparent colored pixels on the bitmap.
	BitBlt(hdcTemp, 0, 0, ptSize.x, ptSize.y, hdcBack, 0, 0, SRCAND);

	// XOR the bitmap with the background on the destination DC.
	BitBlt(hdcMem, 0, 0, ptSize.x, ptSize.y, hdcTemp, 0, 0, SRCPAINT);

	// Copy the destination to the screen.
	BitBlt(dc, xStart, yStart, ptSize.x, ptSize.y, hdcMem, 0, 0,
		SRCCOPY);

	// Place the original bitmap back into the bitmap sent here.
	BitBlt(hdcTemp, 0, 0, ptSize.x, ptSize.y, hdcSave, 0, 0, SRCCOPY);

	// Delete the memory bitmaps.
	DeleteObject(SelectObject(hdcBack, bmBackOld));
	DeleteObject(SelectObject(hdcObject, bmObjectOld));
	DeleteObject(SelectObject(hdcMem, bmMemOld));
	DeleteObject(SelectObject(hdcSave, bmSaveOld));

	// Delete the memory DCs.
	DeleteDC(hdcMem);
	DeleteDC(hdcBack);
	DeleteDC(hdcObject);
	DeleteDC(hdcSave);
	DeleteDC(hdcTemp);
}

//////////////////////////////////////////////////////////////////////

/*static*/extern struct ReposItem ShowTextCtlPosData[]/*={
	{ IDC_CHANGES		, REPOS_SIZERIGHT | REPOS_SIZEDOWN },
	{ IDOK				, REPOS_MOVERIGHT | REPOS_MOVEDOWN },
	{ 0 }
}*/;

extern POINT ShowTextCtlPos[2];

BOOL APIENTRY ShowStreamsReportDlgProc( HWND hDlg, UINT message, UINT wParam, LONG lParam) {
	static RECT rInitial;

    switch (message) {
		case WM_SIZE:
			guiReposResize(hDlg, ShowTextCtlPosData, ShowTextCtlPos);
			return TRUE;

		case WM_GETMINMAXINFO:
			{
				LPMINMAXINFO lpmmi = (LPMINMAXINFO)lParam;

				lpmmi->ptMinTrackSize.x = rInitial.right - rInitial.left;
				lpmmi->ptMinTrackSize.y = rInitial.bottom - rInitial.top;
			}
			return TRUE;

		case WM_INITDIALOG:
			{
				GetWindowRect(hDlg, &rInitial);
				VDStringW report = inputAVI->writeReport();
				ModList<ModStream>::iterator it(g_Streams);
				while(it) {
					if(!(*it)->m_bInputFile) {
						report += L"\r\n\r\n---== External stream ==---\r\n";
						report += (*it)->m_Source->writeReport();
						//report += L"\r\n---=====================---";
					}
					it++;
				}
				SendMessage(hDlg, WM_SETICON, (WPARAM)ICON_SMALL, (LPARAM)LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_VIRTUALDUB)));
				SendMessage(hDlg, WM_SETTEXT, 0, (LPARAM)_T("Report"));
				SendMessage(GetDlgItem(hDlg, IDC_CHANGES), WM_SETFONT, (WPARAM)GetStockObject(ANSI_FIXED_FONT), MAKELPARAM(TRUE, 0));
				if(IsWindowsNT())
					SendMessageW(GetDlgItem(hDlg, IDC_CHANGES), WM_SETTEXT, 0, (LPARAM)report.c_str());
				else
					SendMessageA(GetDlgItem(hDlg, IDC_CHANGES), WM_SETTEXT, 0, (LPARAM)VDFastTextWToA(report.c_str()));
				guiReposInit(hDlg, ShowTextCtlPosData, ShowTextCtlPos);
			}
			return FALSE;

		case WM_COMMAND:
			if (LOWORD(wParam) == IDOK || LOWORD(wParam) == IDCANCEL) {
				EndDialog(hDlg, TRUE);  
				return TRUE;
			}
			break;
	}
	return FALSE;
}

//////////////////////////////////////////////////////////////////////

class ModDialogStreamCommentsW32 : public VDDialogBaseW32 {
public:
	inline ModDialogStreamCommentsW32(ModList<ModStreamComment> &comments)
		: VDDialogBaseW32(g_bUseUnicodeWinAPI ? IDD_STREAM_COMMENTSW : IDD_STREAM_COMMENTSA)
		, m_Comments(comments) {}

	LRESULT Activate(VDGUIHandle hParent) {
		return ActivateDialog(hParent);
	};

protected:
	ModList<ModStreamComment> &m_Comments;
	HWND	m_hListBox;
	BOOL	DlgProc(UINT msg, UINT wParam, LONG lParam);
	void	UpdateListBoxHorizontalExtent(void);
	sint32	PrintCommentInfo(ModStreamComment *comment, HDC hDC, RECT rcItem, bool bGetHorizontalExtent);
	void	UpdateLanguageList(void);
};

void ModDisplayStreamCommentsDialog(VDGUIHandle hwndParent, ModList<ModStreamComment>& comments) {
	ModList<ModStreamComment> tmpComments;
	CopyModList<ModStreamComment>(comments, tmpComments);
	ModDialogStreamCommentsW32 dlg(tmpComments);
	if(dlg.Activate(hwndParent)) {
		comments.DeleteContents(true);
		comments.Clear();
		CopyModList<ModStreamComment>(tmpComments, comments);
	}
}

static LRESULT CALLBACK ListHookMouseProc(int nCode, WPARAM wParam, LPARAM lParam) {
	if(nCode == HC_ACTION) {
		switch(wParam) {
		case WM_LBUTTONDOWN:
			{
				RECT listRect;
				RECT itemRect;
				MOUSEHOOKSTRUCT *pStruct = (MOUSEHOOKSTRUCT *)lParam;
				if(pStruct->hwnd != g_hListBox)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				GetWindowRect(g_hListBox, &listRect);
				long relX = pStruct->pt.x - listRect.left;
				long relY = pStruct->pt.y - listRect.top;
				int index = SendMessage(g_hListBox, LB_ITEMFROMPOINT, 0, MAKEWPARAM(relX, relY));
				if(SendMessage(g_hListBox, LB_GETITEMRECT, index, (WPARAM)&itemRect) == LB_ERR)
					return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
				if( (relX < itemRect.left) || (relX > itemRect.right)
				|| (relY < itemRect.top) || (relY > itemRect.bottom) ) {
					SendMessage(g_hListBox, LB_SETCURSEL, -1, 0);
					RedrawWindow(g_hDlg, NULL, NULL, RDW_INTERNALPAINT | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
					// Inform the window that we unselected the item (so that it can update
					// its state if necessary)
					SendMessage(g_hDlg, WM_COMMAND, MAKEWPARAM(GetDlgCtrlID(g_hListBox), LBN_SELCHANGE), (LPARAM)g_hListBox);
					return TRUE;
				}
			}
		}
	}
	return CallNextHookEx(g_MouseHook, nCode, wParam, lParam);
}

static struct ReposItem CommentListCtlPosData[]={
	{ IDC_COMMENT_LIST		, REPOS_SIZERIGHT | REPOS_SIZEDOWN },
	{ IDOK					, REPOS_MOVERIGHT },
	{ IDCANCEL				, REPOS_MOVERIGHT },
	{ IDC_MOVEUP			, REPOS_MOVERIGHT },
	{ IDC_MOVEDOWN			, REPOS_MOVERIGHT },
	{ IDC_DELETE			, REPOS_MOVERIGHT },
	{ IDC_ADD				, REPOS_MOVERIGHT },
	{ IDC_EDIT2				, REPOS_MOVERIGHT },
	{ IDC_COMMENT_TYPE		, REPOS_MOVEDOWN },
	{ IDC_COMMENT_LANGUAGE	, REPOS_MOVEDOWN },
	{ IDC_FULL_LANGUAGE_LIST, REPOS_MOVEDOWN },
	{ IDC_STATIC_1			, REPOS_MOVEDOWN },
	{ IDC_COMMENT_TAG		, REPOS_MOVEDOWN },
	{ IDC_STATIC_2			, REPOS_MOVEDOWN },
	{ IDC_COMMENT_VALUE		, REPOS_MOVEDOWN | REPOS_SIZERIGHT },
	{ 0 }
};

POINT CommentListCtlPos[15];

void ModDialogStreamCommentsW32::UpdateLanguageList(void) {
	bool bFullLanguageList = !!IsDlgButtonChecked(mhdlg, IDC_FULL_LANGUAGE_LIST);
	HWND hComboLang = GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE);
	int langIndex = SendMessage(hComboLang, CB_GETCURSEL, 0, 0);
	if(langIndex != CB_ERR)
		langIndex = SendMessage(hComboLang, CB_GETITEMDATA, langIndex, 0);
	int newIndex = 0;
	SendMessage(hComboLang, CB_RESETCONTENT, 0, 0);
	int index = SendMessage(hComboLang, CB_ADDSTRING, 0, (LPARAM)_T("User defined"));
	if( (index!=CB_ERR) && (index!=CB_ERRSPACE) )
		SendMessage(hComboLang, CB_SETITEMDATA, index, (LPARAM)0);
	else
		throw MyError("Error initializing Language list");
	VDStringW sLanguage;
	int i = 0;
	for(i=0;
		s_LanguageList[i].m_EnglishName || s_LanguageList[i].m_ISO639_1_Code || s_LanguageList[i].m_ISO639_2_Code;
		i++)
	{
		if(!s_LanguageList[i].m_EnglishName)
			continue;
		if(!s_LanguageList[i].m_ISO639_1_Code && !bFullLanguageList)
			continue;
		sLanguage = s_LanguageList[i].m_EnglishName;
		if(g_bUseUnicodeWinAPI)
			index = SendMessageW(hComboLang, CB_ADDSTRING, 0, (LPARAM)sLanguage.c_str());
		else
			index = SendMessageA(hComboLang, CB_ADDSTRING, 0, (LPARAM)VDTextWToA(sLanguage).c_str());
		if( (index!=CB_ERR) && (index!=CB_ERRSPACE) )
			SendMessage(hComboLang, CB_SETITEMDATA, index, (LPARAM)i+1);
		else
			throw MyError("Error initializing Language list");
		if(langIndex == i+1)
			newIndex = index;
	}
	SendMessage(hComboLang, CB_SETCURSEL, newIndex, 0);
}

BOOL ModDialogStreamCommentsW32::DlgProc(UINT msg, UINT wParam, LONG lParam) {
	using namespace nsModStream;
	TCHAR tchBuffer[1024];
	char szBufferA[1024];
	wchar_t szBufferW[1024];
	TEXTMETRIC tm;
	int y;
	LPMEASUREITEMSTRUCT lpmis;
	LPDRAWITEMSTRUCT lpdis;
	static RECT rInitial;
	static BOOL returnValue;

	switch(msg) {
	case WM_INITDIALOG:
		{
			returnValue = FALSE;
			g_hDlg = mhdlg;
			GetWindowRect(mhdlg, &rInitial);
			g_hListBox = m_hListBox = GetDlgItem(mhdlg, IDC_COMMENT_LIST);
			HWND hComboBox = GetDlgItem(mhdlg, IDC_COMMENT_TYPE);
			VDStringW sTypeName;
			int i=0, index;
			while((s_CommentTypes[i].m_Type!=cVoid) || s_CommentTypes[i].m_Name || s_CommentTypes[i].m_Tip) {
				if(!s_CommentTypes[i].m_bCanUse) {
					i++;
					continue;
				}
				eCommentType sType = s_CommentTypes[i].m_Type;
				sTypeName = GetCommentName(s_CommentTypes[i].m_Type);
				if(g_bUseUnicodeWinAPI)
					index = SendMessageW(hComboBox, CB_ADDSTRING, 0, (LPARAM)sTypeName.c_str());
				else
					index = SendMessageA(hComboBox, CB_ADDSTRING, 0, (LPARAM)VDTextWToA(sTypeName).c_str());
				if( (index!=CB_ERR) && (index!=CB_ERRSPACE) )
					SendMessage(hComboBox, CB_SETITEMDATA, index, sType);
				else
					throw MyError("Error initializing Comment type list");
				i++;
			}
			UpdateLanguageList();
			SendMessage(hComboBox, CB_SETCURSEL, 0, 0);
			InstallMouseHook(ListHookMouseProc, mhdlg, IDC_COMMENT_LIST);
			SendMessage(mhdlg, WM_SETICON, (WPARAM)ICON_SMALL, (LPARAM)LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_VIRTUALDUB)));
			RefreshList<ModStreamComment>(m_hListBox, m_Comments, -1);
			UpdateListBoxHorizontalExtent();
			guiReposInit(mhdlg, CommentListCtlPosData, CommentListCtlPos);
		}
		return TRUE;

	case WM_SIZE:
		guiReposResize(mhdlg, CommentListCtlPosData, CommentListCtlPos);
		return TRUE;

	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO lpmmi = (LPMINMAXINFO)lParam;

			lpmmi->ptMinTrackSize.x = rInitial.right - rInitial.left;
			lpmmi->ptMinTrackSize.y = rInitial.bottom - rInitial.top;
		}
		return TRUE;

	case WM_MEASUREITEM:
		{
			if(wParam != IDC_COMMENT_LIST)
				return FALSE;
			lpmis = (LPMEASUREITEMSTRUCT)lParam;
			HDC hDC = GetDC(m_hListBox);
			HFONT hFont = (HFONT)SendMessage(m_hListBox, WM_GETFONT, 0, 0);
			if(!hFont)
				hFont = (HFONT)GetStockObject(SYSTEM_FONT);
			HGDIOBJ hOld = SelectObject(hDC, hFont);
			GetTextMetrics(hDC, &tm);
			SelectObject(hDC, hOld);
			lpmis->itemHeight = 2 * tm.tmHeight + 3 * 2;
			ReleaseDC(m_hListBox, hDC);
		}
		return TRUE;

	case WM_DRAWITEM:
		if(wParam != IDC_COMMENT_LIST)
			return FALSE;
		lpdis = (LPDRAWITEMSTRUCT) lParam;
		if(lpdis->itemID == -1)
			break;
		switch (lpdis->itemAction) {
		case ODA_SELECT:
		case ODA_DRAWENTIRE:
		case ODA_FOCUS:
			{
				HGDIOBJ hOld;
				ModStreamComment *comment = m_Comments[lpdis->itemID];
				COLORREF color;
				HBRUSH brush;
				if(lpdis->itemState & ODS_SELECTED)
					color = RGB(0xE0, 0xE0, 0xE0);
				else
					color = RGB(0xFF, 0xFF, 0xFF);
				brush = CreateSolidBrush(color);
				FillRect(lpdis->hDC, &lpdis->rcItem, brush);
				DeleteObject(brush);
				SetBkColor(lpdis->hDC, color);
				GetTextMetrics(lpdis->hDC, &tm);
				y = (lpdis->rcItem.top + lpdis->rcItem.bottom - tm.tmHeight) / 2;
				_sntprintf(tchBuffer, sizeof(tchBuffer)-1, _T("%u"), lpdis->itemID+1);
				TextOut(lpdis->hDC, LEFT_INDEX, y, tchBuffer, _tcslen(tchBuffer));
				PrintCommentInfo(comment, lpdis->hDC, lpdis->rcItem, false);
				hOld = SelectObject(lpdis->hDC, CreatePen(PS_SOLID, 2, RGB(0,0,0)));
				MoveToEx(lpdis->hDC, lpdis->rcItem.left, lpdis->rcItem.top, NULL);
				LineTo(lpdis->hDC, lpdis->rcItem.right, lpdis->rcItem.top);
				MoveToEx(lpdis->hDC, lpdis->rcItem.left, lpdis->rcItem.bottom, NULL);
				LineTo(lpdis->hDC, lpdis->rcItem.right, lpdis->rcItem.bottom);
				DeleteObject(SelectObject(lpdis->hDC, CreatePen(PS_DOT, 1, RGB(0xA0,0xA0,0xA0))));
				MoveToEx(lpdis->hDC, LEFT_FIRSTLINE, lpdis->rcItem.top, NULL);
				LineTo(lpdis->hDC, LEFT_FIRSTLINE, lpdis->rcItem.bottom);
				DeleteObject(SelectObject(lpdis->hDC, hOld));
			}
			break;
		}
	return TRUE;

	case WM_VKEYTOITEM:
		{
			if((HWND)lParam == m_hListBox) {
				if(LOWORD(wParam) == VK_DELETE) {
					SendMessage(mhdlg, WM_COMMAND, IDC_DELETE, (LPARAM)GetDlgItem(mhdlg, IDC_DELETE));
					return -2;	// Tell the listbox we did the necessary handling
				}
			}
			return -1; // Tell the listbox to handle the keystroke as usual
		}
		break;

	case WM_COMMAND:
		if(m_hListBox == (HWND)lParam) {
			if(HIWORD(wParam) == LBN_SELCHANGE) {
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=m_Comments.GetCount()) )
					return TRUE;
				ModStreamComment *comment = m_Comments[index];
				if(comment) {
					SendDlgItemMessage(mhdlg, IDC_COMMENT_TAG, EM_SETREADONLY, CanEditCommentTag(comment->m_Type) ? FALSE : TRUE, 0);
					SendDlgItemMessage(mhdlg, IDC_COMMENT_VALUE, EM_SETREADONLY, CanEditCommentValue(comment->m_Type) ? FALSE : TRUE, 0);
					HWND hComboBox = GetDlgItem(mhdlg, IDC_COMMENT_TYPE);
					int nbTags = SendMessage(hComboBox, CB_GETCOUNT, 0, 0);
					if(nbTags != CB_ERR) {
						eCommentType cType;
						for(int i=0 ; i<nbTags ; i++) {
							cType = (eCommentType)SendMessage(hComboBox, CB_GETITEMDATA, i, 0);
							if((comment->m_Type == cType) || ((comment->m_Type>cLanguage) && (cType==cLanguage)) ) {
								SendMessage(hComboBox, CB_SETCURSEL, i, 0);
								break;
							}
						}
						// Verify that this is a valid language
						if((cType == cLanguage) && ((comment->m_Type==cLanguage) || ISO639_IsValidIndex_2(comment->m_Type-cLanguage-1)) ) {
							int tableIndex = comment->m_Type-cLanguage-1;
							if((tableIndex >= 0) && !ISO639_IsValidIndex_1(tableIndex) && !IsDlgButtonChecked(mhdlg, IDC_FULL_LANGUAGE_LIST)) {
								CheckDlgButton(mhdlg, IDC_FULL_LANGUAGE_LIST, BST_CHECKED);
								UpdateLanguageList();
							}
							ShowWindow(GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE), SW_SHOW);
							ShowWindow(GetDlgItem(mhdlg, IDC_FULL_LANGUAGE_LIST), SW_SHOW);
							if(tableIndex >= 0) {
								HWND hComboLang = GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE);
								tableIndex = ISO639_GetRootIndex(tableIndex);
								int nbLangTags = SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETCOUNT, 0, 0);
								if(nbLangTags != CB_ERR) {
									while(nbLangTags--) {
										int langData = SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETITEMDATA, nbLangTags, 0);
										if( (langData != CB_ERR) && (langData-1 == tableIndex) ) {
											SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_SETCURSEL, nbLangTags, 0);
											break;
										}
									}
									if(nbLangTags < 0)
										tableIndex = -1;
								} else
									tableIndex = -1;
							}
							if(tableIndex < 0)
								SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_SETCURSEL, 0, 0);
						} else if(comment->m_Type != cType) {
							ShowWindow(GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE), SW_HIDE);
							ShowWindow(GetDlgItem(mhdlg, IDC_FULL_LANGUAGE_LIST), SW_HIDE);
							SendMessage(hComboBox, CB_SETCURSEL, 0, 0);
						} else {
							ShowWindow(GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE), SW_HIDE);
							ShowWindow(GetDlgItem(mhdlg, IDC_FULL_LANGUAGE_LIST), SW_HIDE);
						}
					} else
						SendMessage(hComboBox, CB_SETCURSEL, 0, 0);
					if(g_bUseUnicodeWinAPI) {
						SetDlgItemTextW(mhdlg, IDC_COMMENT_TAG, comment->m_Tag.c_str());
						SetDlgItemTextW(mhdlg, IDC_COMMENT_VALUE, comment->m_Value.c_str());
					} else {
						SetDlgItemTextA(mhdlg, IDC_COMMENT_TAG, VDTextWToA(comment->m_Tag).c_str());
						SetDlgItemTextA(mhdlg, IDC_COMMENT_VALUE, VDTextWToA(comment->m_Value).c_str());
					}
				}
				return TRUE;
			}
		} else if((HWND)lParam == GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE)) {
			if(HIWORD(wParam) == CBN_SELCHANGE) {
				int index = SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETCURSEL, 0, 0);
				if(index != CB_ERR) {
					int langIndex = (int)SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETITEMDATA, index, 0);
					if(langIndex == 0) {
						SendDlgItemMessage(mhdlg, IDC_COMMENT_TAG, EM_SETREADONLY, TRUE, 0);
						SendDlgItemMessage(mhdlg, IDC_COMMENT_VALUE, EM_SETREADONLY, FALSE, 0);
					} else if((langIndex!=CB_ERR) && (langIndex > 0)) {
						langIndex--;
						SendDlgItemMessage(mhdlg, IDC_COMMENT_TAG, EM_SETREADONLY, TRUE, 0);
						SendDlgItemMessage(mhdlg, IDC_COMMENT_VALUE, EM_SETREADONLY, TRUE, 0);
						VDStringW sValue(s_LanguageList[langIndex].m_EnglishName);
						sValue += L"[";
						sValue += s_LanguageList[langIndex].m_ISO639_2_Code;
						if(s_LanguageList[langIndex].m_ISO639_1_Code) {
							sValue += L"/";
							sValue += s_LanguageList[langIndex].m_ISO639_1_Code;
						}
						sValue += L"]";
						if(g_bUseUnicodeWinAPI)
							SetDlgItemTextW(mhdlg, IDC_COMMENT_VALUE, sValue.c_str());
						else
							SetDlgItemTextA(mhdlg, IDC_COMMENT_VALUE, VDTextWToA(sValue).c_str());
					}
				}
			}
		} else if((HWND)lParam == GetDlgItem(mhdlg, IDC_COMMENT_TYPE)) {
			if(HIWORD(wParam) == CBN_SELCHANGE) {
				int index = SendDlgItemMessage(mhdlg, IDC_COMMENT_TYPE, CB_GETCURSEL, 0, 0);
				if(index != CB_ERR) {
					eCommentType cType = (eCommentType)SendDlgItemMessage(mhdlg, IDC_COMMENT_TYPE, CB_GETITEMDATA, index, 0);
					SendDlgItemMessage(mhdlg, IDC_COMMENT_TAG, EM_SETREADONLY, CanEditCommentTag(cType) ? FALSE : TRUE, 0);
					SendDlgItemMessage(mhdlg, IDC_COMMENT_VALUE, EM_SETREADONLY, CanEditCommentValue(cType) ? FALSE : TRUE, 0);
					if(cType == cLanguage) {
						ShowWindow(GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE), SW_SHOW);
						ShowWindow(GetDlgItem(mhdlg, IDC_FULL_LANGUAGE_LIST), SW_SHOW);
					} else {
						ShowWindow(GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE), SW_HIDE);
						ShowWindow(GetDlgItem(mhdlg, IDC_FULL_LANGUAGE_LIST), SW_HIDE);
					}
				}
			}
		} else if((HWND)lParam == GetDlgItem(mhdlg, IDC_FULL_LANGUAGE_LIST)) {
			UpdateLanguageList();
		} else
		switch(LOWORD(wParam)) {
		case IDC_ADD:
			{
				int iTag, iValue;
				VDStringW sTag;
				VDStringW sValue;
				if(g_bUseUnicodeWinAPI) {
					iTag = GetDlgItemTextW(mhdlg, IDC_COMMENT_TAG, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					sTag = szBufferW;
					iValue = GetDlgItemTextW(mhdlg, IDC_COMMENT_VALUE, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					sValue = szBufferW;
				} else {
					iTag = GetDlgItemTextA(mhdlg, IDC_COMMENT_TAG, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					sTag = VDTextAToW(szBufferA, -1);
					iValue = GetDlgItemTextA(mhdlg, IDC_COMMENT_VALUE, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					sValue = VDTextAToW(szBufferA, -1);
				}
				if( (iTag<=0) && (iValue<=0) )
					return TRUE;
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				HWND hComboBox = GetDlgItem(mhdlg, IDC_COMMENT_TYPE);
				int indexType = SendMessage(hComboBox, CB_GETCURSEL, 0, 0);
				ModStreamComment *newComment = new ModStreamComment();
				newComment->m_Tag = sTag;
				newComment->m_Value = sValue;
				if(indexType != CB_ERR)
					newComment->m_Type = (eCommentType)SendMessage(hComboBox, CB_GETITEMDATA, indexType, 0);
				else
					newComment->m_Type = cGeneric;
				if(newComment->m_Type == cLanguage) {
					int langIndex = SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETCURSEL, 0, 0);
					if(langIndex != CB_ERR)
						langIndex = SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETITEMDATA, langIndex, 0);
					if(langIndex != CB_ERR)
						newComment->m_Type = (eCommentType)(cLanguage + langIndex);
				}
				if(index < 0)
					m_Comments.Append(newComment);
				else
					m_Comments.Insert(newComment, index);
				SetDlgItemText(mhdlg, IDC_COMMENT_TAG, _T(""));
				SetDlgItemText(mhdlg, IDC_COMMENT_VALUE, _T(""));
				if(newComment->m_Type >= cLanguage) {
					SendDlgItemMessage(mhdlg, IDC_COMMENT_TAG, EM_SETREADONLY, FALSE, 0);
					SendDlgItemMessage(mhdlg, IDC_COMMENT_VALUE, EM_SETREADONLY, FALSE, 0);
					ShowWindow(GetDlgItem(mhdlg, IDC_COMMENT_LANGUAGE), SW_HIDE);
					ShowWindow(GetDlgItem(mhdlg, IDC_FULL_LANGUAGE_LIST), SW_HIDE);
					SendMessage(hComboBox, CB_SETCURSEL, 0, 0);
				}
				if( (InsertListBoxItem(m_hListBox, _T(""), (LPARAM)newComment, index)!=index) && (index!=-1))
					RefreshList<ModStreamComment>(m_hListBox, m_Comments, index);
				UpdateListBoxHorizontalExtent();
			}
			return TRUE;

		case IDC_DELETE:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=m_Comments.GetCount()) )
					return TRUE;
				ModStreamComment *comment = m_Comments[index];
				if(comment) {
					m_Comments.DeleteNode(comment);
					if(SendMessage(m_hListBox, LB_DELETESTRING, index, 0) == LB_ERR)
						RefreshList<ModStreamComment>(m_hListBox, m_Comments, index);
					else {
						if(index == m_Comments.GetCount())
							index--;
						SendMessage(m_hListBox, LB_SETCURSEL, index, 0);
					}
					UpdateListBoxHorizontalExtent();
				}
			}
			return TRUE;

		case IDC_MOVEUP:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if(index<=0)
					return TRUE;
				ModStreamComment *previous = m_Comments[index-1];
				ModStreamComment *current = m_Comments[index];
				if(previous && current) {
					m_Comments.Swap(previous, current);
					if( (SendMessage(m_hListBox, LB_SETITEMDATA, index-1, (LPARAM)current)==LB_ERR)
						|| (SendMessage(m_hListBox, LB_SETITEMDATA, index, (LPARAM)previous)==LB_ERR) )
						RefreshList<ModStreamComment>(m_hListBox, m_Comments, index-1);
					else
						SendMessage(m_hListBox, LB_SETCURSEL, index-1, 0);
				}
			}
			return TRUE;

		case IDC_MOVEDOWN:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=m_Comments.GetCount()-1) )
					return TRUE;
				ModStreamComment *current = m_Comments[index];
				ModStreamComment *next = m_Comments[index+1];
				if(current && next) {
					m_Comments.Swap(current, next);
					if( (SendMessage(m_hListBox, LB_SETITEMDATA, index, (LPARAM)next)==LB_ERR)
						|| (SendMessage(m_hListBox, LB_SETITEMDATA, index+1, (LPARAM)current)==LB_ERR) )
						RefreshList<ModStreamComment>(m_hListBox, m_Comments, index+1);
					else
						SendMessage(m_hListBox, LB_SETCURSEL, index+1, 0);
				}
			}
			return TRUE;

		case IDC_EDIT2:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=m_Comments.GetCount()) )
					return TRUE;
				int iTag, iValue;
				VDStringW sTag;
				VDStringW sValue;
				if(g_bUseUnicodeWinAPI) {
					iTag = GetDlgItemTextW(mhdlg, IDC_COMMENT_TAG, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					sTag = szBufferW;
					iValue = GetDlgItemTextW(mhdlg, IDC_COMMENT_VALUE, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					sValue = szBufferW;
				} else {
					iTag = GetDlgItemTextA(mhdlg, IDC_COMMENT_TAG, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					sTag = VDTextAToW(szBufferA, -1);
					iValue = GetDlgItemTextA(mhdlg, IDC_COMMENT_VALUE, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					sValue = VDTextAToW(szBufferA, -1);
				}
				if( (iTag<=0) && (iValue<=0) )
					return TRUE;
				HWND hComboBox = GetDlgItem(mhdlg, IDC_COMMENT_TYPE);
				int indexType = SendMessage(hComboBox, CB_GETCURSEL, 0, 0);
				ModStreamComment *comment = m_Comments[index];
				if(!comment)
					return TRUE;
				comment->m_Tag = sTag;
				comment->m_Value = sValue;
				if(indexType != CB_ERR)
					comment->m_Type = (eCommentType)SendMessage(hComboBox, CB_GETITEMDATA, indexType, 0);
				else
					comment->m_Type = cGeneric;
				if(comment->m_Type == cLanguage) {
					int langIndex = SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETCURSEL, 0, 0);
					if(langIndex != CB_ERR)
						langIndex = SendDlgItemMessage(mhdlg, IDC_COMMENT_LANGUAGE, CB_GETITEMDATA, langIndex, 0);
					if(langIndex != CB_ERR)
						comment->m_Type = (eCommentType)(cLanguage + langIndex);
				}
				UpdateListBoxHorizontalExtent();
				RedrawWindow(mhdlg, NULL, NULL, RDW_INTERNALPAINT | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
			}
			return TRUE;

		case IDOK:
			returnValue = TRUE;
		case IDCANCEL:
			InstallMouseHook(NULL, NULL, NULL);
			End(returnValue);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

sint32 ModDialogStreamCommentsW32::PrintCommentInfo(ModStreamComment *comment, HDC hDC, RECT rcItem, bool bGetHorizontalExtent) {
	TEXTMETRIC tm;
	HGDIOBJ hOld;
	int y;
	sint32 horExtent = LEFT_FIRSTLINE;

	if(bGetHorizontalExtent) {
		HFONT hFont = (HFONT)SendMessage(m_hListBox, WM_GETFONT, 0, 0);
		if(!hFont)
			hFont = (HFONT)GetStockObject(SYSTEM_FONT);
		hOld = SelectObject(hDC, hFont);
	}
	GetTextMetrics(hDC, &tm);
	if(comment) {
		nsModStream::eCommentType cType = comment->m_Type;
		VDStringW cTag(L"Tag : [");
		cTag += comment->m_Tag;
		cTag += L"]";
		VDStringW cValue(L"Value : [");
		cValue += comment->m_Value;
		cValue += L"]";
		VDStringW cTypeName(L"Type : ");
		cTypeName += nsModStream::GetCommentName(cType);
		cTypeName += L"  (";
		cTypeName += cTag;
		cTypeName += L")";
		//VDStringW cTypeTip(nsModStream::GetCommentTip(cType));
		y = rcItem.top;
		VDStringW cDesc;
		if(cType == nsModStream::cGeneric)
			cDesc = cTag;
		else
			cDesc = cTypeName;
		if(bGetHorizontalExtent) {
			GetHorizontalExtent(cDesc, hDC, horExtent, LEFT_FIRSTLINE+5 + 5);
			GetHorizontalExtent(cValue, hDC, horExtent, LEFT_OTHERLINE+5 + 5);
		} else {
			if(g_bUseUnicodeWinAPI) {
				TextOutW(hDC, LEFT_FIRSTLINE+5, y + 2, cDesc.c_str(), cDesc.length());
				TextOutW(hDC, LEFT_OTHERLINE+5, y + 4 + tm.tmHeight, cValue.c_str(), cValue.length());
			} else {
				VDStringA cDescA = VDTextWToA(cDesc);
				VDStringA cValueA = VDTextWToA(cValue);
				TextOutA(hDC, LEFT_FIRSTLINE+5, y + 2, cDescA.c_str(), cDescA.length());
				TextOutA(hDC, LEFT_OTHERLINE+5, y + 4 + tm.tmHeight, cValueA.c_str(), cValueA.length());
			}
		}
	} else {
		y = (rcItem.top + rcItem.bottom - tm.tmHeight) / 2;
		if(bGetHorizontalExtent)
			GetHorizontalExtent(VDStringW(L"Could not retrieve comment"), hDC, horExtent, LEFT_FIRSTLINE+5 + 5);
		else
			TextOut(hDC, LEFT_FIRSTLINE+5, y, _T("Could not retrieve comment"), _tcslen(_T("Could not retrieve comment")));
	}
	if(bGetHorizontalExtent) {
		SelectObject(hDC, hOld);
		return horExtent;
	}
	return -1;
}

void ModDialogStreamCommentsW32::UpdateListBoxHorizontalExtent() {
	HDC hDC = GetDC(m_hListBox);
	sint32 horExtent = 0;
	sint32 tmpExtent = 0;
	RECT rcItem = { 0 };
	ModList<ModStreamComment>::iterator it(m_Comments);
	while(it) {
		ModStreamComment *comment = *it;
		tmpExtent = PrintCommentInfo(comment, hDC, rcItem, true);
		if(tmpExtent > horExtent)
			horExtent = tmpExtent;
		it++;
	}
	SendMessage(m_hListBox, LB_SETHORIZONTALEXTENT, horExtent, 0);
	ReleaseDC(m_hListBox, hDC);
}

//////////////////////////////////////////////////////////////////////

class ModDialogStreamChaptersW32 : public VDDialogBaseW32 {
public:
	inline ModDialogStreamChaptersW32(ModChapterList& chapters)
		: VDDialogBaseW32(g_bUseUnicodeWinAPI ? IDD_STREAM_CHAPTERSW : IDD_STREAM_CHAPTERSA)
		, m_Chapters(chapters) {}

	LRESULT Activate(VDGUIHandle hParent) {
		return ActivateDialog(hParent);
	};

protected:
	ModChapterList &m_Chapters;
	HWND	m_hListBox;
	BOOL	DlgProc(UINT msg, UINT wParam, LONG lParam);
	void	UpdateListBoxHorizontalExtent(void);
	sint32	PrintChapterInfo(ModChapterList::tChapterNode *chapter, HDC hDC, RECT rcItem, bool bGetHorizontalExtent);
};

void ModDisplayStreamChaptersDialog(VDGUIHandle hwndParent, ModChapterList& chapters) {
	chapters.Sort();
	ModChapterList tmpChapters;
	tmpChapters = chapters;
	ModDialogStreamChaptersW32 dlg(tmpChapters);
	if(dlg.Activate(hwndParent)) {
		chapters.clear();
		chapters = tmpChapters;
	}
}

static struct ReposItem ChapterListCtlPosData[]={
	{ IDC_CHAPTER_LIST		, REPOS_SIZERIGHT | REPOS_SIZEDOWN },
	{ IDOK					, REPOS_MOVERIGHT },
	{ IDCANCEL				, REPOS_MOVERIGHT },
	{ IDC_IMPORT			, REPOS_MOVERIGHT },
	{ IDC_DELETE			, REPOS_MOVERIGHT },
	{ IDC_ADD				, REPOS_MOVERIGHT },
	{ IDC_EDIT2				, REPOS_MOVERIGHT },
	{ IDC_STATIC_1			, REPOS_MOVEDOWN },
	{ IDC_CHAPTER_TIME		, REPOS_MOVEDOWN },
	{ IDC_STATIC_2			, REPOS_MOVEDOWN },
	{ IDC_CHAPTER_NAME		, REPOS_MOVEDOWN | REPOS_SIZERIGHT },
	{ 0 }
};

POINT ChapterListCtlPos[11];

BOOL ModDialogStreamChaptersW32::DlgProc(UINT msg, UINT wParam, LONG lParam) {
	TCHAR tchBuffer[1024];
	char szBufferA[1024];
	wchar_t szBufferW[1024];
	TEXTMETRIC tm;
	int y;
	LPMEASUREITEMSTRUCT lpmis;
	LPDRAWITEMSTRUCT lpdis;
	static RECT rInitial;
	static BOOL returnValue;

	switch(msg) {
	case WM_INITDIALOG:
		{
			returnValue = FALSE;
			GetWindowRect(mhdlg, &rInitial);
			g_hListBox = m_hListBox = GetDlgItem(mhdlg, IDC_CHAPTER_LIST);
			SetDlgItemText(mhdlg, IDC_CHAPTER_TIME, _T("00:00:00.000"));
			InstallMouseHook(ListHookMouseProc, mhdlg, IDC_CHAPTER_LIST);
			SendMessage(mhdlg, WM_SETICON, (WPARAM)ICON_SMALL, (LPARAM)LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_VIRTUALDUB)));
			RefreshListData<ModChapterList::tChapterNode>(m_hListBox, m_Chapters.GetPtrs(), -1);
			UpdateListBoxHorizontalExtent();
			CheckDlgButton(mhdlg, IDC_ADJUST_CHAPTERS, m_Chapters.GetAdjustChapters());
			guiReposInit(mhdlg, ChapterListCtlPosData, ChapterListCtlPos);
		}
		return TRUE;

	case WM_SIZE:
		guiReposResize(mhdlg, ChapterListCtlPosData, ChapterListCtlPos);
		return TRUE;

	case WM_GETMINMAXINFO:
		{
			LPMINMAXINFO lpmmi = (LPMINMAXINFO)lParam;

			lpmmi->ptMinTrackSize.x = rInitial.right - rInitial.left;
			lpmmi->ptMinTrackSize.y = rInitial.bottom - rInitial.top;
		}
		return TRUE;

	case WM_MEASUREITEM:
		{
			if(wParam != IDC_CHAPTER_LIST)
				return FALSE;
			lpmis = (LPMEASUREITEMSTRUCT)lParam;
			HDC hDC = GetDC(m_hListBox);
			HFONT hFont = (HFONT)SendMessage(m_hListBox, WM_GETFONT, 0, 0);
			if(!hFont)
				hFont = (HFONT)GetStockObject(SYSTEM_FONT);
			HGDIOBJ hOld = SelectObject(hDC, hFont);
			GetTextMetrics(hDC, &tm);
			SelectObject(hDC, hOld);
			ReleaseDC(m_hListBox, hDC);
			lpmis->itemHeight = 2 * tm.tmHeight + 3 * 2;
		}
		return TRUE;

	case WM_DRAWITEM:
		if(wParam != IDC_CHAPTER_LIST)
			return FALSE;
		lpdis = (LPDRAWITEMSTRUCT) lParam;
		if(lpdis->itemID == -1)
			break;
		switch (lpdis->itemAction) {
		case ODA_SELECT:
		case ODA_DRAWENTIRE:
		case ODA_FOCUS:
			{
				HGDIOBJ hOld;
				ModChapterList::tChapterNode *chapter = NULL;
				if(lpdis->itemID < m_Chapters.size()) {
					ModChapterList::iterator it = m_Chapters.begin();
					std::advance(it, lpdis->itemID);
					chapter = &(*it);
				}
				COLORREF color;
				HBRUSH brush;
				if(lpdis->itemState & ODS_SELECTED)
					color = RGB(0xE0, 0xE0, 0xE0);
				else
					color = RGB(0xFF, 0xFF, 0xFF);
				brush = CreateSolidBrush(color);
				FillRect(lpdis->hDC, &lpdis->rcItem, brush);
				DeleteObject(brush);
				SetBkColor(lpdis->hDC, color);
				GetTextMetrics(lpdis->hDC, &tm);
				y = (lpdis->rcItem.top + lpdis->rcItem.bottom - tm.tmHeight) / 2;
				_sntprintf(tchBuffer, sizeof(tchBuffer)-1, _T("%u"), lpdis->itemID+1);
				TextOut(lpdis->hDC, LEFT_INDEX, y, tchBuffer, _tcslen(tchBuffer));
				PrintChapterInfo(chapter, lpdis->hDC, lpdis->rcItem, false);
				hOld = SelectObject(lpdis->hDC, CreatePen(PS_SOLID, 2, RGB(0,0,0)));
				MoveToEx(lpdis->hDC, lpdis->rcItem.left, lpdis->rcItem.top, NULL);
				LineTo(lpdis->hDC, lpdis->rcItem.right, lpdis->rcItem.top);
				MoveToEx(lpdis->hDC, lpdis->rcItem.left, lpdis->rcItem.bottom, NULL);
				LineTo(lpdis->hDC, lpdis->rcItem.right, lpdis->rcItem.bottom);
				DeleteObject(SelectObject(lpdis->hDC, CreatePen(PS_DOT, 1, RGB(0xA0,0xA0,0xA0))));
				MoveToEx(lpdis->hDC, LEFT_FIRSTLINE, lpdis->rcItem.top, NULL);
				LineTo(lpdis->hDC, LEFT_FIRSTLINE, lpdis->rcItem.bottom);
				DeleteObject(SelectObject(lpdis->hDC, hOld));
			}
			break;
		}
	return TRUE;

	case WM_VKEYTOITEM:
		{
			if((HWND)lParam == m_hListBox) {
				if(LOWORD(wParam) == VK_DELETE) {
					SendMessage(mhdlg, WM_COMMAND, IDC_DELETE, (LPARAM)GetDlgItem(mhdlg, IDC_DELETE));
					return -2;	// Tell the listbox we did the necessary handling
				}
			}
			return -1; // Tell the listbox to handle the keystroke as usual
		}
		break;

	case WM_COMMAND:
		if(m_hListBox == (HWND)lParam) {
			if(HIWORD(wParam) == LBN_SELCHANGE) {
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=m_Chapters.size()) )
					return TRUE;
				ModChapterList::tChapterNode *chapter = NULL;
				{
					ModChapterList::iterator it = m_Chapters.begin();
					std::advance(it, index);
					chapter = &(*it);
				}
				if(chapter) {
					if(g_bUseUnicodeWinAPI) {
						SetDlgItemTextW(mhdlg, IDC_CHAPTER_TIME, chapter->GetTime().c_str());
						SetDlgItemTextW(mhdlg, IDC_CHAPTER_NAME, chapter->GetName());
					} else {
						SetDlgItemTextA(mhdlg, IDC_CHAPTER_TIME, VDTextWToA(chapter->GetTime()).c_str());
						SetDlgItemTextA(mhdlg, IDC_CHAPTER_NAME, VDTextWToA(chapter->GetName()).c_str());
					}
				}
				return TRUE;
			}
		} else
		switch(LOWORD(wParam)) {
		case IDC_ADD:
			{
				int iTime, iName;
				VDStringW sName;
				ModChapterList::tChapterNode newChapter;
				bool bValidTime = false;
				if(g_bUseUnicodeWinAPI) {
					iTime = GetDlgItemTextW(mhdlg, IDC_CHAPTER_TIME, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					bValidTime = newChapter.SetTime(szBufferW);
					iName = GetDlgItemTextW(mhdlg, IDC_CHAPTER_NAME, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					sName = szBufferW;
				} else {
					iTime = GetDlgItemTextA(mhdlg, IDC_CHAPTER_TIME, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					bValidTime = newChapter.SetTime(szBufferA);
					iName = GetDlgItemTextA(mhdlg, IDC_CHAPTER_NAME, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					sName = VDTextAToW(szBufferA, -1);
				}
				if(!bValidTime) {
					guiMessageBox(mhdlg, IDS_ERR_WRONG_TIME_FORMAT, IDS_ERR_WRONG_TIME_FORMAT_CAP, MB_OK | MB_ICONEXCLAMATION);
					SetFocus(GetDlgItem(mhdlg, IDC_CHAPTER_TIME));
					return TRUE;
				}
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				newChapter.SetName(sName.c_str());
				if(index < 0)
					m_Chapters.AddChapter(newChapter);
				else {
					if(index >= m_Chapters.size())
						index = m_Chapters.size() -1;
					ModChapterList::iterator it = m_Chapters.begin();
					std::advance(it, index);
					m_Chapters.insert(it, newChapter);
				}
				SetDlgItemText(mhdlg, IDC_CHAPTER_TIME, _T("00:00:00.000"));
				SetDlgItemText(mhdlg, IDC_CHAPTER_NAME, _T(""));
				m_Chapters.Sort();
				//if( (InsertListBoxItem(m_hListBox, _T(""), (LPARAM)newChapter, index)!=index) && (index!=-1))
				RefreshListData<ModChapterList::tChapterNode>(m_hListBox, m_Chapters.GetPtrs(), index);
				UpdateListBoxHorizontalExtent();
			}
			return TRUE;

		case IDC_DELETE:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=m_Chapters.size()) )
					return TRUE;
				ModChapterList::iterator it = m_Chapters.begin();
				std::advance(it, index);
				m_Chapters.erase(it);
				if(SendMessage(m_hListBox, LB_DELETESTRING, index, 0) == LB_ERR)
					RefreshListData<ModChapterList::tChapterNode>(m_hListBox, m_Chapters.GetPtrs(), index);
				else {
					if(index == m_Chapters.size())
						index--;
					SendMessage(m_hListBox, LB_SETCURSEL, index, 0);
				}
				UpdateListBoxHorizontalExtent();
			}
			return TRUE;

		case IDC_EDIT2:
			{
				int index = SendMessage(m_hListBox, LB_GETCURSEL, 0, 0);
				if( (index<0) || (index>=m_Chapters.size()) )
					return TRUE;
				ModChapterList::tChapterNode *chapter = NULL;
				{
					ModChapterList::iterator it = m_Chapters.begin();
					std::advance(it, index);
					chapter = &(*it);
				}
				if(!chapter)
					return TRUE;
				int iTime, iName;
				VDStringW sName;
				bool bValidTime = false;
				if(g_bUseUnicodeWinAPI) {
					iTime = GetDlgItemTextW(mhdlg, IDC_CHAPTER_TIME, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					bValidTime = chapter->SetTime(szBufferW);
					iName = GetDlgItemTextW(mhdlg, IDC_CHAPTER_NAME, szBufferW, sizeof(szBufferW)/sizeof(szBufferW[0])-1);
					sName = szBufferW;
				} else {
					iTime = GetDlgItemTextA(mhdlg, IDC_CHAPTER_TIME, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					bValidTime = chapter->SetTime(szBufferA);
					iName = GetDlgItemTextA(mhdlg, IDC_CHAPTER_NAME, szBufferA, sizeof(szBufferA)/sizeof(szBufferA[0])-1);
					sName = VDTextAToW(szBufferA, -1);
				}
				if(!bValidTime) {
					guiMessageBox(mhdlg, IDS_ERR_WRONG_TIME_FORMAT, IDS_ERR_WRONG_TIME_FORMAT_CAP, MB_OK | MB_ICONEXCLAMATION);
					SetFocus(GetDlgItem(mhdlg, IDC_CHAPTER_TIME));
					return TRUE;
				}
				chapter->SetName(sName.c_str());
				m_Chapters.Sort();
				UpdateListBoxHorizontalExtent();
				RedrawWindow(mhdlg, NULL, NULL, RDW_INTERNALPAINT | RDW_INVALIDATE | RDW_UPDATENOW | RDW_ALLCHILDREN);
			}
			return TRUE;

		case IDC_IMPORT:
			{
				const VDStringW fileName(VDGetLoadFileName(kFileDialog_ChapterFileIn, (VDGUIHandle)mhdlg, L"Open Chapter file", g_fileFiltersModChapterFile, g_prefs.main.fAttachExtension ? L"txt" : NULL));

				if(!fileName.empty()) {
					if(!ImportModStreamChapters(fileName, m_Chapters))
						guiMessageBox(mhdlg, IDS_ERR_WRONG_CHAPTER_FORMAT, IDS_ERR_WRONG_CHAPTER_FORMAT_CAP, MB_OK | MB_ICONEXCLAMATION);
					else {
						RefreshListData<ModChapterList::tChapterNode>(m_hListBox, m_Chapters.GetPtrs(), -1);
						UpdateListBoxHorizontalExtent();
					}
				}
			}
			return TRUE;

		case IDOK:
			returnValue = TRUE;
			m_Chapters.SetAdjustChapters(!!IsDlgButtonChecked(mhdlg, IDC_ADJUST_CHAPTERS));
		case IDCANCEL:
			InstallMouseHook(NULL, NULL, NULL);
			End(returnValue);
			return TRUE;
		}
		break;
	}
	return FALSE;
}

sint32 ModDialogStreamChaptersW32::PrintChapterInfo(ModChapterList::tChapterNode *chapter, HDC hDC, RECT rcItem, bool bGetHorizontalExtent) {
	TEXTMETRIC tm;
	HGDIOBJ hOld;
	int y;
	sint32 horExtent = LEFT_FIRSTLINE;

	if(bGetHorizontalExtent) {
		HFONT hFont = (HFONT)SendMessage(m_hListBox, WM_GETFONT, 0, 0);
		if(!hFont)
			hFont = (HFONT)GetStockObject(SYSTEM_FONT);
		hOld = SelectObject(hDC, hFont);
	}
	GetTextMetrics(hDC, &tm);

	if(chapter) {
		VDStringW cTime = chapter->GetTime();
		VDStringW cName(L"Name : [");
		cName += chapter->GetName();
		cName += L"]";
		y = rcItem.top;
		if(bGetHorizontalExtent) {
			GetHorizontalExtent(cTime, hDC, horExtent, LEFT_FIRSTLINE+5 + 5);
			GetHorizontalExtent(cName, hDC, horExtent, LEFT_OTHERLINE+5 + 5);
		} else {
			if(g_bUseUnicodeWinAPI) {
				TextOutW(hDC, LEFT_FIRSTLINE+5, y + 2, cTime.c_str(), cTime.length());
				TextOutW(hDC, LEFT_OTHERLINE+5, y + 4 + tm.tmHeight, cName.c_str(), cName.length());
			} else {
				VDStringA cTimeA = VDTextWToA(cTime);
				VDStringA cNameA = VDTextWToA(cName);
				TextOutA(hDC, LEFT_FIRSTLINE+5, y + 2, cTimeA.c_str(), cTimeA.length());
				TextOutA(hDC, LEFT_OTHERLINE+5, y + 4 + tm.tmHeight, cNameA.c_str(), cNameA.length());
			}
		}
	} else {
		y = (rcItem.top + rcItem.bottom - tm.tmHeight) / 2;
		if(bGetHorizontalExtent)
			GetHorizontalExtent(VDStringW(L"Could not retrieve chapter"), hDC, horExtent, LEFT_OTHERLINE+5 + 5);
		else
			TextOut(hDC, LEFT_FIRSTLINE+5, y, _T("Could not retrieve chapter"), _tcslen(_T("Could not retrieve chapter")));
	}

	if(bGetHorizontalExtent) {
		SelectObject(hDC, hOld);
		return horExtent;
	}
	return -1;
}

void ModDialogStreamChaptersW32::UpdateListBoxHorizontalExtent() {
	HDC hDC = GetDC(m_hListBox);
	sint32 horExtent = 0;
	sint32 tmpExtent = 0;
	RECT rcItem = { 0 };
	ModChapterList::iterator it;
	for(it=m_Chapters.begin() ; it!=m_Chapters.end() ; it++) {
		ModChapterList::tChapterNode *chapter = &(*it);
		tmpExtent = PrintChapterInfo(chapter, hDC, rcItem, true);
		if(tmpExtent > horExtent)
			horExtent = tmpExtent;
	}
	SendMessage(m_hListBox, LB_SETHORIZONTALEXTENT, horExtent, 0);
	ReleaseDC(m_hListBox, hDC);
}

//////////////////////////////////////////////////////////////////////

extern const char g_szRegKeyPersistence[]="Persistence";
extern const char g_szRegKeyRunAsJob[]="Run as job";
extern const char g_szRegKeySegmentFrameCount[]="Segment frame limit";
extern const char g_szRegKeyUseSegmentFrameCount[]="Use segment frame limit";
extern const char g_szRegKeySegmentSizeLimit[]="Segment size limit";
extern const char g_szRegKeySaveAsFilterIndex[]="Save As filter index";
extern const char g_szError[];
extern const char g_szWarning[];
extern char g_szFile[MAX_PATH];
extern wchar_t g_szInputAVIFile[];

//extern bool CoachCheckSaveOp(HWND hwndParent, DubOptions *dopt, COMPVARS *vcomp, WAVEFORMATEX *wfex, List *filter_list, char *filename, uint32 outputFileType);

static struct {
	uint32	m_Container;
	wchar_t	*m_Description;
	wchar_t	*m_Extensions;
	wchar_t	*m_PreferredExtension;
} s_SaveAsTypes[] = {
	{SOURCE_AVI,	L"Audio-Video Interleave (*.avi)",	L"*.avi",	L"avi"},
	{SOURCE_OGM,	L"Ogg Media File (*.ogm)",			L"*.ogm",	L"ogm"},
	{SOURCE_MKV,	L"Matroska File (*.mkv)",			L"*.mkv",	L"mkv"},
	{0,				NULL,								NULL,		NULL}
};

static DWORD	g_SaveAsFilterIndex = 0;
static bool		g_bUseAVICompatibility = false;

class SaveAsControlStruct {
public:
	bool	m_bSaveAsJob;
	bool	m_bUseAVICompatibility;
	bool	m_bUseThreshold;
	long	m_ThresholdMB;
	long	m_ThresholdFrames;

	SaveAsControlStruct(void) {
		m_bUseAVICompatibility = g_bUseAVICompatibility;
		m_bUseThreshold = false;
		VDRegistryAppKey key(g_szRegKeyPersistence);
		m_bSaveAsJob = key.getBool(g_szRegKeyRunAsJob, false);
		m_ThresholdFrames = key.getInt(g_szRegKeySegmentFrameCount, 100);
		m_ThresholdMB = key.getInt(g_szRegKeySegmentSizeLimit, 2000);
	};

	~SaveAsControlStruct(void) {
		VDRegistryAppKey key(g_szRegKeyPersistence);
		key.setBool(g_szRegKeyRunAsJob, m_bSaveAsJob);
		g_bUseAVICompatibility = m_bUseAVICompatibility;
		if(m_bUseThreshold) {
			if(m_ThresholdFrames)
				key.setInt(g_szRegKeySegmentFrameCount, m_ThresholdFrames);
			key.setInt(g_szRegKeySegmentSizeLimit, m_ThresholdMB);
		}
	};
};

static struct {
	int		m_VideoMode;
	int		m_MenuID;
} s_VideoModes[4] = {
	{DubVideoOptions::M_NONE,		ID_VIDEO_MODE_DIRECT},
	{DubVideoOptions::M_FASTREPACK,	ID_VIDEO_MODE_FASTRECOMPRESS},
	{DubVideoOptions::M_SLOWREPACK,	ID_VIDEO_MODE_NORMALRECOMPRESS},
	{DubVideoOptions::M_FULL,		ID_VIDEO_MODE_FULL}
};

void ModEnumVideoModes(void (__cdecl *enumFunctionCB)(int iID, LPCTSTR szValue, UINT iState, void *pData), void *pData) {
	TCHAR tchBuffer[1024];
	int i;
	for(i=0 ; i<sizeof(s_VideoModes)/sizeof(s_VideoModes[0]) ; i++) {
		UINT iBufferLength = sizeof(tchBuffer) / sizeof(tchBuffer[0]);
		UINT iMenuState = GetMenuState(g_projectui->getMainMenu(), s_VideoModes[i].m_MenuID, MF_BYCOMMAND);
		if(ModGetMenuItemText(g_projectui->getMainMenu(), s_VideoModes[i].m_MenuID, FALSE, tchBuffer, iBufferLength))
			enumFunctionCB(s_VideoModes[i].m_VideoMode, tchBuffer, iMenuState, pData);
	}
}

struct _ComboBox_VideoMode {
	HWND	m_hComboBox;
	int		m_iVideoMode;
};

void AddModeInComboBox(int iID, LPCTSTR szValue, UINT iState, void *pData) {
	_ComboBox_VideoMode *data = (_ComboBox_VideoMode *)pData;
	if(!data || (iState&MF_DISABLED) || (iState&MF_GRAYED))
		return;
	HWND hComboBox = data->m_hComboBox;
	int iVideoMode = data->m_iVideoMode;
	int index = SendMessage(hComboBox, CB_ADDSTRING, 0, (LPARAM)szValue);
	if((index != CB_ERR) && (index != CB_ERRSPACE)) {
		SendMessage(hComboBox, CB_SETITEMDATA, index, iID);
		if(iVideoMode == iID)
			SendMessage(hComboBox, CB_SETCURSEL, index, 0);
	}
}

void PopulateComboBoxWithVideoModes(HWND hComboBox, int iVideoMode) {
	if( (iVideoMode!=DubVideoOptions::M_NONE)
		&& (iVideoMode!=DubVideoOptions::M_FASTREPACK)
		&& (iVideoMode!=DubVideoOptions::M_SLOWREPACK)
		&& (iVideoMode!=DubVideoOptions::M_FULL))
		iVideoMode = DubVideoOptions::M_FULL;
	SendMessage(hComboBox, CB_RESETCONTENT, 0, 0);
	_ComboBox_VideoMode data;
	data.m_hComboBox = hComboBox;
	data.m_iVideoMode = iVideoMode;
	ModEnumVideoModes(AddModeInComboBox, &data);
}

static void HitVideoMode(int iVideoMode) {
	int i;
	for(i=0 ; i<sizeof(s_VideoModes)/sizeof(s_VideoModes[0]) ; i++) {
		if(s_VideoModes[i].m_VideoMode == iVideoMode) {
			SendMessage(g_hWnd, WM_COMMAND, MAKEWPARAM(s_VideoModes[i].m_MenuID, 0), 0);
			break;
		}
	}
}

static void UpdateVideoCompressionField(HWND hField) {
	if(!(g_Vcompression.dwFlags & ICMF_COMPVARS_VALID)) {
		memset(&g_Vcompression, 0, sizeof g_Vcompression);
		g_Vcompression.dwFlags |= ICMF_COMPVARS_VALID;
		g_Vcompression.lQ = 10000;
	}

	g_Vcompression.cbSize = sizeof(COMPVARS);

	BITMAPINFOHEADER *bihInput = (inputVideoAVI ? inputVideoAVI->getImageFormat() : NULL);

	VDStringW videoComp;

	if(bihInput && (g_dubOpts.video.mode==DubVideoOptions::M_NONE)) {
		union {
			char fccbuf[5];
			FOURCC fcc;
		};
		fcc = bihInput->biCompression;
		fccbuf[4] = 0;
		VDStringW tmpValue(VDFastTextAToW(fccbuf));
		videoComp = VDStringW::setf(L"(No recompression: %s)"
			, tmpValue.c_str());
	} else if(g_Vcompression.hic) {
		ICINFO info = { sizeof(ICINFO) };

		if(ICGetInfo(g_Vcompression.hic, &info, sizeof(ICINFO))) {
			union {
				char fccbuf[5];
				FOURCC fcc;
			};
			fcc = info.fccHandler;
			fccbuf[4] = 0;
			VDStringW tmpValue(VDFastTextAToW(fccbuf));
			//info.szName
			if(info.szDescription)
				videoComp = info.szDescription;
			videoComp += VDStringW::setf(L" (%s)"
				, tmpValue.c_str());
		}
	} else
		videoComp = L"(Uncompressed RGB)";

	if(g_bUseUnicodeWinAPI)
		SetWindowTextW(hField, videoComp.c_str());
	else
		SetWindowTextA(hField, VDTextWToA(videoComp).c_str());
}

UINT CALLBACK SaveAsDlgHookProc(HWND hDlg, UINT uiMsg, WPARAM wParam, LPARAM lParam) {
	OFNOTIFY *ofn;
	BOOL fOk;
	SaveAsControlStruct *controlStruct;

	switch(uiMsg) {
	case WM_COMMAND:
		if((GetDlgItem(hDlg, IDC_VIDEO_MODE) == (HWND)lParam) && (HIWORD(wParam) == LBN_SELCHANGE)) {
			int index = SendDlgItemMessage(hDlg, IDC_VIDEO_MODE, CB_GETCURSEL, 0, 0);
			if(index != CB_ERR)
				index = SendDlgItemMessage(hDlg, IDC_VIDEO_MODE, CB_GETITEMDATA, index, 0);
			if(index != CB_ERR) {
				HitVideoMode(index);
				EnableWindow(GetDlgItem(hDlg, IDC_CHANGE_VIDEOCOMPRESSION), index > DubVideoOptions::M_NONE);
				UpdateVideoCompressionField(GetDlgItem(hDlg, IDC_VIDEO_COMPRESSION));
			}
		} else switch(LOWORD(wParam)) {
		case IDC_FRAMELIMIT:
			if(HIWORD(wParam) == BN_CLICKED)
				EnableWindow(GetDlgItem(hDlg, IDC_EDIT_FRAMELIMIT), SendMessage((HWND)lParam, BM_GETSTATE, 0, 0));
			break;
		case IDC_FILE_SEGMENTATION:
			if(HIWORD(wParam) == BN_CLICKED) {
				if(IsDlgButtonChecked(hDlg, IDC_FILE_SEGMENTATION)) {
					EnableWindow(GetDlgItem(hDlg, IDC_FRAMELIMIT), TRUE);
					EnableWindow(GetDlgItem(hDlg, IDC_EDIT_FRAMELIMIT), IsDlgButtonChecked(hDlg, IDC_FRAMELIMIT));
					EnableWindow(GetDlgItem(hDlg, IDC_LIMIT), TRUE);
				} else {
					EnableWindow(GetDlgItem(hDlg, IDC_FRAMELIMIT), FALSE);
					EnableWindow(GetDlgItem(hDlg, IDC_EDIT_FRAMELIMIT), FALSE);
					EnableWindow(GetDlgItem(hDlg, IDC_LIMIT), FALSE);
				}
			}
			break;
		case IDC_CHANGE_VIDEOCOMPRESSION:
			if (!(g_Vcompression.dwFlags & ICMF_COMPVARS_VALID)) {
				memset(&g_Vcompression, 0, sizeof g_Vcompression);
				g_Vcompression.dwFlags |= ICMF_COMPVARS_VALID;
				g_Vcompression.lQ = 10000;
			}

			g_Vcompression.cbSize = sizeof(COMPVARS);

			ChooseCompressor(hDlg, &g_Vcompression, NULL);
			UpdateVideoCompressionField(GetDlgItem(hDlg, IDC_VIDEO_COMPRESSION));
			break;
		}
		return 0;

	case WM_NOTIFY:
		ofn = (OFNOTIFY *)lParam;

		switch(ofn->hdr.code) {
		case CDN_INITDONE:
			{
				controlStruct = (SaveAsControlStruct *)ofn->lpOFN->lCustData;

				CheckDlgButton(hDlg, IDC_ADD_AS_JOB, controlStruct->m_bSaveAsJob ? BST_CHECKED : BST_UNCHECKED);
				CheckDlgButton(hDlg, IDC_AVI_COMPATIBILITY, controlStruct->m_bUseAVICompatibility ? BST_CHECKED : BST_UNCHECKED);
				CheckDlgButton(hDlg, IDC_FILE_SEGMENTATION, controlStruct->m_bUseThreshold ? BST_CHECKED : BST_UNCHECKED);
				EnableWindow(GetDlgItem(hDlg, IDC_AVI_COMPATIBILITY), ofn->lpOFN->nFilterIndex==1);
				if(controlStruct->m_bUseThreshold) {
					EnableWindow(GetDlgItem(hDlg, IDC_FRAMELIMIT), TRUE);
					EnableWindow(GetDlgItem(hDlg, IDC_EDIT_FRAMELIMIT), FALSE);
					EnableWindow(GetDlgItem(hDlg, IDC_LIMIT), TRUE);
				} else {
					EnableWindow(GetDlgItem(hDlg, IDC_FRAMELIMIT), FALSE);
					EnableWindow(GetDlgItem(hDlg, IDC_EDIT_FRAMELIMIT), FALSE);
					EnableWindow(GetDlgItem(hDlg, IDC_LIMIT), FALSE);
				}
				SetDlgItemInt(hDlg, IDC_EDIT_FRAMELIMIT, controlStruct->m_ThresholdFrames, FALSE);
				SetDlgItemInt(hDlg, IDC_LIMIT, controlStruct->m_ThresholdMB, FALSE);
				PopulateComboBoxWithVideoModes(GetDlgItem(hDlg, IDC_VIDEO_MODE), g_dubOpts.video.mode);
				EnableWindow(GetDlgItem(hDlg, IDC_CHANGE_VIDEOCOMPRESSION), g_dubOpts.video.mode > DubVideoOptions::M_NONE);

				UpdateVideoCompressionField(GetDlgItem(hDlg, IDC_VIDEO_COMPRESSION));
			}
			break;

		case CDN_TYPECHANGE:
			EnableWindow(GetDlgItem(hDlg, IDC_AVI_COMPATIBILITY), ofn->lpOFN->nFilterIndex==1);
			break;

		case CDN_FILEOK:
			controlStruct = (SaveAsControlStruct *)ofn->lpOFN->lCustData;

			controlStruct->m_bSaveAsJob = !!IsDlgButtonChecked(hDlg, IDC_ADD_AS_JOB);
			controlStruct->m_bUseAVICompatibility = !!IsDlgButtonChecked(hDlg, IDC_AVI_COMPATIBILITY);
			controlStruct->m_bUseThreshold = !!IsDlgButtonChecked(hDlg, IDC_FILE_SEGMENTATION);

			if(controlStruct->m_bUseThreshold) {
				controlStruct->m_ThresholdMB = GetDlgItemInt(hDlg, IDC_LIMIT, &fOk, FALSE);

				if (!fOk || controlStruct->m_ThresholdMB<50 || controlStruct->m_ThresholdMB>2048) {
					guiMessageBox(hDlg, IDS_ERR_SEGMENT_SIZE, IDS_ERR_CAPTION, MB_OK | MB_ICONEXCLAMATION);
					SetFocus(GetDlgItem(hDlg, IDC_LIMIT));

					SetWindowLong(hDlg, DWL_MSGRESULT, 1);
					return 1;
				}

				if(IsDlgButtonChecked(hDlg, IDC_FRAMELIMIT)) {
					controlStruct->m_ThresholdFrames = GetDlgItemInt(hDlg, IDC_EDIT_FRAMELIMIT, &fOk, FALSE);

					if (!fOk || controlStruct->m_ThresholdFrames<1) {
						guiMessageBox(hDlg, IDS_ERR_SEGMENT_FRAME, IDS_ERR_CAPTION, MB_OK | MB_ICONEXCLAMATION);
						SetFocus(GetDlgItem(hDlg, IDC_EDIT_FRAMELIMIT));

						SetWindowLong(hDlg, DWL_MSGRESULT, 1);
						return 1;
					}
				} else
					controlStruct->m_ThresholdFrames = 0;
			}

			return 0;
		}
		break;
	}
	return FALSE;
}

void SaveAs(HWND hWnd) {
	SaveAsControlStruct controlStruct;

	SetAudioSource();

	if(!inputVideoAVI) {
		guiMessageBox(hWnd, IDS_ERR_NO_INPUT_VIDEO, IDS_ERR_CAPTION, MB_OK | MB_ICONEXCLAMATION);
		return;
	}

	VDRegistryAppKey key(g_szRegKeyPersistence);
	if(!g_SaveAsFilterIndex)
		g_SaveAsFilterIndex = key.getInt(g_szRegKeySaveAsFilterIndex, 1);
	wchar_t *fileFilters = (wchar_t *)malloc(sizeof(wchar_t));
	sint32 fileFiltersSize = 1;
	*fileFilters = 0;
	VDStringW fileDefaultExt(L"avi");
	int i;
	for(i=0
		;s_SaveAsTypes[i].m_Description || s_SaveAsTypes[i].m_Extensions || s_SaveAsTypes[i].m_PreferredExtension
		; i++)
	{
		if(i+1 == g_SaveAsFilterIndex)
			fileDefaultExt = s_SaveAsTypes[i].m_PreferredExtension;
		sint32 descPos = fileFiltersSize - 1;
		sint32 descSize = wcslen(s_SaveAsTypes[i].m_Description) + 1;
		fileFiltersSize += descSize;
		sint32 extPos = fileFiltersSize - 1;
		sint32 extSize = wcslen(s_SaveAsTypes[i].m_Extensions) + 1;
		fileFiltersSize += extSize;
		fileFilters = (wchar_t *)realloc(fileFilters, fileFiltersSize*sizeof(wchar_t));
		if(!fileFilters)
			throw MyMemoryError();
		memcpy(fileFilters + descPos, s_SaveAsTypes[i].m_Description, descSize*sizeof(wchar_t));
		memcpy(fileFilters + extPos, s_SaveAsTypes[i].m_Extensions, extSize*sizeof(wchar_t));
	}
	*(fileFilters + fileFiltersSize - 1) = 0;

	LPARAM lCustData = (LPARAM)&controlStruct;
	//const VDStringW fileName(VDGetSaveFileName(kFileDialog_SaveAs, (VDGUIHandle)hWnd, L"Save As", fileFiltersSaveAs, g_prefs.main.fAttachExtension ? L"avi" : NULL, &g_SaveAsFilterIndex, IDD_SAVE_AS, g_hInst, (LPOFNHOOKPROC)SaveAsDlgHookProc, (LPARAM)&controlStruct));
	const VDStringW fileName(VDGetSaveFileName(kFileDialog_SaveAs, (VDGUIHandle)hWnd, L"Save As", fileFilters, g_prefs.main.fAttachExtension ? fileDefaultExt.c_str() : NULL, &g_SaveAsFilterIndex, IDD_SAVE_AS, g_hInst, (LPOFNHOOKPROC)SaveAsDlgHookProc, lCustData));

	free(fileFilters);
	if(!fileName.empty()) {
		key.setInt(g_szRegKeySaveAsFilterIndex, g_SaveAsFilterIndex);
		uint32 outputFileType = s_SaveAsTypes[g_SaveAsFilterIndex-1].m_Container;
		wchar_t *preferredFileExtension = s_SaveAsTypes[g_SaveAsFilterIndex-1].m_PreferredExtension;
		VDStringA fileNameA = VDTextWToA(fileName);
		char *szFileNameA = (char *)fileNameA.c_str();
		if(controlStruct.m_bUseThreshold) {
			{
				char szPrefixBuffer[MAX_PATH], szPattern[MAX_PATH*2], *t, *t2, c;
				const char *s;
				int nMatchCount = 0;

				VDStringA fileExtension_0 = VDTextWToA(preferredFileExtension);
				VDStringA fileExtension_1(".");
				fileExtension_1 += fileExtension_0;
				VDStringA fileExtension_2(".*");
				fileExtension_2 += fileExtension_1;
				VDStringA filePattern(".%d");
				filePattern += VDStringA(fileExtension_1.c_str(), fileExtension_1.size()-1);
				filePattern += "%c";
				char filePatternEnd = tolower(*(fileExtension_1.c_str() + fileExtension_1.size()-1));

				t = VDFileSplitPath(szFileNameA);
				t2 = VDFileSplitExt(t);

				if(!stricmp(t2, fileExtension_1.c_str())) {
					while(t2>t && isdigit((unsigned)t2[-1]))
						--t2;

					if(t2>t && t2[-1]=='.') {
						strcpy(t2, fileExtension_0.c_str());
					}
				}

				strcpy(szPrefixBuffer, szFileNameA);
				VDFileSplitExt(szPrefixBuffer)[0] = 0;

				s = VDFileSplitPath(szPrefixBuffer);
				t = szPattern;

				while(*t++ = *s++)
					if (s[-1]=='%')
						*t++ = '%';

				t = szPrefixBuffer;
				while(*t)
					++t;

				strcpy(t, fileExtension_2.c_str());

				WIN32_FIND_DATA wfd;
				HANDLE h;

				h = FindFirstFile(szPrefixBuffer, &wfd);
				if (h != INVALID_HANDLE_VALUE) {
					strcat(szPattern, filePattern.c_str());

					do {
						int n;

						if (2 == sscanf(wfd.cFileName, szPattern, &n, &c) && (tolower(c)==filePatternEnd))
							++nMatchCount;
					
					} while(FindNextFile(h, &wfd));
					FindClose(h);
				}

				if (nMatchCount) {
/*					if (IDOK != guiMessageBoxF(g_hWnd, g_szWarning, MB_OKCANCEL|MB_ICONEXCLAMATION,
						"There %s %d existing file%s which match%s the filename pattern \"%s\". These files "
						"will be erased if you continue, to prevent confusion with the new files."
						,nMatchCount==1 ? "is" : "are"
						,nMatchCount
						,nMatchCount==1 ? "" : "s"
						,nMatchCount==1 ? "es" : ""
						,VDFileSplitPath(szPrefixBuffer)))
						return;*/
					if (IDOK != guiMessageBoxF(g_hWnd, IDS_WARN_SEGAVI_NAME_COLISION, IDS_WARN_CAPTION, MB_OKCANCEL|MB_ICONEXCLAMATION
						,nMatchCount==1 ? "is" : "are"
						,nMatchCount
						,nMatchCount==1 ? "" : "s"
						,nMatchCount==1 ? "es" : ""
						,VDFileSplitPath(szPrefixBuffer)))
						return;

					h = FindFirstFile(szPrefixBuffer, &wfd);
					if (h != INVALID_HANDLE_VALUE) {
						strcat(szPattern, filePattern.c_str());

						t = VDFileSplitPath(szPrefixBuffer);

						do {
							int n;

							if (2 == sscanf(wfd.cFileName, szPattern, &n, &c) && (tolower(c)==filePatternEnd)) {
								strcpy(t, wfd.cFileName);
								DeleteFile(t);
							}
						} while(FindNextFile(h, &wfd));
						FindClose(h);
					}
				}
			}
		}

//		if(!CoachCheckSaveOp(hWnd, &g_dubOpts, &g_Vcompression, g_ACompressionFormat, &g_listFA, g_szInputAVIFile, outputFileType))
//			return;

		if(controlStruct.m_bUseThreshold) {
			if(controlStruct.m_bSaveAsJob) {
				try {
					JobAddConfiguration(&g_dubOpts, g_szInputAVIFile, NULL, fileName.c_str(), true, &inputAVI->listFiles, controlStruct.m_ThresholdMB, controlStruct.m_ThresholdFrames, true, outputFileType);
				} catch(const MyError& e) {
					e.post(g_hWnd, g_szError);
				}
			} else {
				switch(outputFileType) {
				case SOURCE_OGM:
					SaveSegmentedOGM(fileName.c_str(), false, NULL, controlStruct.m_ThresholdMB, controlStruct.m_ThresholdFrames);
					break;
				case SOURCE_MKV:
					SaveSegmentedMKV(fileName.c_str(), false, NULL, controlStruct.m_ThresholdMB, controlStruct.m_ThresholdFrames);
					break;
				default:
					SaveSegmentedAVI(fileName.c_str(), false, NULL, controlStruct.m_ThresholdMB, controlStruct.m_ThresholdFrames);
					break;
				}
			}
		} else {
			if(controlStruct.m_bSaveAsJob) {
				try {
					JobAddConfiguration(&g_dubOpts, g_szInputAVIFile, NULL, fileName.c_str(), controlStruct.m_bUseAVICompatibility, &inputAVI->listFiles, 0, 0, true, outputFileType);
				} catch(const MyError& e) {
					e.post(g_hWnd, g_szError);
				}
			} else {
				switch(outputFileType) {
				case SOURCE_OGM:
					SaveOGM(fileName.c_str(), false, NULL, controlStruct.m_bUseAVICompatibility);
					break;
				case SOURCE_MKV:
					SaveMKV(fileName.c_str(), false, NULL, controlStruct.m_bUseAVICompatibility);
					break;
				default:
					SaveAVI(fileName.c_str(), false, NULL, controlStruct.m_bUseAVICompatibility);
					break;
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////

BOOL CALLBACK VideoJumpLastKeyDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
	case WM_INITDIALOG:
		SendDlgItemMessage(hdlg, IDC_SPIN_MBYTES, UDM_SETRANGE, 0, MAKELONG(4000, 0));
		SendDlgItemMessage(hdlg, IDC_SPIN_MBYTES, UDM_SETPOS, 0, 700);
		SetFocus(GetDlgItem(hdlg, IDC_EDIT_MBYTES));
		SendDlgItemMessage(hdlg, IDC_EDIT_MBYTES, EM_SETSEL, (WPARAM)0, (LPARAM)-1);
		return FALSE;

	case WM_COMMAND:
		switch(LOWORD(wParam)) {
		case IDCANCEL:
			EndDialog(hdlg, -1);
			break;
		case IDOK:
			UINT uiMbytes = SendDlgItemMessage(hdlg, IDC_SPIN_MBYTES, UDM_GETPOS, 0, 0);
			EndDialog(hdlg, uiMbytes);
			break;
		}
		return TRUE;
	}
	return FALSE;
}

sint64 PositionFilePositionCallback(HWND hwnd, void *pvData, long pos) {
	try {
		if(inputAVI && inputVideoAVI) {
			if(inputSubset) {
				if(pos >= inputSubset->getTotalFrames())
					pos = inputSubset->getTotalFrames() - 1;
				pos = inputSubset->lookupFrame(pos);
			}
			return inputVideoAVI->sampleOffset(pos);
		}
	} catch(const MyError&) {
	}
	return 0;
}

long JumpToLastKeyBeforePos(sint64 pos) {
	if(!inputAVI || !inputVideoAVI) return -1;
	LONG lSample = 0;
	LONG lSample2 = 0;
	sint64 curPos = 0;
	bool bEnd = false;
	while(true) {
		lSample = inputVideoAVI->nextKey(lSample);
		if(lSample<0) {
			lSample = inputVideoAVI->getEnd();
			bEnd = true;
		}
		lSample2 = lSample;
		if(inputSubset)
			lSample2 = inputSubset->lookupFrame(lSample);

		curPos = inputVideoAVI->sampleOffset(lSample2);
		if(curPos == pos)
			return lSample;
		else if(curPos > pos) {
			lSample = inputVideoAVI->prevKey(lSample);
			break;
		} else if(bEnd)
			return lSample;
	}
	return lSample;
}

static int	s_VideoMode_Save = -1;
static bool	s_bShowInput_Save = true;
static bool	s_bShowOutput_Save = true;
static bool	s_bShowDecompressed_Save = false;

void ForceDirectStreamCopy(bool bForce) {
	HMENU hMenuNormal = g_projectui->getMainMenu();

	if(bForce) {
		// Force 'Direct stream copy' only, and don't show input / output video
		// Save the current state of all those settings
		s_VideoMode_Save = g_dubOpts.video.mode;
		s_bShowInput_Save = g_dubOpts.video.fShowInputFrame;
		s_bShowOutput_Save = g_dubOpts.video.fShowOutputFrame;
		s_bShowDecompressed_Save = g_drawDecompressedFrame;
		g_dubOpts.video.mode = DubVideoOptions::M_NONE;
		g_dubOpts.video.fShowInputFrame = false;
		g_dubOpts.video.fShowOutputFrame = false;
		g_drawDecompressedFrame = false;
		VDEnableMenuItemW32(hMenuNormal,	ID_VIDEO_MODE_FASTRECOMPRESS,	false);
		VDEnableMenuItemW32(hMenuNormal,	ID_VIDEO_MODE_NORMALRECOMPRESS,	false);
		VDEnableMenuItemW32(hMenuNormal,	ID_VIDEO_MODE_FULL,				false);
	} else {
		// Reallow all video modes, and reload saved settings
		VDEnableMenuItemW32(hMenuNormal,	ID_VIDEO_MODE_FASTRECOMPRESS,	true);
		VDEnableMenuItemW32(hMenuNormal,	ID_VIDEO_MODE_NORMALRECOMPRESS,	true);
		VDEnableMenuItemW32(hMenuNormal,	ID_VIDEO_MODE_FULL,				true);
		if(s_VideoMode_Save != -1) {
			g_dubOpts.video.mode = s_VideoMode_Save;
			g_dubOpts.video.fShowInputFrame = s_bShowInput_Save;
			g_dubOpts.video.fShowOutputFrame = s_bShowOutput_Save;
			g_drawDecompressedFrame = s_bShowDecompressed_Save;
		}
	}
}