//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "ModTreeControl.h"
#include "resource.h"
#include <utility>
#include <windowsx.h>
#include "ModFileSelectControl.h"

extern HINSTANCE	g_hInst;
extern const char	g_szModTreeControlName[] = "vdModTreeControl";

ModTreeControl_Control::ModTreeControl_Control(ModTreeControl& treeControl)
: m_TreeControl(treeControl)
, m_Item(NULL)
, m_bEnabled(true)
, m_Parent(&treeControl)
, m_Label(_T(""))
, m_ToolTip(_T(""))
{
}

ModTreeControl_Control::~ModTreeControl_Control(void) {
	ModTreeControl_Control *curControl = NULL;
	while(m_Children.size()) {
		curControl = m_Children.front();
		m_Children.pop_front();
		delete curControl;
	}
}

void ModTreeControl_Control::SetItem(HTREEITEM item) {
	m_Item = item;
}

HTREEITEM ModTreeControl_Control::GetItem(void) {
	return m_Item;
}

bool ModTreeControl_Control::Expand(bool bRecursive) {
	bool bResult = m_TreeControl.Expand(m_Item);
	if(bRecursive) {
		ModTreeControl_Control *curControl = NULL;
		std::list<ModTreeControl_Control *>::const_iterator it;
		for(it=m_Children.begin() ; it!=m_Children.end() ; it++)
			bResult &= (*it)->Expand(bRecursive);
	}
	m_TreeControl.RefreshNeededSize();
	return bResult;
}

bool ModTreeControl_Control::Collapse(bool bRecursive) {
	bool bResult = true;
	if(bRecursive) {
		ModTreeControl_Control *curControl = NULL;
		std::list<ModTreeControl_Control *>::const_iterator it;
		for(it=m_Children.begin() ; it!=m_Children.end() ; it++)
			bResult &= (*it)->Collapse(bRecursive);
	}
	bResult &= m_TreeControl.Collapse(m_Item);
	m_TreeControl.RefreshNeededSize();
	return bResult;
}

void ModTreeControl_Control::SetParent(ModTreeControl_Control *parent) {
	m_Parent = parent;
	RefreshState();
}

void ModTreeControl_Control::AddChild(ModTreeControl_Control& child) {
	child.SetParent(this);
	m_Children.push_back(&child);
}

void ModTreeControl_Control::AddRadioButton(ModTreeControl_RadioButton& radioButton) {
	m_RadioButtons.push_back(&radioButton);
}

void ModTreeControl_Control::CheckRadioButton(ModTreeControl_RadioButton& radioButton) {
	ModTreeControl_RadioButton *curControl = NULL;
	std::list<ModTreeControl_RadioButton *>::const_iterator it;
	for(it=m_RadioButtons.begin() ; it!=m_RadioButtons.end() ; it++) {
		if((*it) != &radioButton)
			(*it)->Check(false);
	}
}

bool ModTreeControl_Control::BranchEnabled(void) {
	if(!m_bEnabled)
		return false;
	if(m_Parent)
		return (m_Parent->BranchEnabled() & m_Parent->SubBranchEnabled());
	return true;
}

bool ModTreeControl_Control::SubBranchEnabled(void) {
	return BranchEnabled();
}

void ModTreeControl_Control::IconHit(void) {
	ItemHit();
}

void ModTreeControl_Control::LabelHit(void) {
	ItemHit();
}

void ModTreeControl_Control::ItemHit(void) {
}

void ModTreeControl_Control::SpaceKeyHit(void) {
	m_TreeControl.ExpandCollapse(m_Item);
}

bool ModTreeControl_Control::Enabled(void) {
	return m_bEnabled;
}

void ModTreeControl_Control::Enable(void) {
	m_bEnabled = true;
	RefreshState();
}

void ModTreeControl_Control::Disable(void) {
	m_bEnabled = false;
	RefreshState();
}

void ModTreeControl_Control::RefreshState(void) {
	RefreshStateImage();
	std::list<ModTreeControl_Control *>::const_iterator it;
	for(it=m_Children.begin() ; it!=m_Children.end() ; it++)
		(*it)->RefreshState();
}

void ModTreeControl_Control::RefreshStateImage(void) {
	m_TreeControl.SetStateImages(this, std::string("GROUP"), std::string("GROUP"));
}

bool ModTreeControl_Control::RefreshString(void) {
	return m_TreeControl.SetItemLabel(this, m_Label);
}

bool ModTreeControl_Control::SetLabel(LPCTSTR pszLabel) {
	m_Label = (pszLabel ? pszLabel : _T(""));
	return RefreshString();
}

LPCTSTR ModTreeControl_Control::GetLabel(void) {
	return m_Label.c_str();
}

void ModTreeControl_Control::SetToolTip(LPCTSTR pszToolTip) {
	m_ToolTip = (pszToolTip ? pszToolTip : _T(""));
}

LPCTSTR ModTreeControl_Control::GetToolTip(void) {
	return m_ToolTip.c_str();
}

tcstring ModTreeControl_Control::GetInfoTip(void) {
	tcstring infoTip;
	if(m_ToolTip.size() > 0) {
		infoTip = m_Label;
		infoTip += _T("\r\n\r\n");
		infoTip += m_ToolTip;
	}
	return infoTip;
}

COLORREF ModTreeControl_Control::GetTextColor(void) {
	if(BranchEnabled())
		return RGB(0x00, 0x00, 0x00);
	else
		return RGB(0x80, 0x80, 0x80);
}

//////////////////////////////////////////////////////////////////////

ModTreeControl_CheckBox::ModTreeControl_CheckBox(ModTreeControl& treeControl)
: ModTreeControl_Control(treeControl)
, m_bChecked(false)
{
}

bool ModTreeControl_CheckBox::SubBranchEnabled(void) {
	if(!ModTreeControl_Control::SubBranchEnabled())
		return false;
	return m_bChecked;
}

bool ModTreeControl_CheckBox::Checked(void) {
	return m_bChecked;
}

void ModTreeControl_CheckBox::Check(bool bCheck) {
	m_bChecked = bCheck;
	RefreshState();
}

void ModTreeControl_CheckBox::RefreshStateImage(void) {
	if(m_bChecked)
		m_TreeControl.SetStateImages(this, std::string("CHECKBOX_CHECKED"), std::string("CHECKBOX_CHECKED"));
	else
		m_TreeControl.SetStateImages(this, std::string("CHECKBOX_UNCHECKED"), std::string("CHECKBOX_UNCHECKED"));
}

void ModTreeControl_CheckBox::IconHit(void) {
	if(BranchEnabled())
		Check(!m_bChecked);
}

void ModTreeControl_CheckBox::SpaceKeyHit(void) {
	IconHit();
}

//////////////////////////////////////////////////////////////////////

ModTreeControl_RadioButton::ModTreeControl_RadioButton(ModTreeControl& treeControl)
: ModTreeControl_Control(treeControl)
, m_bChecked(false)
{
}

bool ModTreeControl_RadioButton::SubBranchEnabled(void) {
	if(!ModTreeControl_Control::SubBranchEnabled())
		return false;
	return m_bChecked;
}

bool ModTreeControl_RadioButton::Checked(void) {
	return m_bChecked;
}

void ModTreeControl_RadioButton::Check(bool bCheck) {
	if(m_bChecked = bCheck)
		m_Parent->CheckRadioButton(*this);
	RefreshState();
}

void ModTreeControl_RadioButton::RefreshStateImage(void) {
	if(m_bChecked)
		m_TreeControl.SetStateImages(this, std::string("RADIOBUTTON_CHECKED"), std::string("RADIOBUTTON_CHECKED"));
	else
		m_TreeControl.SetStateImages(this, std::string("RADIOBUTTON_UNCHECKED"), std::string("RADIOBUTTON_UNCHECKED"));
}

void ModTreeControl_RadioButton::IconHit(void) {
	if(BranchEnabled())
		Check(true);
}

void ModTreeControl_RadioButton::SpaceKeyHit(void) {
	IconHit();
}

//////////////////////////////////////////////////////////////////////

ModTreeControl_EditBox::ModTreeControl_EditBox(ModTreeControl& treeControl)
: ModTreeControl_Control(treeControl)
, m_Value(_T(""))
, m_Width(100)
, m_hWnd(NULL)
{
}

bool ModTreeControl_EditBox::SetValue(LPCTSTR pszValue) {
	m_Value = pszValue;
	return RefreshString();
}

LPCTSTR ModTreeControl_EditBox::GetValue(void) {
	return m_Value.c_str();
}

bool ModTreeControl_EditBox::RefreshString(void) {
	tcstring label;
	label = m_Label;
	label += _T(" : ");
	if(GetValue())
		label += GetValue();
	return m_TreeControl.SetItemLabel(this, label);
}

void ModTreeControl_EditBox::RefreshStateImage(void) {
	m_TreeControl.SetStateImages(this, std::string("EDITBOX"), std::string("EDITBOX"));
}

void ModTreeControl_EditBox::ItemHit(void) {
	if(BranchEnabled()) {
		m_TreeControl.SetItemLabel(this, m_Label);
		m_TreeControl.EditValue(*this);
	}
}

void ModTreeControl_EditBox::SetWidth(int nWidth) {
	if(nWidth > 0)
		m_Width = nWidth;
}

int ModTreeControl_EditBox::GetWidth(void) {
	return m_Width;
}

HWND ModTreeControl_EditBox::StartEdit(HWND hParent, RECT& rPos) {
	rPos.right = rPos.left + m_Width;
	if(m_hWnd = CreateWindowEx(WS_EX_CLIENTEDGE, _T("EDIT"), _T(""), WS_VISIBLE | WS_CHILD | WS_BORDER | ES_AUTOHSCROLL
		, rPos.left, rPos.top, rPos.right - rPos.left, rPos.bottom - rPos.top, hParent, (HMENU)2, g_hInst, NULL)) {
		HFONT hFont = (HFONT)SendMessage(hParent, WM_GETFONT, 0, 0);
		SendMessage(m_hWnd, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(TRUE, 0));
		SetWindowText(m_hWnd, m_Value.c_str());
	}
	return m_hWnd;
}

void ModTreeControl_EditBox::EndEdit(void) {
	if(m_hWnd) {
		int textLength = GetWindowTextLength(m_hWnd);
		m_Value.resize(textLength);
		GetWindowText(m_hWnd, (LPTSTR)m_Value.c_str(), m_Value.length()+1);
		DestroyWindow(m_hWnd);
		m_hWnd = NULL;
		RefreshString();
	}
}

//////////////////////////////////////////////////////////////////////

ModTreeControl_ComboBox::ModTreeControl_ComboBox(ModTreeControl& treeControl)
: ModTreeControl_EditBox(treeControl)
, m_SelectedValue(0)
, m_ShownValues(4)
{
}

LPCTSTR ModTreeControl_ComboBox::GetValue(void) {
	return GetSelectedValue();
}

void ModTreeControl_ComboBox::RefreshStateImage(void) {
	m_TreeControl.SetStateImages(this, std::string("COMBOBOX"), std::string("COMBOBOX"));
}

void ModTreeControl_ComboBox::AddValue(LPCTSTR pszValue, bool bSelected) {
	if(!pszValue)
		return;
	tcstring newValue = pszValue;
	m_ValueList.push_back(newValue);
	if(bSelected)
		SelectValue(m_ValueList.size() - 1);
}

void ModTreeControl_ComboBox::SelectValue(int nIndex) {
	if( (nIndex>=0) && (nIndex<m_ValueList.size()) && (nIndex!=m_SelectedValue) ) {
		m_SelectedValue = nIndex;
		m_TreeControl.RefreshNeededSize();
	}
	RefreshString();
}

int ModTreeControl_ComboBox::GetSelected(void) {
	return m_SelectedValue;
}

LPCTSTR ModTreeControl_ComboBox::GetSelectedValue(void) {
	if( (m_SelectedValue>=0) && (m_SelectedValue<m_ValueList.size()) )
		return m_ValueList[m_SelectedValue].c_str();
	return NULL;
}

void ModTreeControl_ComboBox::SetShownValues(int nNumber) {
	if(nNumber > 0)
		m_ShownValues = nNumber;
}

HWND ModTreeControl_ComboBox::StartEdit(HWND hParent, RECT& rPos) {
	rPos.right = rPos.left + m_Width;
	if(m_hWnd = CreateWindowEx(WS_EX_CLIENTEDGE, _T("COMBOBOX"), _T(""), WS_VISIBLE | WS_CHILD | WS_BORDER | WS_VSCROLL | CBS_AUTOHSCROLL | CBS_DROPDOWNLIST | CBS_HASSTRINGS
		, rPos.left, rPos.top, rPos.right - rPos.left, (m_ShownValues+1)*(rPos.bottom - rPos.top), hParent, (HMENU)2, g_hInst, NULL)) {
		HFONT hFont = (HFONT)SendMessage(hParent, WM_GETFONT, 0, 0);
		SendMessage(m_hWnd, WM_SETFONT, (WPARAM)hFont, MAKELPARAM(TRUE, 0));
		RECT wRect;
		if(GetWindowRect(m_hWnd, &wRect))
			rPos.bottom = rPos.top + (wRect.bottom - wRect.top);
		int nPosX = rPos.left;
		int nPosY = rPos.top;
		int nWidth = rPos.right - rPos.left;
		int nHeight = rPos.bottom - rPos.top;
		int itemHeight = SendMessage(m_hWnd, CB_GETITEMHEIGHT, 0, 0);
		if(itemHeight!=CB_ERR)
			SetWindowPos(m_hWnd, HWND_TOP, nPosX, nPosY, nWidth, nHeight+m_ShownValues*itemHeight+5, SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | SWP_SHOWWINDOW);
		int iItem = -1;
		while(++iItem < m_ValueList.size())
			SendMessage(m_hWnd, CB_ADDSTRING, 0, (LPARAM)m_ValueList[iItem].c_str());
		int nbItems = SendMessage(m_hWnd, CB_GETCOUNT, 0, 0);
		TCHAR szBuf[1024];
		HGDIOBJ hOld;
		HDC hDC = GetDC(m_hWnd);
		sint32 horExtent = 0;
		hFont = (HFONT)SendMessage(m_hWnd, WM_GETFONT, 0, 0);
		if(!hFont)
			hFont = (HFONT)GetStockObject(SYSTEM_FONT);
		hOld = SelectObject(hDC, hFont);
		if(nbItems!=CB_ERR) {
			while(nbItems--) {
				memset(szBuf, 0, sizeof(szBuf));
				if(SendMessage(m_hWnd, CB_GETLBTEXT, (WPARAM)nbItems, (LPARAM)szBuf) != CB_ERR) {
					SIZE newExtent;
					if(!GetTextExtentPoint32(hDC, szBuf, lstrlen(szBuf), &newExtent))
						newExtent.cx = 0;
					if(newExtent.cx > horExtent)
						horExtent = newExtent.cx;
				}
			}
		}
		if(m_ShownValues<m_ValueList.size())
			horExtent += GetSystemMetrics(SM_CXVSCROLL);
		if(horExtent+5 > nWidth)
			SendMessage(m_hWnd, CB_SETDROPPEDWIDTH, horExtent+5, 0);
		SelectObject(hDC, hOld);
		ReleaseDC(m_hWnd, hDC);
		SendMessage(m_hWnd, CB_SETCURSEL, m_SelectedValue, 0);
	}
	return m_hWnd;
}

void ModTreeControl_ComboBox::EndEdit(void) {
	if(m_hWnd) {
		int nSelected = SendMessage(m_hWnd, CB_GETCURSEL, 0, 0);
		if( (nSelected>=0) && (nSelected!=CB_ERR) )
			SelectValue(nSelected);
		DestroyWindow(m_hWnd);
		m_hWnd = NULL;
	}
}

//////////////////////////////////////////////////////////////////////

ModTreeControl_FileSelect::ModTreeControl_FileSelect(ModTreeControl& treeControl)
: ModTreeControl_EditBox(treeControl)
{
	m_Filters = _T("All files (*.*)\0")	_T("*.*\0");
}

void ModTreeControl_FileSelect::SetFilters(LPCTSTR szFilters) {
	m_Filters = szFilters;
}

void ModTreeControl_FileSelect::RefreshStateImage(void) {
	m_TreeControl.SetStateImages(this, std::string("FILESELECT"), std::string("FILESELECT"));
}

HWND ModTreeControl_FileSelect::StartEdit(HWND hParent, RECT& rPos) {
	rPos.right = rPos.left + m_Width;
	ModFileSelectControl *control = ModFileSelectControl::Create(hParent
		, rPos.left, rPos.top, rPos.right-rPos.left, rPos.bottom-rPos.top, 2);
	if(control) {
		m_hWnd = control->GetWindow();
		HFONT hFont = (HFONT)SendMessage(hParent, WM_GETFONT, 0, 0);
		control->SetFont(hFont, TRUE);
		control->SetFileName(GetValue());
		control->SetFilters(m_Filters);
	}
	return m_hWnd;
}

void ModTreeControl_FileSelect::EndEdit(void) {
	if(m_hWnd) {
		ModFileSelectControl *control = GetModFileSelectControl(m_hWnd);
		m_Value = control->GetFileName();
		DestroyWindow(m_hWnd);
		m_hWnd = NULL;
		RefreshString();
	}
}

//////////////////////////////////////////////////////////////////////

ATOM RegisterModTreeControl() {
	WNDCLASS wc;

	wc.style		= CS_DBLCLKS;
	wc.lpfnWndProc	= ModTreeControl::StaticWndProc;
	wc.cbClsExtra	= 0;
	wc.cbWndExtra	= sizeof(ModTreeControl *);
	wc.hInstance	= g_hInst;
	wc.hIcon		= NULL;
	wc.hCursor		= NULL;
	wc.hbrBackground= /*(HBRUSH)(COLOR_BACKGROUND+1)*/0;
	wc.lpszMenuName	= NULL;
	wc.lpszClassName= g_szModTreeControlName;

	return RegisterClass(&wc);
}

ModTreeControl *GetModTreeControl(HWND hwnd) {
	return reinterpret_cast<ModTreeControl *>(GetWindowLong(hwnd, 0));
}

ModTreeControl *ModTreeControl::Create(HWND hwndParent, int x, int y, int cx, int cy, UINT id) {
	HWND hwnd = CreateWindowEx(WS_EX_CLIENTEDGE | WS_EX_TRANSPARENT | WS_EX_CONTROLPARENT, g_szModTreeControlName, "Moo"
		, WS_VISIBLE | WS_CHILD, x, y, cx, cy, hwndParent, (HMENU)id, g_hInst, NULL);

	if(hwnd)
		return GetModTreeControl(hwnd);

	return NULL;
}

ModTreeControl::ModTreeControl(HWND hwnd)
: ModTreeControl_Control(*this)
, m_hWnd(hwnd)
, m_hTreeList(NULL)
, m_hInputWnd(NULL)
, m_InputEdit(NULL)
, m_LargestItem(NULL)
, m_VScrollBarWidth(GetSystemMetrics(SM_CXVSCROLL))
, m_HScrollBarHeight(GetSystemMetrics(SM_CYHSCROLL))
, m_BKColor(RGB(0xFF, 0xFF, 0xFF))
, m_SelectedColor(RGB(0xC0, 0xC0, 0xC0))
, m_hSelectedItem(NULL)
{
	m_Parent = NULL;
	m_ClientSize.cx = 0;
	m_ClientSize.cy = 0;
	m_TreeListSize.cx = 0;
	m_TreeListSize.cy = 0;
	m_TreeListActualSize.cx = 0;
	m_TreeListActualSize.cy = 0;
	m_ScrollPos.cx = 0;
	m_ScrollPos.cy = 0;
}

ModTreeControl::~ModTreeControl(void) {
}

LRESULT CALLBACK ModTreeControl::StaticWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam) {
	ModTreeControl *pThis = (ModTreeControl *)GetWindowLong(hwnd, 0);

	switch(msg) {
	case WM_NCCREATE:
		if (!(pThis = new ModTreeControl(hwnd)))
			return FALSE;
		SetWindowLong(hwnd, 0, (DWORD)pThis);
		break;

	case WM_NCDESTROY:
		delete pThis;
		return DefWindowProc(hwnd, msg, wParam, lParam);
	}

	return pThis->WndProc(msg, wParam, lParam);
}

LRESULT ModTreeControl::WndProc(UINT msg, WPARAM wParam, LPARAM lParam) {
	switch(msg) {
	case WM_CREATE:
		if(!Init())
			return -1;
		break;
	case WM_DESTROY:
		ClearInputControl();
		break;
	case WM_SIZE:
		{
			RECT wRect;
			if(GetWindowRect(m_hWnd, &wRect)) {
				SIZE newSize;
				newSize.cx = wRect.right - wRect.left - 2*GetSystemMetrics(SM_CXEDGE);
				newSize.cy = wRect.bottom - wRect.top - 2*GetSystemMetrics(SM_CYEDGE);
				if( (newSize.cx!=m_ClientSize.cx) || (newSize.cy!=m_ClientSize.cy) ) {
					m_ClientSize = newSize;
					ClearInputControl();
					SetTreeListSize(m_TreeListSize.cx, m_TreeListSize.cy);
				}
			}
		}
		break;
	case WM_HSCROLL:
		{
			SCROLLINFO si;
			si.cbSize = sizeof(si);
			si.fMask = SIF_ALL;
			GetScrollInfo(m_hWnd, SB_HORZ, &si);
			long newPos = si.nPos;
			switch(LOWORD(wParam)) {
			case SB_LEFT:
				newPos = 0;
				break;
			case SB_PAGELEFT:
				newPos -= 10;
				break;
			case SB_LINELEFT:
				newPos--;
				break;
			case SB_LINERIGHT:
				newPos++;
				break;
			case SB_PAGERIGHT:
				newPos += 10;
				break;
			case SB_RIGHT:
				newPos = si.nMax;
				break;
			case SB_THUMBTRACK:
				newPos = si.nTrackPos;
				break;
			}
			if(newPos < si.nMin)
				newPos = si.nMin;
			if(newPos > si.nMax - (int)si.nPage + 1)
				newPos = si.nMax - (int)si.nPage + 1;
			if(newPos != m_ScrollPos.cx) {
				ScrollWindow(m_hWnd, m_ScrollPos.cx - newPos, 0, NULL, NULL);
				UpdateWindow(m_hWnd);
				m_ScrollPos.cx = newPos;
				si.nPos = newPos;
				SetScrollInfo(m_hWnd, SB_HORZ, &si, TRUE);
			}
		}
		return TRUE;
	case WM_VSCROLL:
		{
			SCROLLINFO si;
			si.cbSize = sizeof(si);
			si.fMask = SIF_ALL;
			GetScrollInfo(m_hWnd, SB_VERT, &si);
			long newPos = si.nPos;
			switch(LOWORD(wParam)) {
			case SB_TOP:
				newPos = 0;
				break;
			case SB_PAGEUP:
				newPos -= 10;
				break;
			case SB_LINEUP:
				newPos--;
				break;
			case SB_LINEDOWN:
				newPos++;
				break;
			case SB_PAGEDOWN:
				newPos += 10;
				break;
			case SB_BOTTOM:
				newPos = si.nMax;
				break;
			case SB_THUMBTRACK:
				newPos = si.nTrackPos;
				break;
			}
			if(newPos < si.nMin)
				newPos = si.nMin;
			if(newPos > si.nMax - (int)si.nPage + 1)
				newPos = si.nMax - (int)si.nPage + 1;
			if(newPos != m_ScrollPos.cy) {
				ScrollWindow(m_hWnd, 0, m_ScrollPos.cy - newPos, NULL, NULL);
				UpdateWindow(m_hWnd);
				m_ScrollPos.cy = newPos;
				si.nPos = newPos;
				SetScrollInfo(m_hWnd, SB_VERT, &si, TRUE);
			}
		}
		return TRUE;
	case WM_NOTIFY:
		{
			LPNMTREEVIEW pnmh = (LPNMTREEVIEW)lParam;
			if(pnmh->hdr.hwndFrom == m_hTreeList) {
				switch(pnmh->hdr.code) {
				case TVN_SELCHANGED:
					m_hSelectedItem = TreeView_GetSelection(m_hTreeList);
					break;
				case TVN_SINGLEEXPAND:
					RefreshNeededSize();
					return TVNRET_DEFAULT;
				case TVN_ITEMEXPANDED:
					RefreshNeededSize();
					break;
				case TVN_KEYDOWN:
					{
						LPNMTVKEYDOWN ptvkd = (LPNMTVKEYDOWN)lParam;
						if(ptvkd->wVKey == VK_SPACE) {
							HTREEITEM hItem = TreeView_GetSelection(m_hTreeList);
							if(!hItem)
								break;
							ModTreeControl_Control *hitControl = (ModTreeControl_Control *)ModTreeControl::GetItemData(m_hTreeList, hItem);
							if(!hitControl)
								break;
							hitControl->SpaceKeyHit();
							break;
						} else if(ptvkd->wVKey == VK_RETURN)
							return TRUE;
					}
					break;
				case NM_CLICK:
				case NM_DBLCLK:
					{
						POINT pPoint;
						if(GetCursorPos(&pPoint)) {
							RECT listRect;
							if(pnmh->hdr.hwndFrom == m_hInputWnd)
								break;
							if(pnmh->hdr.hwndFrom != m_hTreeList) {
								ClearInputControl();
								break;
							}
							GetWindowRect(m_hTreeList, &listRect);
							long relX = pPoint.x - listRect.left;
							long relY = pPoint.y - listRect.top;
							TVHITTESTINFO hitInfo;
							hitInfo.pt.x = relX;
							hitInfo.pt.y = relY;
							HTREEITEM hItem = TreeView_HitTest(m_hTreeList, &hitInfo);
							if(!hItem) {
								ClearInputControl();
								break;
							}
							ModTreeControl_Control *hitControl = (ModTreeControl_Control *)ModTreeControl::GetItemData(m_hTreeList, hItem);
							if(!hitControl) {
								ClearInputControl();
								break;
							}
							if( (hitControl!=m_InputEdit) || (hitInfo.flags&TVHT_ONITEMRIGHT) || (hitInfo.flags&TVHT_ONITEMINDENT) )
								ClearInputControl();
							if(hitInfo.flags & TVHT_ONITEMICON)
								hitControl->IconHit();
							else if(hitInfo.flags & TVHT_ONITEMLABEL)
								hitControl->LabelHit();
							break;
						}
					}
					break;
				case TVN_GETINFOTIP:
					{
						LPNMTVGETINFOTIP lpGetInfoTip = (LPNMTVGETINFOTIP)lParam;
						ModTreeControl_Control *control = (ModTreeControl_Control *)lpGetInfoTip->lParam;
						if(!control)
							break;
						tcstring infoTip = control->GetInfoTip();
						lstrcpyn(lpGetInfoTip->pszText, infoTip.c_str(), lpGetInfoTip->cchTextMax);
					}
					break;
				case NM_CUSTOMDRAW:
					{
						LPNMTVCUSTOMDRAW  lptvcd = (LPNMTVCUSTOMDRAW)lParam;
						switch(lptvcd->nmcd.dwDrawStage) {
						case CDDS_PREPAINT :
							return CDRF_NOTIFYITEMDRAW;
						case CDDS_ITEMPREPAINT:
						case CDDS_SUBITEM | CDDS_ITEMPREPAINT:
							{
								ModTreeControl_Control *control = (ModTreeControl_Control *)lptvcd->nmcd.lItemlParam;
								if(!control)
									break;
								lptvcd->clrText = control->GetTextColor();
								if(m_hSelectedItem == (HTREEITEM)lptvcd->nmcd.dwItemSpec)
									lptvcd->clrTextBk = m_SelectedColor;
								else
									lptvcd->clrTextBk = m_BKColor;
								return CDRF_NEWFONT;
							}
						}
					}
					break;
				}
				break;
			}
		}
	default:
		return DefWindowProc(m_hWnd, msg, wParam, lParam);
	}
	return FALSE;
}

bool ModTreeControl::Init(void) {
	RECT cRect;
	if(!GetClientRect(m_hWnd, &cRect)) {
		cRect.left = 0;
		cRect.top = 0;
		cRect.right = CW_USEDEFAULT;
		cRect.bottom = CW_USEDEFAULT;
	}
	DWORD treeStyle = WS_VISIBLE | WS_CHILD | WS_TABSTOP
		| TVS_NOSCROLL | TVS_INFOTIP | TVS_HASBUTTONS | TVS_HASLINES | TVS_LINESATROOT | TVS_DISABLEDRAGDROP;

	m_ClientSize.cx = m_TreeListSize.cx = cRect.right - cRect.left;
	m_ClientSize.cy = m_TreeListSize.cy = cRect.bottom - cRect.top;
	m_hTreeList = CreateWindowEx(/*WS_EX_TRANSPARENT*/0, WC_TREEVIEW, "Tree View"
		, treeStyle
		, 0, 0, m_TreeListSize.cx, m_TreeListSize.cy
		, m_hWnd, (HMENU)1, g_hInst, NULL);

	if(m_hTreeList) {
		TreeView_SetBkColor(m_hTreeList, m_BKColor);
		HIMAGELIST imageList;
		HBITMAP hBitmap;
		imageList = ImageList_Create(16, 16, ILC_MASK | ILC_COLOR24, 6, 0);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_GROUP));
		m_StateImageList.insert(std::make_pair(std::string("GROUP"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_EDITBOX));
		m_StateImageList.insert(std::make_pair(std::string("EDITBOX"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_COMBOBOX));
		m_StateImageList.insert(std::make_pair(std::string("COMBOBOX"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_FILESELECT));
		m_StateImageList.insert(std::make_pair(std::string("FILESELECT"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_CHECKBOX_UNCHECKED));
		m_StateImageList.insert(std::make_pair(std::string("CHECKBOX_UNCHECKED"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_CHECKBOX_CHECKED));
		m_StateImageList.insert(std::make_pair(std::string("CHECKBOX_CHECKED"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_RADIOBUTTON_UNCHECKED));
		m_StateImageList.insert(std::make_pair(std::string("RADIOBUTTON_UNCHECKED"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		hBitmap = LoadBitmap(g_hInst, MAKEINTRESOURCE(IDB_RADIOBUTTON_CHECKED));
		m_StateImageList.insert(std::make_pair(std::string("RADIOBUTTON_CHECKED"), ImageList_Add(imageList, hBitmap, NULL)));
		DeleteObject(hBitmap);
		TreeView_SetImageList(m_hTreeList, imageList, TVSIL_NORMAL);
	}

	return !!m_hTreeList;
}

bool ModTreeControl::SetItemLabel(HWND hwndTreeView, HTREEITEM hItem, LPCTSTR pszLabel) {
	TVITEM tvItem;

	tvItem.mask = TVIF_HANDLE | TVIF_TEXT;
	tvItem.hItem = hItem;
	tvItem.pszText = (LPTSTR)pszLabel;
	// Shouldn't be necessary, but who knows :>
	tvItem.cchTextMax = (pszLabel ? lstrlen(pszLabel) : 0);

	return !!TreeView_SetItem(hwndTreeView, &tvItem);
}

bool ModTreeControl::SetItemLabel(ModTreeControl_Control *control, tcstring label) {
	if(!control)
		return false;
	return !!SetItemLabel(m_hTreeList, control->GetItem(), label.c_str());
}

bool ModTreeControl::SetItemData(HWND hwndTreeView, HTREEITEM hItem, LPARAM itemData) {
	TVITEM tvItem;

	tvItem.mask = TVIF_HANDLE | TVIF_PARAM;
	tvItem.hItem = hItem;
	tvItem.lParam = itemData;

	return !!TreeView_SetItem(hwndTreeView, &tvItem);
}

LPARAM ModTreeControl::GetItemData(HWND hwndTreeView, HTREEITEM hItem) {
	TVITEM tvItem;

	tvItem.mask = TVIF_HANDLE | TVIF_PARAM;
	tvItem.hItem = hItem;
	tvItem.lParam = 0;

	if(!!TreeView_GetItem(hwndTreeView, &tvItem))
		return tvItem.lParam;

	return 0;
}

bool ModTreeControl::SetStateImages(HWND hwndTreeView, HTREEITEM hItem, int iImage, int iSelectedImage) {
	TVITEM tvItem;

	tvItem.mask = TVIF_HANDLE | TVIF_IMAGE | TVIF_SELECTEDIMAGE;
	tvItem.hItem = hItem;
	tvItem.iImage = iImage;
	tvItem.iSelectedImage = iSelectedImage;

	return !!TreeView_SetItem(hwndTreeView, &tvItem);
}

bool ModTreeControl::SetStateImages(ModTreeControl_Control *control, std::string sImage, std::string sSelectedImage) {
	if(!control)
		return false;
	std::map<std::string, int>::const_iterator it;
	it = m_StateImageList.find(sImage);
	int iImage = (it!=m_StateImageList.end() ? it->second : 0);
	it = m_StateImageList.find(sSelectedImage);
	int iSelectedImage = (it!=m_StateImageList.end() ? it->second : 0);
	return SetStateImages(m_hTreeList, control->GetItem(), iImage, iSelectedImage);
}

ModTreeControl_Control *ModTreeControl::AddGroup(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszToolTip) {
	TVINSERTSTRUCT itemData;
	itemData.hParent = (parent ? parent->GetItem() : NULL);
	itemData.hInsertAfter = TVI_LAST;
	itemData.item.mask = TVIF_STATE;
	itemData.item.stateMask = TVIS_STATEIMAGEMASK;
	itemData.item.state = TVIS_BOLD | TVIS_EXPANDED;

	ModTreeControl_Control *newControl = NULL;
	HTREEITEM newItem = TreeView_InsertItem(m_hTreeList, &itemData);
	if(newItem) {
		newControl = new ModTreeControl_Control(*this);
		newControl->SetItem(newItem);
		SetItemData(m_hTreeList, newItem, (LPARAM)newControl);
		if(!parent)
			parent = this;
		parent->AddChild(*newControl);
		newControl->SetLabel(pszLabel);
		newControl->SetToolTip(pszToolTip);
		newControl->RefreshState();
		RefreshNeededSize();
	}

	return newControl;
}

ModTreeControl_CheckBox *ModTreeControl::AddCheckBox(ModTreeControl_Control *parent, LPCTSTR pszLabel, bool bChecked, LPCTSTR pszToolTip) {
	TVINSERTSTRUCT itemData;
	itemData.hParent = (parent ? parent->GetItem() : NULL);
	itemData.hInsertAfter = TVI_LAST;
	itemData.item.mask = 0;

	ModTreeControl_CheckBox *newControl = NULL;
	HTREEITEM newItem = TreeView_InsertItem(m_hTreeList, &itemData);
	if(newItem) {
		newControl = new ModTreeControl_CheckBox(*this);
		newControl->SetItem(newItem);
		SetItemData(m_hTreeList, newItem, (LPARAM)newControl);
		if(!parent)
			parent = this;
		parent->AddChild(*newControl);
		newControl->SetLabel(pszLabel);
		newControl->SetToolTip(pszToolTip);
		newControl->Check(bChecked);
		RefreshNeededSize();
	}

	return newControl;
}

ModTreeControl_RadioButton *ModTreeControl::AddRadioButton(ModTreeControl_Control *parent, LPCTSTR pszLabel, bool bChecked, LPCTSTR pszToolTip) {
	TVINSERTSTRUCT itemData;
	itemData.hParent = (parent ? parent->GetItem() : NULL);
	itemData.hInsertAfter = TVI_LAST;
	itemData.item.mask = 0;

	ModTreeControl_RadioButton *newControl = NULL;
	HTREEITEM newItem = TreeView_InsertItem(m_hTreeList, &itemData);
	if(newItem) {
		newControl = new ModTreeControl_RadioButton(*this);
		newControl->SetItem(newItem);
		SetItemData(m_hTreeList, newItem, (LPARAM)newControl);
		if(!parent)
			parent = this;
		parent->AddChild(*newControl);
		newControl->SetLabel(pszLabel);
		newControl->SetToolTip(pszToolTip);
		parent->AddRadioButton(*newControl);
		newControl->Check(bChecked);
		RefreshNeededSize();
	}

	return newControl;
}

ModTreeControl_EditBox *ModTreeControl::AddEditBox(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszValue, LPCTSTR pszToolTip) {
	TVINSERTSTRUCT itemData;
	itemData.hParent = (parent ? parent->GetItem() : NULL);
	itemData.hInsertAfter = TVI_LAST;
	itemData.item.mask = 0;

	ModTreeControl_EditBox *newControl = NULL;
	HTREEITEM newItem = TreeView_InsertItem(m_hTreeList, &itemData);
	if(newItem) {
		newControl = new ModTreeControl_EditBox(*this);
		newControl->SetItem(newItem);
		SetItemData(m_hTreeList, newItem, (LPARAM)newControl);
		if(!parent)
			parent = this;
		parent->AddChild(*newControl);
		newControl->SetLabel(pszLabel);
		newControl->SetToolTip(pszToolTip);
		newControl->SetValue(pszValue);
		newControl->RefreshState();
		RefreshNeededSize();
	}

	return newControl;
}

ModTreeControl_ComboBox *ModTreeControl::AddComboBox(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszToolTip) {
	TVINSERTSTRUCT itemData;
	itemData.hParent = (parent ? parent->GetItem() : NULL);
	itemData.hInsertAfter = TVI_LAST;
	itemData.item.mask = 0;

	ModTreeControl_ComboBox *newControl = NULL;
	HTREEITEM newItem = TreeView_InsertItem(m_hTreeList, &itemData);
	if(newItem) {
		newControl = new ModTreeControl_ComboBox(*this);
		newControl->SetItem(newItem);
		SetItemData(m_hTreeList, newItem, (LPARAM)newControl);
		if(!parent)
			parent = this;
		parent->AddChild(*newControl);
		newControl->SetLabel(pszLabel);
		newControl->SetToolTip(pszToolTip);
		newControl->RefreshState();
		RefreshNeededSize();
	}

	return newControl;
}

ModTreeControl_FileSelect *ModTreeControl::AddFileSelect(ModTreeControl_Control *parent, LPCTSTR pszLabel, LPCTSTR pszValue, LPCTSTR pszToolTip) {
	TVINSERTSTRUCT itemData;
	itemData.hParent = (parent ? parent->GetItem() : NULL);
	itemData.hInsertAfter = TVI_LAST;
	itemData.item.mask = 0;

	ModTreeControl_FileSelect *newControl = NULL;
	HTREEITEM newItem = TreeView_InsertItem(m_hTreeList, &itemData);
	if(newItem) {
		newControl = new ModTreeControl_FileSelect(*this);
		newControl->SetItem(newItem);
		SetItemData(m_hTreeList, newItem, (LPARAM)newControl);
		if(!parent)
			parent = this;
		parent->AddChild(*newControl);
		newControl->SetLabel(pszLabel);
		newControl->SetToolTip(pszToolTip);
		newControl->SetValue(pszValue);
		newControl->RefreshState();
		RefreshNeededSize();
	}

	return newControl;
}

bool ModTreeControl::Expand(bool bRecursive) {
	bool bResult = true;
	std::list<ModTreeControl_Control *>::const_iterator it;
	for(it=m_Children.begin() ; it!=m_Children.end() ; it++)
		bResult &= (*it)->Expand(bRecursive);
	RefreshNeededSize();
	return bResult;
}

bool ModTreeControl::Collapse(bool bRecursive) {
	bool bResult = true;
	std::list<ModTreeControl_Control *>::const_iterator it;
	for(it=m_Children.begin() ; it!=m_Children.end() ; it++)
		bResult &= (*it)->Collapse(bRecursive);
	RefreshNeededSize();
	return bResult;
}

bool ModTreeControl::Expand(HTREEITEM item) {
	bool bResult = false;
	if(item)
		bResult = !!TreeView_Expand(m_hTreeList, item, TVE_EXPAND);
	RefreshNeededSize();
	return bResult;
}

bool ModTreeControl::Collapse(HTREEITEM item) {
	bool bResult = false;
	if(item)
		bResult = !!TreeView_Expand(m_hTreeList, item, TVE_COLLAPSE);
	RefreshNeededSize();
	return bResult;
}

bool ModTreeControl::ExpandCollapse(HTREEITEM item) {
	if(item) {
		if(TreeView_GetItemState(m_hTreeList, item, TVIS_EXPANDED) & TVIS_EXPANDED)
			return Collapse(item);
		else
			return Expand(item);
	}
	return false;
}

void ModTreeControl::EditValue(ModTreeControl_EditBox& editBox) {
	RECT cRect;
	if(TreeView_GetItemRect(m_hTreeList, editBox.GetItem(), &cRect, TRUE)) {
		RECT rPos;
		int nBorder = 2 * GetSystemMetrics(SM_CYEDGE);
		rPos.left = cRect.right + 5;
		rPos.right = rPos.left + editBox.GetWidth();
		rPos.top = cRect.top - nBorder/2;
		rPos.bottom = rPos.top + cRect.bottom - cRect.top + nBorder;
		if(m_InputEdit != &editBox) {
			ClearInputControl();
			m_InputEdit = &editBox;
			if(m_hInputWnd = editBox.StartEdit(m_hTreeList, rPos)) {
				int neededWidth = rPos.right + 5;
				int neededHeight = rPos.bottom + 5;
				if( (neededWidth>m_TreeListSize.cx) || (neededHeight>m_TreeListSize.cy) )
					SetTreeListSize(
					neededWidth>m_TreeListSize.cx ? neededWidth : m_TreeListSize.cx
					, neededHeight>m_TreeListSize.cy ? neededHeight : m_TreeListSize.cy
					, true
					);
				SCROLLINFO sInfo;
				sInfo.cbSize = sizeof(sInfo);
				sInfo.fMask = SIF_ALL;
				if(GetClientRect(m_hWnd, &cRect) && GetScrollInfo(m_hWnd, SB_HORZ, &sInfo) && (sInfo.nMax>0) && (sInfo.nPage>0) ) {
					int nPos = sInfo.nPos;
					if(rPos.right > nPos + sInfo.nPage)
						nPos = rPos.right - sInfo.nPage;
					if(nPos > rPos.left)
						nPos = rPos.left;
					if(nPos != sInfo.nPos) {
						sInfo.nPos = nPos - 1;
						SetScrollInfo(m_hWnd, SB_HORZ, &sInfo, TRUE);
						SendMessage(m_hWnd, WM_HSCROLL, MAKEWPARAM(SB_LINERIGHT, 0), (LPARAM)0);
					}
				}
			}
		}
	}
}

void ModTreeControl::ClearInputControl(void) {
	if(m_InputEdit) {
		m_InputEdit->EndEdit();
		m_hInputWnd = NULL;
		m_InputEdit = NULL;
		RefreshNeededSize();
	}
}

void ModTreeControl::UpdateScrollBars(void) {
	//RECT wRect;
	int clientWidth = m_ClientSize.cx;
	int clientHeight = m_ClientSize.cy;
	//GetWindowRect(m_hTreeList, &wRect);
	//int listWidth = wRect.right - wRect.left;
	//int listHeight = wRect.bottom - wRect.top;
	int listWidth = m_TreeListActualSize.cx;
	int listHeight = m_TreeListActualSize.cy;
	bool bShowHScroll = false;
	bool bShowVScroll = false;
	if(clientWidth < listWidth) {
		clientHeight -= m_HScrollBarHeight;
		bShowHScroll = true;
	}
	if(clientHeight<listHeight) {
		clientWidth -= m_VScrollBarWidth;
		bShowVScroll = true;
	}
	if( !bShowHScroll && (clientWidth<listWidth) ) {
		clientHeight -= m_HScrollBarHeight;
		bShowHScroll = true;
	}

	// Update the horizontal scrollbar
	SCROLLINFO sInfo;
	sInfo.cbSize = sizeof(sInfo);
	sInfo.fMask = SIF_RANGE | SIF_POS | SIF_PAGE;
	sInfo.nMin = 0;
	if(bShowHScroll) {
		sInfo.nPos = m_ScrollPos.cx;
		sInfo.nMax = listWidth - 1;
		sInfo.nPage = clientWidth;
		if(sInfo.nPos > sInfo.nMax - (int)sInfo.nPage + 1)
			sInfo.nPos = sInfo.nMax - (int)sInfo.nPage + 1;
		ShowScrollBar(m_hWnd, SB_HORZ, TRUE);
		SetScrollInfo(m_hWnd, SB_HORZ, &sInfo, TRUE);
	} else {
		sInfo.nPos = 0;
		ShowScrollBar(m_hWnd, SB_HORZ, FALSE);
	}
	// Update the window according to the change
	sInfo.fMask = SIF_ALL;
	if(sInfo.nPos != m_ScrollPos.cx) {
		ScrollWindow(m_hWnd, m_ScrollPos.cx - sInfo.nPos, 0, NULL, NULL);
		UpdateWindow(m_hWnd);
		m_ScrollPos.cx = sInfo.nPos;
		SetScrollInfo(m_hWnd, SB_HORZ, &sInfo, TRUE);
	}

	// Update the vertical scrollbar
	sInfo.fMask = SIF_RANGE | SIF_POS | SIF_PAGE;
	if(bShowVScroll) {
		sInfo.nPos = m_ScrollPos.cy;
		sInfo.nMax = listHeight - 1;
		sInfo.nPage = clientHeight;
		if(sInfo.nPos > sInfo.nMax - (int)sInfo.nPage + 1)
			sInfo.nPos = sInfo.nMax - (int)sInfo.nPage + 1;
		ShowScrollBar(m_hWnd, SB_VERT, TRUE);
		SetScrollInfo(m_hWnd, SB_VERT, &sInfo, TRUE);
	} else {
		sInfo.nPos = 0;
		ShowScrollBar(m_hWnd, SB_VERT, FALSE);
	}
	sInfo.fMask = SIF_ALL;
	if(sInfo.nPos != m_ScrollPos.cy) {
		ScrollWindow(m_hWnd, 0, m_ScrollPos.cy - sInfo.nPos, NULL, NULL);
		UpdateWindow(m_hWnd);
		m_ScrollPos.cy = sInfo.nPos;
		SetScrollInfo(m_hWnd, SB_VERT, &sInfo, TRUE);
	}
}

void ModTreeControl::SetTreeListSize(int nWidth, int nHeight, bool bOnlyShowOff) {
	if(!bOnlyShowOff) {
		m_TreeListSize.cx = nWidth;
		m_TreeListSize.cy = nHeight;
	}
	m_TreeListActualSize.cx = nWidth;
	m_TreeListActualSize.cy = nHeight;
	if(nWidth < m_ClientSize.cx)
		nWidth = m_ClientSize.cx;
	if(nHeight < m_ClientSize.cy)
		nHeight = m_ClientSize.cy;
	SetWindowPos(m_hTreeList, HWND_TOP, 0, 0, nWidth, nHeight, SWP_NOMOVE | SWP_NOOWNERZORDER | SWP_NOZORDER | SWP_SHOWWINDOW);
	UpdateScrollBars();
}

void ModTreeControl::RefreshNeededSize(void) {
	HTREEITEM hItem = NULL;
	RECT iRect;
	m_LargestItem = NULL;
	m_TreeListSize.cx = 0;
	m_TreeListSize.cy = 0;
	hItem = TreeView_GetLastVisible(m_hTreeList);
	if(hItem && TreeView_GetItemRect(m_hTreeList, hItem, &iRect, TRUE)
		&& (iRect.bottom + 5 > m_TreeListSize.cy))
		m_TreeListSize.cy = iRect.bottom + 5;
	hItem = TreeView_GetFirstVisible(m_hTreeList);
	while(hItem) {
		if(TreeView_GetItemRect(m_hTreeList, hItem, &iRect, TRUE)) {
			if(iRect.right + 5 > m_TreeListSize.cx) {
				m_TreeListSize.cx = iRect.right + 5;
				m_LargestItem = hItem;
			}
		}
		hItem = TreeView_GetNextVisible(m_hTreeList, hItem);
	}
	if(!m_TreeListSize.cx)
		m_TreeListSize.cx = m_ClientSize.cx;
	if(!m_TreeListSize.cy)
		m_TreeListSize.cy = m_ClientSize.cy;
	SetTreeListSize(m_TreeListSize.cx, m_TreeListSize.cy);
}

void ModTreeControl::Flush(void) {
	ClearInputControl();
}
