//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include "SRTOutputFile.h"
#include "ModClassInterfaces.h"


SRTOutputStream::SRTOutputStream(SRTOutputFile *parent, eTextType inputType, eTextType outputType)
: AVIOutputStream(NULL)
, m_TextUtil(inputType, outputType, false, false) {
	m_Parent		= parent;
	//m_InputInType	= inputType;
	m_OutputInType	= outputType;
	m_SamplesProcessed = 0;
	m_SubtitleIndex	= 0;
}

bool SRTOutputStream::QueryInterface(InterfaceID ID, void** destination) {
	if(ID == IOverlappingOutput::GetInterfaceID()) {
		*destination = static_cast<IOverlappingOutput *>(this);
		return true;
	}
	return false;
}

void SRTOutputStream::write(uint32 flags, const void *pBuffer, uint32 cbBuffer, uint32 samples) {
	m_SamplesProcessed += samples;
	WriteData(flags, (void *)pBuffer, cbBuffer, m_SamplesProcessed - samples, samples);
}

BOOL SRTOutputStream::WriteData(LONG dwIndexFlags, LPVOID lpBuffer, LONG cbBuffer, LONG lSampleStart, LONG lSampleCount) {
	void *data = NULL;
	uint32 dataLen = 0;

	VDStringW subtitleW = m_TextUtil.GetText(lpBuffer, cbBuffer);
	
	if(subtitleW.size()) {
		VDStringW srtFormat;
		if(m_SubtitleIndex++)
			srtFormat = L"\r\n";
		srtFormat += VDStringW::setf(L"%lu\r\n", m_SubtitleIndex);
		//sint64 startTimeMS = m_SamplesProcessed;
		//sint64 endTimeMS = startTimeMS + lSamples;
		sint64 startTimeMS = lSampleStart;
		sint64 endTimeMS = startTimeMS + lSampleCount;
		int startHour, startMinute, startSecond, startMillisecond;
		int endHour, endMinute, endSecond, endMillisecond;
		startMillisecond = startTimeMS % 1000;
		startSecond = (startTimeMS / 1000) % 60;
		startMinute = (startTimeMS / 1000 / 60) % 60;
		startHour = (startTimeMS / 1000 / 60 / 60);
		endMillisecond = endTimeMS % 1000;
		endSecond = (endTimeMS / 1000) % 60;
		endMinute = (endTimeMS / 1000 / 60) % 60;
		endHour = (endTimeMS / 1000 / 60 / 60);
		srtFormat += VDStringW::setf(L"%02d:%02d:%02d,%03d --> %02d:%02d:%02d,%03d\r\n"
			, startHour, startMinute, startSecond, startMillisecond
			, endHour, endMinute, endSecond, endMillisecond);
		srtFormat += VDStringW::setf(L"%s\r\n", subtitleW.c_str());

		VDStringA srtFormatA;

		switch(m_OutputInType) {
		case kTextType_ASCII:
			srtFormatA = VDTextWToA(srtFormat);
			data = (void *)srtFormatA.c_str();
			dataLen = srtFormatA.size();
			break;
		case kTextType_UTF8:
			srtFormatA = VDTextWToU8(srtFormat);
			data = (void *)srtFormatA.c_str();
			dataLen = srtFormatA.size();
			break;
		default:
			data = (void *)srtFormat.c_str();
			dataLen = sizeof(wchar_t) * srtFormat.size();
			break;
		}

		if(data && dataLen)
			m_Parent->write(data, dataLen);
	}
	return TRUE;
}

void SRTOutputStream::setEndOfStream(void) {
}

void SRTOutputStream::setStreamLag(sint32 samplesLag) {
}

//////////////////////////////////////////////////////////////////////

char SRTOutputFile::szME[] = "SRTOutputFile";

SRTOutputFile::SRTOutputFile(eTextType outputType, bool bWriteBOM) {
	m_hFile			= INVALID_HANDLE_VALUE;
	m_InputInType	= kTextType_ASCII;
	m_OutputInType	= outputType;
	m_bWriteBOM		= bWriteBOM;
}

SRTOutputFile::~SRTOutputFile() {
	finalize();
}

IVDMediaOutputStream *SRTOutputFile::getOutputStream(uint32 serial) {
	if(serial == 1)
		return audioOut;
	else
		throw MyError("Attempt to access another output stream (%lu) when saving to SRT", serial);
	return NULL;
}

bool SRTOutputFile::initOutputStreams(ModVideoStream& videoStream, ModList<ModDubStream>& streams) {
	if(streams.GetFirst()) {
		ModDubStream *inStream = streams.GetFirst();
		ISubtitleHandler *pSubtitle;
		if(inStream->m_Source->QueryInterface(ISubtitleHandler::GetInterfaceID(), (void **)&pSubtitle))
			m_InputInType = pSubtitle->GetOutputType();
		else
			throw MyError("Subtitle stream doesn't expose the ISubtitleHandler interface!");
		inStream->m_Serial = 1;
	}

	if (!(audioOut = new SRTOutputStream(this, m_InputInType, m_OutputInType))) return FALSE;

	return TRUE;
}

IVDMediaOutputStream *SRTOutputFile::createVideoStream() {
	return NULL;
}

IVDMediaOutputStream *SRTOutputFile::createAudioStream() {
	VDASSERT(!audioOut);
	if (!(audioOut = new_nothrow SRTOutputStream(this, m_InputInType, m_OutputInType)))
		throw MyMemoryError();
	return audioOut;
}

/*BOOL SRTOutputFile::initOutputStreams() {
	if (!(audioOut = new SRTOutputStream(this, m_InputInType, m_OutputInType))) return FALSE;

	return TRUE;
}*/

bool SRTOutputFile::init(const wchar_t *szFile) {
	if (!audioOut) return FALSE;

	m_File.open(szFile, nsVDFile::kWrite | nsVDFile::kDenyNone | nsVDFile::kOpenAlways | nsVDFile::kSequential);

	if(m_bWriteBOM) {
		void *BOM = NULL;
		uint8 BOMLength = 0;
		GetBOMInformation(m_OutputInType, BOM, BOMLength);
		if(BOM && BOMLength)
			write(BOM, BOMLength);
	}

	return TRUE;
}

void SRTOutputFile::finalize() {
	m_File.truncateNT();
	m_File.close();
}

void SRTOutputFile::write(const void *pBuffer, uint32 cbBuffer) {
	m_File.write(pBuffer, cbBuffer);
}
