//	VirtualDubMod - Video processing and capture application based on VirtualDub by Avery Lee
//	Copyright (C) 1998-2003 VirtualDubMod Developers
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#include <vd2/system/error.h>
#include "SRTOutputFileSystem.h"
#include "SRTOutputFile.h"


SRTOutputFileSystem::SRTOutputFileSystem(const wchar_t *pszFilename, eTextType outputType, bool bWriteBOM)
	: mFilename(pszFilename)
{
	m_Stream		= NULL;
	m_OutputInType	= outputType;
	m_bWriteBOM		= bWriteBOM;
}

AVIOutput *SRTOutputFileSystem::CreateSegment() {
	vdautoptr<SRTOutputFile> pOutput(new SRTOutputFile(m_OutputInType, m_bWriteBOM));

	if(m_Stream) {
		ModList<ModDubStream> tmpList;
		tmpList.DeleteContents(false);
		tmpList.Append(m_Stream);
		ModVideoStream voidVideo;
		pOutput->initOutputStreams(voidVideo, tmpList);
	} else
		pOutput->createAudioStream()->setFormat(&mAudioFormat[0], mAudioFormat.size());

	pOutput->init(mFilename.c_str());

	return pOutput.release();
}

void SRTOutputFileSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	pSegment->finalize();
	delete pSegment;
}

void SRTOutputFileSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
}

void SRTOutputFileSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool SRTOutputFileSystem::AcceptsVideo() {
	return false;
}

bool SRTOutputFileSystem::AcceptsAudio() {
	return false;
}

void SRTOutputFileSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
}

void SRTOutputFileSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	if(m_Stream)
		throw MyError("Attempt to output more than one stream when saving to SRT");
	m_Stream = &stream;
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool SRTOutputFileSystem::AcceptsText() {
	return true;
}
