//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2001 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <windows.h>

#include <vd2/system/file.h>
#include <vd2/system/error.h>
#include <vd2/system/filesys.h>

#include "PositionControl.h"

#include "InputFile.h"
#include "InputFileImages.h"
#include "AudioSource.h"
#include "VideoSource.h"
#include "AVIOutput.h"
#include "AVIOutputWAV.h"
#include "AVIOutputImages.h"
#include "AVIOutputStriped.h"
#include "Dub.h"
#include "DubOutput.h"
#include "AudioFilterSystem.h"
#include "FrameSubset.h"
#include "ProgressDialog.h"
#include "oshelper.h"

#include "mpeg.h"
#include "gui.h"
#include "prefs.h"
#include "command.h"
#include "project.h"
// ********************************************************************
// *** VirtualDubMod												***
// *** Oct 2003														***
// BEGIN **************************************************************
#include "resource.h"
#include "ModStreamsSetup.h"
#include "ogg&co/InputFileOGM.h"
#include "ogg&co/OGMOutputFileSystem.h"
#include "ModTextSource.h"
#include "SRTOutputFileSystem.h"
#include "matroska/InputFileMatroska.h"
#include "matroska/MatroskaOutputFileSystem.h"

#include <vector>
#include "gui.h"
using namespace std;

extern int	g_iSelectedStream;

vector<int> baddies;

///////////////////////////////////////////////////////////////////////////

extern HWND					g_hWnd;
extern DubOptions			g_dubOpts;

extern wchar_t g_szInputWAVFile[MAX_PATH];

extern DubSource::ErrorMode	g_videoErrorMode;
extern DubSource::ErrorMode	g_audioErrorMode;

vdrefptr<InputFile>		inputAVI;
InputFileOptions	*g_pInputOpts			= NULL;

vdrefptr<VideoSource>	inputVideoAVI;
vdrefptr<AudioSource>	inputAudio;
FrameSubset			*inputSubset			= NULL;
// END ****************************************************************

IDubber				*g_dubber				= NULL;

COMPVARS			g_Vcompression;
WAVEFORMATEX		*g_ACompressionFormat		= NULL;
DWORD				g_ACompressionFormatSize	= 0;

VDAudioFilterGraph	g_audioFilterGraph;

extern VDProject *g_project;


bool				g_drawDecompressedFrame	= FALSE;
bool				g_showStatusWindow		= TRUE;

///////////////////////////////////////////////////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void SetAudioSource() {
	ModStream *selectedStream = g_Streams[g_iSelectedStream];
	if(selectedStream && selectedStream->m_Source && (selectedStream->m_Source->getSourceMainType()==SOURCE_AUDIO))
		inputAudio = (AudioSource *)selectedStream->m_Source;
	else
		inputAudio = NULL;
}
// END ****************************************************************

void AppendAVI(const wchar_t *pszFile) {
	if (inputAVI) {
		long lTail = inputAVI->videoSrc->getEnd();

		if (inputAVI->Append(pszFile)) {
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
			RefreshStreamsComments();
// END ****************************************************************
			if (inputSubset)
				inputSubset->insert(inputSubset->end(), FrameSubsetNode(lTail, inputAVI->videoSrc->getEnd() - lTail, false));
			RemakePositionSlider();
		}
	}
}

void AppendAVIAutoscan(const wchar_t *pszFile) {
	wchar_t buf[MAX_PATH];
	wchar_t *s = buf, *t;
	int count = 0;

	if (!inputAVI)
		return;

	VDPosition originalCount = inputAVI->videoSrc->getEnd();

	wcscpy(buf, pszFile);

	t = VDFileSplitExt(VDFileSplitPath(s));

	if (t>buf)
		--t;

	try {
		for(;;) {
			if (!VDDoesPathExist(VDStringW(buf)))
				break;
		
			if (!inputAVI->Append(buf))
				break;

			++count;

			s = t;

			for(;;) {
				if (s<buf || !isdigit(*s)) {
					memmove(s+2, s+1, sizeof(wchar_t) * wcslen(s));
					s[1] = L'1';
					++t;
				} else {
					if (*s == L'9') {
						*s-- = L'0';
						continue;
					}
					++*s;
				}
				break;
			}
		}
	} catch(const MyError& e) {
		// if the first segment failed, turn the warning into an error
		if (!count)
			throw;

		// log append errors, but otherwise eat them
		VDLog(kVDLogWarning, VDTextAToW(e.gets()));
	}

	guiSetStatus("Appended %d segments (stopped at \"%s\")", 255, count, VDTextWToA(buf).c_str());

	if (count) {
		if (inputSubset)
			inputSubset->insert(inputSubset->end(), FrameSubsetNode(originalCount, inputAVI->videoSrc->getEnd() - originalCount, false));
		RemakePositionSlider();
	}
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void CloseAVI(bool bKeepModStreams) {
	if (g_pInputOpts) {
		delete g_pInputOpts;
		g_pInputOpts = NULL;
	}

	if(inputAVI && !bKeepModStreams)
		RemoveModStreams();
	ClearVideoStream();
	inputAudio = NULL;	/* be sure the audio stream is released (to close the input file) */
// END ****************************************************************
	inputVideoAVI = NULL;
	inputAVI = NULL;

	if (inputSubset) {
		delete inputSubset;
		inputSubset = NULL;
	}

	filters.DeinitFilters();
	filters.DeallocateBuffers();
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void SaveWAV(const wchar_t *szFilename, bool fProp, DubOptions *quick_opts, bool bWriteHeader, uint32 outputFileType, eTextType outputTextType) {
	if(!inputVideoAVI)
		throw MyError("No input file to process.");

	SetAudioSource();

	outputFileType = SOURCE_WAV;
	ModList<ModStream>::iterator it(g_Streams);
	while(it) {
		if((*it)->m_bToDemux) {
			if( ((*it)->m_Source->getSourceType() == SOURCE_AUDIO_VORBIS)
			&& ((*it)->m_Opts.mode == DubAudioOptions::M_NONE) ) {
				outputFileType = SOURCE_OGM;
				break;
			} else if((*it)->m_Source->getSourceMainType() == SOURCE_TEXT) {
				outputFileType = SOURCE_TEXT_SRT;
				break;
			}
		}
		it++;
	}

	switch(outputFileType) {
	case SOURCE_OGM:
		{
			OGMOutputFileSystem fileout;

			fileout.SetCaching(false);
			fileout.SetFilename(szFilename);
			fileout.SetBuffer(g_dubOpts.perf.outputBufferSize);
			fileout.SetAcceptsVideo(false);
			g_project->RunOperation(&fileout, TRUE, quick_opts, 0, fProp, 0, 0, true, outputFileType);
		}
		break;
	case SOURCE_TEXT_SRT:
		{
			SRTOutputFileSystem wavout(szFilename, outputTextType, bWriteHeader);
			g_project->RunOperation(&wavout, TRUE, quick_opts, 0, fProp, 0, 0, true, outputFileType);
		}
		break;
	default:
		{
			VDAVIOutputWAVSystem wavout(szFilename, bWriteHeader);
			g_project->RunOperation(&wavout, TRUE, quick_opts, 0, fProp, 0, 0, true, outputFileType);
		}
		break;
	}
}
// END ****************************************************************

///////////////////////////////////////////////////////////////////////////

void SaveAVI(const wchar_t *szFilename, bool fProp, DubOptions *quick_opts, bool fCompatibility) {
	VDAVIOutputFileSystem fileout;

	fileout.Set1GBLimit(g_prefs.fAVIRestrict1Gb != 0);
	fileout.SetCaching(false);
	fileout.SetIndexing(!fCompatibility);
	fileout.SetFilename(szFilename);
	fileout.SetBuffer(g_dubOpts.perf.outputBufferSize);

	g_project->RunOperation(&fileout, FALSE, quick_opts, g_prefs.main.iDubPriority, fProp);
}

void SaveStripedAVI(const wchar_t *szFile) {
	AVIStripeSystem *stripe_def = NULL;

	///////////////

	SetAudioSource();

	if (!inputVideoAVI)
		throw MyError("No input video stream to process.");

	VDAVIOutputStripedSystem outstriped(szFile);

	outstriped.Set1GBLimit(g_prefs.fAVIRestrict1Gb != 0);

	g_project->RunOperation(&outstriped, FALSE, NULL, g_prefs.main.iDubPriority, false, 0, 0);
}

void SaveStripeMaster(const wchar_t *szFile) {
	AVIStripeSystem *stripe_def = NULL;

	///////////////

	SetAudioSource();

	if (!inputVideoAVI)
		throw MyError("No input video stream to process.");

	VDAVIOutputStripedSystem outstriped(szFile);

	outstriped.Set1GBLimit(g_prefs.fAVIRestrict1Gb != 0);

	g_project->RunOperation(&outstriped, 2, NULL, g_prefs.main.iDubPriority, false, 0, 0);
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2003											***
// BEGIN **************************************************************
void SaveSegmentedAVI(const wchar_t *szFilename, bool fProp, DubOptions *quick_opts, long lSpillThreshold, long lSpillFrameThreshold, int mode) {
	if (!inputVideoAVI)
		throw MyError("No input file to process.");

	DubOptions *options = (quick_opts ? quick_opts : &g_dubOpts);
	long iEndOffsetMS = options->video.lEndOffsetMS;
	int iFileIndex = 0;
	long frameSplit_Start = 0;
	long frameSplit_End = inputVideoAVI->msToSamples(options->video.lStartOffsetMS);
	sint64 framePos_End = 0;
	if(frameSplit_End)
		framePos_End = inputVideoAVI->samplesToMs((inputSubset ? inputSubset->getTotalFrames() : inputVideoAVI->getLength()) - frameSplit_End);
	long iThreshold = 0;
	if( (mode==1) && (options->video.mode==DubVideoOptions::M_NONE) )
		iThreshold = lSpillThreshold * 1024 * 1024;
	do {
		if(iThreshold>0) {
			lSpillThreshold = 0;
			frameSplit_Start = frameSplit_End;
			frameSplit_End = JumpToLastKeyBeforePos(framePos_End + iThreshold);
			if(frameSplit_End >= inputVideoAVI->getStart())
				frameSplit_End = -1;
			if(frameSplit_End > 0)
				framePos_End = ((VideoSourceAVI*)inputVideoAVI->asStream())->sampleOffset(frameSplit_End);
			options->video.lStartOffsetMS = inputVideoAVI->samplesToMs(frameSplit_Start);
			if(frameSplit_End > 0)
				options->video.lEndOffsetMS = inputVideoAVI->samplesToMs((inputSubset ? inputSubset->getTotalFrames() : inputVideoAVI->getLength()) - frameSplit_End);
			else
				options->video.lEndOffsetMS = 0;
			if(options->video.lEndOffsetMS < iEndOffsetMS) {
				options->video.lEndOffsetMS = iEndOffsetMS;
				framePos_End = -1;
			}
		}

		VDAVIOutputFileSystem outfile;

		outfile.SetIndexing(false);
		outfile.SetCaching(false);
		outfile.SetBuffer(g_dubOpts.perf.outputBufferSize);

		const VDStringW filename(szFilename);
		outfile.SetFilenamePattern(VDFileSplitExtLeft(filename).c_str(), VDFileSplitExtRight(filename).c_str(), 2, iFileIndex++, frameSplit_End <= 0);

		g_project->RunOperation(&outfile, FALSE, quick_opts, g_prefs.main.iDubPriority, fProp, lSpillThreshold, lSpillFrameThreshold);
	} while(frameSplit_End > 0);
}

void SaveOGM(const wchar_t *szFilename, bool fProp, DubOptions *quick_opts, bool fCompatibility) {
	OGMOutputFileSystem fileout;

	fileout.SetCaching(false);
	fileout.SetFilename(szFilename);
	fileout.SetBuffer(g_dubOpts.perf.outputBufferSize);

	g_project->RunOperation(&fileout, FALSE, quick_opts, g_prefs.main.iDubPriority, fProp, 0, 0, true, SOURCE_OGM);
}

void SaveSegmentedOGM(const wchar_t *szFilename, bool fProp, DubOptions *quick_opts, long lSpillThreshold, long lSpillFrameThreshold, int mode) {
	if (!inputVideoAVI)
		throw MyError("No input file to process.");

	DubOptions *options = (quick_opts ? quick_opts : &g_dubOpts);
	long iEndOffsetMS = options->video.lEndOffsetMS;
	int iFileIndex = 0;
	long frameSplit_Start = 0;
	long frameSplit_End = inputVideoAVI->msToSamples(options->video.lStartOffsetMS);
	sint64 framePos_End = 0;
	if(frameSplit_End)
		framePos_End = inputVideoAVI->samplesToMs((inputSubset ? inputSubset->getTotalFrames() : inputVideoAVI->getLength()) - frameSplit_End);
	long iThreshold = 0;
	if( (mode==1) && (options->video.mode==DubVideoOptions::M_NONE) )
		iThreshold = lSpillThreshold * 1024 * 1024;
	do {
		if(iThreshold>0) {
			lSpillThreshold = 0;
			frameSplit_Start = frameSplit_End;
			frameSplit_End = JumpToLastKeyBeforePos(framePos_End + iThreshold);
			if(frameSplit_End >= inputVideoAVI->getEnd())
				frameSplit_End = -1;
			if(frameSplit_End > 0)
				framePos_End = ((VideoSourceAVI*)inputVideoAVI->asStream())->sampleOffset(frameSplit_End);
			options->video.lStartOffsetMS = inputVideoAVI->samplesToMs(frameSplit_Start);
			if(frameSplit_End > 0)
				options->video.lEndOffsetMS = inputVideoAVI->samplesToMs((inputSubset ? inputSubset->getTotalFrames() : inputVideoAVI->getLength()) - frameSplit_End);
			else
				options->video.lEndOffsetMS = 0;
			if(options->video.lEndOffsetMS < iEndOffsetMS) {
				options->video.lEndOffsetMS = iEndOffsetMS;
				framePos_End = -1;
			}
		}

		OGMOutputFileSystem outfile;

		outfile.SetCaching(false);
		outfile.SetBuffer(g_dubOpts.perf.outputBufferSize);

		const VDStringW filename(szFilename);
		outfile.SetFilenamePattern(VDFileSplitExtLeft(filename).c_str(), VDFileSplitExtRight(filename).c_str(), 2, iFileIndex++, frameSplit_End <= 0);

		g_project->RunOperation(&outfile, FALSE, quick_opts, g_prefs.main.iDubPriority, fProp, lSpillThreshold, lSpillFrameThreshold, true, SOURCE_OGM);
	} while(frameSplit_End > 0);
}

void SaveMKV(const wchar_t *szFilename, bool fProp, DubOptions *quick_opts, bool fCompatibility) {
	MatroskaOutputFileSystem fileout;

	fileout.SetFilename(szFilename);
	fileout.SetBuffer(g_dubOpts.perf.outputBufferSize);

	g_project->RunOperation(&fileout, FALSE, quick_opts, g_prefs.main.iDubPriority, fProp, 0, 0, true, SOURCE_MKV);
}

void SaveSegmentedMKV(const wchar_t *szFilename, bool fProp, DubOptions *quick_opts, long lSpillThreshold, long lSpillFrameThreshold, int mode) {
	if (!inputVideoAVI)
		throw MyError("No input file to process.");

	DubOptions *options = (quick_opts ? quick_opts : &g_dubOpts);
	long iEndOffsetMS = options->video.lEndOffsetMS;
	int iFileIndex = 0;
	long frameSplit_Start = 0;
	long frameSplit_End = inputVideoAVI->msToSamples(options->video.lStartOffsetMS);
	sint64 framePos_End = 0;
	if(frameSplit_End)
		framePos_End = inputVideoAVI->samplesToMs((inputSubset ? inputSubset->getTotalFrames() : inputVideoAVI->getLength()) - frameSplit_End);
	long iThreshold = 0;
	if( (mode==1) && (options->video.mode==DubVideoOptions::M_NONE) )
		iThreshold = lSpillThreshold * 1024 * 1024;
	do {
		if(iThreshold>0) {
			lSpillThreshold = 0;
			frameSplit_Start = frameSplit_End;
			frameSplit_End = JumpToLastKeyBeforePos(framePos_End + iThreshold);
			if(frameSplit_End >= inputVideoAVI->getEnd())
				frameSplit_End = -1;
			if(frameSplit_End > 0)
				framePos_End = ((VideoSourceAVI*)inputVideoAVI->asStream())->sampleOffset(frameSplit_End);
			options->video.lStartOffsetMS = inputVideoAVI->samplesToMs(frameSplit_Start);
			if(frameSplit_End > 0)
				options->video.lEndOffsetMS = inputVideoAVI->samplesToMs((inputSubset ? inputSubset->getTotalFrames() : inputVideoAVI->getLength()) - frameSplit_End);
			else
				options->video.lEndOffsetMS = 0;
			if(options->video.lEndOffsetMS < iEndOffsetMS) {
				options->video.lEndOffsetMS = iEndOffsetMS;
				framePos_End = -1;
			}
		}

		MatroskaOutputFileSystem outfile;

		outfile.SetBuffer(g_dubOpts.perf.outputBufferSize);

		const VDStringW filename(szFilename);
		outfile.SetFilenamePattern(VDFileSplitExtLeft(filename).c_str(), VDFileSplitExtRight(filename).c_str(), 2, iFileIndex++, frameSplit_End <= 0);

		g_project->RunOperation(&outfile, FALSE, quick_opts, g_prefs.main.iDubPriority, fProp, lSpillThreshold, lSpillFrameThreshold, true, SOURCE_MKV);
	} while(frameSplit_End > 0);
}
// END ****************************************************************

void SaveImageSequence(const wchar_t *szPrefix, const wchar_t *szSuffix, int minDigits, bool fProp, DubOptions *quick_opts, int targetFormat) {
	VDAVIOutputImagesSystem outimages;

	outimages.SetFilenamePattern(szPrefix, szSuffix, minDigits);
	outimages.SetFormat(targetFormat);
		
	g_project->RunOperation(&outimages, FALSE, quick_opts, g_prefs.main.iDubPriority, fProp, 0, 0);
}

///////////////////////////////////////////////////////////////////////////


void SetSelectionStart(long ms) {
	if (!inputVideoAVI)
		return;

	g_dubOpts.video.lStartOffsetMS = ms;
	SendDlgItemMessage(g_hWnd, IDC_POSITION, PCM_SETSELSTART, (WPARAM)TRUE, inputVideoAVI->msToSamples(ms));
}

void SetSelectionEnd(long ms) {
	if (!inputVideoAVI)
		return;

	g_dubOpts.video.lEndOffsetMS = ms;
	SendDlgItemMessage(g_hWnd, IDC_POSITION, PCM_SETSELEND, (WPARAM)TRUE,
		(inputSubset ? inputSubset->getTotalFrames() : inputVideoAVI->getLength()) - inputVideoAVI->msToSamples(ms));
}

void RemakePositionSlider() {
	if (!inputAVI) return;

	HWND hwndPosition = GetDlgItem(g_hWnd, IDC_POSITION);

	SendMessage(hwndPosition, PCM_SETRANGEMIN, (BOOL)FALSE, 0);
	SendMessage(hwndPosition, PCM_SETRANGEMAX, (BOOL)TRUE , inputSubset->getTotalFrames());

	if (g_dubOpts.video.lStartOffsetMS || g_dubOpts.video.lEndOffsetMS) {
		SendMessage(hwndPosition, PCM_SETSELSTART, (BOOL)FALSE, inputVideoAVI->msToSamples(g_dubOpts.video.lStartOffsetMS));
		SendMessage(hwndPosition, PCM_SETSELEND, (BOOL)TRUE , inputSubset->getTotalFrames() - inputVideoAVI->msToSamples(g_dubOpts.video.lEndOffsetMS));
	} else {
		SendMessage(hwndPosition, PCM_CLEARSEL, (BOOL)TRUE, 0);
	}
}

void RecalcPositionTimeConstant() {
	HWND hwndPosition = GetDlgItem(g_hWnd, IDC_POSITION);

	DubVideoStreamInfo vInfo;
	DubAudioStreamInfo aInfo;

#pragma message(__TODO__ "Replace by the new InitStreamValueStatic ? (anyway what is the use of this function here ?)")
	if (inputVideoAVI) {
		try {
			InitStreamValuesStatic(vInfo, aInfo, inputVideoAVI, inputAudio, &g_dubOpts, NULL);
			SendMessage(hwndPosition, PCM_SETFRAMERATE, vInfo.frameRateIn.getHi(), vInfo.frameRateIn.getLo());
		} catch(const MyError&) {
			// The input stream may throw an error here trying to obtain the nearest key.
			// If so, bail.
		}
	} else {
		SendMessage(hwndPosition, PCM_SETFRAMERATE, 0, 0);
	}
}

void ScanForUnreadableFrames(FrameSubset *pSubset, VideoSource *pVideoSource) {
	const VDPosition lFirst = pVideoSource->getStart();
	const VDPosition lLast = pVideoSource->getEnd();
	VDPosition lFrame = lFirst;
	void *pBuffer = NULL;
	int cbBuffer = 0;

	try {
		ProgressDialog pd(g_hWnd, "Frame scan", "Scanning for unreadable frames", lLast-lFrame, true);
		bool bLastValid = true;
		long lRangeFirst;
		long lDeadFrames = 0;
		long lMaskedFrames = 0;

// *******************************************************************
// *** VirtualDubMod											   ***
// *** Tobias Minich, Jan 2003									   ***
// BEGIN *************************************************************
		baddies.clear();
// END ***************************************************************

		pd.setValueFormat("Frame %d of %d");

		pVideoSource->streamBegin(false);

		while(lFrame <= lLast) {
			LONG lActualBytes, lActualSamples;
			int err;
			bool bValid;

			pd.advance(lFrame - lFirst);
			pd.check();

			do {
				bValid = false;

				if (!bLastValid && !pVideoSource->isKey(lFrame))
					break;

				if (lFrame < lLast) {
					err = pVideoSource->read(lFrame, 1, NULL, 0, &lActualBytes, &lActualSamples);

					if (err)
						break;

					if (cbBuffer < lActualBytes) {
						int cbNewBuffer = (lActualBytes + 65535) & ~65535;
						void *pNewBuffer = realloc(pBuffer, cbNewBuffer);

						if (!pNewBuffer)
							throw MyMemoryError();

						cbBuffer = cbNewBuffer;
						pBuffer = pNewBuffer;
					}

					err = pVideoSource->read(lFrame, 1, pBuffer, cbBuffer, &lActualBytes, &lActualSamples);

					if (err)
						break;

					try {
						pVideoSource->streamGetFrame(pBuffer, lActualBytes, pVideoSource->isKey(lFrame), FALSE, lFrame);
					} catch(...) {
						++lDeadFrames;
						break;
					}
				}

				bValid = true;
			} while(false);

			if (!bValid)
				++lMaskedFrames;

			if (bValid ^ bLastValid) {
// *******************************************************************
// *** VirtualDubMod											   ***
// *** Tobias Minich, Jan 2003									   ***
// BEGIN *************************************************************
/*				if (!bValid)
					lRangeFirst = lFrame;
				else
					pSubset->setRange(lRangeFirst, lFrame - lRangeFirst, true);
*/
				int *newFrame = new int;
				*newFrame = lFrame;
				baddies.push_back(*newFrame);
// END ***************************************************************
				bLastValid = bValid;
			}

			++lFrame;
		}

		pVideoSource->streamEnd();

		guiSetStatus("%ld frames masked (%ld frames bad, %ld frames good but undecodable)", 255, lMaskedFrames, lDeadFrames, lMaskedFrames-lDeadFrames);

	} catch(...) {
		free(pBuffer);
		throw;
	}

	free(pBuffer);
}

// *******************************************************************
// *** VirtualDubMod											   ***
// *** Tobias Minich, Jan 2003									   ***
// BEGIN *************************************************************

bool BadFrames() {
	return !baddies.empty();
}

void MaskBadFrames(FrameSubset *pSubset) {
	int a,b;
	if (BadFrames()) {
		vector<int>::iterator walkit;
		walkit = baddies.begin();
		do {
			a = *walkit;
			b = *(++walkit);
			walkit++;
			pSubset->setRange(a, b - a, true);
		} while (walkit != baddies.end());
		baddies.clear();
	}
}

void DeleteBadFrames(FrameSubset *pSubset) {
	int a,b;
	if (BadFrames()) {
		vector<int>::iterator walkit;
		walkit = baddies.begin();
		do {
			a = *walkit;
			b = *(++walkit);
			walkit++;
			pSubset->deleteRange(a, b - a);
		} while (walkit != baddies.end());
		baddies.clear();
	}
}

void ShowBadFrames() {
	int a,b;
	char *buf;

	if (BadFrames()) {
		buf = new char[(baddies.size()*15)+200];

		strcpy(buf, "The following frames could not be decoded:");

		vector<int>::iterator walkit;
		walkit = baddies.begin();
		do {
			a = *walkit;
			b = *(++walkit);
			walkit++;
			sprintf((buf+strlen(buf)),"\n%d-%d",a,b-1);
		} while (walkit != baddies.end());
		guiMessageBoxF(g_hWnd, IDS_GENERAL_STRING, IDS_INFO_BAD_FRAMES, MB_OK|MB_ICONINFORMATION, buf);
	}
}
// END ***************************************************************
