//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2003 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"

#define f_DUB_CPP


#include <process.h>
#include <time.h>
#include <vector>
#include <deque>
#include <utility>

#include <windows.h>
#include <vfw.h>

#include "resource.h"

#include "crash.h"
#include <vd2/system/thread.h>
#include <vd2/system/tls.h>
#include <vd2/system/time.h>
#include <vd2/system/atomic.h>
#include <vd2/system/fraction.h>
#include <vd2/system/vdalloc.h>
#include <vd2/system/VDRingBuffer.h>
#include <vd2/system/profile.h>
#include <vd2/Dita/resources.h>
#include "AudioFilterSystem.h"
#include "convert.h"
#include "filters.h"
#include "gui.h"
#include "prefs.h"
#include "command.h"
#include "misc.h"

//#define new new(_NORMAL_BLOCK, __FILE__, __LINE__)

#include <vd2/system/error.h>
#include "VideoSequenceCompressor.h"
#include "AsyncBlitter.h"
#include "AVIOutputPreview.h"
#include "AVIOutput.h"
#include "AVIOutputWAV.h"
#include "AVIOutputImages.h"
#include "AVIOutputStriped.h"
#include "AudioSource.h"
#include "VideoSource.h"
#include "AVIPipe.h"
#include "VBitmap.h"
#include "FrameSubset.h"
#include "InputFile.h"
#include "VideoTelecineRemover.h"
#include "VideoDisplay.h"

#include "Dub.h"
#include "DubOutput.h"
#include "DubStatus.h"
#include "DubUtils.h"
#include "DubIO.h"

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
#include "ModAudio.h"
#include "ModList.h"
#include "ModDubStream.h"
#include "ModAudioSourceMP3.h"	// for MPA_MAX_LENGTH
#include "ModAudioStreamSourceAC3.h"
#include "ModAudioStreamSourceMP3.h"
#include "ogg&co/OggDS.h"
#include "ogg&co/ModAudioStreamSourceVorbis.h"

// jonny's mods
extern int g_OptionNoWrite;
// END ****************************************************************

using namespace nsVDDub;

///////////////////////////////////////////////////////////////////////////

extern const char g_szError[];
extern HWND g_hWnd;
extern bool g_fWine;

///////////////////////////////////////////////////////////////////////////

namespace {
	enum { kVDST_Dub = 1 };

	enum {
		kVDM_SegmentOverflowOccurred,
		kVDM_BeginningNextSegment,
		kVDM_IOThreadLivelock,
		kVDM_ProcessingThreadLivelock,
		kVDM_CodecDelayedDuringDelayedFlush,
		kVDM_CodecLoopingDuringDelayedFlush,
	};
};

///////////////////////////////////////////////////////////////////////////

DubOptions g_dubOpts = {
	{
		0,				// no amp
		500,			// preload by 500ms
		1,				// every frame
		0,				// no new rate
		0,				// offset: 0ms
		false,			// period is in frames
		true,			// audio interleaving enabled
		true,			// yes, offset audio with video
		true,			// yes, clip audio to video length
		false,			// no integral change
		false,			// no high quality
		false,			// use fixed-function audio pipeline
		DubAudioOptions::P_NOCHANGE,		// no precision change
		DubAudioOptions::C_NOCHANGE,		// no channel change
		DubAudioOptions::M_NONE,
	},

	{
		DubVideoOptions::D_24BIT,	// input: 24bit
		DubVideoOptions::D_24BIT,	// output: 24bit
		DubVideoOptions::M_FULL,	// mode: full
		TRUE,						// show input video
		TRUE,						// show output video
		FALSE,						// decompress output video before display
		FALSE,						// histograms (unused)
		TRUE,						// sync to audio
		1,							// no frame rate decimation
		0,0,						// no target
		0,							// no change in frame rate
		0,							// start offset: 0ms
		0,							// end offset: 0ms
		false,						// No inverse telecine
		false,						// (IVTC mode)
		-1,							// (IVTC offset)
		false,						// (IVTC polarity)
		0,							// progressive preview
	},

	{
		2097152,				// 2Mb AVI output buffer
		65536,					// 64K WAV input buffer
		32,						// 32 pipe buffers
		true,					// dynamic enable
		false,
		false,					// directdraw,
		true,					// drop frames
	},

	true,			// show status
	false,			// move slider
};

static const int g_iPriorities[][2]={

	// I/O							processor
	{ THREAD_PRIORITY_IDLE,			THREAD_PRIORITY_IDLE,			},
	{ THREAD_PRIORITY_LOWEST,		THREAD_PRIORITY_LOWEST,			},
	{ THREAD_PRIORITY_BELOW_NORMAL,	THREAD_PRIORITY_LOWEST,			},
	{ THREAD_PRIORITY_NORMAL,		THREAD_PRIORITY_BELOW_NORMAL,	},
	{ THREAD_PRIORITY_NORMAL,		THREAD_PRIORITY_NORMAL,			},
	{ THREAD_PRIORITY_ABOVE_NORMAL,	THREAD_PRIORITY_NORMAL,			},
	{ THREAD_PRIORITY_HIGHEST,		THREAD_PRIORITY_ABOVE_NORMAL,	},
	{ THREAD_PRIORITY_HIGHEST,		THREAD_PRIORITY_HIGHEST,		}
};

/////////////////////////////////////////////////
void AVISTREAMINFOtoAVIStreamHeader(AVIStreamHeader_fixed *dest, const AVISTREAMINFO *src) {
	dest->fccType			= src->fccType;
	dest->fccHandler		= src->fccHandler;
	dest->dwFlags			= src->dwFlags;
	dest->wPriority			= src->wPriority;
	dest->wLanguage			= src->wLanguage;
	dest->dwInitialFrames	= src->dwInitialFrames;
	dest->dwStart			= src->dwStart;
	dest->dwScale			= src->dwScale;
	dest->dwRate			= src->dwRate;
	dest->dwLength			= src->dwLength;
	dest->dwSuggestedBufferSize = src->dwSuggestedBufferSize;
	dest->dwQuality			= src->dwQuality;
	dest->dwSampleSize		= src->dwSampleSize;
	dest->rcFrame.left		= (short)src->rcFrame.left;
	dest->rcFrame.top		= (short)src->rcFrame.top;
	dest->rcFrame.right		= (short)src->rcFrame.right;
	dest->rcFrame.bottom	= (short)src->rcFrame.bottom;

	VDDEBUG("scale %ld, rate %ld, length %ld\n",src->dwScale,src->dwRate, src->dwLength);
}

///////////////////////////////////////////////////////////////////////////

namespace {
	bool VDIsDC565(HDC hdc) {
		bool bIs565 = false;

		int bitsPerPel = GetDeviceCaps(hdc, BITSPIXEL);

		if (bitsPerPel==15 || bitsPerPel==16) {
			COLORREF crTmp;
			crTmp = GetPixel(hdc, 0,0);
			SetPixel(hdc, 0, 0, RGB(0x80, 0x88, 0x80));

			if (GetPixel(hdc,0,0) == RGB(0x80, 0x88, 0x80))
				bIs565 = true;

			SetPixel(hdc, 0, 0, crTmp);
		}

		return bIs565;
	}
}


///////////////////////////////////////////////////////////////////////////
//
//	Dubber
//
///////////////////////////////////////////////////////////////////////////

class Dubber : public IDubber, protected VDThread, protected IVDTimerCallback {
private:
	void TimerCallback();

	void NextSegment();

	enum VideoWriteResult {
		kVideoWriteOK,
		kVideoWriteDelayed,
		kVideoWriteBuffered,
		kVideoWriteDiscarded
	};

	VideoWriteResult WriteVideoFrame(void *buffer, int exdata, int droptype, LONG lastSize, VDPosition sampleFrame, VDPosition displayFrame, VDPosition timelineFrame);
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	void WriteStream(ModDubStream& stream, void *buffer, long lActualBytes, long lActualSamples);
	void UpdateTimeSensitiveInformation(void);
	void UpdateInfo_NextSegment(void);
	void CheckSpill(long videopt, ModDubStream *stream, long streampt);
	void SetVideoEOS(void);
	void SetEOS(ModDubStream& stream);

	void ThreadRun();

	MyError				err;
	bool				fError;

	VDAtomicInt			mStopLock;

	DubOptions			*opt;
	IVDVideoSource			*vSrc;
	InputFile			*pInput;
	AVIOutput			*AVIout;
	IVDMediaOutputStream		*mpVideoOut;			// alias: AVIout->videoOut
	IVDDubberOutputSystem	*mpOutputSystem;
	COMPVARS			*compVars;

	DubVideoStreamInfo	vInfo;

	bool				fUseVideoCompression;
	bool				fPreview;
	volatile bool		fAbort;
	volatile bool		fUserAbort;
	bool				fADecompressionOk;
	bool				fVDecompressionOk;
	bool				fFiltersOk;
	bool				fNoProcessingPriority;

	VDCallbackTimer		mFrameTimer;
	long				mPulseClock;

	VDDubIOThread		*mpIOThread;
	VDAtomicInt			mIOThreadCounter;

	VideoSequenceCompressor	*pVideoPacker;

	AVIPipe *			mpVideoPipe;

	AsyncBlitter *		blitter;

	HIC					outputDecompressor;

	IVDVideoDisplay *	mpInputDisplay;
	IVDVideoDisplay *	mpOutputDisplay;
	bool				mbInputDisplayInitialized;
	bool				fShowDecompressedFrame;
	bool				fDisplay565;

	int					iOutputDepth;
	VDFormatStruct<BITMAPINFOHEADER>	mpCompressorVideoFormat;
	VDFormatStruct<BITMAPINFOHEADER>	mpDisplayFormat;

	FrameSubset				*inputSubsetActive;
	FrameSubset				*inputSubsetAlloc;
	VideoTelecineRemover	*pInvTelecine;
	int					nVideoLagPreload;

	VDAtomicInt			mRefreshFlag;
	HWND				hwndStatus;

	bool				fSyncToAudioEvenClock;
	bool				mbAudioFrozen;
	bool				mbAudioFrozenValid;

	long				lDropFrames;

	FilterStateInfo		fsi;

	bool				fPhantom;

	IDubStatusHandler	*pStatusHandler;

	long				lVideoSizeEstimate;
	bool				fEnableSpill;
	long				lSegmentFrameLimit;
	sint64				mSegmentByteLimit;

	// interleaving
	VDStreamInterleaver		mInterleaver;
	VDRenderFrameMap		mVideoFrameMap;
	VDRenderFrameIterator	mVideoFrameIterator;

	VDAtomicInt			mProcessingThreadCounter;
	VDRTProfileChannel	mProcessingProfileChannel;

	///////

	int					mLastProcessingThreadCounter;
	int					mProcessingThreadFailCount;
	int					mLastIOThreadCounter;
	int					mIOThreadFailCount;

	///////

	ModVideoStream			m_VideoStream;
	ModChapterList	m_VideoChapters;
	ModList<ModDubStream>	*m_Streams;
	VDSignal				m_AudioReadSignal;
	bool					m_bHandleVBR;
	uint32					m_OutputFormat;
	bool					m_bVideoEOSSet;

public:
	Dubber(DubOptions *, bool bHandleVBR=false, uint32 outputFormat=SOURCE_AVI);
	~Dubber();

	void SetPhantomVideoMode();
	void SetInputDisplay(IVDVideoDisplay *);
	void SetOutputDisplay(IVDVideoDisplay *);
	void SetInputFile(InputFile *pInput);
	void EnableSpill(sint64 threshold, long framethreshold);

	void InitAudioConversionChains();
	void InitOutputFile();
	bool AttemptInputOverlay(BITMAPINFOHEADER *pbih);
	bool AttemptInputOverlays();

	void InitDirectDraw();
	void InitInputDisplay();
	void InitOutputDisplay();
	bool NegotiateFastFormat(BITMAPINFOHEADER *pbih);
	bool NegotiateFastFormat(int depth);
	void InitSelectInputFormat();
	void Init(ModVideoStream& video, ModList<ModDubStream>& streams, IVDDubberOutputSystem *outsys, COMPVARS *videoCompVars);
// END ****************************************************************
	void Go(int iPriority = 0);
	void Stop();

	void RealizePalette();
	void Abort();
	void ForceAbort();
	bool isRunning();
	bool isAbortedByUser();
	bool IsPreviewing();

	void SetStatusHandler(IDubStatusHandler *pdsh);
	void SetPriority(int index);
	void UpdateFrames();
};


///////////////////////////////////////////////////////////////////////////

IDubber::~IDubber() {
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
IDubber *CreateDubber(DubOptions *xopt, bool bHandleVBR, uint32 outputFormat) {
	return new Dubber(xopt, bHandleVBR, outputFormat);
}

Dubber::Dubber(DubOptions *xopt, bool bHandleVBR, uint32 outputFormat)
	: VDThread("Processing")
	, mpIOThread(0)
	, mIOThreadCounter(0)
	, mStopLock(0)
	, mpVideoPipe(NULL)
	, mSegmentByteLimit(0)
	, mVideoFrameIterator(mVideoFrameMap)
	, mProcessingThreadCounter(0)
	, mProcessingProfileChannel("Processor")
	, mLastProcessingThreadCounter(0)
	, mProcessingThreadFailCount(0)
	, mLastIOThreadCounter(0)
	, mIOThreadFailCount(0)
{
	m_OutputFormat	= outputFormat;
	m_bHandleVBR	= bHandleVBR;
	opt				= xopt;
	vSrc			= NULL;
	pInput			= NULL;

	// clear the workin' variables...

	fError				= false;

	fAbort				= false;
	fUserAbort			= false;

	pVideoPacker		= NULL;
	pStatusHandler		= NULL;

	fADecompressionOk	= false;
	fVDecompressionOk	= false;
	fFiltersOk			= false;

	blitter				= NULL;
	outputDecompressor	= NULL;
	hwndStatus			= NULL;
	vInfo.total_size	= 0;
	vInfo.fAudioOnly	= false;

	inputSubsetActive	= NULL;
	inputSubsetAlloc	= NULL;

	fPhantom = false;

	pInvTelecine		= NULL;

	AVIout				= NULL;
	mpVideoOut			= NULL;
	fEnableSpill		= false;

	lSegmentFrameLimit	= 0;

	fSyncToAudioEvenClock = false;
	mbAudioFrozen = false;
	mbAudioFrozenValid = false;

	m_Streams = NULL;
	m_bVideoEOSSet = false;
}

Dubber::~Dubber() {
	VDDEBUG("Dubber: destructor called.\n");

	Stop();
}

/////////////////////////////////////////////////

void Dubber::SetPhantomVideoMode() {
	fPhantom = true;
	vInfo.fAudioOnly = true;
}

void Dubber::SetInputFile(InputFile *pInput) {
	this->pInput = pInput;
}

void Dubber::SetStatusHandler(IDubStatusHandler *pdsh) {
	pStatusHandler = pdsh;
}


/////////////

void Dubber::SetInputDisplay(IVDVideoDisplay *pDisplay) {
	mpInputDisplay = pDisplay;
}

void Dubber::SetOutputDisplay(IVDVideoDisplay *pDisplay) {
	mpOutputDisplay = pDisplay;
}

/////////////

void Dubber::EnableSpill(sint64 segsize, long framecnt) {
	fEnableSpill = true;
	mSegmentByteLimit = segsize;
	lSegmentFrameLimit = framecnt;
}

// END ****************************************************************

void InitStreamValuesStatic(DubVideoStreamInfo& vInfo, DubAudioStreamInfo& aInfo, IVDVideoSource *video, AudioSource *audio, DubOptions *opt, FrameSubset *pfs) {
	if (!pfs)
		pfs = inputSubset;

	IVDStreamSource *pVideoStream = 0;
	
	if (video) {
		pVideoStream = video->asStream();

		if (pfs) {
			vInfo.start_src		= 0;
			vInfo.end_src		= pfs->getTotalFrames();
		} else {
			vInfo.start_src		= pVideoStream->getStart();
			vInfo.end_src		= pVideoStream->getEnd();
		}
	} else {
		vInfo.start_src		= 0;
		vInfo.end_src		= 0;
	}
	vInfo.cur_dst		= 0;
	vInfo.end_dst		= 0;
	vInfo.cur_proc_dst	= 0;
	vInfo.end_proc_dst	= 0;

	if (audio) {
		aInfo.start_src		= audio->getStart();
		aInfo.end_src		= audio->getEnd();
	} else {
		aInfo.start_src		= 0;
		aInfo.end_src		= aInfo.end_dst		= 0;
	}

	vInfo.cur_src			= vInfo.start_src;
	aInfo.cur_src			= aInfo.start_src;

	if (video) {
		// compute new frame rate

		VDFraction framerate(pVideoStream->getRate());

		if (opt->video.frameRateNewMicroSecs == DubVideoOptions::FR_SAMELENGTH) {
			if (audio && audio->getLength()) {
				framerate = VDFraction::reduce64(pVideoStream->getLength() * (sint64)1000, audio->samplesToMs(audio->getLength()));
			}
		} else if (opt->video.frameRateNewMicroSecs)
			framerate = VDFraction(1000000, opt->video.frameRateNewMicroSecs);

		// are we supposed to offset the video?

		if (opt->video.lStartOffsetMS) {
			vInfo.start_src += pVideoStream->msToSamples(opt->video.lStartOffsetMS); 
		}

		if (opt->video.lEndOffsetMS)
			vInfo.end_src -= pVideoStream->msToSamples(opt->video.lEndOffsetMS);

		vInfo.frameRateIn	= framerate;

		if (opt->video.frameRateDecimation==1 && opt->video.frameRateTargetLo)
			vInfo.frameRate	= VDFraction(opt->video.frameRateTargetHi, opt->video.frameRateTargetLo);
		else
			vInfo.frameRate	= framerate / opt->video.frameRateDecimation;

		vInfo.usPerFrameIn	= (long)vInfo.frameRateIn.scale64ir(1000000);
		vInfo.usPerFrame	= (long)vInfo.frameRate.scale64ir(1000000);

		// make sure we start reading on a key frame

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
		if (opt->video.mode == DubVideoOptions::M_NONE) {
			vInfo.start_src	= video->nearestKey(vInfo.start_src);
			long end = vInfo.end_src + pVideoStream->getStart();
			while(video->isB(end-1))
				end++;
			vInfo.end_src = end - pVideoStream->getStart();
		}
// END ****************************************************************

		vInfo.cur_src		= vInfo.start_src;
		vInfo.end_dst		= (long)(vInfo.frameRate / vInfo.frameRateIn).scale64t(vInfo.end_src - vInfo.start_src);
		vInfo.end_proc_dst	= vInfo.end_dst;
	}

	if (audio) {
		// offset the start of the audio appropriately...

		aInfo.start_us = -1000*opt->audio.offset;

		if (video) {
			const sint64 video_start	= vInfo.start_src - pVideoStream->getStart();
			const sint64 video_length	= vInfo.end_src - vInfo.start_src;


			if (opt->audio.fStartAudio && video && opt->video.lStartOffsetMS) {
				if (!pfs) {
					aInfo.start_us += vInfo.frameRateIn.scale64ir(1000000 * video_start);
				}
			}
		}

		aInfo.start_src += audio->msToSamples((long)(aInfo.start_us / 1000));

		// clip the end of the audio if supposed to...

		if (video && opt->audio.fEndAudio) {
			const sint64 video_length	= vInfo.end_src - vInfo.start_src;
			long lMaxLength;

			const VDFraction audiorate(audio->getRate());

			lMaxLength = (long)(audiorate / vInfo.frameRateIn).scale64r(video_length);

			if (aInfo.end_src - aInfo.start_src > lMaxLength)
				aInfo.end_src = aInfo.start_src + lMaxLength;
		}

		// resampling audio?

		aInfo.resampling = false;
		aInfo.converting = false;

		if (opt->audio.mode > DubAudioOptions::M_NONE) {
			if (opt->audio.new_rate) {
				aInfo.resampling = true;
			}

			if (opt->audio.newPrecision != DubAudioOptions::P_NOCHANGE || opt->audio.newChannels != DubAudioOptions::C_NOCHANGE) {
				aInfo.converting = true;

				aInfo.is_16bit = opt->audio.newPrecision==DubAudioOptions::P_16BIT
								|| (opt->audio.newPrecision==DubAudioOptions::P_NOCHANGE && audio->getWaveFormat()->wBitsPerSample>8);
				aInfo.is_stereo = opt->audio.newChannels==DubAudioOptions::C_STEREO
								|| (opt->audio.newChannels==DubAudioOptions::C_NOCHANGE && audio->getWaveFormat()->nChannels>1);
				aInfo.is_right = (opt->audio.newChannels==DubAudioOptions::C_MONORIGHT);
				aInfo.single_channel = (opt->audio.newChannels==DubAudioOptions::C_MONOLEFT || opt->audio.newChannels==DubAudioOptions::C_MONORIGHT);
				aInfo.bytesPerSample = (aInfo.is_16bit ? 2 : 1) * (aInfo.is_stereo ? 2 : 1);

			}
		}

		aInfo.cur_src		= aInfo.start_src;
		aInfo.cur_dst		= 0;
	}

	vInfo.cur_proc_src = vInfo.cur_src;
	aInfo.cur_proc_src = aInfo.cur_src;

	VDDEBUG("Dub: Audio is from (%ld,%ld) starting at %ld\n", aInfo.start_src, aInfo.end_src, aInfo.cur_src);
	VDDEBUG("Dub: Video is from (%ld,%ld) starting at %ld\n", vInfo.start_src, vInfo.end_src, vInfo.cur_src);
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void InitStreamValuesStatic(DubVideoStreamInfo& vInfo, ModVideoStream& videoStream, ModList<ModDubStream>& streams, DubOptions *opt, FrameSubset *pfs, bool bPreview) {
	if (!pfs)
		pfs = inputSubset;

	VideoSource *video = (VideoSource *)videoStream.m_Source->asStream();
#pragma message(__TODO__ "See what streams we are refering to when stretching video")
	AudioSource *audio = (AudioSource *)(streams.GetFirst() ? streams.GetFirst()->m_Source : NULL);

	IVDStreamSource *pVideoStream = 0;
	
	if (video) {
		pVideoStream = video->asStream();

		if (pfs) {
			vInfo.start_src		= 0;
			vInfo.end_src		= pfs->getTotalFrames();
		} else {
			vInfo.start_src		= pVideoStream->getStart();
			vInfo.end_src		= pVideoStream->getEnd();
		}
	} else {
		vInfo.start_src		= 0;
		vInfo.end_src		= 0;
	}
	vInfo.cur_dst		= 0;
	vInfo.end_dst		= 0;
	vInfo.cur_proc_dst	= 0;
	vInfo.end_proc_dst	= 0;

	ModList<ModDubStream>::iterator it(streams);
	ModDubStream *curStream;
	while(it) {
		curStream = *it;

		AudioSource *curAudioSource = (AudioSource *)curStream->m_Source;
		curAudioSource->extend(curStream->m_bHandleVBR, curStream->m_GroupFrames);

		curStream->m_bText = (curAudioSource->getSourceMainType()==SOURCE_TEXT);

		curStream->m_bSeemsVBR = false;
		long aSourceBlockAlign = curAudioSource->getWaveFormat()->nBlockAlign;
		//long aSourceRate = curAudioSource->streamInfo.dwRate;
		//long aSourceScale = curAudioSource->streamInfo.dwScale;
		// Is this a VBR stream ?
		// Simple test : request a small amount of samples (1 and 2 for example)
		// If we get exactly what we requested, in the expected number of bytes
		// then it is more likely to be a CBR stream
		LONG bytesRead, samplesRead;
		if( (curAudioSource->_read(0, 1, NULL, 0x7FFFFFFF, &bytesRead, &samplesRead)==AVIERR_OK)
			&& ((samplesRead!=1) || (bytesRead!=aSourceBlockAlign)) )
			curStream->m_bSeemsVBR = true;
		if( !curStream->m_bSeemsVBR && (curAudioSource->_read(0, 2, NULL, 0x7FFFFFFF, &bytesRead, &samplesRead)==AVIERR_OK)
			&& ((samplesRead!=2) || (bytesRead!=2*aSourceBlockAlign)) )
			curStream->m_bSeemsVBR = true;
		//VDASSERT( (curStream->m_bSeemsVBR && ((aSourceBlockAlign!=aSourceScale) || (curAudioSource->getWaveFormat()->nAvgBytesPerSec!=aSourceRate)))
		//	|| (!curStream->m_bSeemsVBR && (aSourceBlockAlign==aSourceScale) && (curAudioSource->getWaveFormat()->nAvgBytesPerSec==aSourceRate)) );
		
		if (curStream->m_Source) {
			curStream->m_aInfo.start_src	= curStream->m_Source->getStart();
			curStream->m_aInfo.end_src		= curStream->m_Source->getEnd();
		} else {
			curStream->m_aInfo.start_src	= 0;
			curStream->m_aInfo.end_src		= curStream->m_aInfo.end_dst	= 0;
		}
		curStream->m_aInfo.cur_src			= curStream->m_aInfo.start_src;
		it++;
	}

	vInfo.cur_src = vInfo.start_src;

	if (video) {
		// compute new frame rate

		VDFraction framerate(pVideoStream->getRate());

		if (opt->video.frameRateNewMicroSecs == DubVideoOptions::FR_SAMELENGTH) {
			if (audio && audio->getLength()) {
				framerate = VDFraction::reduce64(video->getLength() * (sint64)1000, audio->samplesToMs(audio->getLength()));
			}
		} else if (opt->video.frameRateNewMicroSecs)
			framerate = VDFraction(1000000, opt->video.frameRateNewMicroSecs);

		// are we supposed to offset the video?

		if (opt->video.lStartOffsetMS) {
			vInfo.start_src += video->msToSamples(opt->video.lStartOffsetMS); 
		}

		if (opt->video.lEndOffsetMS)
			vInfo.end_src -= video->msToSamples(opt->video.lEndOffsetMS);

		vInfo.frameRateIn	= framerate;

		if (opt->video.frameRateDecimation==1 && opt->video.frameRateTargetLo)
			vInfo.frameRate	= VDFraction(opt->video.frameRateTargetHi, opt->video.frameRateTargetLo);
		else
			vInfo.frameRate	= framerate / opt->video.frameRateDecimation;

		vInfo.usPerFrameIn	= (long)vInfo.frameRateIn.scale64ir(1000000);
		vInfo.usPerFrame	= (long)vInfo.frameRate.scale64ir(1000000);

		// make sure we start reading on a key frame

		if (opt->video.mode == DubVideoOptions::M_NONE) {
			// THIS WORKS WELL _ONLY_ IF YOU DON'T DELETE SUBSETS BEFORE YOUR START
			// POINT OR AFTER YOUR END POINT !!!
			if(pfs) {
				vInfo.start_src	= video->nearestKey(vInfo.start_src + pVideoStream->getStart()) - pVideoStream->getStart();

				long real_end = pfs->lookupFrame(vInfo.end_src + pVideoStream->getStart());
				long new_real_end = real_end;
				while(video->isB(new_real_end-1))
					new_real_end++;
				if((real_end>=0) && (new_real_end>=0))
					vInfo.end_src += (new_real_end - real_end);
			} else {
				vInfo.start_src	= video->nearestKey(vInfo.start_src + pVideoStream->getStart()) - pVideoStream->getStart();
				long end = vInfo.end_src + pVideoStream->getEnd();
				while(video->isB(end-1))
					end++;
				vInfo.end_src = end - pVideoStream->getStart();
			}
		}

		vInfo.cur_src		= vInfo.start_src;
		vInfo.end_dst		= (long)(vInfo.frameRate / vInfo.frameRateIn).scale64t(vInfo.end_src - vInfo.start_src);
		vInfo.end_proc_dst	= vInfo.end_dst;
	}

	vInfo.cur_proc_src = vInfo.cur_src;
	VDDEBUG("Dub: Video is from (%ld,%ld) starting at %ld\n", vInfo.start_src, vInfo.end_src, vInfo.cur_src);

	it.begin(streams);
	int nb = 1;
	while(it) {
		curStream = *it;
		if (curStream->m_Source) {
			// offset the start of the audio appropriately...

			curStream->m_aInfo.start_us = -1000*curStream->m_Opts.offset;

			if (video) {
				const sint64 video_start	= vInfo.start_src - pVideoStream->getStart();
				const sint64 video_length	= vInfo.end_src - vInfo.start_src;

				if (curStream->m_Opts.fStartAudio && video && opt->video.lStartOffsetMS) {
					if (!pfs) {
						curStream->m_aInfo.start_us += vInfo.frameRateIn.scale64ir(1000000 * video_start);
					}
				}
			}

			long newStartSrc = curStream->m_aInfo.start_src + curStream->m_Source->msToSamples((long)(curStream->m_aInfo.start_us / 1000));
			curStream->m_InitialSamplesLag = curStream->m_Source->samplesLag(newStartSrc);
			newStartSrc += curStream->m_InitialSamplesLag;
			
			curStream->m_OffsetSamples = newStartSrc - curStream->m_aInfo.start_src;
			curStream->m_aInfo.start_us = VD64(1000) * curStream->m_Source->samplesToMs(curStream->m_OffsetSamples);

			if(curStream->m_OffsetSamples > 0)
				curStream->m_OffsetSamples = 0;

			//curStream->m_StartUS = curStream->m_aInfo.start_us;
			if( (curStream->m_aInfo.start_us>=0) || !curStream->m_bHandleVBR || bPreview)
				curStream->m_aInfo.start_src = newStartSrc;

			// clip the end of the audio if supposed to...

			if (video && curStream->m_Opts.fEndAudio) {
				const sint64 video_length	= vInfo.end_src - vInfo.start_src;
				long lMaxLength;

				const VDFraction audiorate(curStream->m_Source->getRate());

				lMaxLength = (long)(audiorate / vInfo.frameRateIn).scale64r(video_length);

				if (curStream->m_aInfo.end_src - curStream->m_aInfo.start_src > lMaxLength)
					curStream->m_aInfo.end_src = curStream->m_aInfo.start_src + lMaxLength;
			}

			// resampling audio?

			curStream->m_aInfo.resampling = false;
			curStream->m_aInfo.converting = false;

			if (curStream->m_Opts.mode > DubAudioOptions::M_NONE) {
				if (curStream->m_Opts.new_rate) {
					curStream->m_aInfo.resampling = true;
				}

				if (curStream->m_Opts.newPrecision != DubAudioOptions::P_NOCHANGE || curStream->m_Opts.newChannels != DubAudioOptions::C_NOCHANGE) {
					curStream->m_aInfo.converting = true;

					curStream->m_aInfo.is_16bit = curStream->m_Opts.newPrecision==DubAudioOptions::P_16BIT
									|| (curStream->m_Opts.newPrecision==DubAudioOptions::P_NOCHANGE && ((AudioSource *)curStream->m_Source)->getWaveFormat()->wBitsPerSample>8);
					curStream->m_aInfo.is_stereo = curStream->m_Opts.newChannels==DubAudioOptions::C_STEREO
									|| (curStream->m_Opts.newChannels==DubAudioOptions::C_NOCHANGE && ((AudioSource *)curStream->m_Source)->getWaveFormat()->nChannels>1);
					curStream->m_aInfo.is_right = (curStream->m_Opts.newChannels==DubAudioOptions::C_MONORIGHT);
					curStream->m_aInfo.single_channel = (curStream->m_Opts.newChannels==DubAudioOptions::C_MONOLEFT || curStream->m_Opts.newChannels==DubAudioOptions::C_MONORIGHT);
					curStream->m_aInfo.bytesPerSample = (curStream->m_aInfo.is_16bit ? 2 : 1) * (curStream->m_aInfo.is_stereo ? 2 : 1);

				}
			}

			curStream->m_aInfo.cur_src	= curStream->m_aInfo.start_src;
			curStream->m_aInfo.cur_dst	= 0;
		}
		curStream->m_aInfo.cur_proc_src = curStream->m_aInfo.cur_src;
		VDDEBUG("Dub: Extra stream %d (%s) is from (%ld,%ld) starting at %ld\n"
			, nb++, curStream->m_Source ? curStream->m_Source->getSourceMainType()==SOURCE_AUDIO ?
				"Audio" : curStream->m_Source->getSourceMainType()==SOURCE_TEXT ?
				"Text" : curStream->m_Source->getSourceMainType()==SOURCE_VIDEO ?
				"Video" : "Unknown"
			: "No source!"
			, curStream->m_aInfo.start_src, curStream->m_aInfo.end_src, curStream->m_aInfo.cur_src);
		it++;
	}

}

//////////////////////////////////////////////////////////////////////////////

// may be called at any time in Init() after streams setup

void Dubber::InitAudioConversionChains() {

	// ready the audio stream for streaming operation
	ModList<ModDubStream>& streams = *m_Streams;

	ModList<ModDubStream>::iterator it(streams);
	AudioStream *audioStream = NULL;
	while(it) {
		ModDubStream *aStream = *it;
		if(!aStream->m_Source /*|| (aStream->m_Source->getSourceMainType()!=SOURCE_AUDIO) */) {
			it++;
			continue;
		}

		AudioSource *curAudioSource = (AudioSource *)aStream->m_Source;
		WAVEFORMATEX *curWFEX = curAudioSource->getWaveFormat();
		aStream->m_bIsNandubMP3VBR = (
			((curWFEX->wFormatTag==WAVE_FORMAT_MPEGLAYER3) || (curWFEX->wFormatTag==WAVE_FORMAT_MPEG))
			&& (curWFEX->nBlockAlign==1152)
			);
		// There are 3 cases in which we can find an nBlockalign of 1152 for an MP3 stream :
		// 1. MP3 VBR stream coming from ModAudioSourceMP3 and using the Nandub trick
		// 2. MP3 stream coming from an AVI file, and using the Nandub trick
		// 3. MP3 stream coming from an AVI file (chunks length is a multiple of 1152 then)
		// Case 3 is quite rare (and weird) but I once found such a file (apparently made with
		// Nandub). Since applying Nandub tricks to a stream that doesn't use them lead to serious
		// out-of-synch issues, we do an extra test here :
		// IsVBR() returns true in Case 1&2, and false in case 3
		IMP3Source *sMP3;
		if(aStream->m_Source->QueryInterface(IMP3Source::GetInterfaceID(), (void **)&sMP3)) {
			if(aStream->m_bIsNandubMP3VBR)
				aStream->m_bIsNandubMP3VBR = sMP3->IsVBR();
		}

		if(aStream->m_bSeemsVBR &&
			((aStream->m_Opts.mode>DubAudioOptions::M_NONE) || fPreview))
			aStream->m_bSeemsVBR = false;

		aStream->m_Source->streamBegin(fPreview);
		fADecompressionOk = true;

		// Initialize audio conversion chain

		bool bUseAudioFilterGraph = (aStream->m_Opts.mode > DubAudioOptions::M_NONE && aStream->m_AudioFilterGraph);

		if (bUseAudioFilterGraph) {
			// The input filter takes the globally defined inputAudio pointer as its audio source
			AudioSource *savedSource = inputAudio;
			inputAudio = (AudioSource *)aStream->m_Source;
			audioStream = new_nothrow AudioFilterSystemStream(*aStream->m_AudioFilterGraph, aStream->m_aInfo.start_us);
			inputAudio = savedSource;
			if (!audioStream)
				throw MyMemoryError();

			aStream->ChainAudioStream(audioStream);
		} else {
			// First, create a source.

			/*if( (fPreview || (aStream->m_Opts.mode > DubAudioOptions::M_NONE))
				&& (curWFEX->wFormatTag==0x2000)) {
				if (!(audioStream = new_nothrow ModAudioStreamSourceAC3(curAudioSource, aStream->m_aInfo.start_src, aStream->m_Source->getEnd() - aStream->m_aInfo.start_src)))
					throw MyError("Dub: Unable to create AC3 audio stream source");
			} else */if( (fPreview || (aStream->m_Opts.mode > DubAudioOptions::M_NONE))
				&& ((curWFEX->wFormatTag==WAVE_FORMAT_MPEGLAYER3) || (curWFEX->wFormatTag==WAVE_FORMAT_MPEG))
				&& g_VDMPrefs.m_bPrissRocks) {
				if (!(audioStream = new_nothrow ModAudioStreamSourceMP3(curAudioSource, aStream->m_aInfo.start_src, aStream->m_Source->getEnd() - aStream->m_aInfo.start_src)))
					throw MyError("Dub: Unable to create MP3 audio stream source");
				// the stream will be decompressed
				aStream->m_bIsNandubMP3VBR = false;
			} else if( (fPreview || (aStream->m_Opts.mode > DubAudioOptions::M_NONE))
				&& (curWFEX->wFormatTag==WAVE_FORMAT_EXTENSIBLE)
				&& (curAudioSource->getFormatLen() >= sizeof(WAVEFORMATEXTENSIBLE))
				&& (((WAVEFORMATEXTENSIBLE *)curWFEX)->SubFormat==MEDIASUBTYPE_Vorbis)) {
				if (!(audioStream = new_nothrow ModAudioStreamSourceVorbis(curAudioSource, aStream->m_aInfo.start_src, aStream->m_Source->getEnd() - aStream->m_aInfo.start_src)))
					throw MyError("Dub: Unable to create Vorbis audio stream source");
			} else if (!(audioStream = new_nothrow AudioStreamSource(curAudioSource, aStream->m_aInfo.start_src, aStream->m_Source->getEnd() - aStream->m_aInfo.start_src, aStream->m_Opts.mode > DubAudioOptions::M_NONE)))
				throw MyMemoryError();

			aStream->ChainAudioStream(audioStream);
		}

		// Tack on a subset filter as well...

		if (inputSubsetActive) {
			sint64 offset = 0;
		
			if (aStream->m_Opts.fStartAudio)
				offset = vInfo.frameRateIn.scale64ir((sint64)1000000 * vInfo.start_src);

			if((aStream->m_bHandleVBR && aStream->m_bSeemsVBR) || (aStream->m_Source->getSourceMainType()==SOURCE_TEXT)) {
				if (!(audioStream = new_nothrow ModAudioSubset(audioStream, inputSubsetActive, vInfo.frameRateIn, offset, aStream->m_Source->getSourceMainType()==SOURCE_TEXT, aStream)))
					throw MyError("Dub: Unable to create audio subset filter");
			} else {
				if (!(audioStream = new_nothrow AudioSubset(audioStream, inputSubsetActive, vInfo.frameRateIn, offset, aStream)))
					throw MyError("Dub: Unable to create audio subset filter");
			}

			aStream->ChainAudioStream(audioStream);
		}

		if (!bUseAudioFilterGraph) {
			// Attach a converter if we need to...

			if (aStream->m_aInfo.converting) {
				if (aStream->m_aInfo.single_channel)
					audioStream = new_nothrow AudioStreamConverter(audioStream, aStream->m_aInfo.is_16bit, aStream->m_aInfo.is_right, true);
				else
					audioStream = new_nothrow AudioStreamConverter(audioStream, aStream->m_aInfo.is_16bit, aStream->m_aInfo.is_stereo, false);

				if (!audioStream)
					throw MyMemoryError();

				aStream->ChainAudioStream(audioStream);
			}

			// Attach a converter if we need to...

			if (aStream->m_aInfo.resampling) {
				if (!(audioStream = new_nothrow AudioStreamResampler(audioStream, aStream->m_Opts.new_rate ? aStream->m_Opts.new_rate : ((AudioSource *)aStream->m_Source)->getWaveFormat()->nSamplesPerSec, aStream->m_Opts.integral_rate, aStream->m_Opts.fHighQuality)))
					throw MyMemoryError();

				aStream->ChainAudioStream(audioStream);
			}

			// Attach an amplifier if needed...

			if (aStream->m_Opts.mode > DubAudioOptions::M_NONE && aStream->m_Opts.volume) {
				if (!(audioStream = new_nothrow AudioStreamAmplifier(audioStream, aStream->m_Opts.volume)))
					throw MyMemoryError();

				aStream->ChainAudioStream(audioStream);
			}
		}

		// Make sure we only get what we want...

		if (vSrc && aStream->m_Opts.fEndAudio) {
			const WAVEFORMATEX *pAudioFormat = audioStream->GetFormat();
			const sint64 nFrames = (sint64)(vInfo.end_src - vInfo.start_src);
			VDFraction audioRate(pAudioFormat->nAvgBytesPerSec, pAudioFormat->nBlockAlign);
			if((aStream->m_Opts.mode == DubAudioOptions::M_NONE) && !fPreview && ((aStream->m_bHandleVBR && aStream->m_bSeemsVBR) || aStream->m_bIsNandubMP3VBR))
				audioRate = VDFraction(aStream->m_Source->getRate());
			const VDFraction audioPerVideo(audioRate / vInfo.frameRateIn);

			audioStream->SetLimit((long)audioPerVideo.scale64r(nFrames));
		}

		aStream->m_AudioTimingStream = audioStream;
		aStream->m_AudioStatusStream = audioStream;

		// Tack on a compressor if we want...

		AudioCompressor *pCompressor = NULL;

		if (aStream->m_Opts.mode > DubAudioOptions::M_NONE && !aStream->m_AudioCompressionFormat.empty()) {
			if (!(pCompressor = new_nothrow AudioCompressor(audioStream, &*aStream->m_AudioCompressionFormat, aStream->m_AudioCompressionFormat.size())))
				throw MyMemoryError();

			audioStream = pCompressor;
			aStream->ChainAudioStream(audioStream);
		}

		// Check the output format, and if we're compressing to
		// MPEG Layer III, compensate for the lag and create a bitrate corrector

		if (!fEnableSpill && !g_prefs.fNoCorrectLayer3 && pCompressor && pCompressor->GetFormat()->wFormatTag == WAVE_FORMAT_MPEGLAYER3) {
			pCompressor->CompensateForMP3();

			if (!(aStream->m_AudioCorrector = new_nothrow AudioStreamL3Corrector(audioStream)))
				throw MyMemoryError();

			audioStream = aStream->m_AudioCorrector;
			aStream->ChainAudioStream(audioStream);
		}

		aStream->m_bIsNandubMP3VBR = (aStream->m_bIsNandubMP3VBR && (aStream->m_Opts.mode == DubAudioOptions::M_NONE));
		aStream->m_SamplesPerBlock = VDFraction(1);
		curWFEX = audioStream->GetFormat();
		if(aStream->m_bIsNandubMP3VBR)
			aStream->m_SamplesPerBlock = VDFraction(curAudioSource->getStreamInfo().dwScale);
		else if(!aStream->m_bSeemsVBR && (curWFEX->nAvgBytesPerSec && curWFEX->nBlockAlign))
			aStream->m_SamplesPerBlock = VDFraction(curWFEX->nSamplesPerSec * curWFEX->nBlockAlign
				, curWFEX->nAvgBytesPerSec);

		if(curAudioSource->getSourceMainType() == SOURCE_TEXT)
			aStream->m_GroupFrames = 0;
		if( (!fPreview && (aStream->m_Opts.mode == DubAudioOptions::M_NONE))
			&& (curAudioSource->getWaveFormat()->wFormatTag==WAVE_FORMAT_EXTENSIBLE)
			&& (curAudioSource->getFormatLen() >= sizeof(WAVEFORMATEXTENSIBLE))
			&& (((WAVEFORMATEXTENSIBLE *)curAudioSource->getFormat())->SubFormat==MEDIASUBTYPE_Vorbis) )
			aStream->m_GroupFrames = 0;
		if(!aStream->m_bHandleVBR && aStream->m_bIsNandubMP3VBR)
			aStream->m_GroupFrames = 0;
		if(fPreview) {
			if( aStream->m_bIsNandubMP3VBR || (curAudioSource->getSourceType()==SOURCE_AUDIO_AC3) || (curAudioSource->getSourceType()==SOURCE_AUDIO_DTS) )
				aStream->m_GroupFrames = 0;
		}

		it++;
	}

}
// END ****************************************************************
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void Dubber::InitOutputFile() {

	ModList<ModDubStream>& streams = *m_Streams;

	// Do streams.
	ModList<ModDubStream>::iterator_safer it(streams);
	ModDubStream *curStream = NULL;
	while(it) {
		curStream = *it;
		if(curStream->m_Source) {
			if ( ((curStream->m_Source->getSourceMainType()==SOURCE_AUDIO) && mpOutputSystem->AcceptsAudio())
				|| ((curStream->m_Source->getSourceMainType()==SOURCE_TEXT) && mpOutputSystem->AcceptsText()) ) {
				// initialize AVI parameters...

				AVIStreamHeader_fixed	hdr;

				AVISTREAMINFOtoAVIStreamHeader(&hdr, &curStream->m_Source->getStreamInfo());
				hdr.dwStart			= 0;
				hdr.dwInitialFrames	= curStream->m_Opts.preload ? 1 : 0;

				if (curStream->m_Opts.mode > DubAudioOptions::M_NONE) {
					const WAVEFORMATEX *outputAudioFormat = curStream->m_AudioStream->GetFormat();
					hdr.dwSampleSize	= outputAudioFormat->nBlockAlign;
					hdr.dwRate			= outputAudioFormat->nAvgBytesPerSec;
					hdr.dwScale			= outputAudioFormat->nBlockAlign;
					hdr.dwLength		= MulDiv(hdr.dwLength, outputAudioFormat->nSamplesPerSec, ((AudioSource *)curStream->m_Source)->getWaveFormat()->nSamplesPerSec);
				}

				mpOutputSystem->AddStream(*curStream, hdr, curStream->m_AudioStream->GetFormat(), curStream->m_AudioStream->GetFormatLen(), curStream->m_Opts.enabled);
			}
		} else
			streams.DeleteNode(curStream);
		it++;
	}

	// Do video.

	if (vSrc && mpOutputSystem->AcceptsVideo()) {
		VBitmap outputBitmap;
		
		if (opt->video.mode >= DubVideoOptions::M_FULL)
			outputBitmap = *filters.OutputBitmap();
		else
			outputBitmap.init(vSrc->getFrameBuffer(), vSrc->getDecompressedFormat());

		AVIStreamHeader_fixed hdr;

		AVISTREAMINFOtoAVIStreamHeader(&hdr, &vSrc->asStream()->getStreamInfo());

		hdr.dwSampleSize = 0;

		if (opt->video.mode > DubVideoOptions::M_NONE) {
			if (fUseVideoCompression) {
				hdr.fccHandler	= compVars->fccHandler;
				hdr.dwQuality	= compVars->lQ;
			} else {
				hdr.fccHandler	= mmioFOURCC('D','I','B',' ');
			}
		}

		hdr.dwRate			= vInfo.frameRate.getHi();
		hdr.dwScale			= vInfo.frameRate.getLo();
		hdr.dwLength		= vInfo.end_dst;

		hdr.rcFrame.left	= 0;
		hdr.rcFrame.top		= 0;
		hdr.rcFrame.right	= (short)outputBitmap.w;
		hdr.rcFrame.bottom	= (short)outputBitmap.h;

		// initialize compression

		if (opt->video.mode >= DubVideoOptions::M_FASTREPACK) {
			if (opt->video.mode <= DubVideoOptions::M_SLOWREPACK) {
				const BITMAPINFOHEADER *pFormat = vSrc->getDecompressedFormat();

				mpCompressorVideoFormat.assign(pFormat, VDGetSizeOfBitmapHeaderW32(pFormat));
			} else {
				mpCompressorVideoFormat.resize(sizeof(BITMAPINFOHEADER));

				mpCompressorVideoFormat->biSize				= sizeof(BITMAPINFOHEADER);
				mpCompressorVideoFormat->biWidth			= outputBitmap.w;
				mpCompressorVideoFormat->biHeight			= outputBitmap.h;
				mpCompressorVideoFormat->biPlanes			= 1;
				mpCompressorVideoFormat->biBitCount			= iOutputDepth;
				mpCompressorVideoFormat->biCompression		= BI_RGB;
				mpCompressorVideoFormat->biSizeImage		= outputBitmap.pitch * outputBitmap.h;
				mpCompressorVideoFormat->biXPelsPerMeter	= 0;
				mpCompressorVideoFormat->biYPelsPerMeter	= 0;
				mpCompressorVideoFormat->biClrUsed			= 0;
				mpCompressorVideoFormat->biClrImportant		= 0;
			}
		} else {
			const BITMAPINFOHEADER *pFormat = vSrc->getImageFormat();

			mpCompressorVideoFormat.assign(pFormat, vSrc->asStream()->getFormatLen());
		}

		// Initialize output compressor.

		VDDEBUG("Dub: Initializing output compressor.\n");

		VDFormatStruct<BITMAPINFOHEADER>	outputFormat;

		if (fUseVideoCompression) {
			LONG formatSize;
			DWORD icErr;

			formatSize = ICCompressGetFormatSize(compVars->hic, (LPBITMAPINFO)&*mpCompressorVideoFormat);
			if (formatSize < ICERR_OK)
				throw "Error getting compressor output format size.";

			VDDEBUG("Video compression format size: %ld\n",formatSize);

			outputFormat.resize(formatSize);

			// Huffyuv doesn't initialize a few padding bytes at the end of its format
			// struct, so we clear them here.
			// Some XviD versions do the same (reserves 108 bytes - BITMAPV4HEADER - but only set the 40 of BITMAPINFOHEADER)
			memset(&*outputFormat, 0, outputFormat.size());

			if (ICERR_OK != (icErr = ICCompressGetFormat(compVars->hic,
								&*mpCompressorVideoFormat,
								&*outputFormat)))
				throw MyICError("Output compressor", icErr);

			if (!(pVideoPacker = new VideoSequenceCompressor()))
				throw MyMemoryError();

			pVideoPacker->init(compVars->hic, (LPBITMAPINFO)&*mpCompressorVideoFormat, (BITMAPINFO *)&*outputFormat, compVars->lQ, compVars->lKey);
			pVideoPacker->setDataRate(compVars->lDataRate*1024, vInfo.usPerFrame, vInfo.end_src - vInfo.start_src);
			pVideoPacker->start();

			lVideoSizeEstimate = pVideoPacker->getMaxSize();

			// attempt to open output decompressor

			if (opt->video.mode <= DubVideoOptions::M_FASTREPACK)
				fShowDecompressedFrame = false;
			else if (fShowDecompressedFrame = !!opt->video.fShowDecompressedFrame) {
				DWORD err;

				if (!(outputDecompressor = ICLocate(
							'CDIV',
							hdr.fccHandler,
							&*outputFormat,
							&*mpCompressorVideoFormat, ICMODE_DECOMPRESS))) {

					MyError("Output video warning: Could not locate output decompressor.").post(NULL,g_szError);

				} else if (ICERR_OK != (err = ICDecompressBegin(
						outputDecompressor,
						&*outputFormat,
						&*mpCompressorVideoFormat))) {

					MyICError("Output video warning", err).post(NULL,g_szError);

					ICClose(outputDecompressor);
					outputDecompressor = NULL;

					fShowDecompressedFrame = false;
				}
			}

		} else {
			if (opt->video.mode < DubVideoOptions::M_FASTREPACK) {

				if ((vSrc->getImageFormat()->biCompression == 0xFFFFFFFF) && (m_OutputFormat!=SOURCE_MKV) )
					throw MyError("The source video stream uses a compression algorithm which is not compatible with AVI files. "
								"Direct stream copy cannot be used with this video stream.");

				IVDStreamSource *pVideoStream = vSrc->asStream();

				outputFormat.assign(vSrc->getImageFormat(), pVideoStream->getFormatLen());

				// cheese
				const VDPosition videoFrameStart	= pVideoStream->getStart();
				const VDPosition videoFrameEnd		= pVideoStream->getEnd();

				lVideoSizeEstimate = 0;
				for(VDPosition frame = videoFrameStart; frame < videoFrameEnd; ++frame) {
					long bytes = 0;

					if (!pVideoStream->read(frame, 1, 0, 0, &bytes, 0))
						if (lVideoSizeEstimate < bytes)
							lVideoSizeEstimate = bytes;
				}
			} else {
				outputFormat.assign(vSrc->getDecompressedFormat(), sizeof(BITMAPINFOHEADER));

				if (opt->video.mode == DubVideoOptions::M_FULL) {
					outputFormat->biCompression= BI_RGB;
					outputFormat->biWidth		= outputBitmap.w;
					outputFormat->biHeight		= outputBitmap.h;
					outputFormat->biBitCount	= iOutputDepth;
					outputFormat->biSizeImage	= outputBitmap.pitch * outputBitmap.h;
				}

				lVideoSizeEstimate = outputFormat->biSizeImage;
				lVideoSizeEstimate = (lVideoSizeEstimate+1) & -2;
			}
		}

		mpOutputSystem->SetVideo(m_VideoStream, hdr, &*outputFormat, outputFormat.size());
	}

	VDDEBUG("Dub: Creating output file.\n");

	AVIout = mpOutputSystem->CreateSegment();
	mpVideoOut = AVIout->getVideoOutput();
}
// END ****************************************************************

bool Dubber::AttemptInputOverlay(BITMAPINFOHEADER *pbih) {
//	return vSrc->setDecompressedFormat(pbih) && mInputDisplay.InitDirectDrawYUVOverlay(mhwndInputDisplay, pbih, vSrc->getFrameBuffer());

	if (!vSrc->setDecompressedFormat(pbih))
		return false;

	int format;

	switch(pbih->biCompression) {
		case 'YVYU':
			format = IVDVideoDisplay::kFormatYUV422_UYVY;
			break;
		case '2YUY':
			format = IVDVideoDisplay::kFormatYUV422_YUYV;
			break;
		default:
			return false;
	}

	return mpInputDisplay->SetSource(vSrc->getFrameBuffer(), (pbih->biWidth*2 + 3) & ~3, pbih->biWidth, pbih->biHeight, format, 0, 0, false, opt->video.nPreviewFieldMode>0);
}

bool Dubber::AttemptInputOverlays() {
	BITMAPINFOHEADER bih;

	memcpy(&bih, vSrc->getImageFormat(), sizeof(BITMAPINFOHEADER));

	bih.biSize			= sizeof(BITMAPINFOHEADER);
	bih.biPlanes		= 1;
	bih.biBitCount		= 16;
	bih.biSizeImage		= (bih.biWidth+(bih.biWidth&1))*2*bih.biHeight;
	bih.biXPelsPerMeter	= 0;
	bih.biYPelsPerMeter	= 0;
	bih.biClrUsed		= 0;
	bih.biClrImportant	= 0;

	//---- begin 16-bit YUV negotiation ----

	// Attempt CYUV (YUV 4:2:2, Y?Y? ordering)

//	bih.biCompression = 'VUYC';
//	if (AttemptInputOverlay(&bih))
//		break;

	// Attempt UYVY (YUV 4:2:2)

	bih.biCompression = 'YVYU';
	if (AttemptInputOverlay(&bih))
		return true;

	// Attempt YUYV (YUV 4:2:2)
//	bih.biCompression = 'VYUY';
//	if (AttemptInputOverlay(&bih))
//		break;

	// Attempt YUY2 (YUV 4:2:2, YUYV ordering)

	bih.biCompression = '2YUY';

	if (AttemptInputOverlay(&bih))
		return true;

	//---- begin 12-bit YUV negotiation ----
#if 0
	// Attempt YV12 (YUV 4:2:0)

	bih.biCompression = '21VY';
	bih.biSizeImage		= (bih.biWidth/2)*(bih.biHeight/2)*6;
	bih.biBitCount		= 12;

	if (AttemptInputOverlay(&bih))
		return true;
#endif

	return false;
}

void Dubber::InitDirectDraw() {

	if (!opt->perf.useDirectDraw)
		return;

	// Should we try and establish a DirectDraw overlay?

	if (opt->video.mode == DubVideoOptions::M_SLOWREPACK) {
//		blitter->sendAFC(0x80000000, AttemptInputOverlays2, this);
		if (AttemptInputOverlays())
			mbInputDisplayInitialized = true;
	}
}

void Dubber::InitInputDisplay() {
	VDDEBUG("Dub: Initializing input window display.\n");


	// Check color depth of output device.  If it is 8-bit, we're
	// stuck with DrawDibDraw().  If it's at least 15 bits, then
	// we should create a DIBSection, select it into a memory
	// context, and BltBlt() to the screen instead.  It's about
	// 5-10% faster under Win95 and about 400x (!) faster under WINE.
	//
	// Okay, never mind... WINE still doesn't support DIBSection
	// windows. :(

	const BITMAPINFOHEADER& hdr = *vSrc->getDecompressedFormat();
	const void *pFB = vSrc->getFrameBuffer();
	const ptrdiff_t stride = ((hdr.biWidth * hdr.biBitCount + 31) >> 3) & ~3;
	int format = 0;

	if (hdr.biCompression == BI_RGB && hdr.biBitCount == 16)
		format = IVDVideoDisplay::kFormatRGB1555;
	else if (hdr.biCompression == BI_RGB && hdr.biBitCount == 24)
		format = IVDVideoDisplay::kFormatRGB888;
	else if (hdr.biCompression == BI_RGB && hdr.biBitCount == 32)
		format = IVDVideoDisplay::kFormatRGB8888;
	else if (hdr.biCompression == BI_BITFIELDS && hdr.biBitCount == 16)
		format = IVDVideoDisplay::kFormatRGB565;

	mpInputDisplay->SetSource((const char *)pFB + stride * (hdr.biHeight-1), -stride, hdr.biWidth, hdr.biHeight, format, 0, 0, true, opt->video.nPreviewFieldMode>0);

//	mInputDisplay.SetInterlacing(opt->video.nPreviewFieldMode != 0, opt->video.nPreviewFieldMode > 1);
}

void Dubber::InitOutputDisplay() {
	VDDEBUG("Dub: Initializing output window display.\n");

	if (opt->video.mode == DubVideoOptions::M_FULL) {
		mpDisplayFormat = mpCompressorVideoFormat;

		// check to see if DC is 565 16-bit, the only mode that does not support a line
		// of grays... hmm... is 15 possible for bitsPerPel?

		fDisplay565 = false;
		if (HDC hdc = GetDC(NULL)) {
			fDisplay565 = VDIsDC565(hdc);
			ReleaseDC(NULL, hdc);
		}

		if (fDisplay565 && fPreview && mpDisplayFormat->biBitCount == 16) {
			mpDisplayFormat.resize(sizeof(BITMAPV4HEADER));

			BITMAPV4HEADER *pV4Hdr = (BITMAPV4HEADER *)&*mpDisplayFormat;

			pV4Hdr->bV4Size				= sizeof(BITMAPV4HEADER);
			pV4Hdr->bV4V4Compression	= BI_BITFIELDS;
			pV4Hdr->bV4RedMask			= 0xf800;
			pV4Hdr->bV4GreenMask		= 0x07e0;
			pV4Hdr->bV4BlueMask			= 0x001f;
			pV4Hdr->bV4AlphaMask		= 0x0000;
			pV4Hdr->bV4CSType			= 0;
		}

		HANDLE hMapObject;
		LONG lMapOffset;

		filters.getOutputMappingParams(hMapObject, lMapOffset);

		int format = 0;
		if (mpDisplayFormat->biCompression == BI_RGB && mpDisplayFormat->biBitCount == 16)
			format = IVDVideoDisplay::kFormatRGB1555;
		else if (mpDisplayFormat->biCompression == BI_RGB && mpDisplayFormat->biBitCount == 24)
			format = IVDVideoDisplay::kFormatRGB888;
		else if (mpDisplayFormat->biCompression == BI_RGB && mpDisplayFormat->biBitCount == 32)
			format = IVDVideoDisplay::kFormatRGB8888;
		else if (mpDisplayFormat->biCompression == BI_BITFIELDS && mpDisplayFormat->biBitCount == 16)
			format = IVDVideoDisplay::kFormatRGB565;

		const VBitmap& vbm = *filters.OutputBitmap();

		mpOutputDisplay->SetSource((const char *)vbm.data + vbm.pitch * (vbm.h - 1), -vbm.pitch, vbm.w, vbm.h, format, hMapObject, lMapOffset, true, opt->video.nPreviewFieldMode>0);

//		mOutputDisplay.InitGDI(mhwndOutputDisplay, &*mpCompressorVideoFormat, hMapObject, lMapOffset);
	}

//	mOutputDisplay.SetInterlacing(opt->video.nPreviewFieldMode != 0, opt->video.nPreviewFieldMode > 1);
}

bool Dubber::NegotiateFastFormat(BITMAPINFOHEADER *pbih) {
	return vSrc->setDecompressedFormat(pbih) &&
			ICERR_OK == ICCompressQuery(compVars->hic, pbih, NULL);
}

bool Dubber::NegotiateFastFormat(int depth) {
	return vSrc->setDecompressedFormat(depth) &&
			ICERR_OK == ICCompressQuery(compVars->hic, vSrc->getDecompressedFormat(), NULL);
}

void Dubber::InitSelectInputFormat() {
	//	DIRECT:			Don't care.
	//	FASTREPACK:		Negotiate with output compressor.
	//	SLOWREPACK:		[Dub]		Use selected format.
	//					[Preview]	Negotiate with display driver.
	//	FULL:			Use selected format.

	if (opt->video.mode == DubVideoOptions::M_NONE)
		return;

	if (opt->video.mode == DubVideoOptions::M_FASTREPACK && fUseVideoCompression) {
		BITMAPINFOHEADER bih;

		// Begin decompressor to compressor format negotiation.
		//
		memcpy(&bih, vSrc->getImageFormat(), sizeof(BITMAPINFOHEADER));

		// Don't use odd-width YUV formats.  They may technically be allowed, but
		// a lot of codecs crash.  For instance, Huffyuv in "Convert to YUY2"
		// mode will accept a 639x360 format for compression, but crashes trying
		// to decompress it.
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Oct 2003												***
// BEGIN **************************************************************
		if (!(bih.biWidth & 1)) {
			bih.biSize			= sizeof(BITMAPINFOHEADER);
			bih.biPlanes		= 1;
			bih.biXPelsPerMeter	= 0;
			bih.biYPelsPerMeter	= 0;
			bih.biClrUsed		= 0;
			bih.biClrImportant	= 0;

			// Attempt YV12

			if (!(bih.biHeight & 1)) {
				bih.biBitCount		= 12;
				bih.biCompression	= '21VY';
				bih.biSizeImage		= ((bih.biWidth+1)>>1) * ((bih.biHeight+1)>>1) * 6;

				if (NegotiateFastFormat(&bih)) {
					mpInputDisplay->Reset();		// VideoDisplay doesn't handle planar formats (yet).
					mbInputDisplayInitialized = true;
					return;
				}
			}

			// Attempt UYVY.

			bih.biBitCount		= 16;
			bih.biCompression	= 'YVYU';
			bih.biSizeImage		= (bih.biWidth+(bih.biWidth&1))*2*bih.biHeight;

			if (NegotiateFastFormat(&bih)) {
				mpInputDisplay->SetSource(vSrc->getFrameBuffer(), (bih.biWidth*2 + 3) & ~3, bih.biWidth, bih.biHeight, IVDVideoDisplay::kFormatYUV422_UYVY);
				mbInputDisplayInitialized = true;
				return;
			}

			// Attempt YVYU.
			bih.biCompression	= 'UYVY';

			if (NegotiateFastFormat(&bih)) {
				mpInputDisplay->Reset();		// No support for YVYU yet
				mbInputDisplayInitialized = true;
				return;
			}

			// Attempt YUY2.

			bih.biCompression	= '2YUY';

			if (NegotiateFastFormat(&bih)) {
				mpInputDisplay->SetSource(vSrc->getFrameBuffer(), (bih.biWidth*2 + 3) & ~3, bih.biWidth, bih.biHeight, IVDVideoDisplay::kFormatYUV422_YUYV);
				mbInputDisplayInitialized = true;
				return;
			}
		}
// END ****************************************************************

		// Attempt RGB format negotiation.

		if (NegotiateFastFormat(16+8*opt->video.inputDepth))
			return;

		if (NegotiateFastFormat(24))
			return;
		if (NegotiateFastFormat(32))
			return;
		if (NegotiateFastFormat(16))
			return;
		if (NegotiateFastFormat(8))
			return;

		// Attempt source format.

		if (NegotiateFastFormat(vSrc->getImageFormat())) {
			mpInputDisplay->Reset();
			mbInputDisplayInitialized = true;
			return;
		}

		throw MyError("Video format negotiation failed: use slow-repack or full mode.");
	}

	// Negotiate RGB format.

	if (!vSrc->setDecompressedFormat(16+8*opt->video.inputDepth))
		if (!vSrc->setDecompressedFormat(32))
			if (!vSrc->setDecompressedFormat(24))
				if (!vSrc->setDecompressedFormat(16))
					if (!vSrc->setDecompressedFormat(8))
						throw MyError("The decompression codec cannot decompress to an RGB format. This is very unusual. Check that any \"Force YUY2\" options are not enabled in the codec's properties.");
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void Dubber::Init(ModVideoStream& videoStream, ModList<ModDubStream>& streams, IVDDubberOutputSystem *pOutputSystem, COMPVARS *videoCompVars) {

	vSrc				= videoStream.m_Source;
	mpOutputSystem		= pOutputSystem;

	m_VideoStream		= videoStream;
	m_VideoStream.m_Opts = opt->video;
	m_VideoChapters		= videoStream.m_Chapters;
	m_Streams			= &streams;

	fPreview			= mpOutputSystem->IsRealTime();

	compVars			= videoCompVars;
	fUseVideoCompression = !fPreview && opt->video.mode>DubVideoOptions::M_NONE && compVars && (compVars->dwFlags & ICMF_COMPVARS_VALID) && compVars->hic;
//	fUseVideoCompression = opt->video.mode>DubVideoOptions::M_NONE && compVars && (compVars->dwFlags & ICMF_COMPVARS_VALID) && compVars->hic;

	// check the mode; if we're using DirectStreamCopy mode, we'll need to
	// align the subset to keyframe boundaries!

	if (vSrc && inputSubset) {
		inputSubsetActive = inputSubset;

		if (opt->video.mode == DubVideoOptions::M_NONE) {
			if (!(inputSubsetActive = inputSubsetAlloc = new FrameSubset()))
				throw MyMemoryError();

			long lastEnd = -1;
			IVDStreamSource *pVideoStream = vSrc->asStream();

			const VDPosition videoFrameStart	= pVideoStream->getStart();

			for(FrameSubset::iterator it(inputSubset->begin()), itEnd(inputSubset->end()); it!=itEnd; ++it) {
				long real_start = it->start;
				long real_len = it->len;
				long end = it->start + it->len;
				long start = vSrc->nearestKey(it->start + vSrc->asStream()->getStart()) - vSrc->asStream()->getStart();
				if(it->bMask && (real_start < lastEnd)) {
					real_len -= (lastEnd - real_start);
					real_start = lastEnd;
				}

				// Go take the forward reference of B frames
				if(!it->bMask) {
					while(vSrc->isB(end-1)) {
						lastEnd = ++end;
						real_len++;
					}
				}

				VDDEBUG("   subset: %5d[%5d]-%-5d\n", it->start, start, it->start+it->len-1);
				inputSubsetActive->addRange(real_start, real_len, it->bMask);

				// Mask ranges never need to be extended backwards, because they don't hold any
				// data of their own.  If an include range needs to be extended backwards, though,
				// it may need to extend into a previous merge range.  To avoid this problem,
				// we do a delete of the range before adding the tail.

				if (!it->bMask) {
					if (start < it->start) {
						FrameSubset::iterator it2(it);

						while(it2 != inputSubset->begin()) {
							--it2;

							long prevtail = it2->start + it2->len;

							if (prevtail < start || prevtail > it->start + it->len)
								break;

							if (it2->start >= start || !it2->bMask) {	// within extension range: absorb
								it->len += it->start - it2->start;
								it->start = it2->start;
								it2 = inputSubset->erase(it2);
							} else {									// before extension range and masked: split merge
								int offset = start - it->start;
								it2->len -= offset;
								it->start -= offset;
								it->len += offset;
								break;
							}
						}

						int left = it->start - start;
						
						if (left > 0) {
							it->start = start;
							it->len += left;
						}
					}
				}
			}
		}
	}

	ModList<ModDubStream>::iterator streamIt(*m_Streams);
	while(streamIt) {
		ModDubStream *aStream = (*streamIt);
		bool bHandleVBRHere = m_bHandleVBR;
		bool bAllowGroupFrames = true;
		/* When using the audio filter graph don't forget to set in CBR mode */
		if(aStream->m_Opts.mode > DubAudioOptions::M_NONE && aStream->m_AudioFilterGraph) {
			bHandleVBRHere = false;
			bAllowGroupFrames = false;
		}
		/* Same when decompressing using ACM */
		else if(fPreview || (aStream->m_Opts.mode > DubAudioOptions::M_NONE)) {
			AudioSource *curAudioSource = (AudioSource *)aStream->m_Source;
			WAVEFORMATEX *curWFEX = curAudioSource->getWaveFormat();
			bool bCustomHandling = false;
			if( ((curWFEX->wFormatTag==WAVE_FORMAT_MPEGLAYER3) || (curWFEX->wFormatTag==WAVE_FORMAT_MPEG))
				&& g_VDMPrefs.m_bPrissRocks) {
				bCustomHandling = true;
			} else if( (curWFEX->wFormatTag==WAVE_FORMAT_EXTENSIBLE)
				&& (curAudioSource->getFormatLen() >= sizeof(WAVEFORMATEXTENSIBLE))
				&& (((WAVEFORMATEXTENSIBLE *)curWFEX)->SubFormat==MEDIASUBTYPE_Vorbis)) {
				bCustomHandling = true;
			}
			if(!bCustomHandling) {
				bHandleVBRHere = false;
				bAllowGroupFrames = false;
			}
		}
		aStream->m_bHandleVBR = bHandleVBRHere;
		if(!bAllowGroupFrames)
			aStream->m_GroupFrames = 1;
		streamIt++;
	}

	// initialize stream values

	InitStreamValuesStatic(vInfo, videoStream, streams, opt, inputSubsetActive, fPreview);

	vInfo.frameRateNoTelecine = vInfo.frameRate;
	vInfo.usPerFrameNoTelecine = vInfo.usPerFrame;
	if (opt->video.mode >= DubVideoOptions::M_FULL && opt->video.fInvTelecine) {
		vInfo.frameRate = vInfo.frameRate * VDFraction(4, 5);
		vInfo.usPerFrame = (long)vInfo.frameRate.scale64ir(1000000);

		vInfo.end_proc_dst	= (long)(vInfo.frameRate / vInfo.frameRateIn).scale64t(vInfo.end_src - vInfo.start_src);
		vInfo.end_dst += 4;
		vInfo.end_dst -= vInfo.end_dst % 5;
	}

	UpdateTimeSensitiveInformation();

	// initialize blitter

	VDDEBUG("Dub: Creating blitter.\n");

	if (!fPreview)
		blitter = new AsyncBlitter();
	else
		blitter = new AsyncBlitter(8);

	if (!blitter) throw MyError("Couldn't create AsyncBlitter");

	blitter->pulse();

	// initialize directdraw display if in preview

	mbInputDisplayInitialized = false;
	if (fPreview)
		InitDirectDraw();

	// Select an appropriate input format.  This is really tricky...

	vInfo.fAudioOnly = true;
	if (vSrc && mpOutputSystem->AcceptsVideo()) {
		if (!mbInputDisplayInitialized)
			InitSelectInputFormat();
		vInfo.fAudioOnly = false;
	}

	iOutputDepth = 16+8*opt->video.outputDepth;

	// Initialize filter system.

	int nVideoLagOutput = 0;		// Frames that will be buffered in the output frame space (video filters)
	int nVideoLagTimeline = 0;		// Frames that will be buffered in the timeline frame space (IVTC)

	if (vSrc && opt->video.mode >= DubVideoOptions::M_FULL) {
		BITMAPINFOHEADER *bmih = vSrc->getDecompressedFormat();

		filters.initLinearChain(&g_listFA, (Pixel *)(bmih+1), bmih->biWidth, bmih->biHeight, 32 /*bmih->biBitCount*/, iOutputDepth);

		fsi.lMicrosecsPerFrame		= vInfo.usPerFrame;
		fsi.lMicrosecsPerSrcFrame	= vInfo.usPerFrameIn;
		fsi.lCurrentFrame			= 0;
		fsi.flags					= fPreview ? FilterStateInfo::kStateRealTime | FilterStateInfo::kStatePreview : 0;

		if (filters.ReadyFilters(&fsi))
			throw "Error readying filters.";

		fFiltersOk = true;

		nVideoLagTimeline = nVideoLagOutput = filters.getFrameLag();

		// Inverse telecine?

		if (opt->video.fInvTelecine) {
			if (!(pInvTelecine = CreateVideoTelecineRemover(filters.InputBitmap(), !opt->video.fIVTCMode, opt->video.nIVTCOffset, opt->video.fIVTCPolarity)))
				throw MyMemoryError();

			nVideoLagTimeline = 10 + ((nVideoLagOutput+3)&~3)*5;
		}
	}

	nVideoLagPreload = nVideoLagOutput;
	vInfo.cur_dst = -nVideoLagTimeline;

	// initialize input decompressor

	if (vSrc && mpOutputSystem->AcceptsVideo()) {

		VDDEBUG("Dub: Initializing input decompressor.\n");

		vSrc->streamBegin(fPreview);
		fVDecompressionOk = true;

	}

	// Initialize audio.

	VDDEBUG("Dub: Initializing streams.\n");

	InitAudioConversionChains();

	// Initialize input window display.

	if (!mbInputDisplayInitialized) {
		if (vSrc && mpOutputSystem->AcceptsVideo())
			InitInputDisplay();
	}

	// initialize output parameters and output file

	InitOutputFile();
	UpdateInfo_NextSegment();

	// Initialize output window display.

	if (vSrc && mpOutputSystem->AcceptsVideo())
		InitOutputDisplay();


	// initialize interleaver

	int bAudio = 0;
	bool bInterleave = false;
	streamIt.begin(*m_Streams);
	while(streamIt) {
		bAudio++;
		bInterleave |= (*streamIt)->m_Opts.enabled;
		streamIt++;
	}

	mInterleaver.Init(bAudio + 1);
#pragma message(__TODO__ "Good ?")
	mInterleaver.EnableInterleaving(bInterleave);
	mInterleaver.InitStream(0, lVideoSizeEstimate, 0, 1, 1, 1);

	streamIt.begin(*m_Streams);
	int nIndex = 1;
	while(streamIt) {
		ModDubStream *curStream = *streamIt;

		Fraction audioBlocksPerVideoFrame;

		if (curStream->m_Opts.is_ms) {
			// blocks / frame = (ms / frame) / (ms / block)
			audioBlocksPerVideoFrame = Fraction(vInfo.usPerFrame, 1000) / Fraction(curStream->m_Opts.interval, 1);
		} else {
			audioBlocksPerVideoFrame = Fraction(1, curStream->m_Opts.interval);
		}

		// (bytes/sec) / (bytes/samples) = (samples/sec)
		// (samples/sec) / (frames/sec) = (samples/frame)
		// (samples/frame) / (blocks/frame) = (samples/block)

		const WAVEFORMATEX *pwfex = curStream->m_AudioStream->GetFormat();
		Fraction samplesPerSec(pwfex->nAvgBytesPerSec, pwfex->nBlockAlign);
		curStream->m_AudioPipe.m_bCBRPipe = true;
		if((curStream->m_Opts.mode == DubAudioOptions::M_NONE) && !fPreview && ((curStream->m_bHandleVBR && curStream->m_bSeemsVBR) || curStream->m_bIsNandubMP3VBR)) {
			samplesPerSec = Fraction(curStream->m_Source->getRate());
			curStream->m_AudioPipe.m_bCBRPipe = false;
		}

		if( (curStream->m_Source->getSourceMainType()!=SOURCE_AUDIO) || (curStream->m_bHandleVBR && !fPreview && g_VDMPrefs.m_bDisablePreload) )
			curStream->m_Opts.preload = 0;

		sint32 preload = (sint32)(samplesPerSec * Fraction(curStream->m_Opts.preload, 1000)).roundup32ul();
		curStream->m_aInfo.lPreloadSamples = preload;

		double samplesPerFrame = (double)samplesPerSec / (double)vInfo.frameRate;

		mInterleaver.InitStream(nIndex++
			, (curStream->m_AudioPipe.m_bCBRPipe ? pwfex->nBlockAlign : 0)
			, preload, samplesPerFrame, (double)audioBlocksPerVideoFrame
			, 262144);		// don't write TOO many samples at once

		AVIout->setStreamLag(curStream->m_Serial, curStream->m_SamplesPerBlock.scale64r(curStream->m_InitialSamplesLag));

		streamIt++;
	}

	if (fEnableSpill) {
		if (lSegmentFrameLimit)
			mInterleaver.SetSegmentFrameLimit(lSegmentFrameLimit);

		if (mSegmentByteLimit)
			mInterleaver.SetSegmentByteLimit(mSegmentByteLimit, 64);
	}

	// initialize frame iterator

	if (vSrc && mpOutputSystem->AcceptsVideo()) {
		mVideoFrameMap.Init(vSrc, vInfo.start_src, vInfo.frameRateIn / vInfo.frameRateNoTelecine, inputSubsetActive, vInfo.end_dst, opt->video.mode == DubVideoOptions::M_NONE);
		mVideoFrameIterator.Init(vSrc, opt->video.mode == DubVideoOptions::M_NONE);

		if (opt->video.mode == DubVideoOptions::M_NONE && fEnableSpill && mSegmentByteLimit)
			mInterleaver.SetCutEstimator(0, &mVideoFrameIterator);
	} else {
		mInterleaver.EndStream(0);
	}

	// Create data pipes.

	if (!(mpVideoPipe = new_nothrow AVIPipe(opt->perf.pipeBufferCount, 16384)))
		throw MyMemoryError();

	streamIt.begin(*m_Streams);
	while(streamIt) {
		ModDubStream *aStream = *streamIt;

		const WAVEFORMATEX *pwfex = aStream->m_AudioStream->GetFormat();

		uint32 bytes = pwfex->nAvgBytesPerSec * 2;		// 2 seconds

#pragma message(__TODO__ "Verify this is correct")
		uint32 samples = ceil(2. * double(aStream->m_Source->getRate()));

		if(aStream->m_AudioPipe.m_bCBRPipe)
			aStream->m_AudioPipe.Init(bytes - bytes % pwfex->nBlockAlign);
		else
			aStream->m_AudioPipe.Init(samples);
		
		streamIt++;
	}
}

static void UpdateChapters(ModChapterList& srcChapters, ModChapterList& dstChapters, sint64 iTimeStartMS, sint64 iTimeEndMS, FrameSubset *pSubset, VDFraction& frameRate) {
	ModChapterList::iterator chapterIt;

	dstChapters.clear();

	if(pSubset) {
		sint64 iTotalFrames = 0;
		FrameSubset::iterator subsetIt;
		for(subsetIt=pSubset->begin() ; subsetIt!=pSubset->end() ; subsetIt++) {
			sint64 iTimeSubsetStart = frameRate.scale64ir(VD64(1000) * (*subsetIt).start);
			sint64 iTimeSubsetEnd = frameRate.scale64ir(VD64(1000) * ((*subsetIt).start + (*subsetIt).len));

			for(chapterIt=srcChapters.begin() ; chapterIt!=srcChapters.end() ; chapterIt++) {
				ModChapterList::tChapterNode *pChapter = &(*chapterIt);

				sint32 iTimeChapterMS = sint32(pChapter->GetTimeMS());

				if((iTimeChapterMS>=iTimeSubsetStart) && (iTimeChapterMS<iTimeSubsetEnd)) {
					// This chapter belongs to the current subset
					iTimeChapterMS -= iTimeSubsetStart;					// Relative time in the subset
					iTimeChapterMS += frameRate.scale64ir(VD64(1000) * iTotalFrames);	// Absolute time in the global timeline
					if( (iTimeChapterMS>=iTimeStartMS) && (iTimeChapterMS<iTimeEndMS) )
						dstChapters.AddChapter(iTimeChapterMS - iTimeStartMS, pChapter->GetName());
				}
			}

			iTotalFrames += (*subsetIt).len;
		}
	} else {
		for(chapterIt=srcChapters.begin() ; chapterIt!=srcChapters.end() ; chapterIt++) {
			ModChapterList::tChapterNode *pChapter = &(*chapterIt);

			sint32 iTimeChapterMS = sint32(pChapter->GetTimeMS());
			if( (iTimeChapterMS>=iTimeStartMS) && (iTimeChapterMS<iTimeEndMS) )
				dstChapters.AddChapter(iTimeChapterMS - iTimeStartMS, pChapter->GetName());
		}
	}

	dstChapters.Sort();
}

void Dubber::UpdateTimeSensitiveInformation(void) {
	sint64 iTimeStartMS = opt->video.lStartOffsetMS + vInfo.frameRate.scale64ir((sint64)1000*vInfo.cur_proc_dst);
	sint64 iTimeEndMS = (vSrc->asStream())->samplesToMs(inputSubsetActive ? inputSubsetActive->getTotalFrames() : vSrc->asStream()->getLength())
		- opt->video.lEndOffsetMS;

	UpdateChapters(m_VideoChapters, m_VideoStream.m_Chapters, iTimeStartMS, iTimeEndMS, (m_VideoChapters.GetAdjustChapters() ? inputSubsetActive : NULL), vInfo.frameRate);

	ModList<ModDubStream>::iterator streamIt(*m_Streams);
	while(streamIt) {
		ModDubStream *pStream = *streamIt;

		UpdateChapters(pStream->m_SourceChapters, pStream->m_Chapters, iTimeStartMS, iTimeEndMS, (pStream->m_SourceChapters.GetAdjustChapters() ? inputSubsetActive : NULL), vInfo.frameRate);

		streamIt++;
	}
}
// END ****************************************************************

void Dubber::Go(int iPriority) {
	OSVERSIONINFO ovi;

	// check the version.  if NT, don't touch the processing priority!

	ovi.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);

	fNoProcessingPriority = GetVersionEx(&ovi) && ovi.dwPlatformId == VER_PLATFORM_WIN32_NT;

	if (!iPriority)
		iPriority = fNoProcessingPriority || !mpOutputSystem->IsRealTime() ? 5 : 6;

	// Reset timer.

	VDDEBUG("Dub: Starting multimedia timer.\n");

	if (fPreview) {
		int timerInterval = vInfo.usPerFrame / 1000;

		if (opt->video.fSyncToAudio || opt->video.nPreviewFieldMode) {
			timerInterval /= 2;
		}

		if (!mFrameTimer.Init(this, timerInterval))
			throw MyError("Couldn't initialize timer!");
	}

	// Initialize threads.

	VDDEBUG("Dub: Kickstarting threads.\n");

	if (!ThreadStart())
		throw MyError("Couldn't create processing thread");

//	if (fPreview && !fNoProcessingPriority)
	SetThreadPriority(getThreadHandle(), g_iPriorities[iPriority-1][1]);

	// Continue with other threads.

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
	if (!(mpIOThread = new_nothrow VDDubIOThread(
				fPhantom,
				mpOutputSystem->AcceptsVideo() ? vSrc : NULL,
				mVideoFrameIterator,
				*m_Streams,
				mpVideoPipe,
				fAbort,
				vInfo,
				mIOThreadCounter,
				m_AudioReadSignal,
				m_bHandleVBR)))
		throw MyMemoryError();
// END ****************************************************************

	if (!mpIOThread->ThreadStart())
		throw MyError("Couldn't start I/O thread");

	SetThreadPriority(mpIOThread->getThreadHandle(), g_iPriorities[iPriority-1][0]);

	// Create status window during the dub.

	VDDEBUG("Dub: Creating status window.\n");

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	ModDubStream *dubStream = m_Streams->GetFirst();
	AudioSource *audioSource = NULL;
	AudioStream *audioStatusStream = NULL;
	DubAudioStreamInfo aInfo;
	memset(&aInfo, 0, sizeof(aInfo));
	DubAudioStreamInfo *audioInfo = &aInfo;
	if(dubStream) {
		audioSource = (AudioSource *)dubStream->m_Source;
		audioStatusStream = dubStream->m_AudioStatusStream;
		audioInfo = &dubStream->m_aInfo;
	}
	pStatusHandler->InitLinks(&m_VideoStream, &vInfo, m_Streams, pInput, this, opt);
// END ****************************************************************

	pStatusHandler->Display(NULL, iPriority);
}

//////////////////////////////////////////////

void Dubber::Stop() {
	bool fSkipDXShutdown = false;

	if (mStopLock.xchg(1))
		return;

	fAbort = true;

	VDDEBUG("Dub: Beginning stop process.\n");

	if (mpVideoPipe)
		mpVideoPipe->abort();

	ModList<ModDubStream>::iterator streamIt(*m_Streams);
	ModDubStream *curStream = NULL;
	while(streamIt) {
		curStream = *streamIt;
		curStream->m_AudioPipe.m_CBRPipe.Abort();
		curStream->m_AudioPipe.m_VBRPipe.abort();
		streamIt++;
	}
	m_AudioReadSignal.signal();

	if (blitter)
		blitter->beginFlush();

	VDDEBUG("Dub: Killing threads.\n");

	int nObjectsToWaitOn = 0;
	HANDLE hObjects[3];

	if (VDSignal *pBlitterSigComplete = blitter->getFlushCompleteSignal())
		hObjects[nObjectsToWaitOn++] = pBlitterSigComplete->getHandle();

	if (this->isThreadAttached())
		hObjects[nObjectsToWaitOn++] = this->getThreadHandle();

	if (mpIOThread && mpIOThread->isThreadAttached())
		hObjects[nObjectsToWaitOn++] = mpIOThread->getThreadHandle();

	uint32 startTime = VDGetCurrentTick();

	while(nObjectsToWaitOn > 0) {
		DWORD dwRes;

		dwRes = MsgWaitForMultipleObjects(nObjectsToWaitOn, hObjects, FALSE, 10000, QS_ALLINPUT);

		if (WAIT_OBJECT_0 + nObjectsToWaitOn == dwRes) {
			guiDlgMessageLoop(hwndStatus);
			continue;
		}
		
		uint32 currentTime = VDGetCurrentTick();

		if ((dwRes -= WAIT_OBJECT_0) < nObjectsToWaitOn) {
			if (dwRes+1 < nObjectsToWaitOn)
				hObjects[dwRes] = hObjects[nObjectsToWaitOn - 1];
			--nObjectsToWaitOn;
			startTime = currentTime;
			continue;
		}

		if (currentTime - startTime > 10000) {
			if (IDOK == guiMessageBox(g_hWnd, IDS_ERR_DEADLOCK, IDS_ERR_DEADLOCK_CAP, MB_ICONEXCLAMATION|MB_OK)) {
				vdprotected("aborting process due to a thread deadlock") {
					ExitProcess(0);
				}
			}

			startTime = currentTime;
		}


		VDDEBUG("\tDub: %ld threads active\n", nObjectsToWaitOn);

#ifdef _DEBUG
		if (blitter) VDDEBUG("\t\tBlitter locks active: %08lx\n", blitter->lock_state);
#endif
	}

	if (!fError && mpIOThread)
		fError = mpIOThread->GetError(err);

	delete mpIOThread;
	mpIOThread = 0;

	if (blitter)
		blitter->abort();

	VDDEBUG("Dub: Freezing status handler.\n");

	if (pStatusHandler)
		pStatusHandler->Freeze();

	VDDEBUG("Dub: Killing timers.\n");

	mFrameTimer.Shutdown();

	if (pVideoPacker) {
		VDDEBUG("Dub: Ending frame compression.\n");

		delete pVideoPacker;

		pVideoPacker = NULL;
	}

	VDDEBUG("Dub: Deallocating resources.\n");

	if (mpVideoPipe)	{ delete mpVideoPipe; mpVideoPipe = NULL; }
	streamIt.begin(*m_Streams);
	while(streamIt) {
		curStream = *streamIt;
		curStream->m_AudioPipe.Shutdown();
		streamIt++;
	}

	if (blitter)		{ delete blitter; blitter=NULL; }

	filters.DeinitFilters();

	if (fVDecompressionOk)	{ vSrc->asStream()->streamEnd(); }
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	streamIt.begin(*m_Streams);
	while(streamIt) {
		curStream = *streamIt;

		if (fADecompressionOk && curStream->m_Source)	{ curStream->m_Source->streamEnd(); }

		curStream->RemoveAudioStreams();
		streamIt++;
	}
// BEGIN **************************************************************

	if (inputSubsetAlloc)		{ delete inputSubsetAlloc; inputSubsetAlloc = NULL; }

	VDDEBUG("Dub: Releasing display elements.\n");

	// deinitialize DirectDraw

	filters.DeallocateBuffers();
	
	delete pInvTelecine;	pInvTelecine = NULL;

	if (outputDecompressor)	{
		ICDecompressEnd(outputDecompressor);
		ICClose(outputDecompressor);
		outputDecompressor = NULL;
	}

	if (AVIout) {
		delete AVIout;
		AVIout = NULL;
		mpVideoOut = NULL;
	}

	if (pStatusHandler)
		pStatusHandler->SetLastPosition(vInfo.cur_proc_src);

	if (fError) {
		throw err;
		fError = false;
	}
}

///////////////////////////////////////////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
// This timer is only used in Preview mode. It needs to access the output audio
// stream created (even if there is no audio as input) otherwise the preview
// freeze
void Dubber::TimerCallback() {
	if (opt->video.fSyncToAudio) {
		IVDMediaOutputStream *audioOut = NULL;
		if(m_Streams->GetCount()) {
			ModList<ModDubStream>::iterator it(*m_Streams);
			ModDubStream *curStream = NULL;
			while(it) {
				curStream = *it;
				if(curStream->m_Source->getSourceMainType() == SOURCE_AUDIO) {
					audioOut = AVIout->getOutputStream(curStream->m_Serial);
					if(!audioOut)
						throw MyError("Unable to retrieve output stream %lu", curStream->m_Serial);
					break;
				}
				it++;
			}
			if(!audioOut)
				audioOut = AVIout->getAudioOutput();
		} else
			audioOut = AVIout->getAudioOutput();

		if (audioOut) {
			long lActualPoint;
			AVIAudioPreviewOutputStream *pAudioOut = static_cast<AVIAudioPreviewOutputStream *>(audioOut);

			lActualPoint = pAudioOut->getPosition();

			if (!pAudioOut->isFrozen()) {
				if (opt->video.nPreviewFieldMode) {
					mPulseClock = MulDiv(lActualPoint, 2000, vInfo.usPerFrame);
				} else {
					mPulseClock = MulDiv(lActualPoint, 1000, vInfo.usPerFrame);
				}

				if (mPulseClock<0)
					mPulseClock = 0;

				if (lActualPoint != -1) {
					blitter->setPulseClock(mPulseClock);
					fSyncToAudioEvenClock = false;
					mbAudioFrozen = false;
					return;
				}
			}
		}

		// Audio's frozen, so we have to free-run.  When 'sync to audio' is on
		// and field-based preview is off, we have to divide the 2x clock down
		// to 1x.

		mbAudioFrozen = true;

		if (fSyncToAudioEvenClock || opt->video.nPreviewFieldMode) {
			if (blitter) {
				blitter->pulse();
			}
		}

		fSyncToAudioEvenClock = !fSyncToAudioEvenClock;

		return;
	}

	// When 'sync to audio' is off we use a 1x clock.
	if (blitter)
		blitter->pulse();
}
// END ****************************************************************

/////////////////////////////////////////////////////////////

void Dubber::NextSegment() {
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	UpdateTimeSensitiveInformation();

	vdautoptr<AVIOutput> AVIout_new(mpOutputSystem->CreateSegment());
	mpOutputSystem->CloseSegment(AVIout, false);
	AVIout = AVIout_new.release();
	mpVideoOut = AVIout->getVideoOutput();

	UpdateInfo_NextSegment();

	LONG lFrame = vInfo.cur_proc_dst;
	long lAudioPoint;

	ModList<ModDubStream>::iterator it(*m_Streams);
	while(it) {
		ModDubStream *curStream = *it;

		if(curStream->m_bIsNandubMP3VBR) {
			const VDFraction audioRate(curStream->m_AudioTimingStream->GetFormat()->nSamplesPerSec, curStream->m_SamplesPerBlock);
			const VDFraction audioPerVideo(audioRate / vInfo.frameRateNoTelecine);
			lAudioPoint = (long)(/*curStream->m_aInfo.start_src + */audioPerVideo.scale64r(lFrame));
		} else if(curStream->m_bHandleVBR && curStream->m_bSeemsVBR) {
			const VDFraction audioRate(curStream->m_AudioTimingStream->GetFormat()->nSamplesPerSec);
			const VDFraction audioPerVideo(audioRate / vInfo.frameRateNoTelecine);
			lAudioPoint = (long)(/*curStream->m_aInfo.start_src + */audioPerVideo.scale64r(lFrame));
		} else {
			const VDFraction audioRate(curStream->m_AudioTimingStream->GetFormat()->nAvgBytesPerSec, curStream->m_AudioTimingStream->GetFormat()->nBlockAlign);
			const VDFraction audioPerVideo(audioRate / vInfo.frameRateNoTelecine);
			lAudioPoint = (long)(/*curStream->m_aInfo.start_src + */audioPerVideo.scale64r(lFrame));
		}

		VDDEBUG("Stream %ld was splitted at position %ld. In comparison the video stream was splitted at position %ld. Informing the output if necessary.\n"
			, curStream->m_Serial, curStream->m_aInfo.cur_proc_dst, lAudioPoint);

		if(curStream->m_aInfo.cur_proc_dst != lAudioPoint) {
			if( (m_OutputFormat==SOURCE_OGM)
			&& (curStream->m_Source->getSourceMainType()==SOURCE_TEXT)
			//&& !(curStream->m_bAudioPipeFinalizeAcked)
			&& (curStream->m_aInfo.cur_proc_dst>lAudioPoint)
			&& (curStream->m_FrameSize>0) )
				WriteStream(*curStream, &curStream->m_Frame[0], curStream->m_FrameSize, curStream->m_SamplesPerBlock.scale64r(curStream->m_aInfo.cur_proc_dst - lAudioPoint));
			else
				AVIout->setStreamLag(curStream->m_Serial, curStream->m_SamplesPerBlock.scale64r(curStream->m_aInfo.cur_proc_dst - lAudioPoint));
		}

		it++;
	}
// END ****************************************************************
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, July 2003											***
// BEGIN **************************************************************
void Dubber::UpdateInfo_NextSegment(void) {
	ModList<ModDubStream>::iterator it(*m_Streams);
	while(it) {
		ModDubStream *curStream = *it;

		sint64 iOffsetMS = -curStream->m_Opts.offset;
		curStream->m_OverlappingOutput = NULL;
		ISubtitleHandler *pSub;
		IOverlappingOutput *pOutput;
		IVDMediaOutputStream *outputStream = AVIout->getOutputStream(curStream->m_Serial);
		if(!outputStream)
			throw MyError("Unable to retrieve output stream %lu", curStream->m_Serial);
		if( (curStream->m_Source->getSourceMainType()==SOURCE_TEXT)
			&& curStream->m_Source->QueryInterface(ISubtitleHandler::GetInterfaceID(), (void **)&pSub)
			&& outputStream->QueryInterface(IOverlappingOutput::GetInterfaceID(), (void **)&pOutput) ) {
			ModSubtitleProvider *subProvider = pSub->GetSubtitleProvider();
			if(subProvider) {
				if(!curStream->m_bShortcutSubtitles) {
					curStream->m_SubtitleProvider = *subProvider;
					curStream->m_SubtitleProvider.ApplySubsets(
						iOffsetMS + opt->video.lStartOffsetMS
						, iOffsetMS + vSrc->asStream()->samplesToMs(inputSubsetActive ? inputSubsetActive->getTotalFrames() : vSrc->asStream()->getLength()) - opt->video.lEndOffsetMS
						, inputSubsetActive
						, vInfo.frameRate);
				}
				curStream->m_bShortcutSubtitles = true;
				curStream->m_OverlappingOutput = pOutput;
			}
		}

		it++;
	}
}
// END ****************************************************************

///////////////////////////////////////////////////////////////////

#define BUFFERID_INPUT (1)
#define BUFFERID_OUTPUT (2)

namespace {
	bool AsyncUpdateCallback(int pass, void *pDisplayAsVoid, void *pInterlaced) {
		IVDVideoDisplay *pVideoDisplay = (IVDVideoDisplay *)pDisplayAsVoid;
		int nFieldMode = *(int *)pInterlaced;

		if (nFieldMode) {
			if ((pass^nFieldMode)&1)
				pVideoDisplay->Update(IVDVideoDisplay::kEvenFieldOnly | IVDVideoDisplay::kVisibleOnly);
			else
				pVideoDisplay->Update(IVDVideoDisplay::kOddFieldOnly | IVDVideoDisplay::kVisibleOnly);

			return !pass;
		} else {
			pVideoDisplay->Update(IVDVideoDisplay::kAllFields | IVDVideoDisplay::kVisibleOnly);
			return false;
		}
	}
}

Dubber::VideoWriteResult Dubber::WriteVideoFrame(void *buffer, int exdata, int droptype, LONG lastSize, VDPosition sample_num, VDPosition display_num, VDPosition timeline_num) {
	LONG dwBytes;
	bool isKey;
	void *frameBuffer;
	LPVOID lpCompressedData;

	if (timeline_num >= 0)		// exclude injected frames
		vInfo.cur_proc_src = timeline_num;

	// Preview fast drop -- if there is another keyframe in the pipe, we can drop
	// all the frames to it without even decoding them!
	//
	// Anime song played during development of this feature: "Trust" from the
	// Vandread OST.

//	VDDEBUG2("writing sample %d (display %d), type = %s, drop = %d\n", (int)sample_num, (int)display_num, droptype==AVIPipe::kDroppable ? "droppable" : droptype == AVIPipe::kDependant ? "dependant" : "independent", lDropFrames);

	if (fPreview && opt->perf.fDropFrames) {

		// If audio is frozen, force frames to be dropped.

		bool bDrop = true;

		bDrop = !vSrc->isDecodable(sample_num);
//		if (bDrop)
//			VDDEBUG2("Dubber: Forced drop (undecodable)\n");

		if (mbAudioFrozen && mbAudioFrozenValid) {
			lDropFrames = 1;
		}

		if (lDropFrames) {
			long lCurrentDelta = blitter->getFrameDelta();
			if (opt->video.nPreviewFieldMode)
				lCurrentDelta >>= 1;
			if (lDropFrames > lCurrentDelta) {
				lDropFrames = lCurrentDelta;
				if (lDropFrames < 0)
					lDropFrames = 0;
//				VDDEBUG2("Dubber: Dropping frame skip to %d\n", lDropFrames);
			}
		}

		if (lDropFrames && !bDrop) {

			// Attempt to drop a frame before the decoder.  Droppable frames (zero-byte
			// or B-frames) can be dropped without any problem without question.  Dependant
			// (P-frames or delta frames) and independent frames (I-frames or keyframes)
			// should only be dropped if there is a reasonable expectation that another
			// independent frame will arrive around the time that we want to stop dropping
			// frames, since we'll basically kill decoding until then.

			if (droptype == AVIPipe::kDroppable) {
				bDrop = true;
			} else {
				int total, indep;

				mpVideoPipe->getDropDistances(total, indep);

				// Do a blind drop if we know a keyframe will arrive within two frames.

				if (indep == 0x3FFFFFFF && vSrc->nearestKey(display_num + opt->video.frameRateDecimation*2) > display_num)
					indep = 0;

				if (indep > 0 && indep < lDropFrames) {
//					VDDEBUG2("Dubber: Proactive drop (%d < %d)\n", indep, lDropFrames);
					bDrop = true;
				}
			}
		}

		if (bDrop) {
			if (!(exdata&2)) {
				blitter->nextFrame(opt->video.nPreviewFieldMode ? 2 : 1);
//				VDDEBUG2("Dubber: skipped frame\n");
			}
			++fsi.lCurrentFrame;
			if (lDropFrames)
				--lDropFrames;

			pStatusHandler->NotifyNewFrame(0);

			return kVideoWriteDiscarded;
		}
	}

	// With Direct mode, write video data directly to output.

	if (opt->video.mode == DubVideoOptions::M_NONE || fPhantom) {
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
		mpVideoOut->write(
			(exdata & kBufferFlagDelta) ? (exdata & kBufferFlagB) ? AVIIF_BFRAME : 0 : AVIIF_KEYFRAME
			, (char *)buffer, lastSize, 1);
// END ****************************************************************

		vInfo.total_size += lastSize + 24;
		++vInfo.processed;
		pStatusHandler->NotifyNewFrame(lastSize | (exdata&1 ? 0x80000000 : 0));
		mInterleaver.AddVBRCorrection(0, lastSize);

		return kVideoWriteOK;
	}

	// Fast Repack: Decompress data and send to compressor (possibly non-RGB).
	// Slow Repack: Decompress data and send to compressor.
	// Full:		Decompress, process, filter, convert, send to compressor.

	mProcessingProfileChannel.Begin(0xe0e0e0, "V-Lock1");
	blitter->lock(BUFFERID_INPUT);
	mProcessingProfileChannel.End();

	if (exdata & kBufferFlagPreload)
		mProcessingProfileChannel.Begin(0xfff0f0, "V-Preload");
	else
		mProcessingProfileChannel.Begin(0xffe0e0, "V-Decode");

	VDCHECKPOINT;
	CHECK_FPU_STACK
	vSrc->streamGetFrame(buffer, lastSize, !(exdata&kBufferFlagDelta), false, sample_num);
	CHECK_FPU_STACK

	VDCHECKPOINT;

	mProcessingProfileChannel.End();

	if (exdata & kBufferFlagPreload) {
		blitter->unlock(BUFFERID_INPUT);
		return kVideoWriteBuffered;
	}

	if (lDropFrames && fPreview) {
		blitter->unlock(BUFFERID_INPUT);
		blitter->nextFrame(opt->video.nPreviewFieldMode ? 2 : 1);
		++fsi.lCurrentFrame;

//		VDDEBUG2("Dubber: skipped frame\n");
		--lDropFrames;

		pStatusHandler->NotifyNewFrame(0);

		return kVideoWriteDiscarded;
	}

	// Process frame to backbuffer for Full video mode.  Do not process if we are
	// running in Repack mode only!
	if (opt->video.mode == DubVideoOptions::M_FULL) {
		VBitmap *initialBitmap = filters.InputBitmap();
		VBitmap *lastBitmap = filters.LastBitmap();
		VBitmap *outputBitmap = filters.OutputBitmap();
		VBitmap destbm;
		long lInputFrameNum, lInputFrameNum2;

		lInputFrameNum = display_num - vSrc->asStream()->getStart();

		if (pInvTelecine) {
			lInputFrameNum2 = pInvTelecine->ProcessOut(initialBitmap);
			pInvTelecine->ProcessIn(&VBitmap(vSrc->getFrameBuffer(), vSrc->getDecompressedFormat()), lInputFrameNum);

			lInputFrameNum = lInputFrameNum2;

			if (lInputFrameNum < 0) {
				blitter->unlock(BUFFERID_INPUT);
				return kVideoWriteBuffered;
			}
		} else
			initialBitmap->BitBlt(0, 0, &VBitmap(vSrc->getFrameBuffer(), vSrc->getDecompressedFormat()), 0, 0, -1, -1);

		// process frame

		fsi.lCurrentSourceFrame	= lInputFrameNum;
		fsi.lSourceFrameMS		= (long)vInfo.frameRateIn.scale64ir(fsi.lCurrentSourceFrame * (sint64)1000);
		fsi.lDestFrameMS		= (long)vInfo.frameRateIn.scale64ir(fsi.lCurrentFrame * (sint64)1000);

		mProcessingProfileChannel.Begin(0x008000, "V-Filter");
		filters.RunFilters();
		mProcessingProfileChannel.End();

		++fsi.lCurrentFrame;

		if (nVideoLagPreload>0) {
			--nVideoLagPreload;
			blitter->unlock(BUFFERID_INPUT);
			return kVideoWriteBuffered;
		}


		mProcessingProfileChannel.Begin(0xe0e0e0, "V-Lock2");
		blitter->lock(BUFFERID_OUTPUT);
		mProcessingProfileChannel.End();

//		if (!outputDecompressor)
//			outputBitmap.data = outputBuffer;

		do {
//			IDDrawSurface *pds = mOutputDisplay.GetDD();
//
//			if (pds) {
//				if (!pds->LockInverted(&destbm))
//					break;
//
//				outputBitmap = &destbm;
//			}

			if (fPreview && g_prefs.fDisplay & Preferences::kDisplayDither16)
				outputBitmap->BitBltDither(0, 0, lastBitmap, 0, 0, -1, -1, fDisplay565);
			else if (mpDisplayFormat->biCompression == BI_BITFIELDS)
				outputBitmap->BitBlt565(0, 0, lastBitmap, 0, 0, -1, -1);
			else
				outputBitmap->BitBlt(0, 0, lastBitmap, 0, 0, -1, -1);

//			if (pds)
//				pds->Unlock();

		} while(false);
	}

	// write it to the file
	
	frameBuffer = 		/*(opt->video.mode == DubVideoOptions::M_FASTREPACK ? buffer : */
						opt->video.mode <= DubVideoOptions::M_SLOWREPACK ? vSrc->getFrameBuffer()
						:filters.OutputBitmap()->data;


	if (pVideoPacker) {
		mProcessingProfileChannel.Begin(0x80c080, "V-Compress");
		lpCompressedData = pVideoPacker->packFrame(frameBuffer, &isKey, &dwBytes);
		mProcessingProfileChannel.End();

		// Check if codec buffered a frame.

		if (!lpCompressedData) {
			return kVideoWriteDelayed;
		}

		if (fShowDecompressedFrame && outputDecompressor && dwBytes) {
			DWORD err;
			VBitmap *outputBitmap = filters.OutputBitmap();
			Pixel *outputBuffer = outputBitmap->data;

			CHECK_FPU_STACK

			DWORD dwSize = mpCompressorVideoFormat->biSizeImage;

			mpCompressorVideoFormat->biSizeImage = dwBytes;

			VDCHECKPOINT;
			if (ICERR_OK != (err = ICDecompress(outputDecompressor,
				isKey ? 0 : ICDECOMPRESS_NOTKEYFRAME,
				(BITMAPINFOHEADER *)mpVideoOut->getFormat(),
				lpCompressedData,
				&*mpCompressorVideoFormat,
				outputBuffer
				)))

				fShowDecompressedFrame = false;
			VDCHECKPOINT;

			mpCompressorVideoFormat->biSizeImage = dwSize;

			CHECK_FPU_STACK
		}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
		// jonny's mods
		if (g_OptionNoWrite == 0)
			mpVideoOut->write(isKey ? AVIIF_KEYFRAME : 0, (char *)lpCompressedData, dwBytes, 1);
// END ****************************************************************
	} else {

		dwBytes = ((const BITMAPINFOHEADER *)mpVideoOut->getFormat())->biSizeImage;

		VDCHECKPOINT;
		mpVideoOut->write(AVIIF_KEYFRAME, (char *)frameBuffer, dwBytes, 1);
		VDCHECKPOINT;

		isKey = true;
	}

	vInfo.total_size += dwBytes + 24;
	mInterleaver.AddVBRCorrection(0, dwBytes + (dwBytes&1));

	VDCHECKPOINT;

	if (fPreview || mRefreshFlag.xchg(0)) {
		if (opt->video.fShowInputFrame) {
			blitter->postAPC(BUFFERID_INPUT, AsyncUpdateCallback, mpInputDisplay, &opt->video.nPreviewFieldMode);
		} else
			blitter->unlock(BUFFERID_INPUT);

		if (opt->video.mode == DubVideoOptions::M_FULL && opt->video.fShowOutputFrame && (!outputDecompressor || dwBytes)) {
			blitter->postAPC(BUFFERID_OUTPUT, AsyncUpdateCallback, mpOutputDisplay, &opt->video.nPreviewFieldMode);
		} else
			blitter->unlock(BUFFERID_OUTPUT);
	} else {
		blitter->unlock(BUFFERID_OUTPUT);
		blitter->unlock(BUFFERID_INPUT);
	}

	if (opt->perf.fDropFrames && fPreview) {
		long lFrameDelta;

		lFrameDelta = blitter->getFrameDelta();

		if (opt->video.nPreviewFieldMode)
			lFrameDelta >>= 1;

		if (lFrameDelta < 0) lFrameDelta = 0;
		
		if (lFrameDelta > 0) {
			lDropFrames = lFrameDelta;
//			VDDEBUG2("Dubber: Skipping %d frames\n", lDropFrames);
		}
	}


	blitter->nextFrame(opt->video.nPreviewFieldMode ? 2 : 1);

	++vInfo.processed;

	pStatusHandler->NotifyNewFrame(isKey ? dwBytes : dwBytes | 0x80000000);

	VDCHECKPOINT;
	return kVideoWriteOK;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void Dubber::WriteStream(ModDubStream& stream, void *buffer, long lActualBytes, long lActualSamples) {
	if (!lActualBytes) return;
	IVDMediaOutputStream *streamOut = AVIout->getOutputStream(stream.m_Serial);
	if(streamOut) {
		streamOut->write(AVIIF_KEYFRAME, (char *)buffer, lActualBytes, lActualSamples);

		if((m_OutputFormat==SOURCE_OGM) && (stream.m_Source->getSourceMainType()==SOURCE_TEXT)) {
			if(stream.m_Frame.size() < lActualBytes)
				stream.m_Frame.resize(lActualBytes);
			memcpy(&stream.m_Frame[0], buffer, lActualBytes);
			stream.m_FrameSize = lActualBytes;
		}
	} else
		throw MyError("Unable to retrieve output stream %lu", stream.m_Serial);

	stream.m_aInfo.cur_proc_src += lActualBytes;
}
// END ****************************************************************

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
void Dubber::ThreadRun() {
	bool quit = false;
	bool firstPacket = true;
	bool bVideoEnded = !(vSrc && mpOutputSystem->AcceptsVideo());
	bool bVideoNonDelayedFrameReceived = false;
	//bool bAudioEnded = !(aSrc && mpOutputSystem->AcceptsAudio());
	int iAudioActive = m_Streams->GetCount();
	bool bAudio = !!iAudioActive;
	ModList<ModDubStream>::iterator streamIt;
	ModDubStream *pDubStream;
	bool bOverflowReportedThisSegment = false;
	uint32	nVideoFramesDelayed = 0;

	lDropFrames = 0;
	vInfo.processed = 0;

	VDDEBUG("Dub/Processor: start\n");

	std::vector<char>	audioBuffer;

	try {
		DEFINE_SP(sp);

		for(;;) {
			int stream;
			sint32 count;

			VDStreamInterleaver::Action nextAction = mInterleaver.GetNextAction(stream, count);

			if (!bOverflowReportedThisSegment && mInterleaver.HasSegmentOverflowed()) {
				VDLogAppMessage(kVDLogWarning, kVDST_Dub, kVDM_SegmentOverflowOccurred);
				bOverflowReportedThisSegment = true;
			}

			++mProcessingThreadCounter;

			if (nextAction == VDStreamInterleaver::kActionFinish)
				break;
			else if (nextAction == VDStreamInterleaver::kActionNextSegment) {
				VDLogAppMessage(kVDLogMarker, kVDST_Dub, kVDM_BeginningNextSegment);
				NextSegment();
				bOverflowReportedThisSegment = false;
			} else if (nextAction == VDStreamInterleaver::kActionWrite) {
				if (stream == 0) {
					if (vInfo.cur_proc_dst >= vInfo.end_proc_dst) {
						if (fPreview && bAudio) {
							streamIt.begin(*m_Streams);
							while(streamIt) {
								if((*streamIt)->m_Source->getSourceMainType() == SOURCE_AUDIO) {
									IVDMediaOutputStream *streamOut = AVIout->getOutputStream((*streamIt)->m_Serial);
									if(!streamOut)
										throw MyError("Unable to retrieve output stream %lu", (*streamIt)->m_Serial);
									static_cast<AVIAudioPreviewOutputStream *>(streamOut)->start();
								}
								streamIt++;
							}
							mbAudioFrozenValid = true;
						}
						mInterleaver.EndStream(0);
						bVideoEnded = true;
						SetVideoEOS();
					} else {
						// We cannot wrap the entire loop with a profiling event because typically
						// involves a nice wait in preview mode.

						uint32 nFramesPushedTryingToFlushCodec = 0;

						for(;;) {
							void *buf;
							long len;
							VDPosition	rawFrame;
							VDPosition	displayFrame;
							VDPosition	timelineFrame;
							int exdata;
							int handle;
							int droptype;
							bool bAttemptingToFlushCodecBuffer = false;

							if (fAbort)
								goto abort_requested;

							buf = mpVideoPipe->getReadBuffer(len, rawFrame, displayFrame, timelineFrame, &exdata, &droptype, &handle);
							if (!buf) {
								if (nVideoFramesDelayed > 0) {
									--nVideoFramesDelayed;

									buf			= "";
									len			= 0;
									rawFrame		= -1;
									displayFrame	= -1;
									timelineFrame	= -1;
									exdata		= 0;
									droptype	= AVIPipe::kDroppable;
									handle		= -1;
									bAttemptingToFlushCodecBuffer = true;
								} else {
									mInterleaver.EndStream(0);
									bVideoEnded = true;
									SetVideoEOS();
									break;
								}
							}

							if (firstPacket && fPreview && !bAudio) {
								blitter->enablePulsing(true);
								firstPacket = false;
							}

							VideoWriteResult result = WriteVideoFrame(buf, exdata, droptype, len, rawFrame, displayFrame, timelineFrame);

							if (result == kVideoWriteDelayed) {
								if (bAttemptingToFlushCodecBuffer) {
									const int kReasonableBFrameBufferLimit = 100;

									++nFramesPushedTryingToFlushCodec;

									// DivX 5.0.5 seems to have a bug where in the second pass of a multipass operation
									// it outputs an endless number of delay frames at the end!  This causes us to loop
									// infinitely trying to flush a codec delay that never ends.  Unfortunately, there is
									// one case where such a string of delay frames is valid: when the length of video
									// being compressed is shorter than the B-frame delay.  We attempt to detect when
									// this situation occurs and avert the loop.

									if (bVideoNonDelayedFrameReceived) {
										VDLogAppMessage(kVDLogWarning, kVDST_Dub, kVDM_CodecDelayedDuringDelayedFlush);
										--nVideoFramesDelayed;	// cancel increment below (might underflow but that's OK)
									} else if (nFramesPushedTryingToFlushCodec > kReasonableBFrameBufferLimit) {
										VDLogAppMessage(kVDLogWarning, kVDST_Dub, kVDM_CodecLoopingDuringDelayedFlush, 1, &kReasonableBFrameBufferLimit);
										nVideoFramesDelayed = 0;
										continue;
									}
								}
								++nVideoFramesDelayed;
							}

							bVideoNonDelayedFrameReceived = true;

							if (fPreview && bAudio) {
								streamIt.begin(*m_Streams);
								while(streamIt) {
									if((*streamIt)->m_Source->getSourceMainType() == SOURCE_AUDIO) {
										IVDMediaOutputStream *streamOut = AVIout->getOutputStream((*streamIt)->m_Serial);
										if(!streamOut)
											throw MyError("Unable to retrieve output stream %lu", (*streamIt)->m_Serial);
										static_cast<AVIAudioPreviewOutputStream *>(streamOut)->start();
									}
									streamIt++;
								}
								mbAudioFrozenValid = true;
							}

							if (handle >= 0)
								mpVideoPipe->releaseBuffer(handle);

							if (result == kVideoWriteOK || result == kVideoWriteDiscarded)
								break;
						}
						++vInfo.cur_proc_dst;
					}
				} else /*if (stream == 1) */{
#pragma message(__TODO__ "Use another string ?")
					mProcessingProfileChannel.Begin(0xe0e0ff, "Audio");
					pDubStream = (*m_Streams)[stream-1];
					if(!pDubStream)
						throw MyError("Attemp to access Stream %ld that hasn't been defined", stream);

					if(pDubStream->m_AudioPipe.m_bCBRPipe) {
						const int nBlockAlign = pDubStream->m_AudioStream->GetFormat()->nBlockAlign;
						int bytes = count * nBlockAlign;
						int bytesread = 0;

						if (audioBuffer.size() < bytes)
							audioBuffer.resize(bytes);

						while(bytesread < bytes) {
							int tc = pDubStream->m_AudioPipe.m_CBRPipe.Read(&audioBuffer[bytesread], bytes-bytesread);

							m_AudioReadSignal.signal();

							if (fAbort)
								goto abort_requested;

							if (!tc) {
								if (pDubStream->m_AudioPipe.m_CBRPipe.isInputClosed()) {
									pDubStream->m_AudioPipe.m_CBRPipe.CloseOutput();
									m_AudioReadSignal.signal();
									bytesread -= bytesread % nBlockAlign;
									count = bytesread / nBlockAlign;
									mInterleaver.AddCBRCorrection(stream, count);
									mInterleaver.EndStream(stream);
									iAudioActive--;
									SetEOS(*pDubStream);
									break;
								}

								pDubStream->m_AudioPipe.m_CBRPipe.ReadWait();
							}

							bytesread += tc;
						}

						// apply audio correction on the fly if we are doing L3
						//
						// NOTE: Don't begin correction until we have at least 20 MPEG frames.  The error
						//       is generally under 5% and we don't want the beginning of the stream to go
						//       nuts.
						if (pDubStream->m_AudioCorrector && pDubStream->m_AudioCorrector->GetFrameCount() >= 20) {
							IVDMediaOutputStream *streamOut = AVIout->getOutputStream(pDubStream->m_Serial);
							if(!streamOut)
								throw MyError("Unable to retrieve output stream %lu", pDubStream->m_Serial);
							VDFormatStruct<WAVEFORMATEX> wfex((const WAVEFORMATEX *)streamOut->getFormat(), streamOut->getFormatLen());

							double bytesPerSec = pDubStream->m_AudioCorrector->ComputeByterateDouble(wfex->nSamplesPerSec);

							mInterleaver.AdjustStreamRate(stream, bytesPerSec / (double)vInfo.frameRate);
						}

						mProcessingProfileChannel.End();
						if (count > 0) {
							mProcessingProfileChannel.Begin(0xe0e0ff, "A-Write");
							// For OGM and MKV outputs, take into account the real sample count
							if(pDubStream->m_bHandleVBR || (m_OutputFormat==SOURCE_OGM) || (m_OutputFormat==SOURCE_MKV)) {
								sint64 samplesRead = pDubStream->m_AudioPipe.m_SamplesRead;
								WriteStream(*pDubStream, &audioBuffer.front(), bytesread
									, pDubStream->m_SamplesPerBlock.scale64r(samplesRead+count) - pDubStream->m_SamplesPerBlock.scale64r(samplesRead));
								pDubStream->m_AudioPipe.m_SamplesRead += count;
							} else
								WriteStream(*pDubStream, &audioBuffer.front(), bytesread, count);

							if (firstPacket && fPreview) {
								IVDMediaOutputStream *streamOut = AVIout->getOutputStream(pDubStream->m_Serial);
								if(!streamOut)
									throw MyError("Unable to retrieve output stream %lu", pDubStream->m_Serial);
								streamOut->flush();
								blitter->enablePulsing(true);
								firstPacket = false;
								mbAudioFrozen = false;
							}
							mProcessingProfileChannel.End();
						}
					} else if(pDubStream->m_bShortcutSubtitles) {
						int bytesread = 0;
						long actualBytes;
						VDPosition previousStart = pDubStream->m_aInfo.cur_proc_dst;
						VDPosition actualStart;
						VDPosition actualSamples;
						VDPosition actualRealSamples;
						int exdata;
						int droptype;
						void *srcBuffer;
						int handle;
						sint32 samplesread = 0;
						sint32 realsamplesread = 0;
					
						while(count > 0) {
							if(!(srcBuffer = pDubStream->m_AudioPipe.RetrieveBuffer(actualBytes, actualStart, actualRealSamples, handle)))
								srcBuffer = pDubStream->m_AudioPipe.m_VBRPipe.getReadBuffer(actualBytes, actualSamples, actualRealSamples, actualStart, &exdata, &droptype, &handle);
							else
								actualSamples = actualRealSamples;

							if (fAbort)
								goto abort_requested;

							if(!srcBuffer) {
								if (pDubStream->m_AudioPipe.m_CBRPipe.isInputClosed()) {
									pDubStream->m_AudioPipe.m_CBRPipe.CloseOutput();
									m_AudioReadSignal.signal();
									mInterleaver.EndStream(stream);
									iAudioActive--;
									SetEOS(*pDubStream);
									break;
								}

								pDubStream->m_AudioPipe.m_CBRPipe.ReadWait();
								continue;
							}

							if(actualSamples <= 0)
								break;
							if(actualStart >= pDubStream->m_aInfo.cur_proc_dst + count) {
								samplesread = count;	// Need to make as if we processed those samples
								pDubStream->m_AudioPipe.StoreBuffer(srcBuffer, actualBytes, actualStart, actualRealSamples, handle);
								break;
							}

							actualSamples = actualStart - previousStart;

							pDubStream->m_AudioPipe.Read(actualRealSamples);
							// apply audio correction on the fly if we are doing L3
							//
							// NOTE: Don't begin correction until we have at least 20 MPEG frames.  The error
							//       is generally under 5% and we don't want the beginning of the stream to go
							//       nuts.
							if (pDubStream->m_AudioCorrector && pDubStream->m_AudioCorrector->GetFrameCount() >= 20) {
								IVDMediaOutputStream *streamOut = AVIout->getOutputStream(pDubStream->m_Serial);
								if(!streamOut)
									throw MyError("Unable to retrieve output stream %lu", pDubStream->m_Serial);
								VDFormatStruct<WAVEFORMATEX> wfex((const WAVEFORMATEX *)streamOut->getFormat(), streamOut->getFormatLen());

								double bytesPerSec = pDubStream->m_AudioCorrector->ComputeByterateDouble(wfex->nSamplesPerSec);

								mInterleaver.AdjustStreamRate(stream, bytesPerSec / (double)vInfo.frameRate);
							}

							mProcessingProfileChannel.End();
							if(actualSamples >= 0) {
								mProcessingProfileChannel.Begin(0xe0e0ff, "S-Write");
								//count -= actualSamples;
								previousStart = actualStart;
								samplesread += actualSamples;
								realsamplesread += actualRealSamples;
								bytesread += actualBytes;

								if(!pDubStream->m_OverlappingOutput->WriteData(AVIIF_KEYFRAME, srcBuffer, actualBytes, actualStart, actualRealSamples))
									throw MyError("Error writing text data.");
								mProcessingProfileChannel.End();
							} else
								VDLog(kVDLogError, VDStringW(L"One subtitle starting time is lower than the actual dubbing time."));

							pDubStream->m_AudioPipe.m_VBRPipe.releaseBuffer(handle);
							m_AudioReadSignal.signal();
						}

						pDubStream->m_aInfo.cur_proc_dst += samplesread;
						mInterleaver.AddVBRCorrection(stream, bytesread);
						mInterleaver.AddSampleCorrection(stream, samplesread);
					} else {
						int bytesread = 0;
						long actualBytes;
						VDPosition actualSamples;
						VDPosition actualRealSamples;
						VDPosition tFrame;
						int exdata;
						int droptype;
						void *srcBuffer;
						int handle;
						sint32 samplesread = 0;
						sint32 realsamplesread = 0;

						while(count > 0) {

							srcBuffer = pDubStream->m_AudioPipe.m_VBRPipe.getReadBuffer(actualBytes, actualSamples, actualRealSamples, tFrame, &exdata, &droptype, &handle);

							if (fAbort)
								goto abort_requested;

							if(!srcBuffer) {
								if (pDubStream->m_AudioPipe.m_CBRPipe.isInputClosed()) {
									pDubStream->m_AudioPipe.m_CBRPipe.CloseOutput();
									m_AudioReadSignal.signal();
									mInterleaver.EndStream(stream);
									iAudioActive--;
									SetEOS(*pDubStream);
									break;
								}

								pDubStream->m_AudioPipe.m_CBRPipe.ReadWait();
								continue;
							}

							if(actualSamples <= 0)
								break;
							count -= actualSamples;
							samplesread += actualSamples;
							realsamplesread += actualRealSamples;
							bytesread += actualBytes;

							pDubStream->m_AudioPipe.Read(actualRealSamples);
							if(!pDubStream->m_GroupFrames) {
								mProcessingProfileChannel.End();
								mProcessingProfileChannel.Begin(0xe0e0ff, "A-Write");
								WriteStream(*pDubStream, srcBuffer, actualBytes, (pDubStream->m_bHandleVBR ? actualRealSamples : actualSamples));
								mProcessingProfileChannel.End();
							} else {
								if (audioBuffer.size() < bytesread)
									audioBuffer.resize(bytesread);

								memcpy(&audioBuffer[bytesread-actualBytes], srcBuffer, actualBytes);
							}

							pDubStream->m_AudioPipe.m_VBRPipe.releaseBuffer(handle);
							m_AudioReadSignal.signal();

							if (firstPacket && fPreview) {
								if(!pDubStream->m_GroupFrames || (count<=0) ) {
									IVDMediaOutputStream *streamOut = AVIout->getOutputStream(pDubStream->m_Serial);
									if(!streamOut)
										throw MyError("Unable to retrieve output stream %lu", pDubStream->m_Serial);
									streamOut->flush();
									blitter->enablePulsing(true);
									firstPacket = false;
									mbAudioFrozen = false;
								}
							}
						}

						if(pDubStream->m_GroupFrames) {
							mProcessingProfileChannel.End();
							mProcessingProfileChannel.Begin(0xe0e0ff, "A-Write");
							WriteStream(*pDubStream, &audioBuffer.front(), bytesread, (pDubStream->m_bHandleVBR ? realsamplesread : samplesread));
							mProcessingProfileChannel.End();
						}

						pDubStream->m_aInfo.cur_proc_dst += samplesread;
						mInterleaver.AddVBRCorrection(stream, bytesread);
						mInterleaver.AddSampleCorrection(stream, samplesread);
					}

					mProcessingProfileChannel.End();
				}/* else {
					VDNEVERHERE;
				}*/
			}

			CHECK_STACK(sp);

			if (fAbort)
				break;

			if (bVideoEnded && /*bAudioEnded*/!iAudioActive)
				break;
		}
abort_requested:
		;

	} catch(MyError& e) {
		if (!fError) {
			err.TransferFrom(e);
			fError = true;
		}
		mpVideoPipe->abort();
		fAbort = true;
	}

	fAbort = true;

	mpVideoPipe->finalizeAck();
	streamIt.begin(*m_Streams);
	while(streamIt) {
		pDubStream = *streamIt;
		pDubStream->m_AudioPipe.m_CBRPipe.CloseOutput();
		pDubStream->m_AudioPipe.m_VBRPipe.finalizeAck();
		streamIt++;
	}
	m_AudioReadSignal.signal();

	// attempt a graceful shutdown at this point...
	try {
		// if preview mode, choke the audio

		if (m_Streams->GetCount() && mpOutputSystem->IsRealTime()) {
			streamIt.begin(*m_Streams);
			while(streamIt) {
				if((*streamIt)->m_Source->getSourceMainType() == SOURCE_AUDIO) {
					IVDMediaOutputStream *streamOut = AVIout->getOutputStream((*streamIt)->m_Serial);
					if(!streamOut)
						throw MyError("Unable to retrieve output stream %lu", (*streamIt)->m_Serial);
					static_cast<AVIAudioPreviewOutputStream *>(streamOut)->stop();	
				}
				streamIt++;
			}
		}

		// finalize the output.. if it's not a preview...
		if (!mpOutputSystem->IsRealTime()) {
			// update audio rate...

			streamIt.begin(*m_Streams);
			while(streamIt) {
				pDubStream = *streamIt;
				if (pDubStream->m_AudioCorrector) {
					IVDMediaOutputStream *audioOut = AVIout->getOutputStream(pDubStream->m_Serial);
					if(!audioOut)
						throw MyError("Unable to retrieve output stream %lu", pDubStream->m_Serial);
				
					VDFormatStruct<WAVEFORMATEX> wfex((const WAVEFORMATEX *)audioOut->getFormat(), audioOut->getFormatLen());

					wfex->nAvgBytesPerSec = pDubStream->m_AudioCorrector->ComputeByterate(wfex->nSamplesPerSec);

					audioOut->setFormat(&*wfex, wfex.size());

					AVIStreamHeader_fixed hdr(audioOut->getStreamInfo());
					hdr.dwRate = wfex->nAvgBytesPerSec * hdr.dwScale;
					audioOut->updateStreamInfo(hdr);
				}
				streamIt++;
			}

			// finalize avi
			mpOutputSystem->CloseSegment(AVIout, true);
			AVIout = NULL;
			mpVideoOut = NULL;
			VDDEBUG("Dub/Processor: finalized.\n");
		}
	} catch(MyError& e) {
		if (!fError) {
			err.TransferFrom(e);
			fError = true;
		}
	}

	VDDEBUG("Dub/Processor: end\n");
}

void Dubber::SetVideoEOS(void) {
	if(!m_bVideoEOSSet) {
		m_bVideoEOSSet = true;
		AVIout->getVideoOutput()->setEndOfStream();
	}
}

void Dubber::SetEOS(ModDubStream& stream) {
	AVIout->setEndOfStream(stream.m_Serial);
	stream.m_bEndReached = true;
}
// END ****************************************************************

///////////////////////////////////////////////////////////////////

void Dubber::Abort() {
	fUserAbort = true;
	fAbort = true;
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
	ModList<ModDubStream>::iterator streamIt(*m_Streams);
	while(streamIt) {
		ModDubStream *curStream = *streamIt;
		curStream->m_AudioPipe.m_CBRPipe.Abort();
		curStream->m_AudioPipe.m_VBRPipe.abort();
		streamIt++;
	}
	mpVideoPipe->abort();
	m_AudioReadSignal.signal();
// END ****************************************************************
	PostMessage(g_hWnd, WM_USER, 0, 0);
}

bool Dubber::isRunning() {
	return !fAbort;
}

bool Dubber::isAbortedByUser() {
	return fUserAbort;
}

bool Dubber::IsPreviewing() {
	return fPreview;
}

void Dubber::RealizePalette() {
//	if (HDRAWDIB hddOutput = mOutputDisplay.GetHDD())
//		DrawDibRealize(hddOutput, hDCWindow, FALSE);
}

void Dubber::SetPriority(int index) {
	SetThreadPriority(mpIOThread->getThreadHandle(), g_iPriorities[index][0]);
	SetThreadPriority(getThreadHandle(), g_iPriorities[index][1]);
}

void Dubber::UpdateFrames() {
	mRefreshFlag = 1;

	if (!mStopLock) {
		uint32 curTime = VDGetCurrentTick();

		int iocount = mIOThreadCounter;
		int prcount = mProcessingThreadCounter;

		if (mLastIOThreadCounter != iocount) {
			mLastIOThreadCounter = iocount;
			mIOThreadFailCount = curTime;
		} else if (mLastIOThreadCounter && (curTime - mIOThreadFailCount - 10000) < 3600000) {		// 10s to 1hr
			if (mpIOThread->isThreadActive()) {
				void *eip = ThreadLocation();
				VDLogAppMessage(kVDLogWarning, kVDST_Dub, kVDM_IOThreadLivelock, 1, &eip);
			}
			mLastIOThreadCounter = 0;
		}

		if (mLastProcessingThreadCounter != prcount) {
			mLastProcessingThreadCounter = prcount;
			mProcessingThreadFailCount = curTime;
		} else if (mLastProcessingThreadCounter && (curTime - mProcessingThreadFailCount - 10000) < 3600000) {		// 10s to 1hr
			if (isThreadActive()) {
				void *eip = ThreadLocation();
				VDLogAppMessage(kVDLogWarning, kVDST_Dub, kVDM_ProcessingThreadLivelock, 1, &eip);

				HANDLE h = (HANDLE)getThreadHandle();

				CONTEXT ctx;
				ctx.ContextFlags = CONTEXT_CONTROL;
				SuspendThread(h);
				ctx.Eip = 0;
				ResumeThread(h);
			}
			mLastProcessingThreadCounter = 0;
		}
	}
}
