//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2003 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include <vd2/system/profile.h>
#include <vd2/system/VDRingBuffer.h>
#include "DubIO.h"
#include "Dub.h"
#include "DubUtils.h"
#include "VideoSource.h"
#include "Audio.h"
#include "AVIPipe.h"
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
#include "ModAudioSourceMP3.h"	// for MPA_MAX_LENGTH

using namespace nsVDDub;

VDDubIOThread::VDDubIOThread(
		bool				bPhantom,
		IVDVideoSource		*pVideo,
		VDRenderFrameIterator& videoFrameIterator,
		ModList<ModDubStream>& streams,
		AVIPipe				*const pVideoPipe,
		volatile bool&		bAbort,
		DubVideoStreamInfo& _vInfo,
		VDAtomicInt&		threadCounter,
		VDSignal&			audioReadSignal,
		bool				bHandleVBR
							 )
	: VDThread("Dub-I/O")
	, mbError(false)
	, mbPhantom(bPhantom)
	, mpVideo(pVideo)
	, mpVideoStream(pVideo ? pVideo->asStream() : NULL)
	, mVideoFrameIterator(videoFrameIterator)
	, m_Streams(streams)
	, mpVideoPipe(pVideoPipe)
	, mbAbort(bAbort)
	, vInfo(_vInfo)
	, mThreadCounter(threadCounter)
	, m_AudioReadSignal(audioReadSignal)
	, m_bHandleVBR(bHandleVBR)
{
}
// END ****************************************************************

VDDubIOThread::~VDDubIOThread() {
}

void VDDubIOThread::ThreadRun() {

	///////////

	VDDEBUG("Dub/Main: Start.\n");

	VDRTProfileChannel profchan("I/O");		// sensei?

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
	ModList<ModDubStream>::iterator streamIt;
	ModDubStream *curStream = NULL;

	bool bVideoActive = (mpVideo != 0);
	int iAudioActive = 0;
	streamIt.begin(m_Streams);
	while(streamIt) {
		curStream = *streamIt;
		if(curStream->m_AudioStream != 0)
			iAudioActive++;
		streamIt++;
	}

	double nVideoRate = bVideoActive ? vInfo.frameRateNoTelecine : 0;

	try {
		try {
			while(!mbAbort && (iAudioActive || bVideoActive)) { 
				bool bBlocked = true;

				++mThreadCounter;

				bool bCanWriteVideo = bVideoActive && !mpVideoPipe->full();
				bool bCanWriteAudio = false;
				if(iAudioActive > 0) {
					streamIt.begin(m_Streams);
					while(streamIt) {
						curStream = *streamIt;
						if(!curStream->m_bEndReached && !curStream->m_AudioPipe.full()) {
							bCanWriteAudio = true;
							break;
						}
						streamIt++;
					}
				}

				if (bCanWriteVideo && bCanWriteAudio) {
					int nVideoTotal, nVideoFinalQueued;
					mpVideoPipe->getQueueInfo(nVideoTotal, nVideoFinalQueued);
					streamIt.begin(m_Streams);
					while(streamIt) {
						curStream = *streamIt;
						if(!curStream->m_bEndReached && !curStream->m_AudioPipe.full()) {
							double nAudioRate = curStream->m_Source->getWaveFormat()->nAvgBytesPerSec;
							const int nAudioLevel = curStream->m_AudioPipe.getLevel();
							if (nAudioLevel * nVideoRate < nVideoFinalQueued * nAudioRate) {
								bCanWriteVideo = false;
								break;
							}
						}
						streamIt++;
					}
				}

				if (bCanWriteVideo) {
					bBlocked = false;

					profchan.Begin(0xffe0e0, "Video");

					if (!MainAddVideoFrame() && vInfo.cur_dst >= vInfo.end_dst) {
						bVideoActive = false;
						mpVideoPipe->finalize();
					}

					profchan.End();
					continue;
				}

				if (bCanWriteAudio) {
					streamIt.begin(m_Streams);
					while(streamIt) {
						curStream = *streamIt;
						if (!curStream->m_bEndReached && !curStream->m_AudioPipe.full()) {
							bBlocked = false;

#pragma message(__TODO__ "Use another name ?")
							profchan.Begin(0xe0e0ff, "Audio");

							if (curStream->m_bHandleVBR ? !MainAddStreamFrameVBR(*curStream) : !MainAddStreamFrame(*curStream)) {
								if(curStream->m_bShortcutSubtitles)
									curStream->m_bEndReached = curStream->m_SubtitleProvider.IsEnd();
								else
									curStream->m_bEndReached = curStream->m_AudioStream->isEnd();
								if (curStream->m_bEndReached) {
									iAudioActive--;
									curStream->EndReached();
								}
							}

							profchan.End();
						}
						streamIt++;
					}
					continue;
				}

				if (bBlocked) {
					if (iAudioActive) {
						bool bContinue = false;
						streamIt.begin(m_Streams);
						while(streamIt) {
							curStream = *streamIt;
							if (!curStream->m_bEndReached) {
								bool bEndReached = false;
								if(curStream->m_AudioPipe.m_bCBRPipe)
									bEndReached = curStream->m_AudioPipe.m_CBRPipe.isOutputClosed();
								else
									bEndReached = curStream->m_AudioPipe.m_VBRPipe.isFinalizeAcked();
								if(bEndReached) {
									iAudioActive--;
									curStream->EndReached();
									bContinue = true;
									break;
								}
							}
							streamIt++;
						}
						if (bContinue)
							continue;
					}

					if (bVideoActive && mpVideoPipe->isFinalizeAcked()) {
						bVideoActive = false;
						continue;
					}

					profchan.Begin(0xe0e0e0, "Idle");
					if (iAudioActive) {
						if (bVideoActive)
							mpVideoPipe->getReadSignal().wait(&m_AudioReadSignal);
						else
							m_AudioReadSignal.wait();
					} else
						mpVideoPipe->getReadSignal().wait();

					profchan.End();
				}
			}
// END ****************************************************************
		} catch(MyError& e) {
			if (!mbError) {
				mError.TransferFrom(e);
				mbError = true;
			}
//			e.post(NULL, "Dub Error (will attempt to finalize)");
		}

		// wait for the pipelines to clear...

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
		if (!mbAbort) {
			// must finalize all pipes before waiting on any of them
			streamIt.begin(m_Streams);
			while(streamIt) {
				ModDubStream *curStream = *streamIt;
				curStream->m_AudioPipe.m_CBRPipe.CloseInput();
				// Finalize the VBR pipes now (without waiting)
				// This make sure the dubbing thread won't be waiting
				// for us to write data in the pipe, thus live locking
				// in the video finalizeAndWait (which waits for the
				// dubbing thread to read data from the pipe)
				curStream->m_AudioPipe.m_VBRPipe.finalize();
				streamIt++;
			}
			if (mpVideo)
				mpVideoPipe->finalizeAndWait();
			streamIt.begin(m_Streams);
			while(streamIt) {
				ModDubStream *curStream = *streamIt;
				curStream->m_AudioPipe.m_CBRPipe.WaitUntilOutputClosed();
				curStream->m_AudioPipe.m_VBRPipe.finalizeAndWait();
				streamIt++;
			}
		}

		// kill everyone else...

		//mbAbort = true;
// END ****************************************************************

	} catch(MyError& e) {
//		e.post(NULL,"Dub Error");

		if (!mbError) {
			mError.TransferFrom(e);
			mbError = true;
		}
		mbAbort = true;
	}

	// All done, time to get the pooper-scooper and clean up...

	VDDEBUG("Dub/Main: End.\n");
}

void VDDubIOThread::ReadVideoFrame(VDPosition stream_frame, VDPosition display_frame, VDPosition timeline_frame, bool preload) {
	LONG lActualBytes;
	int hr;

	void *buffer;
	int handle;

	LONG lSize;

	if (mbPhantom) {
		buffer = mpVideoPipe->getWriteBuffer(0, &handle);
		if (!buffer) return;	// hmm, aborted...

		bool bIsKey = !!mpVideo->isKey(display_frame);
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
		bool bIsB = !!mpVideo->isB(display_frame);

		mpVideoPipe->postBuffer(0, stream_frame, display_frame, timeline_frame,
			(bIsKey ? 0 : bIsB ? kBufferTypeB : kBufferFlagDelta)
			+(preload ? kBufferFlagPreload : 0),
			0,
			handle,
			!preload);
// END ****************************************************************

		return;
	}

//	VDDEBUG("Reading frame %ld (%s)\n", lVStreamPos, preload ? "preload" : "process");

	hr = mpVideoStream->read(stream_frame, 1, NULL, 0x7FFFFFFF, &lSize, NULL);
	if (hr) {
		if (hr == AVIERR_FILEREAD)
			throw MyError("Video frame %d could not be read from the source. The file may be corrupt.", stream_frame);
		else
			throw MyAVIError("Dub/IO-Video", hr);
	}

	// Add 4 bytes -- otherwise, we can get crashes with uncompressed video because
	// the bitmap routines expect to be able to read 4 bytes out.

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2004											***
// BEGIN **************************************************************
	// Add 16 bytes for the ASV2 codec (see below)
	/* Note from Michael Koester (mightyknife at sourceforge) :
	 * The Asus driver seems to use MMX copy commands to later copy the frame
	 * data to the memory that is allocated here. It uses a command which
	 * copies 16 byte of data at once. The driver sometimes "forgets the
	 * last MMX command" which results in the buffer being 12 byte too
	 * short.
	 */
	buffer = mpVideoPipe->getWriteBuffer(lSize+16, &handle);
// END ****************************************************************
	if (!buffer) return;	// hmm, aborted...

	hr = mpVideoStream->read(stream_frame, 1, buffer, lSize,	&lActualBytes,NULL); 
	if (hr) {
		if (hr == AVIERR_FILEREAD)
			throw MyError("Video frame %d could not be read from the source. The file may be corrupt.", stream_frame);
		else
			throw MyAVIError("Dub/IO-Video", hr);
	}

	display_frame = mpVideo->streamToDisplayOrder(stream_frame);

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
	mpVideoPipe->postBuffer(lActualBytes, stream_frame, display_frame, timeline_frame,
		(mpVideo->isKey(display_frame) ? 0 : mpVideo->isB(display_frame) ? kBufferTypeB : kBufferFlagDelta)
		+(preload ? kBufferFlagPreload : 0),
		mpVideo->getDropType(display_frame),
		handle,
		!preload);


}

void VDDubIOThread::ReadNullVideoFrame(VDPosition displayFrame, VDPosition timelineFrame) {
	void *buffer;
	int handle;

	buffer = mpVideoPipe->getWriteBuffer(1, &handle);
	if (!buffer) return;	// hmm, aborted...

	if (displayFrame >= 0) {
		VDPosition streamFrame = mpVideo->displayToStreamOrder((long)displayFrame);
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
		bool bIsKey = mpVideo->isKey((long)displayFrame);
		bool bIsB = mpVideo->isB((long)displayFrame);

		mpVideoPipe->postBuffer(0, streamFrame, displayFrame, timelineFrame,
			(bIsKey ? 0 : bIsB ? kBufferTypeB : kBufferFlagDelta),
			mpVideo->getDropType((long)displayFrame),
			handle,
			true);
	} else {
		mpVideoPipe->postBuffer(0, displayFrame, displayFrame, timelineFrame,
			kBufferFlagDelta,
			AVIPipe::kDroppable,
			handle,
			true);
	}
}

//////////////////////

bool VDDubIOThread::MainAddVideoFrame() {
	if (vInfo.cur_dst >= vInfo.end_dst)
		return false;

	VDPosition streamFrame, displayFrame, timelineFrame;
	bool bIsPreroll;
	
	mVideoFrameIterator.Next(streamFrame, displayFrame, timelineFrame, bIsPreroll);

	vInfo.cur_src = vInfo.start_src + timelineFrame;

	if (streamFrame<0)
		ReadNullVideoFrame(displayFrame, timelineFrame);
	else
		ReadVideoFrame(streamFrame, displayFrame, timelineFrame, bIsPreroll);

	if (!bIsPreroll)
		++vInfo.cur_dst;

	return true;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, June 2003											***
// BEGIN **************************************************************
bool VDDubIOThread::MainAddStreamFrame(ModDubStream& stream) {
	LONG lActualBytes=0;
	LONG lActualSamples=0;

	const int blocksize = stream.m_AudioStream->GetFormat()->nBlockAlign;
	int samples = stream.m_AudioPipe.getSpace();
	int handle;
	LONG ltActualBytes, ltActualSamples;
	int tc, len;

	if (stream.m_bIsNandubMP3VBR) {
		// Handle MP3 VBR ala Nandub here
		VDASSERT(!stream.m_AudioPipe.m_bCBRPipe);

		// We already filled our buffers ?
		if(samples < 0)
			return false;
		while(samples > 0) {
			// samples = number of mpa frames we want
			// This one is just in case some frames are really big (OGM / MKV
			// , or corrupted MP3 VBR stream muxed with Nandub or old versions of vdubmod)
			len = stream.m_Source->getStreamInfo().dwSuggestedBufferSize;
			if(len < MPA_MAX_LENGTH)
				len = MPA_MAX_LENGTH;

			if(stream.m_AudioPipe.full())
				break;
			void *destBuffer = stream.m_AudioPipe.BeginWrite(len, tc, handle);

			if (!destBuffer)
				return false;

			if (mbAbort)
				break;

			ltActualSamples = stream.m_AudioStream->Read(destBuffer, 1, &ltActualBytes);

			if (ltActualSamples <= 0)
				break;

			stream.m_AudioPipe.EndWrite(ltActualBytes, ltActualSamples, ltActualSamples, handle);

			stream.m_aInfo.total_size += ltActualBytes;
			stream.m_aInfo.cur_src += ltActualSamples;

			lActualBytes += ltActualBytes;
			lActualSamples += ltActualSamples;

			samples -= ltActualSamples;
		}
	} else {
		// If all went well we should use the new pipe introduced since 1.5.2
		// and working great ... but only for CBR stream :p
		VDASSERT(stream.m_AudioPipe.m_bCBRPipe);

		while(samples > 0) {
			len = samples * blocksize;

			void *dst;
		
			dst = stream.m_AudioPipe.BeginWrite(len, tc, handle);

			if (!tc)
				break;

			if (mbAbort)
				break;

			ltActualSamples = stream.m_AudioStream->Read(dst, tc / blocksize, &ltActualBytes);

			if (ltActualSamples <= 0)
				break;

			int residue = ltActualBytes % blocksize;

			if (residue) {
				VDASSERT(false);	// This is bad -- it means the input file has partial samples.

				ltActualBytes += blocksize - residue;
			}

			stream.m_AudioPipe.EndWrite(ltActualBytes, ltActualSamples, ltActualSamples, handle);

			stream.m_aInfo.total_size += ltActualBytes;
			stream.m_aInfo.cur_src += ltActualSamples;

			lActualBytes += ltActualBytes;
			lActualSamples += ltActualSamples;

			samples -= ltActualSamples;
		}
	}

	return lActualSamples > 0;
}

bool VDDubIOThread::MainAddStreamFrameVBR(ModDubStream& stream) {
	LONG lActualBytes=0;
	LONG lActualSamples=0;

	const int blocksize = stream.m_AudioStream->GetFormat()->nBlockAlign;
	int samples = stream.m_AudioPipe.getSpace();
	int handle;
	LONG ltActualBytes, ltActualSamples;
	int tc, len;

#pragma message(__TODO__ "Is m_SamplesPerBlock valid when using a compressor ?!?")
	if(stream.m_bShortcutSubtitles) {
		VDASSERT(!stream.m_AudioPipe.m_bCBRPipe);

		while(samples > 0) {
			if(stream.m_AudioPipe.full())
				break;
			VDStringW subtitle;
			sint32 startMS, durationMS;
			if(!stream.m_SubtitleProvider.GetNextSubtitle(subtitle, startMS, durationMS)
				|| (durationMS<=0))
				break;
			VDStringA subUTF8 = VDTextWToU8(subtitle);

			len = subUTF8.length() + 1;
			void *destBuffer = stream.m_AudioPipe.BeginWrite(len, tc, handle);

			if (!destBuffer)
				return false;

			if (mbAbort)
				break;

			memcpy(destBuffer, subUTF8.c_str(), len);

			stream.m_AudioPipe.EndWrite(len
				, durationMS
				, durationMS
				, handle
				, startMS);

			stream.m_aInfo.total_size += len;
			stream.m_aInfo.cur_src = startMS + durationMS;

			lActualBytes += len;
			lActualSamples = startMS + durationMS;

			samples -= durationMS;
		}
	} else if (stream.m_bIsNandubMP3VBR) {
		// Handle MP3 VBR ala Nandub here
		VDASSERT(!stream.m_AudioPipe.m_bCBRPipe);

		while(samples > 0) {
			// samples = number of mpa frames we want
			// This one is just in case some frames are really big (OGM / MKV
			// , or corrupted MP3 VBR stream muxed with Nandub or old versions of vdubmod)
			len = stream.m_Source->getStreamInfo().dwSuggestedBufferSize;
			if(len < MPA_MAX_LENGTH)
				len = MPA_MAX_LENGTH;
		
			if(stream.m_AudioPipe.full())
				break;
			void *destBuffer = stream.m_AudioPipe.BeginWrite(len, tc, handle);

			if (!destBuffer)
				return false;

			if (mbAbort)
				break;

			ltActualSamples = stream.m_AudioStream->Read(destBuffer, 1, &ltActualBytes, false);

			if (ltActualSamples <= 0)
				break;

			stream.m_AudioPipe.EndWrite(ltActualBytes
				, ltActualSamples
				, stream.m_SamplesPerBlock.scale64r(stream.m_aInfo.cur_src + ltActualSamples) - stream.m_SamplesPerBlock.scale64r(stream.m_aInfo.cur_src)
				, handle);

			stream.m_aInfo.total_size += ltActualBytes;
			stream.m_aInfo.cur_src += ltActualSamples;

			lActualBytes += ltActualBytes;
			lActualSamples += ltActualSamples;

			samples -= ltActualSamples;
		}
	} else if(!stream.m_AudioPipe.m_bCBRPipe) {
		// Here we receive audio data frame by frame
		while(samples > 0) {
			// Prepare a bigger buffer in case of _real_ VBR stream
			// This one is just in case some frames are really big (OGM / MKV)
			len = (samples * 3 * stream.m_AudioStream->GetFormat()->nAvgBytesPerSec) / stream.m_AudioStream->GetFormat()->nSamplesPerSec;
			if(len < stream.m_Source->getStreamInfo().dwSuggestedBufferSize)
				len = stream.m_Source->getStreamInfo().dwSuggestedBufferSize;
			
			if(stream.m_AudioPipe.full())
				break;
			void *destBuffer = stream.m_AudioPipe.BeginWrite(len, tc, handle);

			if (!destBuffer)
				return false;

			if (mbAbort)
				break;

			ltActualSamples = stream.m_AudioStream->Read(destBuffer, 1, &ltActualBytes, false);

			if (ltActualSamples <= 0)
				break;

			if(stream.m_bText) {
				if(stream.m_aInfo.cur_src + ltActualSamples > stream.m_aInfo.end_src)
					ltActualSamples = stream.m_aInfo.end_src - stream.m_aInfo.cur_src;
			}

			stream.m_AudioPipe.EndWrite(ltActualBytes
				, ltActualSamples
				, stream.m_SamplesPerBlock.scale64r(stream.m_aInfo.cur_src + ltActualSamples) - stream.m_SamplesPerBlock.scale64r(stream.m_aInfo.cur_src)
				, handle);

			stream.m_aInfo.total_size += ltActualBytes;
			stream.m_aInfo.cur_src += ltActualSamples;

			lActualBytes += ltActualBytes;
			lActualSamples += ltActualSamples;

			samples -= ltActualSamples;
		}
	} else {
		while(samples > 0) {
			len = samples * blocksize;

			void *dst;
		
			dst = stream.m_AudioPipe.BeginWrite(len, tc, handle);

			if (!tc)
				break;

			if (mbAbort)
				break;

			ltActualSamples = stream.m_AudioStream->Read(dst, tc / blocksize, &ltActualBytes);

			if (ltActualSamples <= 0)
				break;

			int residue = ltActualBytes % blocksize;

			if (residue) {
				VDASSERT(false);	// This is bad -- it means the input file has partial samples.

				ltActualBytes += blocksize - residue;
			}

			stream.m_AudioPipe.EndWrite(ltActualBytes, ltActualSamples, ltActualSamples, handle);

			stream.m_aInfo.total_size += ltActualBytes;
			stream.m_aInfo.cur_src += ltActualSamples;

			lActualBytes += ltActualBytes;
			lActualSamples += ltActualSamples;

			samples -= ltActualSamples;
		}
	}

	return lActualSamples > 0;
}
// END ****************************************************************

