//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2003 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

#include "stdafx.h"
#include "DubOutput.h"
#include <vd2/system/error.h>
#include "AVIOutput.h"
#include "AVIOutputWAV.h"
#include "AVIOutputImages.h"
#include "AVIOutputStriped.h"
#include "AVIOutputPreview.h"

///////////////////////////////////////////////////////////////////////////
//
//	VDAVIOutputFileSystem
//
///////////////////////////////////////////////////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
VDAVIOutputFileSystem::VDAVIOutputFileSystem()
	: mbAllowCaching(false)
	, mbAllowIndexing(true)
	, mbUse1GBLimit(false)
	, mCurrentSegment(0)
{
	m_VideoStream = NULL;
	m_bLastSegment = false;
}

VDAVIOutputFileSystem::~VDAVIOutputFileSystem() {
	m_Streams.DeleteContents(true);
	m_Streams.Clear();
// END ****************************************************************
}

void VDAVIOutputFileSystem::SetCaching(bool bAllowOSCaching) {
	mbAllowCaching = bAllowOSCaching;
}

void VDAVIOutputFileSystem::SetIndexing(bool bAllowHierarchicalExtensions) {
	mbAllowIndexing = bAllowHierarchicalExtensions;
}

void VDAVIOutputFileSystem::Set1GBLimit(bool bUse1GBLimit) {
	mbUse1GBLimit = bUse1GBLimit;
}

void VDAVIOutputFileSystem::SetBuffer(int bufferSize) {
	mBufferSize = bufferSize;
}

AVIOutput *VDAVIOutputFileSystem::CreateSegment() {
	vdautoptr<AVIOutputFile> pOutput(new AVIOutputFile);

	if (!mbAllowCaching)
		pOutput->disable_os_caching();

	if (!mbAllowIndexing)
		pOutput->disable_extended_avi();

	VDStringW s(mSegmentBaseName);

	if (mSegmentDigits) {
		s += VDFastTextPrintfW(L".%0*d", mSegmentDigits, mCurrentSegment++);
		VDFastTextFree();
		s += mSegmentExt;

		pOutput->setSegmentHintBlock(true, NULL, 1);
	}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	ModList<ModDubStream> tmpList;
	tmpList.DeleteContents(false);
	ModList<ModAVIDubStream>::iterator it(m_Streams);
	while(it) {
		tmpList.Append((*it)->m_Stream);
		it++;
	}
	pOutput->initOutputStreams(*m_VideoStream, tmpList);

	if (!mVideoFormat.empty()) {
		IVDMediaOutputStream *pVideoOut = pOutput->getVideoOutput();
		pVideoOut->setFormat(&mVideoFormat[0], mVideoFormat.size());
		pVideoOut->setStreamInfo(mVideoStreamInfo);
	}

	it.begin(m_Streams);
	ModAVIDubStream *curStream = NULL;
	while(it) {
		curStream = *it;
		IVDMediaOutputStream *streamOut = pOutput->getOutputStream(curStream->m_Stream->m_Serial);
		if(!streamOut)
			throw MyError("Unable to retrieve output stream %lu", curStream->m_Stream->m_Serial);
		if (!curStream->m_Format.empty()) {
			streamOut->setFormat(&curStream->m_Format[0], curStream->m_Format.size());
			streamOut->setStreamInfo(curStream->m_StreamInfo);
		}
		it++;
	}

	pOutput->setBuffering(mBufferSize, mBufferSize >> 2);
	pOutput->setInterleaved(mbInterleaved);

	pOutput->init(s.c_str());

	return pOutput.release();
}

void VDAVIOutputFileSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	AVIOutputFile *pFile = (AVIOutputFile *)pSegment;
	if (mSegmentDigits)
		pFile->setSegmentHintBlock(bLast & m_bLastSegment, NULL, 1);
	pFile->finalize();
	delete pFile;
}
// END ****************************************************************

void VDAVIOutputFileSystem::SetFilename(const wchar_t *pszFilename) {
	mSegmentBaseName	= pszFilename;
	mSegmentDigits		= 0;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, August 2003											***
// BEGIN **************************************************************
void VDAVIOutputFileSystem::SetFilenamePattern(const wchar_t *pszFilename, const wchar_t *pszExt, int nMinimumDigits, int nCurrentSegment, bool bLastSegment) {
	mSegmentBaseName	= pszFilename;
	mSegmentExt			= pszExt;
	mSegmentDigits		= nMinimumDigits;
	mCurrentSegment		= nCurrentSegment;
	m_bLastSegment		= bLastSegment;
}
// END ****************************************************************

void VDAVIOutputFileSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void VDAVIOutputFileSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
	mbInterleaved = bInterleaved;
}

bool VDAVIOutputFileSystem::AcceptsVideo() {
	return true;
}

bool VDAVIOutputFileSystem::AcceptsAudio() {
	return true;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void VDAVIOutputFileSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	m_VideoStream = &stream;
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
	CopyModList<ModStreamComment>(stream.m_Comments, m_Comments);
}

void VDAVIOutputFileSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	ModAVIDubStream *newStream = m_Streams.AddLast();
	newStream->m_Stream = &stream;
	newStream->m_StreamInfo = asi;
	newStream->m_Format.resize(cbFormat);
	memcpy(&newStream->m_Format[0], pFormat, cbFormat);
	if(bInterleaved)
		mbInterleaved = true;
}

bool VDAVIOutputFileSystem::AcceptsText() {
	return false;
}
// END ****************************************************************

///////////////////////////////////////////////////////////////////////////
//
//	VDAVIOutputStripedSystem
//
///////////////////////////////////////////////////////////////////////////

VDAVIOutputStripedSystem::VDAVIOutputStripedSystem(const wchar_t *filename)
	: mbUse1GBLimit(false)
	, mpStripeSystem(new AVIStripeSystem(VDTextWToA(filename, -1).c_str()))
{
}

VDAVIOutputStripedSystem::~VDAVIOutputStripedSystem() {
}

void VDAVIOutputStripedSystem::Set1GBLimit(bool bUse1GBLimit) {
	mbUse1GBLimit = bUse1GBLimit;
}

AVIOutput *VDAVIOutputStripedSystem::CreateSegment() {
	vdautoptr<AVIOutputStriped> pFile(new AVIOutputStriped(mpStripeSystem));

	if (!pFile)
		throw MyMemoryError();

	pFile->set_1Gb_limit();
	return pFile.release();
}

void VDAVIOutputStripedSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	AVIOutputFile *pFile = (AVIOutputFile *)pSegment;
	pFile->setSegmentHintBlock(bLast, NULL, 1);
	pFile->finalize();
	delete pFile;
}

void VDAVIOutputStripedSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void VDAVIOutputStripedSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool VDAVIOutputStripedSystem::AcceptsVideo() {
	return true;
}

bool VDAVIOutputStripedSystem::AcceptsAudio() {
	return true;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void VDAVIOutputStripedSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
	//CopyModList<ModStreamComment>(stream.m_Comments, m_Comments);
}

void VDAVIOutputStripedSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	/*ModAVIDubStream *newStream = m_Streams.AddLast();
	newStream->m_Stream = &stream;
	newStream->m_StreamInfo = asi;
	newStream->m_Format.resize(cbFormat);
	memcpy(&newStream->m_Format[0], pFormat, cbFormat);*/
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool VDAVIOutputStripedSystem::AcceptsText() {
	return false;
}
// END ****************************************************************

///////////////////////////////////////////////////////////////////////////
//
//	VDAVIOutputWAVSystem
//
///////////////////////////////////////////////////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
VDAVIOutputWAVSystem::VDAVIOutputWAVSystem(const wchar_t *pszFilename, bool bWriteWAVHeader)
	: mFilename(pszFilename)
{
	m_Stream = NULL;
	m_bWriteWAVHeader = bWriteWAVHeader;
}
// END ****************************************************************

VDAVIOutputWAVSystem::~VDAVIOutputWAVSystem() {
}

AVIOutput *VDAVIOutputWAVSystem::CreateSegment() {
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	vdautoptr<AVIOutputWAV> pOutput(new AVIOutputWAV(m_bWriteWAVHeader));

	// In case someone is accessing directly the good old VDub methods
	// Anyway if something is wrong an error will be thrown ...
	if(m_Stream) {
		ModList<ModDubStream> tmpList;
		tmpList.DeleteContents(false);
		tmpList.Append(m_Stream);
		ModVideoStream voidVideo;
		pOutput->initOutputStreams(voidVideo, tmpList);
		pOutput->getAudioOutput()->setFormat(&mAudioFormat[0], mAudioFormat.size());
	} else
		pOutput->createAudioStream()->setFormat(&mAudioFormat[0], mAudioFormat.size());
// END ****************************************************************

	pOutput->setBufferSize(65536);

	pOutput->init(mFilename.c_str());

	return pOutput.release();
}

void VDAVIOutputWAVSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	AVIOutputWAV *pFile = (AVIOutputWAV *)pSegment;
	pFile->finalize();
	delete pFile;
}

void VDAVIOutputWAVSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
}

void VDAVIOutputWAVSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool VDAVIOutputWAVSystem::AcceptsVideo() {
	return false;
}

bool VDAVIOutputWAVSystem::AcceptsAudio() {
	return true;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void VDAVIOutputWAVSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
}

void VDAVIOutputWAVSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	if(m_Stream)
		throw MyError("Attempt to output more than one stream when saving to WAV");
	m_Stream = &stream;
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool VDAVIOutputWAVSystem::AcceptsText() {
	return false;
}
// END ****************************************************************

///////////////////////////////////////////////////////////////////////////
//
//	VDAVIOutputImagesSystem
//
///////////////////////////////////////////////////////////////////////////

VDAVIOutputImagesSystem::VDAVIOutputImagesSystem()
{
}

VDAVIOutputImagesSystem::~VDAVIOutputImagesSystem() {
}

AVIOutput *VDAVIOutputImagesSystem::CreateSegment() {
	vdautoptr<AVIOutputImages> pOutput(new AVIOutputImages(mSegmentPrefix.c_str(), mSegmentSuffix.c_str(), mSegmentDigits, mFormat));

	if (!mVideoFormat.empty())
		pOutput->createVideoStream()->setFormat(&mVideoFormat[0], mVideoFormat.size());

	if (!mAudioFormat.empty())
		pOutput->createAudioStream()->setFormat(&mAudioFormat[0], mAudioFormat.size());

	pOutput->init(NULL);

	return pOutput.release();
}

void VDAVIOutputImagesSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	AVIOutputImages *pFile = (AVIOutputImages *)pSegment;
	pFile->finalize();
	delete pFile;
}

void VDAVIOutputImagesSystem::SetFilenamePattern(const wchar_t *pszPrefix, const wchar_t *pszSuffix, int nMinimumDigits) {
	mSegmentPrefix		= pszPrefix;
	mSegmentSuffix		= pszSuffix;
	mSegmentDigits		= nMinimumDigits;
}

void VDAVIOutputImagesSystem::SetFormat(int format) {
	mFormat = format;
}

void VDAVIOutputImagesSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void VDAVIOutputImagesSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool VDAVIOutputImagesSystem::AcceptsVideo() {
	return true;
}

bool VDAVIOutputImagesSystem::AcceptsAudio() {
	return false;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void VDAVIOutputImagesSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void VDAVIOutputImagesSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
}

bool VDAVIOutputImagesSystem::AcceptsText() {
	return false;
}
// END ****************************************************************

///////////////////////////////////////////////////////////////////////////
//
//	VDAVIOutputPreviewSystem
//
///////////////////////////////////////////////////////////////////////////

VDAVIOutputPreviewSystem::VDAVIOutputPreviewSystem()
{
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	m_Stream = NULL;
	m_VideoStream = NULL;
// END ****************************************************************
}

VDAVIOutputPreviewSystem::~VDAVIOutputPreviewSystem() {
}

AVIOutput *VDAVIOutputPreviewSystem::CreateSegment() {
	vdautoptr<AVIOutputPreview> pOutput(new AVIOutputPreview);

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	if(m_Stream) {
		ModList<ModDubStream> tmpList;
		tmpList.DeleteContents(false);
		tmpList.Append(m_Stream);
		pOutput->initOutputStreams(*m_VideoStream, tmpList);
		pOutput->getAudioOutput()->setFormat(&mAudioFormat[0], mAudioFormat.size());
	} else if (!mAudioFormat.empty())
		pOutput->createAudioStream()->setFormat(&mAudioFormat[0], mAudioFormat.size());

	if (!mVideoFormat.empty()) {
		if(!pOutput->getVideoOutput())
			pOutput->createVideoStream()->setFormat(&mVideoFormat[0], mVideoFormat.size());
		else
			pOutput->getVideoOutput()->setFormat(&mVideoFormat[0], mVideoFormat.size());
	}
// END ****************************************************************

	pOutput->init(NULL);

	return pOutput.release();
}

void VDAVIOutputPreviewSystem::CloseSegment(AVIOutput *pSegment, bool bLast) {
	AVIOutputPreview *pFile = (AVIOutputPreview *)pSegment;
	pFile->finalize();
	delete pFile;
}

void VDAVIOutputPreviewSystem::SetVideo(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
}

void VDAVIOutputPreviewSystem::SetAudio(const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool VDAVIOutputPreviewSystem::AcceptsVideo() {
	return true;
}

bool VDAVIOutputPreviewSystem::AcceptsAudio() {
	return true;
}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void VDAVIOutputPreviewSystem::SetVideo(ModVideoStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat) {
	m_VideoStream = &stream;
	mVideoStreamInfo = asi;
	mVideoFormat.resize(cbFormat);
	memcpy(&mVideoFormat[0], pFormat, cbFormat);
	//CopyModList<ModStreamComment>(stream.m_Comments, m_Comments);
}

void VDAVIOutputPreviewSystem::AddStream(ModDubStream& stream, const AVIStreamHeader_fixed& asi, const void *pFormat, int cbFormat, bool bInterleaved) {
	/*ModAVIDubStream *newStream = m_Streams.AddLast();
	newStream->m_Stream = &stream;
	newStream->m_StreamInfo = asi;
	newStream->m_Format.resize(cbFormat);
	memcpy(&newStream->m_Format[0], pFormat, cbFormat);*/
	if(m_Stream)
		throw MyError("Attempt to output more than one stream when previewing");
	m_Stream = &stream;
	mAudioStreamInfo = asi;
	mAudioFormat.resize(cbFormat);
	memcpy(&mAudioFormat[0], pFormat, cbFormat);
}

bool VDAVIOutputPreviewSystem::AcceptsText() {
	return false;	// not for the moment
}
// END ****************************************************************
