//	VirtualDub - Video processing and capture application
//	Copyright (C) 1998-2003 Avery Lee
//
//	This program is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//
//	This program is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//
//	You should have received a copy of the GNU General Public License
//	along with this program; if not, write to the Free Software
//	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// ************************************************************************
// *** VirtualDubMod													***
// *** Tobias Minich, Feb 2003											***
// BEGIN ******************************************************************
// Added to disable std warnings.
#pragma warning(disable : 4786)
// END ********************************************************************

#include "stdafx.h"

#include <windows.h>

#include "resource.h"
#include "prefs.h"
#include "oshelper.h"
#include "gui.h"
#include <vd2/system/cpuaccel.h>
#include <vd2/system/memory.h>
#include <vd2/system/filesys.h>
#include <vd2/system/registry.h>
#include <vd2/system/vdalloc.h>
#include <vd2/Dita/services.h>
#include "Dub.h"
#include "DubOutput.h"
#include "command.h"
#include "job.h"
#include "project.h"
#include "projectui.h"
#include "crash.h"

#include "PositionControl.h"
#include "InputFile.h"
#include <vd2/system/error.h>
// ********************************************************************
// *** VirtualDubMod												***
// *** Oct 2003														***
// BEGIN **************************************************************
#include "ModStreamsSetup.h"
#include "corona.h"
#include "AVIOutputImages.h"
#include "ogg&co/OggDS.h"
#include <vorbis/codec.h>
#include "ebml/EbmlVersion.h"
#include "matroska/KaxVersion.h"

#include "avs/AVSViewer.h"
#include "avs/cmdline.h"
#include "avs/genscript.h"
#include "avs/template.h"
#include "avs/CAviSynth.h"
#include <cderr.h>
#include "log_handling.h"
#include "Tdll.h"

#include "virtualdubmod_messages.h"


HACCEL		g_hAccelAVS;
extern ATOM g_classAVS;
StringList	gFileList;
extern TemplateInfoList gTemplateList;
extern Tdll *g_dllScintilla;
extern CAviSynth *g_dllAviSynth;

extern char g_szPerformanceOptions[];

LPNNSTATS	g_pNNStats = NULL;
int			g_nNNStats = 0;
int			g_nTickMode = 1;

wchar_t g_szInputAVIFileLastWorking[MAX_PATH] = L"";


// END ********************************************************************

enum {
	kFileDialog_Config			= 'conf',
	kFileDialog_ImageDst		= 'imgd',
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
	kFileDialog_SaveSnapshot	= 'snap'
};
HACCEL		g_hAccelAudio;
// END ****************************************************************

///////////////////////////////////////////////////////////////////////////

HINSTANCE	g_hInst;
HWND		g_hWnd =NULL;
HACCEL		g_hAccelMain;

bool				g_fDropFrames			= false;
bool				g_fSwapPanes			= false;
bool				g_bExit					= false;

vdautoptr<VDProjectUI> g_projectui;
VDProject *g_project;

wchar_t g_szInputAVIFile[MAX_PATH];
wchar_t g_szInputWAVFile[MAX_PATH];
wchar_t g_szFile[MAX_PATH];

extern const char g_szError[]="VirtualDub Error";
extern const char g_szWarning[]="VirtualDub Warning";

///////////////////////////

extern bool Init(HINSTANCE hInstance, LPCWSTR lpCmdLine, int nCmdShow);
extern void Deinit();

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void OpenAVI(int index, bool extended_opt);
void PreviewAVI(HWND, DubOptions *, int iPriority=0, bool fProp=false, bool fPreviewVBR=false);
void CPUTest();
void SaveImageSeq(HWND);
void SaveConfiguration(HWND);
// END ****************************************************************

//
//  FUNCTION: WinMain(HANDLE, HANDLE, LPSTR, int)
//
//  PURPOSE: Entry point for the application.
//
//  COMMENTS:
//
//	This function initializes the application and processes the
//	message loop.
//

int APIENTRY WinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance,
    LPSTR /*lpCmdLine*/, int nCmdShow )
{
	MSG msg;

	Init(hInstance, GetCommandLineW(), nCmdShow);

	// Load a file on the command line.

	if (*g_szFile)
		try {
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, July 2003											***
// BEGIN **************************************************************
			VDAutoLogDisplay logDisp;

			g_project->Open(g_szFile);

			logDisp.Post((VDGUIHandle)g_hWnd);
// END ****************************************************************
		} catch(const MyError& e) {
			e.post(g_hWnd, g_szError);
		}

    // Acquire and dispatch messages until a WM_QUIT message is received.

	for(;;) {
		while (PeekMessage(&msg, 0, 0, 0, PM_REMOVE)) {
			if (msg.message == WM_QUIT)
				goto wm_quit_detected;

			if (guiCheckDialogs(&msg))
				continue;

			HWND hwndRoot = VDGetAncestorW32(msg.hwnd, GA_ROOT);

// *******************************************************************
// *** AVS Modification 										   ***
// *** Tobias Minich, August 2002								   ***
// BEGIN *************************************************************
			if (hwndRoot == g_hWnd) {
				if (TranslateAccelerator(g_hWnd, g_hAccelMain, &msg)) continue;
			} else {
				if (GetClassLong(hwndRoot, GCW_ATOM) == g_classAVS) {
					if (TranslateAccelerator(hwndRoot, g_hAccelAVS, &msg)) continue;
				} else {
					if (TranslateAccelerator(hwndRoot, g_hAccelAudio, &msg)) continue;
				}
			}
// END ***************************************************************

			TranslateMessage(&msg);
			DispatchMessage(&msg);
		}

		if (!g_project->Tick())
			WaitMessage();
	}
wm_quit_detected:

	Deinit();

	VDCHECKPOINT;

    return (msg.wParam);           // Returns the value from PostQuitMessage.

}


//////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////


//////////////////////////////////////////////////////////////////////

char PositionFrameTypeCallback(HWND hwnd, void *pvData, long pos) {
	try {
		if (inputVideoAVI) {
			bool bMasked;
			long nFrame = inputSubset->lookupFrame(pos, bMasked);

			return bMasked ? 'M' : inputVideoAVI->getFrameTypeChar(nFrame);
		} else
			return 0;
	} catch(const MyError&) {
		return 0;
	}
}


//////////////////////////////////////////////////////////////////////






extern const wchar_t fileFilters0[]=
		L"Audio-Video Interleave (*.avi)\0"			L"*.avi\0"
		L"All files (*.*)\0"						L"*.*\0"
		;

// *******************************************************************
// *** Redraw+AVS Modification									   ***
// *** Tobias Minich, September 2002							   ***
// BEGIN *************************************************************
static const char fileFiltersAVS[]=
		"All usable types\0"						"*.avi;*.avs;*.mpg;*.mpeg;*.mpv;*.m1v;*.dat;*.vob;*.d2v;*.lst;*.m2v;*.asf;*.wmv\0"
		"Audio-Video Interleave (*.avi)\0"			"*.avi\0"
		"MPEG-1 video file (*.mpeg,*.mpg,*.mpv,*.dat)\0"	"*.mpg;*.mpeg;*.mpv;*.m1v;*.dat\0"
		"MPEG-2 video file (*.mpeg,*.mpg,*.mpv,*.vob)\0"	"*.mpg;*.mpeg;*.mpv;*.m2v;*.vob\0"
		"DVD2AVI project file (*.d2v)\0"		"*.d2v\0"
		"Microsoft media fomats (*.asf,*.wmv)\0"	"*.asf;*.wmv\0"
		"List file for Nic's MPEGSource (*.lst)\0"				"*.lst\0"
		"All files (*.*)\0"							"*.*\0"
		;
// END ***************************************************************
extern const wchar_t fileFiltersAppend[]=
		L"VirtualDub/AVI_IO video segment (*.avi)\0"	L"*.avi\0"
		L"All files (*.*)\0"							L"*.*\0"
		;

static const wchar_t fileFiltersSaveConfig[]=
		L"VirtualDub configuration (*.vcf)\0"		L"*.vcf\0"
		L"Sylia script for VirtualDub (*.syl)\0"	L"*.syl\0"
		L"All files (*.*)\0"						L"*.*\0"
		;


// *******************************************************************
// *** Redraw+AVS Modification									   ***
// *** Tobias Minich, September 2002							   ***
// BEGIN *************************************************************
/*UINT CALLBACK OpenAVSDlgHookProc(  HWND hDlg, UINT uiMsg, WPARAM wParam, LPARAM lParam) {
	switch(uiMsg) {
	case WM_NOTIFY:
		if (((NMHDR *)lParam)->code == CDN_INITDONE) {
			SendDlgItemMessage(hDlg, IDC_OPEN_TEMPLATE, WM_CLEAR, 0, 0);
			TemplateInfoList::iterator walkit;
			for(walkit = gTemplateList.begin(); walkit != gTemplateList.end(); walkit++) {
				SendDlgItemMessage(hDlg, IDC_OPEN_TEMPLATE, CB_ADDSTRING, 0, (long) walkit->sDescription_.data());
			}
			SendDlgItemMessage(hDlg, IDC_OPEN_TEMPLATE, CB_SELECTSTRING, -1, (long) "Default");
		} else if (((NMHDR *)lParam)->code == CDN_FILEOK) {
			OFNOTIFY *ofn = (OFNOTIFY *)lParam;

			ofn->lpOFN->lCustData	= SendDlgItemMessage(hDlg, IDC_OPEN_TEMPLATE, CB_GETCURSEL, 0, 0);
		}
		break;
	}
	return FALSE;
}*/

void OpenAVI(bool ext_opt) {
	bool fExtendedOpen = false;
	bool fAutoscan = false;

	IVDInputDriver *pDriver = 0;

	std::vector<int> xlat;
	tVDInputDrivers inputDrivers;

	VDGetInputDrivers(inputDrivers, IVDInputDriver::kF_Video);

	VDStringW fileFilters(VDMakeInputDriverFileFilter(inputDrivers, xlat));

	std::vector<void *> templates;
	templates.clear();

	templates.push_back(L"Use AviSynth template");
	templates.push_back(_T("[none]"));

	gTemplateList.clear();
	if (GetTemplateInfo()) {
		TemplateInfoList::iterator walkit;
		for(walkit = gTemplateList.begin(); walkit != gTemplateList.end(); walkit++) {
			templates.push_back((void *)walkit->sDescription_.data());
		}
	}
	templates.push_back(NULL);
	const VDFileDialogOption sOptions[]={
		{ VDFileDialogOption::kBool, 0, L"Ask for e&xtended options after this dialog", 0, 0 },
		{ VDFileDialogOption::kBool, 1, L"Automatically load linked segments", 0, 0 },
		{ VDFileDialogOption::kSelectedFilter, 2, 0, 0, 0 },
		{ VDFileDialogOption::kComboBox, 3, (const wchar_t *)&templates.front(), 0, 0 },
		{0}
	};

	int optVals[4]={0,1,0,0};

	VDStringW fname(VDGetLoadFileName(VDFSPECKEY_LOADVIDEOFILE, (VDGUIHandle)g_hWnd, L"Open video file", fileFilters.c_str(), NULL, sOptions, optVals));

	if (fname.empty())
		return;

	fExtendedOpen = !!optVals[0];
	fAutoscan = !!optVals[1];

	if (xlat[optVals[2]-1] >= 0)
		pDriver = inputDrivers[xlat[optVals[2]-1]];

	if (optVals[3]>0) {
		gFileList.clear();
		gFileList.push_back(VDTextWToA(fname.c_str()).c_str());

		TemplateInfoList::iterator theTemplate;

		theTemplate = gTemplateList.begin();

		theTemplate += (optVals[3]-1);

		char filetemp[MAX_PATH];

		if ((theTemplate>=gTemplateList.end()) || !CreateScript(theTemplate->sFileName_.c_str(), filetemp)) {
			gFileList.clear();
			guiMessageBox(g_hWnd, IDS_ERR_CREATE_SCRIPT, IDS_ERR_CAPTION, MB_ICONERROR|MB_OK);
			return;
		}

		fname = VDTextAToW(filetemp);

		gFileList.clear();
	}

// *******************************************************************
// *** Redraw Modification										   ***
// *** Tobias Minich, August 2002								   ***
// BEGIN *************************************************************
	wcscpy(g_szInputAVIFileLastWorking, fname.c_str());
// END ***************************************************************

	VDAutoLogDisplay logDisp;
	g_project->Open(fname.c_str(), pDriver, fExtendedOpen, false, fAutoscan);
	logDisp.Post((VDGUIHandle)g_hWnd);

// *******************************************************************
// *** VirtualDubMod											   ***
// *** Tobias Minich, Aug 2003									   ***
// BEGIN *************************************************************
	if (IsScriptType(g_szInputAVIFile,SCRIPTTYPE_AVS) && g_VDMPrefs.m_bScriptEditorAutoPopup && g_dllScintilla->ok) {
		HWND tmp = AVSEdit(NULL, g_hWnd, false);
		if (g_VDMPrefs.m_bScriptEditorSingleInstance) {
			VDM_FILENAME *fn = new VDM_FILENAME;
			fn->name = new wchar_t[MAX_PATH];
			if (g_szInputAVIFile[0] != 0) {
				wcscpy(fn->name, g_szInputAVIFile);
			} else {
				*fn->name = 0;
			}
			PostMessage(tmp, WM_VDM_SENDFILENAME, 0, (LPARAM) fn);
		}
	}
// END ***************************************************************
/*	OPENFILENAME ofn;
	char *		files;
	LPTSTR		pfiles;
	LPTSTR		pfilesstart;
	int			a;

	files = (char *) calloc(64000,1);

	gTemplateList.clear();
	GetTemplateInfo();

	ofn.lStructSize			= OPENFILENAME_SIZE_VERSION_400;
	ofn.hwndOwner			= g_hWnd;
	ofn.lpstrFilter			= fileFiltersAVS;
	ofn.lpstrCustomFilter	= NULL;
	ofn.nFilterIndex		= 1;
	ofn.lpstrFile			= files;
	ofn.nMaxFile			= 64000;
	ofn.lpstrFileTitle		= g_szInputAVIFileTitle;
	ofn.nMaxFileTitle		= sizeof g_szInputAVIFileTitle;
	ofn.lpstrInitialDir		= NULL;
	ofn.lpstrTitle			= "Open video file(s) via AVISynth";
	ofn.Flags				= OFN_EXPLORER | OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST | OFN_HIDEREADONLY | OFN_ENABLETEMPLATE | OFN_ENABLEHOOK | OFN_ENABLESIZING | OFN_ALLOWMULTISELECT;
	ofn.lpstrDefExt			= NULL;
	ofn.hInstance			= g_hInst;
	ofn.lpTemplateName		= MAKEINTRESOURCE(IDD_OPEN_AVS);
	ofn.lpfnHook			= (LPOFNHOOKPROC)OpenAVSDlgHookProc;

	if (!GetOpenFileName(&ofn)) {
		a = CommDlgExtendedError();
		free(files);
		return;
	}

	pfiles = files;
	pfilesstart = files;
	gFileList.clear();
	pfiles += strlen(pfiles)+1;

	if (*ofn.lpstrFileTitle != 0) {
		gFileList.push_back(files);
	} else {
		while (!(*pfiles == 0)) {
			strcpy(g_szInputAVIFile, files);
			strcat(g_szInputAVIFile, "\\");
			strcat(g_szInputAVIFile, pfiles);
			gFileList.push_back(g_szInputAVIFile);
			pfiles += strlen(pfiles)+1;
		}
	}


	TemplateInfoList::iterator theTemplate;

	theTemplate = gTemplateList.begin();

	if (ofn.lCustData >= 0) {
		theTemplate += ofn.lCustData;
	}

	if ((theTemplate>=gTemplateList.end()) || !CreateScript(theTemplate->sFileName_.c_str(), g_szInputAVIFile)) {
		gFileList.clear();
		free(files);
		guiMessageBox(g_hWnd, IDS_ERR_CREATE_SCRIPT, IDS_ERR_CAPTION, MB_ICONERROR|MB_OK);
		return;
	}

	gFileList.clear();
	free(files);

	try {

		strcpy(g_szInputAVIFileLastWorking, g_szInputAVIFile);

		OpenAVI(g_szInputAVIFile, FILETYPE_AUTODETECT, false, false, false);

		mru_list->add(g_szInputAVIFile);
		SetTickMode(g_nTickMode);
		if (IsScriptType(g_szInputAVIFile,SCRIPTTYPE_AVS) && g_VDMPrefs.m_bScriptEditorAutoPopup && g_dllScintilla->ok) {
			HWND tmp = AVSEdit(NULL, g_hWnd, false);
			if (g_VDMPrefs.m_bScriptEditorSingleInstance) {
				VDM_FILENAME *fn = new VDM_FILENAME;
				fn->name = new char[MAX_PATH];
				if (g_szInputAVIFile[0] != 0) {
					strcpy(fn->name, g_szInputAVIFile);
				} else {
					*fn->name = 0;
				}
				PostMessage(tmp, WM_VDM_SENDFILENAME, 0, (LPARAM) fn);
			}
		}
	} catch(const MyError& e) {
		e.post(g_hWnd, g_szError);
	}*/
}



#if 0
// END ***************************************************************

void OpenAVI(bool ext_opt) {
	bool fExtendedOpen = false;
	bool fAutoscan = false;

	IVDInputDriver *pDriver = 0;

	std::vector<int> xlat;
	tVDInputDrivers inputDrivers;

	VDGetInputDrivers(inputDrivers, IVDInputDriver::kF_Video);

	VDStringW fileFilters(VDMakeInputDriverFileFilter(inputDrivers, xlat));

	static const VDFileDialogOption sOptions[]={
		{ VDFileDialogOption::kBool, 0, L"Ask for e&xtended options after this dialog", 0, 0 },
		{ VDFileDialogOption::kBool, 1, L"Automatically load linked segments", 0, 0 },
		{ VDFileDialogOption::kSelectedFilter, 2, 0, 0, 0 },
		{0}
	};

	int optVals[3]={0,1,0};

	VDStringW fname(VDGetLoadFileName(VDFSPECKEY_LOADVIDEOFILE, (VDGUIHandle)g_hWnd, L"Open video file", fileFilters.c_str(), NULL, sOptions, optVals));

	if (fname.empty())
		return;

	fExtendedOpen = !!optVals[0];
	fAutoscan = !!optVals[1];

	if (xlat[optVals[2]-1] >= 0)
		pDriver = inputDrivers[xlat[optVals[2]-1]];

// *******************************************************************
// *** Redraw Modification										   ***
// *** Tobias Minich, August 2002								   ***
// BEGIN *************************************************************
	wcscpy(g_szInputAVIFileLastWorking, fname.c_str());
// END ***************************************************************

	VDAutoLogDisplay logDisp;
	g_project->Open(fname.c_str(), pDriver, fExtendedOpen, false, fAutoscan);
	logDisp.Post((VDGUIHandle)g_hWnd);

// *******************************************************************
// *** VirtualDubMod											   ***
// *** Tobias Minich, Aug 2003									   ***
// BEGIN *************************************************************
	if (IsScriptType(g_szInputAVIFile,SCRIPTTYPE_AVS) && g_VDMPrefs.m_bScriptEditorAutoPopup && g_dllScintilla->ok) {
		HWND tmp = AVSEdit(NULL, g_hWnd, false);
		if (g_VDMPrefs.m_bScriptEditorSingleInstance) {
			VDM_FILENAME *fn = new VDM_FILENAME;
			fn->name = new wchar_t[MAX_PATH];
			if (g_szInputAVIFile[0] != 0) {
				wcscpy(fn->name, g_szInputAVIFile);
			} else {
				*fn->name = 0;
			}
			PostMessage(tmp, WM_VDM_SENDFILENAME, 0, (LPARAM) fn);
		}
	}
// END ***************************************************************
}

////////////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2003											***
// BEGIN **************************************************************
#endif
#if 0
void SaveAVI(HWND hWnd, bool fUseCompatibility) {
	if (!inputVideoAVI) {
		guiMessageBox(hWnd, IDS_ERR_NO_INPUT_VIDEO, IDS_ERR_CAPTION, MB_OK);
		return;
	}

	static const VDFileDialogOption sOptions[]={
		{ VDFileDialogOption::kBool, 0, L"Don't run this job now; &add it to job control so I can run it in batch mode.", 0, 0 },
		{0}
	};

	int optVals[1]={0};

	VDStringW fname(VDGetSaveFileName(VDFSPECKEY_SAVEVIDEOFILE, (VDGUIHandle)hWnd, fUseCompatibility ? L"Save AVI 1.0 File" : L"Save AVI 2.0 File", fileFilters0, L"avi", sOptions, optVals));
	if (!fname.empty()) {
		bool fAddAsJob = !!optVals[0];

		if (fAddAsJob) {
			JobAddConfiguration(&g_dubOpts, g_szInputAVIFile, NULL, fname.c_str(), fUseCompatibility, &inputAVI->listFiles, 0, 0, true, SOURCE_AVI);
		} else {
			SaveAVI(fname.c_str(), false, NULL, fUseCompatibility);
		}
	}
}
#endif

///////////////////////////////////////////////////////////////////////////

extern const char g_szRegKeyPersistence[]/*="Persistence"*/;
static const char g_szRegKeySaveSelectionAndEditList[]="Save edit list";

#if 0
void SaveSegmentedAVI(HWND hWnd) {
	if (!inputVideoAVI) {
		MessageBox(hWnd, "No input video stream to process.", g_szError, MB_OK);
		return;
	}

	static const VDFileDialogOption sOptions[]={
		{ VDFileDialogOption::kBool, 0, L"Don't run this job now; &add it to job control so I can run it in batch mode.", 0, 0 },
		{ VDFileDialogOption::kEnabledInt, 1, L"&Limit number of video frames per segment:", 1, 0x7fffffff },
		{ VDFileDialogOption::kInt, 3, L"File segment &size limit in MB (50-2048):", 50, 2048 },
		{0}
	};

	VDRegistryAppKey key(g_szRegKeyPersistence);
	int optVals[4]={
		key.getBool(g_szRegKeyRunAsJob, false),
		key.getBool(g_szRegKeyUseSegmentFrameCount, false),
		key.getInt(g_szRegKeySegmentFrameCount, 100),
		key.getInt(g_szRegKeySegmentSizeLimit, 2000),
	};

	VDStringW fname(VDGetSaveFileName(VDFSPECKEY_SAVEVIDEOFILE, (VDGUIHandle)hWnd, L"Save segmented AVI", fileFiltersAppend, L"avi", sOptions, optVals));

	if (!fname.empty()) {
		key.setBool(g_szRegKeyRunAsJob, !!optVals[0]);
		key.setBool(g_szRegKeyUseSegmentFrameCount, !!optVals[1]);
		if (optVals[1])
			key.setInt(g_szRegKeySegmentFrameCount, optVals[2]);
		key.setInt(g_szRegKeySegmentSizeLimit, optVals[3]);

		char szFile[MAX_PATH];

		strcpy(szFile, VDTextWToA(fname).c_str());

		{
			char szPrefixBuffer[MAX_PATH], szPattern[MAX_PATH*2], *t, *t2, c;
			const char *s;
			int nMatchCount = 0;

			t = VDFileSplitPath(szFile);
			t2 = VDFileSplitExt(t);

			if (!stricmp(t2, ".avi")) {
				while(t2>t && isdigit((unsigned)t2[-1]))
					--t2;

				if (t2>t && t2[-1]=='.')
					strcpy(t2, "avi");
			}

			strcpy(szPrefixBuffer, szFile);
			VDFileSplitExt(szPrefixBuffer)[0] = 0;

			s = VDFileSplitPath(szPrefixBuffer);
			t = szPattern;

			while(*t++ = *s++)
				if (s[-1]=='%')
					*t++ = '%';

			t = szPrefixBuffer;
			while(*t)
				++t;

			strcpy(t, ".*.avi");

			WIN32_FIND_DATA wfd;
			HANDLE h;

			h = FindFirstFile(szPrefixBuffer, &wfd);
			if (h != INVALID_HANDLE_VALUE) {
				strcat(szPattern, ".%d.av%c");

				do {
					int n;

					if (2 == sscanf(wfd.cFileName, szPattern, &n, &c) && tolower(c)=='i')
						++nMatchCount;
					
				} while(FindNextFile(h, &wfd));
				FindClose(h);
			}

			if (nMatchCount) {
				if (IDOK != guiMessageBoxF(g_hWnd, g_szWarning, MB_OKCANCEL|MB_ICONEXCLAMATION,
					"There %s %d existing file%s which match%s the filename pattern \"%s\". These files "
					"will be erased if you continue, to prevent confusion with the new files."
					,nMatchCount==1 ? "is" : "are"
					,nMatchCount
					,nMatchCount==1 ? "" : "s"
					,nMatchCount==1 ? "es" : ""
					,VDFileSplitPath(szPrefixBuffer)))
					return;

				h = FindFirstFile(szPrefixBuffer, &wfd);
				if (h != INVALID_HANDLE_VALUE) {
					strcat(szPattern, ".%d.av%c");

					t = VDFileSplitPath(szPrefixBuffer);

					do {
						int n;

						if (2 == sscanf(wfd.cFileName, szPattern, &n, &c) && tolower(c)=='i') {
							strcpy(t, wfd.cFileName);
							DeleteFile(t);
						}
							
						
					} while(FindNextFile(h, &wfd));
					FindClose(h);
				}
			}
		}

		if (!!optVals[0]) {
			JobAddConfiguration(&g_dubOpts, g_szInputAVIFile, NULL, fname.c_str(), true, &inputAVI->listFiles, optVals[3], optVals[1] ? optVals[2] : 0);
		} else {
			SaveSegmentedAVI(fname.c_str(), false, NULL, optVals[3], optVals[1] ? optVals[2] : 0);
		}
	}
}
#endif
// END ****************************************************************

///////////////////////////////////////////////////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2003											***
// BEGIN **************************************************************
void PreviewAVI(HWND hWnd, DubOptions *quick_options, int iPriority, bool fProp, bool fPreviewVBR) {
	if (!inputVideoAVI) {
		guiMessageBox(hWnd, IDS_ERR_NO_INPUT_VIDEO, IDS_ERR_CAPTION, MB_OK);
		return;
	}

	bool bPreview = g_dubOpts.audio.enabled;
	g_dubOpts.audio.enabled = TRUE;

	VDAVIOutputPreviewSystem outpreview;

	g_project->RunOperation(&outpreview, FALSE, quick_options, iPriority, fProp, 0, 0, fPreviewVBR, SOURCE_AVI);

	g_dubOpts.audio.enabled = bPreview;
}
// END ****************************************************************

void CPUTest() {
	long lEnableFlags;

	lEnableFlags = g_prefs.main.fOptimizations;

	if (!(g_prefs.main.fOptimizations & PreferencesMain::OPTF_FORCE)) {
		SYSTEM_INFO si;

		lEnableFlags = CPUCheckForExtensions();

		GetSystemInfo(&si);

		if (si.wProcessorArchitecture == PROCESSOR_ARCHITECTURE_INTEL)
			if (si.wProcessorLevel < 4)
				lEnableFlags &= ~CPUF_SUPPORTS_FPU;		// Not strictly true, but very slow anyway
	}

	// Enable FPU support...

	long lActualEnabled = CPUEnableExtensions(lEnableFlags);

	VDFastMemcpyAutodetect();
}

/////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2003											***
// BEGIN **************************************************************
class VDSaveImageSeqDialogW32 : public VDDialogBaseW32 {
public:
	VDSaveImageSeqDialogW32();
	~VDSaveImageSeqDialogW32();

	BOOL DlgProc(UINT message, WPARAM wParam, LPARAM lParam);
	void UpdateFilenames();

	char szPrefix[MAX_PATH];
	char szPostfix[MAX_PATH];
	char szDirectory[MAX_PATH];
	char szFormat[MAX_PATH];
	int digits;
	long lFirstFrame, lLastFrame;
	corona::FileFormat	m_FileFormat;
	bool bRunAsJob;
};

static corona::FileFormat g_DefaultFileFormat = corona::FF_TGA;

static struct SaveImageSeqFormats {
	corona::FileFormat	m_FileFormat;
	char				*m_ShortDescription;
	char				*m_Extension;
} s_ImageSequenceFormats[] = {
	{ corona::FF_BMP,			"Windows BMP",	".bmp"},	// not supported as output in Corona
	{ corona::FF_TGA,			"TARGA",		".tga"},	// not supported as output in Corona
	{ corona::FF_PNG,			"PNG",			".png"},
//	{ corona::FF_JPEG,			"Jpeg",			".jpg"},	// not supported as output in Corona
//	{ corona::FF_GIF,			"GIF",			".gif"},	// not supported as output in Corona
//	{ corona::FF_PCX,			"PCX",			".pcx"},	// not supported as output in Corona
	{ corona::FF_AUTODETECT,	NULL,			NULL }
};

static sint16 GetImageFormatIndex(corona::FileFormat format) {
	int i;
	for(i=0
		; (s_ImageSequenceFormats[i].m_FileFormat!=corona::FF_AUTODETECT) || s_ImageSequenceFormats[i].m_ShortDescription || s_ImageSequenceFormats[i].m_Extension
		; i++)
		if(s_ImageSequenceFormats[i].m_FileFormat == format)
			return i;
	return -1;
}
// END ****************************************************************

VDSaveImageSeqDialogW32::VDSaveImageSeqDialogW32()
	: VDDialogBaseW32(IDD_AVIOUTPUTIMAGES_FORMAT)
	, digits(0)
	, lFirstFrame(0)
	, lLastFrame(0)
{
	szPrefix[0] = 0;
	szPostfix[0] = 0;
	szDirectory[0] = 0;
	szFormat[0] = 0;
}
VDSaveImageSeqDialogW32::~VDSaveImageSeqDialogW32() {}

void VDSaveImageSeqDialogW32::UpdateFilenames() {
	char buf[512], *s;

	strcpy(szFormat, szDirectory);

	s = szFormat;
	while(*s) ++s;
	if (s>szFormat && s[-1]!=':' && s[-1]!='\\')
		*s++ = '\\';

	strcpy(s, szPrefix);
	while(*s) ++s;

	char *pCombinedPrefixPt = s;

	*s++ = '%';
	*s++ = '0';
	*s++ = '*';
	*s++ = 'l';
	*s++ = 'd';

	strcpy(s, szPostfix);

	sprintf(buf, szFormat, digits, lFirstFrame);
	SetDlgItemText(mhdlg, IDC_STATIC_FIRSTFRAMENAME, buf);
	sprintf(buf, szFormat, digits, lLastFrame);
	SetDlgItemText(mhdlg, IDC_STATIC_LASTFRAMENAME, buf);

	*pCombinedPrefixPt = 0;
}

BOOL VDSaveImageSeqDialogW32::DlgProc(UINT message, WPARAM wParam, LPARAM lParam) {
	UINT uiTemp;
	BOOL fSuccess;

	switch(message) {
// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2003											***
// BEGIN **************************************************************
	case WM_INITDIALOG:
		SetDlgItemText(mhdlg, IDC_FILENAME_PREFIX, szPrefix);
		SetDlgItemText(mhdlg, IDC_FILENAME_SUFFIX, szPostfix);
		SetDlgItemInt(mhdlg, IDC_FILENAME_DIGITS, digits, FALSE);
		SetDlgItemText(mhdlg, IDC_DIRECTORY, szDirectory);
		{
			int i;
			for(i=0
				; (s_ImageSequenceFormats[i].m_FileFormat!=corona::FF_AUTODETECT) || s_ImageSequenceFormats[i].m_ShortDescription || s_ImageSequenceFormats[i].m_Extension
				; i++)
				SendDlgItemMessage(mhdlg, IDC_FORMAT, CB_ADDSTRING, 0, (LPARAM)s_ImageSequenceFormats[i].m_ShortDescription);
			int index = GetImageFormatIndex(m_FileFormat);
			if(index < 0)
				index = 0;
			SendDlgItemMessage(mhdlg, IDC_FORMAT, CB_SETCURSEL, (WPARAM)index, 0);
		}
		UpdateFilenames();

		return TRUE;

	case WM_COMMAND:
		if((HWND)lParam == GetDlgItem(mhdlg, IDC_FORMAT)) {
			if(HIWORD(wParam) == CBN_SELCHANGE) {
				int index = SendDlgItemMessage(mhdlg, IDC_FORMAT, CB_GETCURSEL, 0, 0);
				if(index != CB_ERR) {
					m_FileFormat = s_ImageSequenceFormats[index].m_FileFormat;
					char *szPostfixExt = szPostfix + strlen(szPostfix);
					while( (szPostfixExt>szPostfix) && (*szPostfixExt!='.'))
						szPostfixExt--;
					if( (*szPostfixExt=='.') && stricmp(szPostfixExt, s_ImageSequenceFormats[index].m_Extension)) {
						int length_1 = szPostfixExt-szPostfix;
						int length_2 = strlen(s_ImageSequenceFormats[index].m_Extension);
						char *newPostfix = new char[length_1+length_2+16];
						memcpy(newPostfix, szPostfix, length_1);
						sprintf(newPostfix + length_1, "%s", s_ImageSequenceFormats[index].m_Extension);
						SetDlgItemText(mhdlg, IDC_FILENAME_SUFFIX, newPostfix);
						delete[] newPostfix;
					}
				}
			}
		} else
// END ****************************************************************
		switch(LOWORD(wParam)) {

		case IDC_FILENAME_PREFIX:
			if (HIWORD(wParam) != EN_CHANGE) break;
			SendMessage((HWND)lParam, WM_GETTEXT, sizeof szPrefix, (LPARAM)szPrefix);
			UpdateFilenames();
			return TRUE;

		case IDC_FILENAME_SUFFIX:
			if (HIWORD(wParam) != EN_CHANGE) break;
			SendMessage((HWND)lParam, WM_GETTEXT, sizeof szPostfix, (LPARAM)szPostfix);
			UpdateFilenames();
			return TRUE;

		case IDC_FILENAME_DIGITS:
			if (HIWORD(wParam) != EN_CHANGE) break;
			uiTemp = GetDlgItemInt(mhdlg, IDC_FILENAME_DIGITS, &fSuccess, FALSE);
			if (fSuccess) {
				digits = uiTemp;

				if (digits > 15)
					digits = 15;

				UpdateFilenames();
			}
			return TRUE;

		case IDC_DIRECTORY:
			if (HIWORD(wParam) != EN_CHANGE) break;
			SendMessage((HWND)lParam, WM_GETTEXT, sizeof szDirectory, (LPARAM)szDirectory);
			UpdateFilenames();
			return TRUE;

		case IDC_SELECT_DIR:
			{
				const VDStringW dir(VDGetDirectory(kFileDialog_ImageDst, (VDGUIHandle)mhdlg, L"Select a directory for saved images"));

				if (!dir.empty())
					SetDlgItemText(mhdlg, IDC_DIRECTORY, VDTextWToA(dir).c_str());
			}
			return TRUE;

		case IDC_ADD_AS_JOB:
			bRunAsJob = !!SendMessage((HWND)lParam, BM_GETCHECK, 0, 0);
			return TRUE;

		case IDOK:
			End(TRUE);
			return TRUE;

		case IDCANCEL:
			End(FALSE);
			return TRUE;
		}
		break;
	}

	return FALSE;
}

void SaveImageSeq(HWND hwnd) {
	VDSaveImageSeqDialogW32 dlg;

	if (!inputVideoAVI) {
		guiMessageBox(hwnd, IDS_ERR_NO_INPUT_VIDEO, IDS_ERR_CAPTION, MB_OK);
		return;
	}

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, March 2003											***
// BEGIN **************************************************************
	dlg.m_FileFormat = g_DefaultFileFormat;
	strcpy(dlg.szPostfix,".tga");
	dlg.lFirstFrame	= 0;
	dlg.lLastFrame	= inputSubset->getTotalFrames()-1;
	dlg.bRunAsJob = false;

	if (dlg.ActivateDialog((VDGUIHandle)hwnd)) {
		g_DefaultFileFormat = dlg.m_FileFormat;

		try {
		if (dlg.bRunAsJob)
			JobAddConfigurationImages(&g_dubOpts, g_szInputAVIFile, NULL, VDTextAToW(dlg.szFormat).c_str(), VDTextAToW(dlg.szPostfix).c_str(), dlg.digits, dlg.m_FileFormat, &inputAVI->listFiles);
		else
			SaveImageSequence(VDTextAToW(dlg.szFormat).c_str(), VDTextAToW(dlg.szPostfix).c_str(), dlg.digits, false, NULL, dlg.m_FileFormat);
		} catch(const MyError& e) {
			e.post(NULL, g_szError);
		}
	}
// END ****************************************************************
}

/////////////////////////////

// ********************************************************************
// *** VirtualDubMod												***
// *** Cyrius, Feb 2003												***
// BEGIN **************************************************************
void SaveFrameToFile(const VBitmap& vbm, sint32 new_width, sint32 new_height) {
	BITMAPINFOHEADER bih;
	long lFormatSize;
	HANDLE hMem;

	vbm.MakeBitmapHeaderNoPadding(&bih);
	lFormatSize = bih.biSize;

	if (hMem = VirtualAlloc(NULL, bih.biSizeImage + lFormatSize, MEM_COMMIT, PAGE_READWRITE)) {
		memcpy(hMem, &bih, lFormatSize);

		VBitmap((char *)hMem + lFormatSize, &bih).BitBlt(0, 0, &vbm, 0, 0, -1, -1);

		wchar_t tFileFilters[1024];
		memset(tFileFilters, 0, sizeof(tFileFilters));
		wchar_t *pPos = tFileFilters;
		wchar_t *pEnd = tFileFilters + sizeof(tFileFilters) - 2*sizeof(wchar_t);
		int i;
		for(i=0
			; (s_ImageSequenceFormats[i].m_FileFormat!=corona::FF_AUTODETECT) || s_ImageSequenceFormats[i].m_ShortDescription || s_ImageSequenceFormats[i].m_Extension
			; i++) {
			pPos += _snwprintf(pPos, (pEnd-pPos)/sizeof(wchar_t), L"%s (*%s)", VDTextAToW(s_ImageSequenceFormats[i].m_ShortDescription).c_str(), VDTextAToW(s_ImageSequenceFormats[i].m_Extension).c_str()) + 1;
			pPos += _snwprintf(pPos, (pEnd-pPos)/sizeof(wchar_t), L"*%s", VDTextAToW(s_ImageSequenceFormats[i].m_Extension).c_str()) + 1;
		}

		int index = GetImageFormatIndex(g_DefaultFileFormat);
		if(index <= 0)
			index = 0;
		DWORD nFilterIndex = index + 1;
		VDStringW prefExt = VDStringW::setf(L"*%s", VDTextAToW(s_ImageSequenceFormats[index].m_Extension).c_str());

		static const VDFileDialogOption sOptions[]={
			{ VDFileDialogOption::kSelectedFilter, 0, NULL, 0, 0 },
			{0}
		};

		int optVals[1]={
			nFilterIndex
		};

		const VDStringW filename(VDGetSaveFileName(kFileDialog_SaveSnapshot, (VDGUIHandle)g_hWnd, L"Save Snapshot image", tFileFilters, prefExt.c_str(), sOptions, optVals));

		nFilterIndex = optVals[0];
		if(!filename.empty()) {
			corona::FileFormat format = s_ImageSequenceFormats[nFilterIndex-1].m_FileFormat;
			g_DefaultFileFormat = format;
			OutputImageToFile oToFile(format);
			oToFile.SaveImage(filename.c_str(), &bih, lFormatSize, (char *)hMem + lFormatSize, bih.biSizeImage);
		}
		VirtualFree(hMem, 0, MEM_RELEASE);
	}
}
// END ****************************************************************

/////////////////////////////////////////////////////////////////////////////////

void SaveConfiguration(HWND hWnd) {
	static const VDFileDialogOption sOptions[]={
		{ VDFileDialogOption::kBool, 0, L"Include selection and edit list", 0, 0 },
		{0}
	};

	VDRegistryAppKey key(g_szRegKeyPersistence);
	int optVals[1]={
		key.getBool(g_szRegKeySaveSelectionAndEditList, false),
	};

	const VDStringW filename(VDGetSaveFileName(kFileDialog_Config, (VDGUIHandle)hWnd, L"Save Configuration", fileFiltersSaveConfig, L"vcf", sOptions, optVals));

	if (!filename.empty()) {
		key.setBool(g_szRegKeySaveSelectionAndEditList, !!optVals[0]);

		FILE *f = NULL;
		try {
			f = fopen(VDTextWToA(filename).c_str(), "w");

			if (!f)
				throw MyError("Cannot open output file: %s.", strerror(errno));

			JobWriteConfiguration(f, &g_dubOpts, !!optVals[0]);

			fclose(f);
			f = NULL;
		} catch(const MyError& e) {
			e.post(NULL, g_szError);
		}

		if (f)
			fclose(f);
	}
}
