//  VirtualDub - Video processing and capture application
//  Copyright (C) 1998-2001 Avery Lee
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//
///////////////////////////////////////////////////////////////////////////

//	This version has been modified by fccHandler, December 2003

#include "stdafx.h"
#include <stdio.h>
#include <process.h>
#include <crtdbg.h>
#include <assert.h>
#include <windows.h>
#include <math.h>
#include <vd2/system/error.h>
#include <vd2/system/cpuaccel.h>

#include "mpeg_decode.h"
#include "mpeg_tables.h"
#include "mpeg_asm.h"

extern void CPUTest();		// Main.cpp

// Start codes
#define SEQUENCE_START_CODE     0x1B3
#define PICTURE_START_CODE      0x100
#define SLICE_START_CODE_MIN    0x101
#define SLICE_START_CODE_MAX    0x1AF
#define EXTENSION_START_CODE    0x1B5
#define USER_DATA_START_CODE    0x1B2

// Extension start code IDs
#define SEQUENCE_EXTENSION_ID           1
#define SEQUENCE_DISPLAY_EXTENSION_ID   2
#define QUANT_MATRIX_EXTENSION_ID       3
#define COPYRIGHT_EXTENSION_ID          4
#define PICTURE_DISPLAY_EXTENSION_ID    7
#define PICTURE_CODING_EXTENSION_ID     8

// Chroma formats
#define CHROMA420       1
#define CHROMA422       2
#define CHROMA444       3
// Picture structures
#define TOP_FIELD       1
#define BOTTOM_FIELD    2
#define FRAME_PICTURE   3
// Motion_types
#define MC_FIELD    1
#define MC_FRAME    2
#define MC_16X8     2
#define MC_DMV      3
// Motion vector types
#define MV_FIELD    0
#define MV_FRAME    1
// flags for MPEGBuffer.bitflags
#define PICTUREFLAG_RFF         1       // repeat_first_field
#define PICTUREFLAG_TFF         2       // top_field_first
#define PICTUREFLAG_PF          4       // progressive_frame

//////////////////////////////////////////////////////////////

typedef unsigned char YUVPixel;

// Forward declarations
static void slice(int type);

///////////////////////////////////////////////////////////////////////////////

static const unsigned char zigzag[2][64] = {
  {     // Zig-Zag scan pattern
     0, 1, 8,16, 9, 2, 3,10,
    17,24,32,25,18,11, 4, 5,
    12,19,26,33,40,48,41,34,
    27,20,13, 6, 7,14,21,28,
    35,42,49,56,57,50,43,36,
    29,22,15,23,30,37,44,51,
    58,59,52,45,38,31,39,46,
    53,60,61,54,47,55,62,63
  },
  {     // Alternate scan pattern
     0, 8,16,24, 1, 9, 2,10,
    17,25,32,40,48,56,57,49,
    41,33,26,18, 3,11, 4,12,
    19,27,34,42,50,58,35,43,
    51,59,20,28, 5,13, 6,14,
    21,29,36,44,52,60,37,45,
    53,61,22,30, 7,15,23,31,
    38,46,54,62,39,47,55,63
  }
};

static const unsigned char intramatrix_default[64] = {
     8, 16, 19, 22, 26, 27, 29, 34,
    16, 16, 22, 24, 27, 29, 34, 37,
    19, 22, 26, 27, 29, 34, 34, 38,
    22, 22, 26, 27, 29, 34, 37, 40,
    22, 26, 27, 29, 32, 35, 40, 48,
    26, 27, 29, 32, 35, 40, 48, 58,
    26, 27, 29, 34, 38, 46, 56, 69,
    27, 29, 35, 38, 46, 56, 69, 83};

// non-linear quantization coefficient table
static const unsigned char non_linear[32] = {
    0,  1,  2,  3,  4,  5,  6,  7,  8, 10, 12, 14, 16, 18, 20, 22,
   24, 28, 32, 36, 40, 44, 48, 52, 56, 64, 72, 80, 88, 96,104,112
};

static struct MPEGBuffer {
    YUVPixel *Y;
    YUVPixel *U;
    YUVPixel *V;
    int frame_num;
	int bitflags;
} buffers[3];

static LPVOID memblock = NULL;

static long old_width = 0;
static long old_height = 0;
static int old_chroma = 0;

static unsigned char intramatrix[64];
static unsigned char nonintramatrix[64];
static unsigned char chromaintramatrix[64];
static unsigned char chromanonintramatrix[64];

static int forw_vector_full_pel;
static int back_vector_full_pel;

static bool MPEG2_Flag;
static int quantizer_scale;
static int matrix_coefficients;
static int chroma_format;
static int block_count;
static int progressive_sequence;
static int f_code[2][2];
static int intra_dc_precision;
static int picture_structure;
//static int top_field_first;
static int frame_pred_frame_dct;
static int concealment_vectors;
static int q_scale_type;
static int intra_vlc_format;
static int alternate_scan;
//static int repeat_first_field;
//static int progressive_frame;
static long PMV[2][2][2];
static long dmvector[2];
static long mvfield_select[2][2];
static int motion_vector_count;
static int motion_type;
static int mv_format;
static int mvscale;
static int dct_type;

MPEGBuffer *current_frame;

YUVPixel *uv422, *u444, *v444;

//static short (*dct_coeff)[12][64];
static short (*dct_coeff)[64];

static int second_field;
static bool fault_flag;

static int frame_type;
static long pelWidth, pelHeight, mbWidth, mbHeight;
static long y_pitch, uv_pitch;

static int dct_dc_y_past, dct_dc_u_past, dct_dc_v_past;
static int macro_block_flags;

////////////////////////////////////////////////////////////////////////////////

static __inline void mpeg_dealloc() {
    if (memblock != NULL) {
        VirtualFree(memblock, 0, MEM_RELEASE);
        memblock = NULL;
    }
}

void mpeg_deinitialize() {
    mpeg_dealloc();
    old_width = 0;
    old_height = 0;
    old_chroma = 0;
}

static __inline bool mpeg_alloc()    // Allocate memory
{
    if ((pelWidth > 0) && (pelHeight > 0)) {
        unsigned long chroma_size;

        long mbw = (pelWidth+15)/16;
        long mbh = 2*((pelHeight+31)/32);   // worst-case

        // size of Y buffer (3 of these)
        unsigned long i = mbw*16 * mbh*16 * 3;

        // size of U/V buffer (6 of these)
        switch (chroma_format) {
            case CHROMA420:
                chroma_size = mbw*8 * mbh*8;
                break;
            case CHROMA422:
                chroma_size = mbw*8 * mbh*16;
                break;
            case CHROMA444:
                chroma_size = mbw*16 * mbh*16;
                break;
            default:
                return false;
        }
        i += chroma_size*6;

		// Intermediate 4:2:2 buffer (1 of these)
		i += mbw*8 * mbh*16;
		// Intermediate 4:4:4 buffers (2 of these)
		i += mbw*16 * mbh*16 * 2;

		// dct coefficient block
		i += 12 * 64 * sizeof(short);

        i += 256;   // 256-byte alignment (necessary?)

        mpeg_dealloc();
        memblock = VirtualAlloc(NULL,i,MEM_COMMIT,PAGE_READWRITE);
        if (memblock != NULL) {
            // Assign memory blocks (align to 256-byte boundary)
            buffers[0].Y = (YUVPixel *)
                (((unsigned long)memblock + 255) & 0xFFFFFF00);
            buffers[1].Y = buffers[0].Y + mbw*16 * mbh*16;
            buffers[2].Y = buffers[1].Y + mbw*16 * mbh*16;
            buffers[0].U = buffers[2].Y + mbw*16 * mbh*16;
            buffers[1].U = buffers[0].U + chroma_size;
            buffers[2].U = buffers[1].U + chroma_size;
            buffers[0].V = buffers[2].U + chroma_size;
            buffers[1].V = buffers[0].V + chroma_size;
            buffers[2].V = buffers[1].V + chroma_size;

			uv422 = buffers[2].V + chroma_size;
			u444 = uv422 + mbw*8 * mbh*16;
			v444 = u444 + mbw*16 * mbh*16;
			//dct_coeff = (short *)(v444 + mbw*16 * mbh*16);
			dct_coeff = (short (*)[64])(v444 + mbw*16 * mbh*16);
            return true;
        }
    }
    return false;
}

void mpeg_initialize(int width, int height,
  char *dummy1, char *dummy2, BOOL dummy3) {

// dummy1 will be NULL if initializing MPEG-1, otherwise
// it points to 3 bytes with the necessary information:

// xxxx.... ........ ........ SEQUENCE_EXTENSION_ID
// ....x... ........ ........ escape_bit
// .....xxx xxxx.... ........ MPEG-2 profile / level
// ........ ....x... ........ progressive_sequence
// ........ .....xx. ........ chroma_format
// ........ .......x x....... horizontal_size_extension
// ........ ........ .xx..... vertical_size_extension
    
    if (dummy1 != NULL) {
        // initializing MPEG-2
        pelWidth |= ((dummy1[1]&1)<<13) | ((dummy1[2]&0x80)<<5);
        pelHeight |= ((dummy1[2]&0x60)<<7);
        chroma_format = (dummy1[1]>>1)&3;
        progressive_sequence = (dummy1[1]>>3)&1;
    } else {
        // initializing MPEG-1
		CPUTest();
        mpeg_deinitialize();
        pelWidth = width;
        pelHeight = height;
        chroma_format = CHROMA420;
        progressive_sequence = 1;
    }

    mbWidth = (pelWidth+15)/16;
    mbHeight = (progressive_sequence)?
        (pelHeight+15)/16: 2*((pelHeight+31)/32);
    y_pitch = mbWidth * 16;
    uv_pitch = (chroma_format==CHROMA444)? y_pitch: y_pitch>>1;
    second_field = 0;

    if (!mpeg_alloc()) throw MyMemoryError();
    old_width = pelWidth;
    old_height = pelHeight;
    old_chroma = chroma_format;
    mpeg_reset();
}

void mpeg_reset() {
    if (memblock != NULL) {
        for (int i=0; i<3; i++) {
//            memset(buffers[i].Y, 0, y_pitch * mbHeight * 16);
//            memset(buffers[i].U, 128, uv_pitch * mbHeight * 8);
//            memset(buffers[i].V, 128, uv_pitch * mbHeight * 8);
            buffers[i].frame_num = -1;
        }
    }
}

// My bits-getting functions (replaces CMemoryBitInput class)
static unsigned char *bitarray;
static unsigned long totbits;
static unsigned long bitpos;

static long __declspec(naked) __fastcall show_bits(int nBits) {

// Correct results are guaranteed only if nBits >= 0 and nBits <= 24

    __asm {
        cmp     ecx, 24
        ja      zero            ; Zero if nBits < 0 or nBits > 24
        mov     eax, bitpos
        or      ch, cl
        mov     edx, totbits
        jz      zero            ; Zero if nBits == 0
        sub     edx, eax        ; edx = number of bits remaining
        mov     cl, al
        cmp     edx, 24         ; Less than 25 bits remain?
        jle     alternate3      ; Use alternate method
        shr     eax, 3          ; eax = bytepos
        and     cl, 7           ; cl = bitpos & 7
        add     eax, bitarray
        neg     ch              ; ch = 32 - nBits
        mov     eax, [eax]      ; eax = [3][2][1][0]
done:   bswap   eax             ; eax = [0][1][2][3]
        shl     eax, cl
        mov     cl, ch
        shr     eax, cl
        ret

zero:   xor     eax, eax
        ret

alternate3:
        cmp     edx, 16
        jle     alternate2
    ; only 3 bytes remain to be read, treat the 4th as zero
        mov     edx, bitarray
        and     cl, 7           ; cl = bitpos & 7
        shr     eax, 3          ; eax = bytepos
        neg     ch              ; ch = 32 - nBits
        add     edx, eax
        xor     eax, eax
        mov     al, byte ptr [edx+2]    ; eax = [.][.][.][2]
        shl     eax, 16                 ; eax = [.][2][.][.]
        mov     ax, word ptr [edx]      ; eax = [.][2][1][0]
        jmp     done

alternate2:
        cmp     edx, 8
        jle     alternate1
    ; only 2 bytes remain to be read, assume the 3rd & 4th are zero
        shr     eax, 3              ; eax = bytepos
        and     cl, 7               ; cl = bitpos & 7
        add     eax, bitarray
        neg     ch                  ; ch = 32 - nBits
        movzx   eax, word ptr [eax] ; eax = [.][.][1][0]
        jmp     done

alternate1:
        test    edx, edx
        jle     zero
    ; only one byte remains to be read, assume the others are zero
        shr     eax, 3              ; eax = bytepos
        and     cl, 7               ; cl = bitpos & 7
        add     eax, bitarray
        neg     ch                  ; ch = 32 - nBits
        movzx   eax, byte ptr [eax] ; eax = [.][.][.][0]
        jmp     done
    }
}

/*
static long __declspec(naked) __cdecl get_bits1() {
    __asm {
        mov     eax, bitpos
		mov		cl, 7
		cmp		eax, totbits
		jnc		zero			; no more bits
		inc		bitpos
        mov     ch, al
        shr     eax, 3          ; eax = bytepos
        and     ch, 7           ; ch = bitpos & 7
        add     eax, bitarray
		sub		cl, ch			; cl = 7 - (bitpos & 7)
		mov		al, [eax]
		shr		al, cl
		and		eax, 1
        ret
zero:	xor		eax, eax
		ret
	}
}
*/

#define flush_buffer(nBits) bitpos += nBits

static long get_bits(int nBits) {
    long i = show_bits(nBits);
    flush_buffer(nBits);
    return i;
}

static int next_start_code() {
    int i;

    //align to next byte
    bitpos += (-(long)bitpos) & 7;

    //set position at next start code and return it
    while (bitpos < totbits) {
        i = get_bits(8);
        if (i == 0) {
            i = show_bits(24);
            if ((i & 0xFFFFFF00) == 0x100) {
                bitpos -= 8;
                return i;
            }
        }
    }
    return 0;
}

static __inline void picture_coding_extension() {

    MPEG2_Flag = true;
	current_frame->bitflags = 0;

    f_code[0][0] = get_bits(4)-1;
    f_code[0][1] = get_bits(4)-1;
    f_code[1][0] = get_bits(4)-1;
    f_code[1][1] = get_bits(4)-1;

// zero is forbidden
	if (f_code[0][0] < 0) f_code[0][0] = 0;
	if (f_code[0][1] < 0) f_code[0][1] = 0;
	if (f_code[1][0] < 0) f_code[1][0] = 0;
	if (f_code[1][1] < 0) f_code[1][1] = 0;

    intra_dc_precision = get_bits(2);
    picture_structure = get_bits(2);

//    top_field_first = get_bits(1);
	if (get_bits(1)) current_frame->bitflags |= PICTUREFLAG_TFF;

    frame_pred_frame_dct = get_bits(1);
    concealment_vectors = get_bits(1);
    q_scale_type = get_bits(1);
    intra_vlc_format = get_bits(1);
    alternate_scan = get_bits(1);

//    repeat_first_field = get_bits(1);
	if (get_bits(1)) current_frame->bitflags |= PICTUREFLAG_RFF;

    flush_buffer(1);    // chroma_420_type (obsolete)

//    progressive_frame = get_bits(1);
	if (get_bits(1)) current_frame->bitflags |= PICTUREFLAG_PF;
}

static __inline void quant_matrix_extension() {
    int i;

    if (get_bits(1)) {
        for (i=0; i<64; i++)
            intramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
    }

    if (get_bits(1)) {
        for (i=0; i<64; i++)
            nonintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
    }

    if (get_bits(1)) {
        for (i=0; i<64; i++)
            chromaintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
    } else
        memcpy(chromaintramatrix, intramatrix, 64);

    if (get_bits(1)) {
        for (i=0; i<64; i++)
            chromanonintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
    } else
        memcpy(chromanonintramatrix, nonintramatrix, 64);

}

static __inline void sequence_extension() {
    MPEG2_Flag = true;
    flush_buffer(8);        // profile / level
    progressive_sequence = get_bits(1);
    chroma_format = get_bits(2);
    pelWidth = (get_bits(2) << 12) | (pelWidth & 0x0FFF);
    pelHeight = (get_bits(2) << 12) | (pelHeight & 0x0FFF);
    matrix_coefficients = 1;
    // skip the rest
    flush_buffer(12+1+8+1+2+5);
}

static void extension_and_user_data() {
    int code;

    while (code = next_start_code()) {
        if (code == EXTENSION_START_CODE) {
            flush_buffer(32);
            switch (get_bits(4)) {
                case SEQUENCE_EXTENSION_ID:
                    sequence_extension();
                    break;
                case SEQUENCE_DISPLAY_EXTENSION_ID:
                    if (get_bits(4) & 1) {
                        flush_buffer(16);   // color_primaries, etc.
                        matrix_coefficients = get_bits(8) & 7;
                    }
                    break;
                case PICTURE_CODING_EXTENSION_ID:
                    picture_coding_extension();
                    break;
                case QUANT_MATRIX_EXTENSION_ID:
                    quant_matrix_extension();
                    break;
            }
        } else {
            if (code != USER_DATA_START_CODE) break;
            flush_buffer(32);
        }
    }
}

static __inline void sequence_header() {
    int i;

    // skip over params that are not supposed to change
    // (width, height, aspect_ratio, frame_rate, etc.)
    // since they were parsed already in Mpeg.cpp...
    flush_buffer(24+8+19+11);

    // load new matrices if applicable
    if (get_bits(1)) {      // load_intra_quantizer_matrix
        for (i=0; i<64; i++)
            intramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
    } else  memcpy(intramatrix, intramatrix_default, 64);

    if (get_bits(1)) {      // load_non_intra_quantizer_matrix
        for (i=0; i<64; i++)
            nonintramatrix[zigzag[0][i]] = (unsigned char)get_bits(8);
    } else  memset(nonintramatrix, 16, 64);

    memcpy(chromaintramatrix, intramatrix, 64);
    memcpy(chromanonintramatrix, nonintramatrix, 64);

    // The following remain at MPEG-1 defaults, unless
    // overwritten by MPEG-2 sequence_extension()...
    MPEG2_Flag = false;
    progressive_sequence = 1;
    chroma_format = CHROMA420;
    matrix_coefficients = 5;
    extension_and_user_data();
}

static __inline void picture_header() {
    flush_buffer(10);   // temporal_reference
    frame_type = get_bits(3);

// Note:  The confusion here stems from the fact that the most
// recently decoded I/P frame is the _forward_ prediction for
// a P frame, but it's the _backward_ prediction for a B frame!

// Rule:  The most recently decoded I/P frame is
// always found in the _backward_ prediction buffer

    switch(frame_type) {

    case MPEG_FRAME_TYPE_I:     // I-frames have no prediction
    case MPEG_FRAME_TYPE_P:     // P-frames predict back to the last I or P
        if (!second_field)
            mpeg_swap_buffers(MPEG_BUFFER_FORWARD,MPEG_BUFFER_BACKWARD);
		current_frame = &(buffers[MPEG_BUFFER_BACKWARD]);
        break;

    case MPEG_FRAME_TYPE_B:     // B-frames predict back to the last I or P and forward to the next P
		current_frame = &(buffers[MPEG_BUFFER_BIDIRECTIONAL]);
        break;

    case MPEG_FRAME_TYPE_D:
        throw MyError("D-type frames not supported");

    default:
        throw MyError("Unknown frame type 0x%d", frame_type);
    }

    flush_buffer(16);   // VBV_delay

    if (frame_type == MPEG_FRAME_TYPE_P || frame_type == MPEG_FRAME_TYPE_B) {
        forw_vector_full_pel = get_bits(1);
        f_code[0][0] = f_code[0][1] = get_bits(3)-1;

        if (frame_type == MPEG_FRAME_TYPE_B) {
            back_vector_full_pel = get_bits(1);
            f_code[1][0] = f_code[1][1] = get_bits(3)-1;
        }
    }

    // The following remain at MPEG-1 defaults, unless
    // overwritten by MPEG-2 picture_coding_extension()...
    MPEG2_Flag = false;
    intra_dc_precision = 0;
    picture_structure = FRAME_PICTURE;
//    top_field_first = 0;
    frame_pred_frame_dct = 1;
    concealment_vectors = 0;
    q_scale_type = 0;
    intra_vlc_format = 0;
    alternate_scan = 0;
//    repeat_first_field = 0;
//    progressive_frame = 1;

	current_frame->bitflags = PICTUREFLAG_PF;

    extension_and_user_data();
}

void mpeg_decode_frame(void *input_data, int len, int frame_num) {
    int code;

// sanity check
    if ((input_data == NULL) || (len < 4)) return;

// Bind bits-getting functions to "input_data"
    bitarray = (unsigned char *)input_data;
    totbits = len << 3;
    bitpos = 0;
    second_field = 0;
    frame_type = 0;

Again:

// Decode headers up to the first slice
	while (code = next_start_code()) {
        if (code >= SLICE_START_CODE_MIN
         && code <= SLICE_START_CODE_MAX) break;

        flush_buffer(32);

        if (code == SEQUENCE_START_CODE)
            sequence_header();
        else if (code == PICTURE_START_CODE)
            picture_header();
    }
	if (!code) return;		// ran out of data

    switch (chroma_format) {
        case CHROMA420: block_count = 6; break;
        case CHROMA422: block_count = 8; break;
        case CHROMA444: block_count = 12; break;
        default: return;    // bad chroma_format
    }

    // Allocate memory if necessary
    if (pelWidth != old_width || pelHeight != old_height || chroma_format != old_chroma) {
        if (!mpeg_alloc()) return;		// out of memory
        old_width = pelWidth;
        old_height = pelHeight;
        old_chroma = chroma_format;
    }

    mbWidth = (pelWidth+15)>>4;
    if (progressive_sequence) {
        mbHeight = (pelHeight+15)>>4;
    } else {
        mbHeight = (pelHeight+31)>>5;
        if (picture_structure == FRAME_PICTURE) mbHeight<<=1;
    }
    y_pitch = mbWidth << 4;
    uv_pitch = (chroma_format==CHROMA444)? y_pitch: y_pitch>>1;

// Decode all slices
    do {
        flush_buffer(32);
        slice(code);
        code = next_start_code();
    } while (code >= SLICE_START_CODE_MIN
          && code <= SLICE_START_CODE_MAX);

	if (!second_field) {
		current_frame->frame_num = frame_num;

		if (picture_structure != FRAME_PICTURE) {
			second_field = !second_field;
			goto Again;
		}
    }

    if (MMX_enabled) __asm emms
}

void mpeg_swap_buffers(int buffer1, int buffer2) {
    if (((unsigned int)buffer1 < 3) &&
      ((unsigned int)buffer2 < 3) && (buffer1 != buffer2)) {
        MPEGBuffer b = buffers[buffer1];
        buffers[buffer1] = buffers[buffer2];
        buffers[buffer2] = b;
    }
}

int mpeg_lookup_frame(int frame) {
    if (frame != -1) {
        if (buffers[0].frame_num == frame) return 0;
        if (buffers[1].frame_num == frame) return 1;
        if (buffers[2].frame_num == frame) return 2;
    }
    return -1;
}

/*********************************************************
  Integer IDCT, based on Chen-Wang code
**********************************************************/

#define CW_W1   2841    // 2048*sqrt(2)*cos(1*pi/16)
#define CW_W2   2676    // 2048*sqrt(2)*cos(2*pi/16)
#define CW_W3   2408    // 2048*sqrt(2)*cos(3*pi/16)
#define CW_W5   1609    // 2048*sqrt(2)*cos(5*pi/16)
#define CW_W6   1108    // 2048*sqrt(2)*cos(6*pi/16)
#define CW_W7   565     // 2048*sqrt(2)*cos(7*pi/16)

static void idctrow(short *blk) {
    long x0,x1,x2,x3,x4,x5,x6,x7,x8;

    // pos is (col*8+#)
    x1 = blk[4]<<11;
    x2 = blk[6];
    x3 = blk[2];
    x4 = blk[1];
    x5 = blk[7];
    x6 = blk[5];
    x7 = blk[3];

    if (!(x1|x2|x3|x4|x5|x6|x7)) {
        // Shortcut...
        x0 = blk[0]<<3;
        blk[0]=blk[1]=blk[2]=blk[3]
          =blk[4]=blk[5]=blk[6]=blk[7]=(short)x0;
        return;
    }

    x0=(blk[0]<<11)+128;    // for proper rounding

    // first stage
    x8 = CW_W7 * (x4 + x5);
    x4 = x8 + (CW_W1 - CW_W7) * x4;
    x5 = x8 - (CW_W1 + CW_W7) * x5;
    x8 = CW_W3 * (x6 + x7);
    x6 = x8 - (CW_W3 - CW_W5) * x6;
    x7 = x8 - (CW_W3 + CW_W5) * x7;
    x8 = CW_W6 * (x3 + x2);
    x2 = x8 - (CW_W2 + CW_W6) * x2;
    x3 = x8 + (CW_W2 - CW_W6) * x3;

    // second stage
    x8 = x0 + x1; x0 -= x1;
    x1 = x4 + x6; x4 -= x6;
    x6 = x5 + x7; x5 -= x7;
    x7 = x8 + x3; x8 -= x3;
    x3 = x0 + x2; x0 -= x2;
    x2 = (181 * (x4 + x5) + 128)>>8;
    x4 = (181 * (x4 - x5) + 128)>>8;

    // final stage
    blk[0] = (short)((x7 + x1)>>8);
    blk[1] = (short)((x3 + x2)>>8);
    blk[2] = (short)((x0 + x4)>>8);
    blk[3] = (short)((x8 + x6)>>8);
    blk[4] = (short)((x8 - x6)>>8);
    blk[5] = (short)((x0 - x4)>>8);
    blk[6] = (short)((x3 - x2)>>8);
    blk[7] = (short)((x7 - x1)>>8);
}

static void idctcol(short *blk) {
    long x0,x1,x2,x3,x4,x5,x6,x7,x8;

    x4 = blk[8];
    x3 = blk[16];
    x7 = blk[24];
    x1 = blk[32]<<8;
    x6 = blk[40];
    x2 = blk[48];
    x5 = blk[56];

    if (!(x1|x2|x3|x4|x5|x6|x7)) {
        // Shortcut...
        x0 = (blk[0]+32) >> 6;
        if (x0 < -256) x0 = -256; else if (x0 > 255) x0 = 255;
        blk[0]=blk[8]=blk[16]=blk[24]=blk[32]
          =blk[40]=blk[48]=blk[56] = (short)x0;
        return;
    }

    x0 = (blk[0] << 8) + 8192;

    // first stage
    x8 = (CW_W7 * (x4 + x5)) + 4;
    x4 = (x8 + (CW_W1 - CW_W7) * x4) >> 3;
    x5 = (x8 - (CW_W1 + CW_W7) * x5) >> 3;
    x8 = (CW_W3 * (x6 + x7)) + 4;
    x6 = (x8 - (CW_W3 - CW_W5) * x6) >> 3;
    x7 = (x8 - (CW_W3 + CW_W5) * x7) >> 3;
    x8 = (CW_W6 * (x3 + x2)) + 4;
    x2 = (x8 - (CW_W2 + CW_W6) * x2) >> 3;
    x3 = (x8 + (CW_W2 - CW_W6) * x3) >> 3;

    // second stage
    x8 = x0 + x1; x0 -= x1;
    x1 = x4 + x6; x4 -= x6;
    x6 = x5 + x7; x5 -= x7;
    x7 = x8 + x3; x8 -= x3;
    x3 = x0 + x2; x0 -= x2;
    x2 = (181 * (x4 + x5) + 128) >> 8;
    x4 = (181 * (x4 - x5) + 128) >> 8;

    // final stage
    x5 = (x7 + x1) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[0] = (short)x5;
    x5 = (x3 + x2) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[8] = (short)x5;
    x5 = (x0 + x4) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[16] = (short)x5;
    x5 = (x8 + x6) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[24] = (short)x5;
    x5 = (x8 - x6) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[32] = (short)x5;
    x5 = (x0 - x4) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[40] = (short)x5;
    x5 = (x3 - x2) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[48] = (short)x5;
    x5 = (x7 - x1) >> 14;
    if (x5 < -256) x5 = -256; else if (x5 > 255) x5 = 255;
    blk[56] = (short)x5;
}

static void Fast_IDCT(void *src) {
    int i;
    short *blk;
// two dimensional inverse discrete cosine transform
    for (i=0,blk=(short *)src; i<8; i++,blk+=8) idctrow(blk);
    for (i=0,blk=(short *)src; i<8; i++,blk++) idctcol(blk);
}

static void Add_Block(int comp, int bx, int by, int addflag) {
    int i,j,v,iincr;
    YUVPixel *rfp;

    if (comp < 4) {     // luminance
        rfp = current_frame->Y;
        if (picture_structure==FRAME_PICTURE) {
            if (dct_type) {
                // field DCT coding
                rfp += y_pitch*(by+((comp&2)>>1)) + bx + ((comp&1)<<3);
                iincr = y_pitch<<1;
            } else {
                // frame DCT coding
                rfp += y_pitch*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
                iincr = y_pitch;
            }
        } else {
            // field picture
			if (picture_structure == BOTTOM_FIELD) rfp += y_pitch;
            rfp += (y_pitch<<1)*(by+((comp&2)<<2)) + bx + ((comp&1)<<3);
            iincr = y_pitch<<1;
        }
    } else {        // chrominance
        if (comp & 1)   rfp = current_frame->V;
        else            rfp = current_frame->U;

        // scale coordinates
        if (chroma_format != CHROMA444) {
            bx >>= 1;
            if (chroma_format == CHROMA420) by >>= 1;
        }

        if (picture_structure==FRAME_PICTURE) {
            if (dct_type && (chroma_format != CHROMA420)) {
                // field DCT coding
                rfp += uv_pitch*(by+((comp&2)>>1)) + bx + (comp&8);
                iincr = uv_pitch<<1;
            } else {
                // frame DCT coding
                rfp += uv_pitch*(by+((comp&2)<<2)) + bx + (comp&8);
                iincr = uv_pitch;
            }
        } else {
            // field picture
			if (picture_structure == BOTTOM_FIELD) rfp += uv_pitch;
            rfp += (uv_pitch<<1)*(by+((comp&2)<<2)) + bx + (comp&8);
            iincr = uv_pitch<<1;
        }
    }

    if (MMX_enabled) {
		mmx_Add_Block(dct_coeff[comp], rfp, iincr, addflag);

    } else {
		short *blk = dct_coeff[comp];

		if (addflag) {
			for (j=0; j<8; j++) {
				for (i=0; i<8; i++) {
					v = *blk++ + rfp[i];
					if (v<0) v=0; else if (v>255) v=255;
					rfp[i] = (YUVPixel)v;
				}
				rfp+= iincr;
			}
		} else {
			for (j=0; j<8; j++) {
				for (i=0; i<8; i++) {
					v = *blk++ + 128;
					if (v<0) v=0; else if (v>255) v=255;
					rfp[i] = (YUVPixel)v;
				}
				rfp+= iincr;
			}
		}
	}
}

static long get_luma_diff() {
    int size;
    // decode length
    int code = show_bits(9);	// 1 1111 ....
	if (code < 0x1F0) {
		code >>= 4;
        size = DClumtab0[code].val;
        flush_buffer(DClumtab0[code].len);
	} else {
		code &= 0xF;
		size = DClumtab1[code].val;
		flush_buffer(DClumtab1[code].len);
    }
    if (size > 0) {
        code = get_bits(size);      // maximum is 11 bits
        size = 1 << (size-1);
        if (code < size) code = (code+1) - (size+size);
    } else code = 0;
    return code;
}

static long get_chroma_diff() {
    int size;
    // decode length
    int code = show_bits(10);
	if (code < 0x3E0) {
		code >>= 5;
        size = DCchromtab0[code].val;
        flush_buffer(DCchromtab0[code].len);
	} else {
		code &= 0x1F;
        size = DCchromtab1[code].val;
        flush_buffer(DCchromtab1[code].len);
    }
    if (size > 0) {
        code = get_bits(size);      // maximum is 11 bits
        size = 1 << (size-1);
        if (code < size) code = (code+1) - (size+size);
    } else code = 0;
    return code;
}

// decode one intra coded MPEG-1 block
static __inline void Decode_MPEG1_Intra_Block(int comp) {
    long code,val,i,j,sign;
    const DCTtab *tab;

    short *blk = dct_coeff[comp];

    // decode DC coefficients
    if (comp < 4)           // Y
        val = (dct_dc_y_past += get_luma_diff());
    else if (comp & 1)      // V
        val = (dct_dc_v_past += get_chroma_diff());
    else                    // U
        val = (dct_dc_u_past += get_chroma_diff());

    blk[0] = (short)(val << 3);

    if (frame_type == MPEG_FRAME_TYPE_D) return;

    // decode AC coefficients
    for (i=1; ; i++) {
        code = show_bits(16);
        if (code>=16384)
            tab = &DCTtabnext[(code>>12)-4];
        else if (code>=1024)
            tab = &DCTtab0[(code>>8)-4];
        else if (code>=512)
            tab = &DCTtab1[(code>>6)-8];
        else if (code>=256)
            tab = &DCTtab2[(code>>4)-16];
        else if (code>=128)
            tab = &DCTtab3[(code>>3)-16];
        else if (code>=64)
            tab = &DCTtab4[(code>>2)-16];
        else if (code>=32)
            tab = &DCTtab5[(code>>1)-16];
        else if (code>=16)
            tab = &DCTtab6[code-16];
        else {
            fault_flag = true;
            break;
        }

        flush_buffer(tab->len);
        val = tab->run;

        if (val == 65) {    // escape
            i += get_bits(6);

            val = get_bits(8);
            if (val==0)
                val = get_bits(8);
            else if (val==128)
                val = get_bits(8) - 256;
            else if (val>128)
                val -= 256;

            if ((sign = (val<0))) val = -val;
        } else {
            if (val == 64) break;
            i+= val;
            val = tab->level;
            sign = get_bits(1);
        }
        if (i>=64) {
            fault_flag = true;
            break;
        }

        j = zigzag[0][i];
        val = (val * quantizer_scale * intramatrix[j]) >> 3;
        if (val) val = (val - 1) | 1;       // mismatch
        if (val >= 2048) val = 2047 + sign; // saturation
        if (sign) val = -val;
        blk[j] = (short)val;
    }
}

// decode one non-intra coded MPEG-1 block
static __inline void Decode_MPEG1_Non_Intra_Block(int comp) {
    long code,val,i,j,sign;
    const DCTtab *tab;

    short *blk = dct_coeff[comp];

    // decode AC coefficients
    for (i=0; ; i++) {
        code = show_bits(16);
        if (code>=16384)
            if (i)  tab = &DCTtabnext[(code>>12)-4];
            else    tab = &DCTtabfirst[(code>>12)-4];
        else if (code>=1024)
            tab = &DCTtab0[(code>>8)-4];
        else if (code>=512)
            tab = &DCTtab1[(code>>6)-8];
        else if (code>=256)
            tab = &DCTtab2[(code>>4)-16];
        else if (code>=128)
            tab = &DCTtab3[(code>>3)-16];
        else if (code>=64)
            tab = &DCTtab4[(code>>2)-16];
        else if (code>=32)
            tab = &DCTtab5[(code>>1)-16];
        else if (code>=16)
            tab = &DCTtab6[code-16];
        else {
            fault_flag = true;
            break;
        }

        flush_buffer(tab->len);
        val = tab->run;

        if (val == 65) {        // escape
            i+= get_bits(6);

            val = get_bits(8);
            if (val==0)
                val = get_bits(8);
            else if (val==128)
                val = get_bits(8) - 256;
            else if (val>128)
                val -= 256;

            if ((sign = (val<0))) val = -val;
        } else {
            if (val == 64) break;
            i+= val;
            val = tab->level;
            sign = get_bits(1);
        }
        if (i >= 64) {
            fault_flag = true;
            break;
        }

        j = zigzag[0][i];
        val = (((val<<1)+1) * quantizer_scale * nonintramatrix[j]) >> 4;
        if (val) val = (val - 1) | 1;       // mismatch
        if (val >= 2048) val = 2047 + sign; // saturation
        if (sign) val = -val;
        blk[j] = (short)val;
    }
}

// decode one intra coded MPEG-2 block
static __inline void Decode_MPEG2_Intra_Block(int comp) {
    long code,val,i,j,sign,sum;
    const DCTtab *tab;

    short *blk = dct_coeff[comp];

    unsigned char *qmat = (comp<4 || chroma_format==CHROMA420)?
        intramatrix: chromaintramatrix;

    // DC coefficient
    if (comp < 4)           // Y
        val = (dct_dc_y_past += get_luma_diff());
    else if (comp & 1)      // V
        val = (dct_dc_v_past += get_chroma_diff());
    else                    // U
        val = (dct_dc_u_past += get_chroma_diff());

    sum = val << (3 - intra_dc_precision);
    blk[0] = (short)sum;

    // AC coefficients
    for (i=1; ; i++) {
        code = show_bits(16);
        if (code>=16384)
            if (intra_vlc_format)   tab = &DCTtab0a[(code>>8)-4];
            else                    tab = &DCTtabnext[(code>>12)-4];
        else if (code>=1024)
            if (intra_vlc_format)   tab = &DCTtab0a[(code>>8)-4];
            else                    tab = &DCTtab0[(code>>8)-4];
        else if (code>=512)
            if (intra_vlc_format)   tab = &DCTtab1a[(code>>6)-8];
            else                    tab = &DCTtab1[(code>>6)-8];
        else if (code>=256)
            tab = &DCTtab2[(code>>4)-16];
        else if (code>=128)
            tab = &DCTtab3[(code>>3)-16];
        else if (code>=64)
            tab = &DCTtab4[(code>>2)-16];
        else if (code>=32)
            tab = &DCTtab5[(code>>1)-16];
        else if (code>=16)
            tab = &DCTtab6[code-16];
        else {
            fault_flag = true;
            break;
        }

        flush_buffer(tab->len);
        val = tab->run;

        if (val == 65) {    // escape
            i += get_bits(6);
            val = get_bits(12);
            if (!(val & 2047)) {
                fault_flag = true;
                break;
            }
            if ((sign = (val >= 2048))) val = 4096 - val;
        } else {
            if (val == 64) break;
            i += val;
            val = tab->level;
            sign = get_bits(1);
        }
        if (i >= 64) {
            fault_flag = true;
            break;
        }

        j = zigzag[alternate_scan][i];
        val = (val * quantizer_scale * qmat[j]) >> 4;
        if (val >= 2048) val = 2047 + sign; // saturation
        if (sign) val = -val;
        blk[j] = (short)val;
        sum ^= val;     // mismatch
    }

    if (!fault_flag)
        if (!(sum & 1)) blk[63] ^= 1;   // mismatch control
}

// decode one non-intra coded MPEG-2 block
static __inline void Decode_MPEG2_Non_Intra_Block(int comp) {
    long code,val,i,j,sign,sum;
    const DCTtab *tab;

    short *blk = dct_coeff[comp];

    unsigned char *qmat = (comp<4 || chroma_format==CHROMA420)?
        nonintramatrix: chromanonintramatrix;

    // AC coefficients
    sum = 0;
    for (i=0; ; i++) {
        code = show_bits(16);
        if (code>=16384)
            if (i)  tab = &DCTtabnext[(code>>12)-4];
            else    tab = &DCTtabfirst[(code>>12)-4];
        else if (code>=1024)
            tab = &DCTtab0[(code>>8)-4];
        else if (code>=512)
            tab = &DCTtab1[(code>>6)-8];
        else if (code>=256)
            tab = &DCTtab2[(code>>4)-16];
        else if (code>=128)
            tab = &DCTtab3[(code>>3)-16];
        else if (code>=64)
            tab = &DCTtab4[(code>>2)-16];
        else if (code>=32)
            tab = &DCTtab5[(code>>1)-16];
        else if (code>=16)
            tab = &DCTtab6[code-16];
        else {
            fault_flag = true;
            break;
        }

        flush_buffer(tab->len);
        val = tab->run;

        if (val == 65) {    // escape
            i += get_bits(6);
            val = get_bits(12);
            if (!(val & 2047)) {
                fault_flag = true;
                break;
            }
            if ((sign = (val >= 2048))) val = 4096 - val;
        } else {
            if (val == 64) break;
            i += val;
            val = tab->level;
            sign = get_bits(1);
        }
        if (i >= 64) {
            fault_flag = true;
            break;
        }

        j = zigzag[alternate_scan][i];
        val = (((val<<1)+1) * quantizer_scale * qmat[j]) >> 5;
        if (val >= 2048) val = 2047 + sign; // saturation
        if (sign) val = -val;
        blk[j] = (short)val;
        sum ^= val;             // mismatch
    }

    if (!fault_flag)
        if (!(sum & 1)) blk[63] ^= 1;   // mismatch control
}


///////////////////////////
// Prediction (replaces a_predict* code)

static void form_component_prediction(YUVPixel* src,
  YUVPixel* dst, int lx, int w, int h,
  int x, int y, int dx, int dy, int average_flag) {
    int i,j,v;

    src += lx * (y + (dy >> 1)) + x + (dx >> 1);
    dst += lx * y + x;

/*  The following are equivalent:

a)  result = (s1 + s2 + 1) >> 1;
b)  result = (s1 >> 1) + (s2 >> 1) + ((s1 | s2) & 1);

a)  result = (d + ((s1 + s2 + 1) >> 1) + 1) >> 1;
b)  result = (s1 + s2 + d + d + 3) >> 2;

a)  result = (d + ((s1 + s2 + s3 + s4 + 2) >> 2) + 1) >> 1;
b)  result = (s1 + s2 + s3 + s4 + d + d + d + d + 6) >> 3;
*/

    if (dx & 1) {
        if (dy & 1) {
            // horizontal and vertical half-pel
            if (average_flag) {
                for (j=0; j<h; j++) {
                    for (i=0; i<w; i++) {
                        v = src[i] + src[i+1] + src[i+lx] + src[i+lx+1];
                        v = dst[i] + ((v+2)>>2);
                        dst[i] = (v+1)>>1;
                    }
                    src+= lx;
                    dst+= lx;
                }
            } else {
                for (j=0; j<h; j++) {
                    for (i=0; i<w; i++) {
                        v = src[i] + src[i+1] + src[i+lx] + src[i+lx+1];
                        dst[i] = (v+2)>>2;
                    }
                    src+= lx;
                    dst+= lx;
                }
            }
        } else {
            // horizontal but no vertical half-pel
            if (average_flag) {
                for (j=0; j<h; j++) {
                    for (i=0; i<w; i++) {
                        v = src[i] + src[i+1];
                        v = dst[i] + ((v+1)>>1);
                        dst[i] = (v+1)>>1;
                    }
                    src+= lx;
                    dst+= lx;
                }
            } else {
                for (j=0; j<h; j++) {
                    for (i=0; i<w; i++) {
                        v = src[i] + src[i+1];
                        dst[i] = (v+1)>>1;
                    }
                    src+= lx;
                    dst+= lx;
                }
            }
        }
    } else if (dy & 1) {    // no horizontal but vertical half-pel
        if (average_flag) {
            for (j=0; j<h; j++) {
                for (i=0; i<w; i++) {
                    v = src[i] + src[i+lx];
                    v = dst[i] + ((v+1)>>1);
                    dst[i]=(v+1)>>1;
                }
                src+= lx;
                dst+= lx;
            }
        } else {
            for (j=0; j<h; j++) {
                for (i=0; i<w; i++) {
                    v = src[i] + src[i+lx];
                    dst[i] = (v+1)>>1;
                }
                src+= lx;
                dst+= lx;
            }
        }
    } else {    // no horizontal nor vertical half-pel
        if (average_flag) {
            for (j=0; j<h; j++) {
                for (i=0; i<w; i++) {
                    v = dst[i]+src[i];
                    dst[i] = (v+1)>>1;
                }
                src+= lx;
                dst+= lx;
            }
        } else {
            for (j=0; j<h; j++) {
                for (i=0; i<w; i++) {
                    dst[i] = src[i];
                }
                src+= lx;
                dst+= lx;
            }
        }
    }
}

static void form_prediction(MPEGBuffer *srcBuf, int sfield,
  MPEGBuffer *dstBuf, int dfield, int lx, int w,
  int h, int x, int y, int dx, int dy, int average_flag) {

// Note: lx is always y_pitch or y_pitch*2
// Note: h is always 8 or 16
// Note: w is always 16

    if (sfield) sfield = lx >> 1;
    if (dfield) dfield = lx >> 1;

	if (MMX_enabled) {
			// Y
		mmx_form_component_prediction(
		  srcBuf->Y+sfield,dstBuf->Y+dfield,
		  lx,w,h,x,y,dx,dy,average_flag);

		if (chroma_format != CHROMA444) {
			sfield>>=1; dfield>>=1;
			lx>>=1; w>>=1; x>>=1; dx/=2;
			if (chroma_format == CHROMA420) {
				h>>=1; y>>=1; dy/=2;
			}
		}

		// Cb
		mmx_form_component_prediction(
		  srcBuf->U+sfield,dstBuf->U+dfield,
		  lx,w,h,x,y,dx,dy,average_flag);

		// Cr
		mmx_form_component_prediction(
		  srcBuf->V+sfield,dstBuf->V+dfield,
		  lx,w,h,x,y,dx,dy,average_flag);

	} else {

		// Y
		form_component_prediction(
		  srcBuf->Y+sfield,dstBuf->Y+dfield,
		  lx,w,h,x,y,dx,dy,average_flag);

		if (chroma_format != CHROMA444) {
			sfield>>=1; dfield>>=1;
			lx>>=1; w>>=1; x>>=1; dx/=2;
			if (chroma_format == CHROMA420) {
				h>>=1; y>>=1; dy/=2;
			}
		}

		// Cb
		form_component_prediction(
		  srcBuf->U+sfield,dstBuf->U+dfield,
		  lx,w,h,x,y,dx,dy,average_flag);

		// Cr
		form_component_prediction(
		  srcBuf->V+sfield,dstBuf->V+dfield,
		  lx,w,h,x,y,dx,dy,average_flag);
	}
}

static void dual_prime(int DMV[][2], int mvx, int mvy) {
    if (picture_structure==FRAME_PICTURE) {
        //if (top_field_first) {
		if (current_frame->bitflags & PICTUREFLAG_TFF) {
            DMV[0][0] = ((mvx  +(mvx>0))>>1) + dmvector[0];
            DMV[0][1] = ((mvy  +(mvy>0))>>1) + dmvector[1] - 1;
            DMV[1][0] = ((3*mvx+(mvx>0))>>1) + dmvector[0];
            DMV[1][1] = ((3*mvy+(mvy>0))>>1) + dmvector[1] + 1;
        } else {
            DMV[0][0] = ((3*mvx+(mvx>0))>>1) + dmvector[0];
            DMV[0][1] = ((3*mvy+(mvy>0))>>1) + dmvector[1] - 1;
            DMV[1][0] = ((mvx  +(mvx>0))>>1) + dmvector[0];
            DMV[1][1] = ((mvy  +(mvy>0))>>1) + dmvector[1] + 1;
        }
    } else {
        DMV[0][0] = ((mvx+(mvx>0))>>1) + dmvector[0];
        DMV[0][1] = ((mvy+(mvy>0))>>1) + dmvector[1];
        if (picture_structure==TOP_FIELD) {
            DMV[0][1]--;
        } else {
            DMV[0][1]++;
        }
    }
}

#define forw_frame (&(buffers[MPEG_BUFFER_FORWARD]))
#define back_frame (&(buffers[MPEG_BUFFER_BACKWARD]))

static void form_predictions(int bx, int by) {
    MPEGBuffer *predframe;
    int DMV[2][2];
	int currentfield;
    int addflag = 0;

    if ((macro_block_flags & MBF_FORWARD) || (frame_type==MPEG_FRAME_TYPE_P)) {
        if (picture_structure==FRAME_PICTURE) {
            if ((motion_type==MC_FRAME) || !(macro_block_flags & MBF_FORWARD)) {
                // frame-based prediction
                form_prediction(forw_frame,0,current_frame,0,
                  y_pitch,16,16,bx,by,
                  PMV[0][0][0],PMV[0][0][1],addflag);

            } else if (motion_type==MC_FIELD) { // field-based prediction
                // top field prediction
                form_prediction(forw_frame,mvfield_select[0][0],
                  current_frame,0,y_pitch<<1,16,8,
                  bx,by>>1,PMV[0][0][0],PMV[0][0][1]>>1,addflag);
                // bottom field prediction
                form_prediction(forw_frame,mvfield_select[1][0],
                  current_frame,1,y_pitch<<1,16,8,
                  bx,by>>1,PMV[1][0][0],PMV[1][0][1]>>1,addflag);

            } else if (motion_type==MC_DMV) {   // dual prime prediction
                // calculate derived motion vectors
                dual_prime(DMV,PMV[0][0][0],PMV[0][0][1]>>1);

                // predict top field from top field
                form_prediction(forw_frame,0,current_frame,0,
                  y_pitch<<1,16,8,bx,by>>1,
                  PMV[0][0][0],PMV[0][0][1]>>1,0);
                // predict and add to top field from bottom field
                form_prediction(forw_frame,1,current_frame,0,
                  y_pitch<<1,16,8,bx,by>>1,
                  DMV[0][0],DMV[0][1],1);

                // predict bottom field from bottom field
                form_prediction(forw_frame,1,current_frame,1,
                  y_pitch<<1,16,8,bx,by>>1,
                  PMV[0][0][0],PMV[0][0][1]>>1,0);
                // predict and add to bottom field from top field */
                form_prediction(forw_frame,0,current_frame,1,
                  y_pitch<<1,16,8,bx,by>>1,
                  DMV[1][0],DMV[1][1],1);
            }

        } else {    // TOP_FIELD or BOTTOM_FIELD
            // field picture
            currentfield = (picture_structure==BOTTOM_FIELD);

            // determine which frame to use for prediction
            if ((frame_type==MPEG_FRAME_TYPE_P) && second_field
              && (currentfield!=mvfield_select[0][0])) {
                predframe = back_frame;     // same frame
            } else {
                predframe = forw_frame;     // previous frame
            }

            if ((motion_type==MC_FIELD)
              || !(macro_block_flags & MBF_FORWARD)) {
                // field-based prediction
                form_prediction(predframe,mvfield_select[0][0],current_frame,currentfield,
                  y_pitch<<1,16,16,bx,by,
                  PMV[0][0][0],PMV[0][0][1],addflag);

            } else if (motion_type==MC_16X8) {
                form_prediction(predframe,mvfield_select[0][0],current_frame,currentfield,
                  y_pitch<<1,16,8,bx,by,
                  PMV[0][0][0],PMV[0][0][1],addflag);

                // determine which frame to use for lower half prediction
                if ((frame_type==MPEG_FRAME_TYPE_P) && second_field
                  && (currentfield!=mvfield_select[1][0])) {
                    predframe = back_frame;     // same frame
                } else {
                    predframe = forw_frame;     // previous frame
                }

                form_prediction(predframe,mvfield_select[1][0],current_frame,currentfield,
                  y_pitch<<1,16,8,bx,by+8,
                  PMV[1][0][0],PMV[1][0][1],addflag);

            } else if (motion_type==MC_DMV) {   // dual prime prediction
                if (second_field) {
                    predframe = back_frame;     // same frame
                } else {
                    predframe = forw_frame;     // previous frame
                }

                // calculate derived motion vectors
                dual_prime(DMV,PMV[0][0][0],PMV[0][0][1]);

                // predict from field of same parity
                form_prediction(forw_frame,currentfield,current_frame,currentfield,
                  y_pitch<<1,16,16,bx,by,
                  PMV[0][0][0],PMV[0][0][1],0);

                // predict from field of opposite parity
                form_prediction(predframe,!currentfield,current_frame,currentfield,
                  y_pitch<<1,16,16,bx,by,
                  DMV[0][0],DMV[0][1],1);
            }
        }
        addflag = 1;
    }

    if (macro_block_flags & MBF_BACKWARD) {
        if (picture_structure==FRAME_PICTURE) {
            if (motion_type==MC_FRAME) {
                // frame-based prediction
                form_prediction(back_frame,0,current_frame,0,
                  y_pitch,16,16,bx,by,
                  PMV[0][1][0],PMV[0][1][1],addflag);

            } else {    // field-based prediction
                // top field prediction
                form_prediction(back_frame,mvfield_select[0][1],
                  current_frame,0,y_pitch<<1,16,8,
                  bx,by>>1,PMV[0][1][0],PMV[0][1][1]>>1,addflag);

                // bottom field prediction
                form_prediction(back_frame,mvfield_select[1][1],
                  current_frame,1,y_pitch<<1,16,8,
                  bx,by>>1,PMV[1][1][0],PMV[1][1][1]>>1,addflag);
            }

        } else {    // TOP_FIELD or BOTTOM_FIELD
            // field picture
            currentfield = (picture_structure==BOTTOM_FIELD);

            if (motion_type==MC_FIELD) {
                // field-based prediction
                form_prediction(back_frame,mvfield_select[0][1],
                  current_frame,currentfield,y_pitch<<1,16,16,
                  bx,by,PMV[0][1][0],PMV[0][1][1],addflag);

            } else if (motion_type==MC_16X8) {
                form_prediction(back_frame,mvfield_select[0][1],
                  current_frame,currentfield,y_pitch<<1,16,8,
                  bx,by,PMV[0][1][0],PMV[0][1][1],addflag);

                form_prediction(back_frame,mvfield_select[1][1],
                  current_frame,currentfield,y_pitch<<1,16,8,
                  bx,by+8,PMV[1][1][0],PMV[1][1][1],addflag);
            }
        }
    }
}

///////////////////////////

static int mpeg_get_motion_component() {
    if (get_bits(1) == 0) {
        int code = show_bits(9);

        if (code >= 64) {
            code >>= 6;
            flush_buffer(MVtab0[code].len);
            code = MVtab0[code].val;
        } else if (code >= 24) {
            code >>= 3;
            flush_buffer(MVtab1[code].len);
            code = MVtab1[code].val;
        } else if (code >= 12) {
            code -= 12;
            flush_buffer(MVtab2[code].len);
            code = MVtab2[code].val;
        } else {
            fault_flag = true;
            return 0;
        }
        if (get_bits(1)) code = -code;
        return code;
    }
    return 0;
}

static int get_mb_address_inc() {
    int code, val;

    for (val = 0;;) {
        code = show_bits(11);
        if (code >= 24) break;
        if (code != 15) {   // if not macroblock_stuffing
            if (code == 8)  // if macroblock_escape
                val += 33;
            else {
                fault_flag = true;
                return 1;
            }
        }
        flush_buffer(11);
    }

    if (code >= 1024) {
        flush_buffer(1);
        val++;
    } else if (code >= 128) {
        code >>= 6;
        flush_buffer(MBAtab0[code].len);
        val += MBAtab0[code].val;
    } else {
        code-= 24;
        flush_buffer(MBAtab1[code].len);
        val += MBAtab1[code].val;
    }
    return val;
}

static __inline int get_I_mb_flags() {
// the only valid codes are 1, or 01
    int code = show_bits(2);
    if (code & 2) {
        flush_buffer(1);
        return MBF_INTRA;
    }
    if (code & 1) {
        flush_buffer(2);
    } else {
        fault_flag = true;
    }
    return (MBF_INTRA|MBF_QUANT);
}

static __inline int get_P_mb_flags() {
    int code = show_bits(6);
    if (code>=8) {
        code>>=3;
        flush_buffer(PMBtab0[code].len);
        code = PMBtab0[code].val;
    } else if (code) {
        flush_buffer(PMBtab1[code].len);
        code = PMBtab1[code].val;
    } else fault_flag = true;
    return(code);
}

static __inline int get_B_mb_flags() {
    int code = show_bits(6);
    if (code>=8) {
        code>>=2;
        flush_buffer(BMBtab0[code].len);
        code = BMBtab0[code].val;
    } else if (code) {
        flush_buffer(BMBtab1[code].len);
        code = BMBtab1[code].val;
    } else fault_flag = true;
    return(code);
}

static __inline int get_D_mb_flags() {
    if (get_bits(1))
        flush_buffer(1);
    else fault_flag = true;
    return(MBF_INTRA);
}

static __inline int get_mb_flags() {
    if (frame_type == MPEG_FRAME_TYPE_I) return get_I_mb_flags();
    if (frame_type == MPEG_FRAME_TYPE_P) return get_P_mb_flags();
    if (frame_type == MPEG_FRAME_TYPE_B) return get_B_mb_flags();
    if (frame_type == MPEG_FRAME_TYPE_D) return get_D_mb_flags();
    fault_flag = true;
    return 0;
}

static __inline int get_mb_cbp() {
    int code = show_bits(9);
    if (code >= 128) {
        code >>= 4;
        flush_buffer(CBPtab0[code].len);
        code = CBPtab0[code].val;
    } else if (code >= 8) {
        code >>= 1;
        flush_buffer(CBPtab1[code].len);
        code = CBPtab1[code].val;
    } else if (code) {
        flush_buffer(CBPtab2[code].len);
        code = CBPtab2[code].val;
    } else fault_flag = true;
    return code;
}

// calculate motion vector component
static void decode_motion_vector(long *pred, int r_size,
  long code, long residual, int full_pel) {
    long lim, vec;

    lim = 16 << r_size;
    vec = full_pel? (*pred >> 1): (*pred);

    if (code > 0) {
        vec += ((code - 1) << r_size) + residual + 1;
        if (vec >= lim) vec -= lim + lim;
    } else if (code < 0) {
        vec -= ((-code - 1) << r_size) + residual + 1;
        if (vec < -lim) vec += lim + lim;
    }
    *pred = full_pel? (vec << 1): vec;
}

static void motion_vector(long* pmv,
  int h_size, int v_size, int full_pel) {
    long code, residual;

    // horizontal component (maximum h_size is 14 bits)

    code = mpeg_get_motion_component();
    residual = (h_size>0 && code!=0)? get_bits(h_size): 0;
    decode_motion_vector(&pmv[0],h_size,code,residual,full_pel);

    if (motion_type==MC_DMV)
        dmvector[0] = get_bits(1)? (get_bits(1)? -1: 1): 0;

    // vertical component (maximum v_size is 14 bits)
    code = mpeg_get_motion_component();
    residual = (v_size>0 && code!=0)? get_bits(v_size): 0;
    if (mvscale) pmv[1] >>= 1;
    decode_motion_vector(&pmv[1],v_size,code,residual,full_pel);
    if (mvscale) pmv[1] <<= 1;

    if (motion_type==MC_DMV)
        dmvector[1] = get_bits(1)? (get_bits(1)? -1: 1): 0;
}

static void motion_vectors(int s, int h_size, int v_size) {
    if (motion_vector_count == 1) {
        if (mv_format==MV_FIELD && motion_type!=MC_DMV) {
            mvfield_select[1][s] = mvfield_select[0][s] = get_bits(1);
        }
        motion_vector(PMV[0][s],h_size,v_size,0);

        PMV[1][s][0] = PMV[0][s][0];
        PMV[1][s][1] = PMV[0][s][1];
    } else {    // motion_vector_count == 2
        mvfield_select[0][s] = get_bits(1);
        motion_vector(PMV[0][s],h_size,v_size,0);

        mvfield_select[1][s] = get_bits(1);
        motion_vector(PMV[1][s],h_size,v_size,0);
    }
}

static __inline void decode_mb(int MBA) {
    int pos_x, pos_y;
    int comp, cbp;

    macro_block_flags = get_mb_flags();
    if (fault_flag) return;

    // get frame/field motion type
    if (macro_block_flags & (MBF_FORWARD|MBF_BACKWARD)) {
        if (picture_structure==FRAME_PICTURE) {
            motion_type = frame_pred_frame_dct? MC_FRAME: get_bits(2);
        } else motion_type = get_bits(2);
    } else if ((macro_block_flags & MBF_INTRA) && concealment_vectors) {
        motion_type = (picture_structure==FRAME_PICTURE)? MC_FRAME: MC_FIELD;
    } else motion_type = 0;  // implied

    // derive motion_vector_count, mv_format
    if (picture_structure==FRAME_PICTURE) {
        motion_vector_count = (motion_type==MC_FIELD)? 2: 1;
        mv_format = (motion_type==MC_FRAME)? MV_FRAME: MV_FIELD;
    } else {
        motion_vector_count = (motion_type==MC_16X8)? 2: 1;
        mv_format = MV_FIELD;
    }

    mvscale = ((mv_format==MV_FIELD)
      && (picture_structure==FRAME_PICTURE));

    // get dct_type (frame DCT / field DCT)
    dct_type = (picture_structure==FRAME_PICTURE)
      && (!frame_pred_frame_dct)
      && (macro_block_flags & (MBF_PATTERN|MBF_INTRA))
      ? get_bits(1): 0;

    if (macro_block_flags & MBF_QUANT) {
        quantizer_scale = get_bits(5);
        if (MPEG2_Flag) {
            if (q_scale_type) {
                quantizer_scale = non_linear[quantizer_scale];
            } else quantizer_scale <<= 1;
        }
    }

    // decode forward motion vectors
    if ((macro_block_flags & MBF_FORWARD) ||
      ((macro_block_flags & MBF_INTRA) && concealment_vectors)) {
        if (MPEG2_Flag)
            motion_vectors(0,f_code[0][0],f_code[0][1]);
        else
            motion_vector(PMV[0][0],f_code[0][0],
              f_code[0][0],forw_vector_full_pel);
        if (fault_flag) return;
    }

    // decode backward motion vectors
    if (macro_block_flags & MBF_BACKWARD) {
        if (MPEG2_Flag)
            motion_vectors(1,f_code[1][0],f_code[1][1]);
        else
            motion_vector(PMV[0][1],f_code[1][0],
              f_code[1][0],back_vector_full_pel);
        if (fault_flag) return;
    }

    if ((macro_block_flags & MBF_INTRA) && concealment_vectors)
        flush_buffer(1);    // remove marker_bit

    if (macro_block_flags & MBF_PATTERN) {
        cbp = get_mb_cbp();
        if (fault_flag) return;
        if (chroma_format == CHROMA422)
            cbp = (cbp << 2) | get_bits(2);
        else if (chroma_format == CHROMA444)
            cbp = (cbp << 6) | get_bits(6);
    } else
        cbp = (macro_block_flags & MBF_INTRA)?
            (1 << block_count)-1: 0;

    // decode blocks
    memset(dct_coeff, 0, sizeof(short)*64*block_count);
    cbp <<= (32-block_count);
    for (comp=0; comp<block_count; comp++) {
        if (cbp & 0x80000000) {
            if (macro_block_flags & MBF_INTRA) {
                if (MPEG2_Flag)
                    Decode_MPEG2_Intra_Block(comp);
                else
                    Decode_MPEG1_Intra_Block(comp);
            } else {
                if (MPEG2_Flag)
                    Decode_MPEG2_Non_Intra_Block(comp);
                else
                    Decode_MPEG1_Non_Intra_Block(comp);
            }
            if (fault_flag) return;

			if (ISSE_enabled)
				sse_IDCT(dct_coeff[comp]);
            else if (MMX_enabled)
                mmx_IDCT(dct_coeff[comp]);
            else
                Fast_IDCT(dct_coeff[comp]);
        }
        cbp<<=1;
    }

    if (frame_type == MPEG_FRAME_TYPE_D) flush_buffer(1);

    // reset intra_dc predictors
    if (!(macro_block_flags & MBF_INTRA))
        dct_dc_y_past = dct_dc_u_past = dct_dc_v_past = 0;
    else if (!concealment_vectors) {
        PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;
        PMV[0][1][0]=PMV[0][1][1]=PMV[1][1][0]=PMV[1][1][1]=0;
    }

    // special "No_MC" macroblock_type case
    if ((frame_type == MPEG_FRAME_TYPE_P)
      && !(macro_block_flags & (MBF_FORWARD|MBF_INTRA))) {
        PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;

        if (picture_structure==FRAME_PICTURE)
            motion_type = MC_FRAME;
        else {
            motion_type = MC_FIELD;
            mvfield_select[0][0] = (picture_structure==BOTTOM_FIELD);
        }
    }

    // derive current position
    pos_x = 16 * (MBA % mbWidth);
    pos_y = 16 * (MBA / mbWidth);

	cbp = ((macro_block_flags & MBF_INTRA) == 0);

    // motion compensation
    try {
        if (cbp)	//(!(macro_block_flags & MBF_INTRA))
            form_predictions(pos_x,pos_y);

        // copy or add block data into picture
        for (comp=0; comp<block_count; comp++)
            Add_Block(comp,pos_x,pos_y,cbp); //(macro_block_flags & MBF_INTRA)==0);

    } catch(...) {
        // Ignore r/w errors
    }
}

static __inline void skipped_mb(int MBA) {
    int pos_x, pos_y;
    int comp;

    memset(dct_coeff, 0, sizeof(short)*64*block_count);
    // reset intra_dc predictors
    dct_dc_y_past = dct_dc_u_past = dct_dc_v_past = 0;
    // reset motion vector predictors
    if (frame_type==MPEG_FRAME_TYPE_P)
        PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;

    // derive motion_type
    if (picture_structure==FRAME_PICTURE)
        motion_type = MC_FRAME;
    else {
        motion_type = MC_FIELD;
        // predict from field of same parity
        mvfield_select[0][0]=mvfield_select[0][1] =
            (picture_structure==BOTTOM_FIELD);
    }

    // skipped macroblock in I-frame is not supported
    if (frame_type == MPEG_FRAME_TYPE_I) {
        fault_flag = 1;
        return;
    }

    // derive current position
    pos_x = 16 * (MBA % mbWidth);
    pos_y = 16 * (MBA / mbWidth);

    // motion compensation
    try {
        form_predictions(pos_x,pos_y);

        // copy block data into picture
        for (comp=0; comp<block_count; comp++)
            Add_Block(comp,pos_x,pos_y,1);

    } catch(...) {
        // Ignore r/w errors
    }
}

static __inline void slice(int type) {
    int MBA, MBAInc, MBAMax;
    int vert_ext;

    vert_ext = (MPEG2_Flag && (pelHeight > 2800))? get_bits(3): 0;

    quantizer_scale = get_bits(5);
    if (MPEG2_Flag) {
        if (q_scale_type)
            quantizer_scale = non_linear[quantizer_scale];
        else quantizer_scale <<= 1;
    }

    while(get_bits(1)) flush_buffer(8);

    fault_flag = false;
    MBAInc = get_mb_address_inc();
    if (fault_flag) return;
    MBA = ((vert_ext << 7) + (type & 0xFF) - 1) * mbWidth + MBAInc - 1;
    MBAInc = 1;     // first macroblock, not skipped
    MBAMax = mbWidth * mbHeight;

    dct_dc_y_past = dct_dc_u_past = dct_dc_v_past = 0;
    PMV[0][0][0]=PMV[0][0][1]=PMV[1][0][0]=PMV[1][0][1]=0;
    PMV[0][1][0]=PMV[0][1][1]=PMV[1][1][0]=PMV[1][1][1]=0;

    while (MBA < MBAMax) {
        if (!MBAInc) {
            MBAInc = get_mb_address_inc();
            if (fault_flag) break;
        }
        if (MBAInc == 1)
            decode_mb(MBA);
        else
            skipped_mb(MBA);

        if (fault_flag) break;
        MBAInc--;
        MBA++;
    }
}

/*************************************************************

    MPEG-1 and MPEG-2 YCbCr to RGB/YUY2/UYVY conversions

*************************************************************/

// Color conversion table (see Table 6-9 of ISO/IEC 13818-2)
// Inverse coefficients are represented as 16:16 fixed point

#define Y_MULT	76309UL		// (255/219)*65536

static const long Inverse_Table_6_9[8][7] = {
//  crv,    cbu,     cgu,    cgv,    radd,     gadd,     badd
  {117504, 138453, -13954, -34903, -16228689, 5065481, -18910197}, // no sequence_display_extension
  {117504, 138453, -13954, -34903, -16228689, 5065481, -18910197}, // ITU-R Rec. 709 (1990)
  {104597, 132201, -25675, -53279, -14576620, 8917830, -18109946}, // unspecified
  {104597, 132201, -25675, -53279, -14576620, 8917830, -18109946}, // reserved
  {104448, 132798, -24759, -53109, -14557520, 8778945, -18186343}, // FCC
  {104597, 132201, -25675, -53279, -14576620, 8917830, -18109946}, // ITU-R Rec. 624-4 System B, G
  {104597, 132201, -25675, -53279, -14576620, 8917830, -18109946}, // SMPTE 170M
  {117579, 136230, -16907, -35559, -16238238, 5527474, -18625621}  // SMPTE 240M (1987)
};

// Confirmed range is -289 to 547, myclipper uses offset -290
static const unsigned char myclipper[848] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,1,2,3,4,5,6,7,8,9,10,11,12,13,
    14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,
    30,31,32,33,34,35,36,37,38,39,40,41,42,43,44,45,
    46,47,48,49,50,51,52,53,54,55,56,57,58,59,60,61,
    62,63,64,65,66,67,68,69,70,71,72,73,74,75,76,77,
    78,79,80,81,82,83,84,85,86,87,88,89,90,91,92,93,
    94,95,96,97,98,99,100,101,102,103,104,105,106,107,108,109,
    110,111,112,113,114,115,116,117,118,119,120,121,122,123,124,125,
    126,127,128,129,130,131,132,133,134,135,136,137,138,139,140,141,
    142,143,144,145,146,147,148,149,150,151,152,153,154,155,156,157,
    158,159,160,161,162,163,164,165,166,167,168,169,170,171,172,173,
    174,175,176,177,178,179,180,181,182,183,184,185,186,187,188,189,
    190,191,192,193,194,195,196,197,198,199,200,201,202,203,204,205,
    206,207,208,209,210,211,212,213,214,215,216,217,218,219,220,221,
    222,223,224,225,226,227,228,229,230,231,232,233,234,235,236,237,
    238,239,240,241,242,243,244,245,246,247,248,249,250,251,252,253,
	254,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,
	255,255,255,255,255,255,255,255,255,255,255,255,255,255,255,255
};

// Another little clip table for 16-bit conversion
// Confirmed range is -37 to 68, myclipper2 uses offset -40
static const unsigned char myclipper2[112] = {
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
    0,0,0,0,0,0,0,0,0,1,2,3,4,5,6,7,
    8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,
    24,25,26,27,28,29,30,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,
    31,31,31,31,31,31,31,31,31,31,31,31,31,31,31,31
};

// Upsample 4:2:0 chroma to 4:2:2
// (See ISO/IEC 13818-2 section 6.1.2.1)

static void mpeg_conv420to422(YUVPixel *src, YUVPixel *dst,
  long w, long h, bool progressive) {

	YUVPixel *sm1, *sp1;
	int x,y;

	if (progressive) {
		// MPEG-1 or MPEG-2 progressive frame

		h -= 1;
        for (y = 0; y <= h; y++) {
            sm1 = (y > 0)? src-w: src;
            sp1 = (y < h)? src+w: src;
            for (x = 0; x < w; x++) {
                YUVPixel s = *src++;
                dst[w] = (YUVPixel)((s + *sp1++ + 1) >> 1);
                *dst++ = (YUVPixel)((s + *sm1++ + 1) >> 1);
            }
            dst += w;
        }

	} else {
		// MPEG-2 interlaced frame
		int ym1, y1, yp1;

		h = (h >> 1) - 1;

		for (y = 0; y <= h; y++) {
			// top field
			sm1 = (y > 0)? src-w*2: src;
			sp1 = (y < h)? src+w*2: src;
			for (x = 0; x < w; x++) {
				ym1 = *sm1++;
				y1  = *src++;
				yp1 = *sp1++;
				dst[w*2] = (YUVPixel)((y1*5 + yp1*3 + 4) >> 3);
				*dst++ = (YUVPixel)((y1*7 + ym1 + 4) >> 3);
			}

			// bottom field
			sm1 = (y > 0)? src-w*2: src;
			sp1 = (y < h)? src+w*2: src;
			for (x = 0; x < w; x++) {
				ym1 = *sm1++;
				y1  = *src++;
				yp1 = *sp1++;
				dst[w*2] = (YUVPixel)((y1*7 + yp1 + 4) >> 3);
				*dst++ = (YUVPixel)((y1*5 + ym1*3 + 4) >> 3);
			}

			dst += w*2;
		}

	}
}

// Upsample 4:2:2 chroma to 4:4:4
// (See ISO/IEC 13818-2 section 6.1.2.2)

static void mpeg_conv422to444(YUVPixel *src, YUVPixel *dst,
  long w, long h, int mpeg2) {

	int xm1, x1, xp1;
	int x,y;

	if (!mpeg2) {

		for (y = 0; y < h; y++) {
			x1 = xp1 = *src++;
			for (x = 1; x < w; x++) {
				xm1 = x1;
				x1 = xp1;
				xp1 = *src++;
				*dst++ = (YUVPixel)((xm1 + x1 + 1) >> 1);
				*dst++ = (YUVPixel)((x1 + xp1 + 1) >> 1);
			}
			*dst++ = (YUVPixel)((x1 + xp1 + 1) >> 1);
			*dst++ = xp1;
		}

	} else {

		for (y = 0; y < h; y++) {
			xp1 = *src++;
			for (x = 1; x < w; x++) {
				x1 = xp1;
				xp1 = *src++;
				*dst++ = (YUVPixel)x1;
				*dst++ = (YUVPixel)((x1 + xp1 + 1) >> 1);
			}
			*dst++ = (YUVPixel)xp1;
			*dst++ = (YUVPixel)xp1;
		}

	}
}

// Convert YCbCr 4:4:4 to RGB 16,24,32
// (See ISO/IEC 13818-2 section 6.3.6)

static __inline void mpeg_conv444toRGB(YUVPixel *srcY, YUVPixel *srcU,
  YUVPixel *srcV, void *_dst, int bpr, int w, int h, int bpp) {

    const long crv = Inverse_Table_6_9[matrix_coefficients][0];
    const long cbu = Inverse_Table_6_9[matrix_coefficients][1];
    const long cgu = Inverse_Table_6_9[matrix_coefficients][2];
    const long cgv = Inverse_Table_6_9[matrix_coefficients][3];
    const long radd = Inverse_Table_6_9[matrix_coefficients][4];
    const long gadd = Inverse_Table_6_9[matrix_coefficients][5];
    const long badd = Inverse_Table_6_9[matrix_coefficients][6];

    unsigned char const *clipper = myclipper+290;
	unsigned char const *clipper2 = myclipper2+40;

    long ht,y,u,v;
	int x;

	ht = (long)h * (long)bpr;

	switch (bpp) {

		case 16: {
		// 16-bit color is going to look crappy no matter what
			unsigned short *dst;
			unsigned short i;

			while (ht > 0) {
				ht -= bpr;
				dst = (unsigned short *)((char *)_dst + ht);

				for (x = 0; x < w; x++) {
					y = srcY[x] * Y_MULT;
					u = srcU[x];
					v = srcV[x];

					i = clipper2[(y + v*crv + radd) >> 19];
					i = (i << 5) + clipper2[(y + u*cgu + v*cgv + gadd) >> 19];
					*dst++ = (i << 5) + clipper2[(y + u*cbu + badd) >> 19];
				}

				srcY += y_pitch;
				srcU += y_pitch;
				srcV += y_pitch;
			}
		}
		break;

		case 24: {
			unsigned char *dst;

			while (ht > 0) {
				ht -= bpr;
				dst = (unsigned char *)_dst + ht;

				for (x = 0; x < w; x++) {
					y = srcY[x] * Y_MULT;
					u = srcU[x];
					v = srcV[x];

					*dst++ = clipper[(y + u*cbu + badd) >> 16];
					*dst++ = clipper[(y + u*cgu + v*cgv + gadd) >> 16];
					*dst++ = clipper[(y + v*crv + radd) >> 16];
				}

				srcY += y_pitch;
				srcU += y_pitch;
				srcV += y_pitch;
			}
		}
		break;

		case 32: {
			unsigned long *dst;
			unsigned long i;

			while (ht > 0) {
				ht -= bpr;
				dst = (unsigned long *)((char *)_dst + ht);

				for (x = 0; x < w; x++) {
					y = srcY[x] * Y_MULT;
					u = srcU[x];
					v = srcV[x];

					i = clipper[(y + v*crv + radd) >> 16];
					i = (i << 8) + clipper[(y + u*cgu + v*cgv + gadd) >> 16];
					*dst++ = (i << 8) + clipper[(y + u*cbu + badd) >> 16];
				}

				srcY += y_pitch;
				srcU += y_pitch;
				srcV += y_pitch;
			}
		}
		break;
	}
}

// Convert YCbCr to RGB 16,24,32

static void mpeg_YUVToRGB(int buffer_ID, void *dst,
	long bpr, long w, long h, int bpp) {

	YUVPixel *Y_ptr = buffers[buffer_ID].Y;
	YUVPixel *U_ptr = buffers[buffer_ID].U;
	YUVPixel *V_ptr = buffers[buffer_ID].V;

	YUVPixel *u444ptr = u444;
	YUVPixel *v444ptr = v444;

	long cpw = (pelWidth + 15) & -16;
	long cph = (pelHeight + 15) & -16;

	bool pf = !!(buffers[buffer_ID].bitflags & PICTUREFLAG_PF);

	switch (chroma_format) {
		case CHROMA420:
			if (MMX_enabled) {
				mmx_conv420to422(U_ptr, uv422, cpw >> 1, cph >> 1, pf);
				mmx_conv422to444(uv422, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				mmx_conv420to422(V_ptr, uv422, cpw >> 1, cph >> 1, pf);
				mmx_conv422to444(uv422, v444ptr, cpw >> 1, cph, MPEG2_Flag);
				if (ISSE_enabled) __asm sfence
				__asm emms
			} else {
				mpeg_conv420to422(U_ptr, uv422, cpw >> 1, cph >> 1, pf);
				mpeg_conv422to444(uv422, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				mpeg_conv420to422(V_ptr, uv422, cpw >> 1, cph >> 1, pf);
				mpeg_conv422to444(uv422, v444ptr, cpw >> 1, cph, MPEG2_Flag);
			}
			break;

		case CHROMA422:
			if (MMX_enabled) {
				mmx_conv422to444(U_ptr, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				mmx_conv422to444(V_ptr, v444ptr, cpw >> 1, cph, MPEG2_Flag);
				__asm emms
			} else {
				mpeg_conv422to444(U_ptr, u444ptr, cpw >> 1, cph, MPEG2_Flag);
				mpeg_conv422to444(V_ptr, v444ptr, cpw >> 1, cph, MPEG2_Flag);
			}
			break;

		case CHROMA444:
			u444ptr = U_ptr;
			v444ptr = V_ptr;
			break;
	}

	mpeg_conv444toRGB(Y_ptr, u444ptr, v444ptr, dst, bpr, w, h, bpp);
}


void mpeg_convert_frame32(void *output_buffer, int buffer_ID) {
    if (memblock != NULL && output_buffer != NULL && (unsigned int)buffer_ID < 3) {
		mpeg_YUVToRGB(buffer_ID, output_buffer, mbWidth*16*4,
			pelWidth, pelHeight, 32);
    }
}

void mpeg_convert_frame24(void *output_buffer, int buffer_ID) {
    if (memblock != NULL && output_buffer != NULL && (unsigned int)buffer_ID < 3) {
		mpeg_YUVToRGB(buffer_ID, output_buffer, mbWidth*16*3,
			pelWidth, pelHeight, 24);
    }
}

void mpeg_convert_frame16(void *output_buffer, int buffer_ID) {
    if (memblock != NULL && output_buffer != NULL && (unsigned int)buffer_ID < 3) {
		mpeg_YUVToRGB(buffer_ID, output_buffer, mbWidth*16*2,
			pelWidth, pelHeight, 16);
    }
}

// Convert YCbCr to UYVY

void mpeg_convert_frameUYVY16(void *output_buffer, int buffer_ID) {

	if (memblock != NULL && output_buffer != NULL && (unsigned int)buffer_ID < 3UL) {

		YUVPixel *Y_ptr = buffers[buffer_ID].Y;
		YUVPixel *U_ptr = buffers[buffer_ID].U;
		YUVPixel *V_ptr = buffers[buffer_ID].V;

		YUVPixel *u444ptr = u444;
		YUVPixel *v444ptr = v444;

		long cpw = (pelWidth + 15) & -16;
		long cph = (pelHeight + 15) & -16;

		bool pf = !!(buffers[buffer_ID].bitflags & PICTUREFLAG_PF);

		switch (chroma_format) {

			case CHROMA420:
				if (MMX_enabled) {
					mmx_conv420to422(U_ptr, u444ptr, cpw >> 1, cph >> 1, pf);
					mmx_conv420to422(V_ptr, v444ptr, cpw >> 1, cph >> 1, pf);
					mmx_conv422toUYVY(Y_ptr, u444ptr, v444ptr, output_buffer, cpw >> 1, cph);
					if (ISSE_enabled) __asm sfence
					__asm emms
					return;
				}
				mpeg_conv420to422(U_ptr, u444ptr, cpw >> 1, cph >> 1, pf);
				mpeg_conv420to422(V_ptr, v444ptr, cpw >> 1, cph >> 1, pf);
				break;

			case CHROMA422:
				if (MMX_enabled) {
					mmx_conv422toUYVY(Y_ptr, U_ptr, V_ptr, output_buffer, cpw >> 1, cph);
					__asm emms
					return;
				}
				u444ptr = U_ptr;
				v444ptr = V_ptr;
				break;

			case CHROMA444:
				// Unsupported in this mode
				return;
		}

		YUVPixel *dst = (YUVPixel *)output_buffer;
		for (long x = (cpw >> 1) * cph; x > 0; x--) {
			*dst++ = *u444ptr++;
			*dst++ = *Y_ptr++;
			*dst++ = *v444ptr++;
			*dst++ = *Y_ptr++;
		}
	}
}

// Convert YCbCr to YUY2

void mpeg_convert_frameYUY216(void *output_buffer, int buffer_ID) {

	if (memblock != NULL && output_buffer != NULL && (unsigned int)buffer_ID < 3UL) {

		YUVPixel *Y_ptr = buffers[buffer_ID].Y;
		YUVPixel *U_ptr = buffers[buffer_ID].U;
		YUVPixel *V_ptr = buffers[buffer_ID].V;

		YUVPixel *u444ptr = u444;
		YUVPixel *v444ptr = v444;

		long cpw = (pelWidth + 15) & -16;
		long cph = (pelHeight + 15) & -16;

		bool pf = !!(buffers[buffer_ID].bitflags & PICTUREFLAG_PF);

		switch (chroma_format) {

			case CHROMA420:
				if (MMX_enabled) {
					mmx_conv420to422(U_ptr, u444ptr, cpw >> 1, cph >> 1, pf);
					mmx_conv420to422(V_ptr, v444ptr, cpw >> 1, cph >> 1, pf);
					mmx_conv422toYUY2(Y_ptr, u444ptr, v444ptr, output_buffer, cpw >> 1, cph);
					if (ISSE_enabled) __asm sfence
					__asm emms
					return;
				}
				mpeg_conv420to422(U_ptr, u444ptr, cpw >> 1, cph >> 1, pf);
				mpeg_conv420to422(V_ptr, v444ptr, cpw >> 1, cph >> 1, pf);
				break;

			case CHROMA422:
				if (MMX_enabled) {
					mmx_conv422toYUY2(Y_ptr, U_ptr, V_ptr, output_buffer, cpw >> 1, cph);
					__asm emms
					return;
				}
				u444ptr = U_ptr;
				v444ptr = V_ptr;
				break;

			case CHROMA444:
				// Unsupported in this mode
				return;
		}

		YUVPixel *dst = (YUVPixel *)output_buffer;
		for (long x = (cpw >> 1) * cph; x > 0; x--) {
			*dst++ = *Y_ptr++;
			*dst++ = *u444ptr++;
			*dst++ = *Y_ptr++;
			*dst++ = *v444ptr++;
		}
	}
}

// Convert YCbCr 4:2:0 to YV12

void mpeg_convert_frameYV12(void *output_buffer, int buffer_ID) {

	if (memblock != NULL && output_buffer != NULL && (unsigned int)buffer_ID < 3UL) {

		// Size of UV planes
		long bytes = (pelWidth >> 1) * (pelHeight >> 1);

		// Y plane
		memcpy(output_buffer, buffers[buffer_ID].Y, bytes << 2);
		output_buffer = (char *)output_buffer + (bytes << 2);

		// Only 4:2:0 is supported in this mode

		if (chroma_format == CHROMA420) {

			memcpy(output_buffer, buffers[buffer_ID].V, bytes);
			output_buffer = (char *)output_buffer + bytes;

			memcpy(output_buffer, buffers[buffer_ID].U, bytes);
		}
	}
}
